/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * MUDInternal.h - internal view of MUDLib.g
 */

#ifndef EXEC_H
struct Exec;
#endif
#ifndef REQUEST_H
struct Request;
#endif
#ifndef EFFECTS_H
struct EffectsInfo;
#endif

#define FIXED_COUNT	8

#if TAG_ALLOCS
#define MEM_EXTRA	(sizeof(ULONG_T) + sizeof(BYTE_T *))
#define MEM_ALLOC(ms, len, what)	memAllocT(ms, len, what)
#else
#define MEM_EXTRA	sizeof(ULONG_T)
#define MEM_ALLOC(ms, len, what)	memAlloc(ms, len)
#endif

/* this must be at least as big as the one in SMUD.c! */

#define INPUT_BUFFER_SIZE	(512 + 1)   /* must be at least TEXT_COLUMNS */

typedef struct MemBlock {
    struct MemBlock *mb_next;
    ULONG_T mb_length;
} MemBlock_t;

typedef struct MemPiece {
#if TAG_ALLOCS
    CHAR_T *mp_what;
#endif
    union {
	BYTE_T mp_tag[4];	/* use high byte of length */
	ULONG_T mp_length;	/* includes the tag */
    } mp_;
    struct MemPiece *mp_next;	/* next in short list or long list */
    struct MemPiece *mp_prev;	/* prev in long list */
} MemPiece_t;
#if ENDIAN == LITTLE_ENDIAN
#define TAG_INDEX   3
#else
#define TAG_INDEX   0
#endif

typedef struct FileListStruct {
    struct FileListStruct *fl_next;
    int fl_fd;
    UINT_T fl_inputTotal, fl_inputPos;
    UINT_T
	fl_lineNumber, fl_prevLine, fl_prevPrevLine,
	fl_column, fl_prevColumn, fl_prevPrevColumn;
    CHAR_T fl_fileName[TEXT_COLUMNS + 1];
    CHAR_T fl_inputLine[INPUT_BUFFER_SIZE];
    CHAR_T fl_lastChar;
} FileList_t;

typedef struct ProcList {
    struct ProcList *prl_next;
    Proc_t *prl_proc;
    ULONG_T prl_key;
    UINT_T prl_lockCount;
} ProcList_t;

typedef struct ConstRefList {
    struct ConstRefList *crl_next;
    ULONG_T crl_key;
    BYTE_T *crl_ptr;
} ConstRefList_t;

typedef struct {
    void *(*ms_AllocMem)(ULONG_T length);
    void (*ms_FreeMem)(void *p, ULONG_T length);
    BOOL_T (*ms_ReclaimMem)(BOOL_T panicing);
    BOOL_T (*ms_RecoverMem)(void);
    BOOL_T (*ms_RetryMem)(void);
    int (*ms_Open)(CHAR_T *fileName, int mode, int prot);
    void (*ms_Close)(int fd);
    LONG_T (*ms_Read)(int fd, void *buffer, ULONG_T len);

    void (*ms_abort)(BOOL_T alreadyGone);
    void (*ms_errorChar)(CHAR_T ch);
    void (*ms_outputChar)(CHAR_T ch);

    /* server only */

    void (*ms_incRef)(ULONG_T key);
    void (*ms_decRef)(ULONG_T key);
    ULONG_T (*ms_lookup)(CHAR_T *name, UINT_T nameLen);
    CHAR_T *(*ms_findAnyName)(ULONG_T key);
    Proc_t *(*ms_findProc)(ULONG_T key, UINT_T **pLockCount);
    ULONG_T (*ms_writeProcNew)(BYTE_T *p, UINT_T len);
    Proc_t *(*ms_findBuiltin)(ULONG_T key);
    ULONG_T (*ms_getIntConstValue)(ULONG_T key);

    /* client only */

    void (*ms_putText)(CHAR_T *ptr, UINT_T len);
    void (*ms_setPrompt)(CHAR_T *newPrompt);
    void (*ms_setEcho)(BOOL_T echoOn);
    void (*ms_serverSend)(struct Request *rq);
    struct Request *(*ms_serverRequest)(struct Request *rq);
    void (*ms_handleAll)(BOOL_T wantInput);
    void (*ms_checkMessages)(void);
    void (*ms_getEffects)(struct EffectsInfo *ei);
    void (*ms_doEffects)(BYTE_T *buf, UINT_T len);
    BOOL_T (*ms_doEdit)(CHAR_T *buf, ULONG_T len, UINT_T mode, BOOL_T isProc,
			BOOL_T isLast, BOOL_T deleteFirst, CHAR_T *prompt);
    void (*ms_editErrorStart)(UINT_T line, UINT_T column);
    void (*ms_editErrorString)(CHAR_T *buf);
    void (*ms_editErrorEnd)(void);
    BOOL_T (*ms_getString)(CHAR_T *buf, CHAR_T *prompt, CHAR_T **pRes);
    void (*ms_defineEffect)(ULONG_T key, BYTE_T *data, UINT_T len);
    void (*ms_flushEffect)(ULONG_T key);

    /* variables - things have been rearranged a bit and some padding
       added, so that long-word variables will end up on long-word
       boundaries, assuming the whole structure does, which users of the
       library should try to arrange for. */

    /* stuff for memory.c */

    MemBlock_t *MemBlocks;
    MemPiece_t *LongList;
    MemPiece_t *(ShortList[FIXED_COUNT]);
    ULONG_T MemoryLimit, MemoryAllocated, MemoryInUse;
    ULONG_T MemoryReclaims, MemoryPanics;
    ULONG_T MemoryAllocs, MemoryFrees;
    BOOL_T Merging, Panicing;
    BYTE_T ms_padding0[2];

    /* stuff for byteIO.c */

    BYTE_T *WriteBuffer;
    UINT_T WritePos, WriteBufferSize;

    /* stuff for procIO.c */

    Proc_t *ReadingProc;
    ConstRefList_t *ConstRefList;

    /* stuff for decl.c */

    Proc_t *ParsingProc;
    ULONG_T ParsingProcKey;

    /* stuff for lex.c */

    CHAR_T *Buffer;
    ULONG_T BufferSize, BufferPos;
    CHAR_T CurrentChar, NextChar;
    UINT_T EofState;
    BOOL_T NewLineSignificant;
    BYTE_T ms_padding1[3];		/* get next long-word aligned */
    Table_t ReservedWordTable;

    /* stuff shared by lex.c and the parsers */

    UINT_T tk_kind;
    UINT_T tk_stringLength;
    LONG_T tk_numericValue;
    CHAR_T *tk_stringValue;

    /* stuff for pexec.c, also peeked and poked elsewhere */

    UINT_T ErrorCount;

    /* stuff for main.c, PlayerKey peeked in decl.c also */

    enum ch_state {
	st_name,	    /* asking for player name */
	st_plPassword,      /* asking for player password */
	st_newPlayerNCP,    /* asking if create new player, no creation pw */
	st_newPlayerWCP,    /* asking if create new player, with creation pw */
	st_crPassword,      /* asking for creation password */
	st_npPassword1,     /* asking for new player password */
	st_npPassword2,     /* re-asking for verification */
	st_parsing,	    /* directly executing in wizard mode */
	st_normal,	    /* normal - input goes to server */
	st_exiting,	    /* waiting for a return to exit */
	st_changePassword1, /* server wants us to enter old password */
	st_changePassword2, /* server wants us to enter new password */
	st_changePassword3, /* verify new password */
    } State, SaveState;
    UINT_T nameCount;	    /* how many attempts at character name */
    UINT_T ynCount;	    /* how many attempts at a Y/N */

    CHAR_T PlayerName[PLAYER_NAME_LEN];
    CHAR_T Password[PASSWORD_LEN];

    Table_t SymbolCache;

    CHAR_T InputLine[INPUT_BUFFER_SIZE] ; 	/* for direct input */
    BOOL_T IsRemote, EditReading, EditReadEof;	/* this is remote connection */
    BOOL_T GoAhead;				/* had runClient from server */
    BOOL_T InputSync;				/* synchronous mode */
    UINT_T InputTotal, InputPos;		/* for direct input */
    UINT_T LockCount;				/* fake lockCount */
    ULONG_T EditProc;				/* proc being edited */
    FileList_t *FileList;
    CHAR_T *InputPtr;				/* for 'parseAction' */
    UINT_T EditLineNumber;

    struct Request *FreeRequests;		/* list of free ones */
    UINT_T RequestLimit, RequestCount;		/* max count, current count */

    ProcList_t **ProcMap;
    ProcList_t *ProcList;
    ULONG_T PlayerKey;
    ULONG_T SessionKey;
    CHAR_T *SuppliedPassword;

    enum {nl_clear, nl_faked, nl_blanked1, nl_blanked2} NewLineState;
    CHAR_T LastChar;
    BOOL_T ParseStarted, HadEof, DefineOnError, WantQuit;
    BOOL_T CacheProcs, CacheSymbols;

} MUDState_t;

/* util.c */

extern void errorString(MUDState_t *ms, CHAR_T *st);
extern void errorHex(MUDState_t *ms, ULONG_T n);
extern void errorDecimal(MUDState_t *ms, ULONG_T n);
extern void debugX(MUDState_t *ms, CHAR_T *m1, ULONG_T n, CHAR_T *m2);
extern void outputString(MUDState_t *ms, CHAR_T *st);
extern void outputDecimal(MUDState_t *ms, ULONG_T n);

extern UINT_T charsLen1(CHAR_T *st);
extern void charsCopy(CHAR_T *dest, CHAR_T *src);
extern CHAR_T *charsDup(MUDState_t *ms, CHAR_T *st);
extern void charsFree(MUDState_t *ms, CHAR_T *st);
extern BOOL_T charsEqual(CHAR_T *s1, CHAR_T *s2);
extern void blockCopy(void *dest, void *src, UINT_T len);

/* main.c */

/* main.c - interfaces exported to public clients */

extern void serverMessage(MUDState_t *ms, struct Request *rq);
extern void inputLine(MUDState_t *ms, CHAR_T *p, ULONG_T len);
extern void syncControl(MUDState_t *ms, ULONG_T flag);
extern void rawKeyEvent(MUDState_t *ms, ULONG_T key);
extern void mouseUpEvent(MUDState_t *ms, ULONG_T n, ULONG_T position);
extern void buttonEvent(MUDState_t *ms, ULONG_T whichButton);
extern void releaseEffect(MUDState_t *ms, ULONG_T key);
extern void textResize(MUDState_t *ms, ULONG_T newRows, ULONG_T newCols);
extern ULONG_T parseActionFromString(MUDState_t *ms, BYTE_T *p, ULONG_T owner);
extern ULONG_T parseProcFromString(MUDState_t *ms, BYTE_T *p, ULONG_T owner);
extern void graphicsFlip(MUDState_t *ms, ULONG_T onOff);
extern void voiceFlip(MUDState_t *ms, ULONG_T onOff);
extern void soundFlip(MUDState_t *ms, ULONG_T onOff);
extern void musicFlip(MUDState_t *ms, ULONG_T onOff);
extern void effectDone(MUDState_t *ms, ULONG_T key, ULONG_T id);
extern void editStringDone(MUDState_t *ms, CHAR_T *buf,ULONG_T len,ULONG_T ok);
extern void editProcStart(MUDState_t *ms);
extern void editProcDone(MUDState_t *ms);
extern void editProcCancel(MUDState_t *ms);
extern BOOL_T MUDInitialize(MUDState_t *ms, ULONG_T flags, ULONG_T rqMax);
extern void MUDConnectMessage(MUDState_t *ms, struct Request *rq);
extern BOOL_T MUDConnected(MUDState_t *ms, struct Request *rq);
extern void MUDMain(MUDState_t *ms, CHAR_T *playerName, CHAR_T *password);
extern void MUDDisconnectMessage(MUDState_t *ms, struct Request *rq);
extern void MUDAbort(MUDState_t *ms);
extern void MUDShutDown(MUDState_t *ms);
extern void MUDTerminate(MUDState_t *ms);
extern BOOL_T MUDSource(MUDState_t *ms, CHAR_T *fileName);
extern void MUDCrypt(CHAR_T *p, ULONG_T decrypt, ULONG_T sessionKey);

/* main.c - interfaces exported to private clients */

extern struct Request *allocRequestE(MUDState_t *ms, ULONG_T len);
extern void freeRequestE(MUDState_t *ms, struct Request *rq);
extern void requestInitE(MUDState_t *ms, ULONG_T rqLimit);
extern void requestTermE(MUDState_t *ms);
extern ULONG_T libInitE(MUDState_t *ms, ULONG_T memLimit, ULONG_T rqMax);
extern void libTermE(MUDState_t *ms);

/* main.c - interfaces exported within the library only */

extern void errorHeader(MUDState_t *ms);
extern void getCurrentPos(MUDState_t *ms, CHAR_T *pFileName,
			  UINT_T *pLine, UINT_T *pCol);
extern void errorMessage(MUDState_t *ms, CHAR_T *m);
extern void errorTrailer(MUDState_t *ms);
extern void parseError(MUDState_t *ms, CHAR_T *message);
extern void libSymbolError(MUDState_t *ms, CHAR_T *symbol, CHAR_T *message);
extern void freeRequests(MUDState_t *ms);
extern void libAbort(MUDState_t *ms, BOOL_T alreadyGone);
extern void deleteProcFromList(MUDState_t *ms, ULONG_T key);
extern void memAbort(MUDState_t *ms, BOOL_T alreadyGone);
extern BOOL_T getChar(MUDState_t *ms, CHAR_T *pCh);
extern void libIncRef(MUDState_t *ms, ULONG_T key);
extern void libDecRef(MUDState_t *ms, ULONG_T key);
extern ULONG_T libLookup(MUDState_t *ms, CHAR_T *name, UINT_T nameLen);
extern CHAR_T *libFindName(MUDState_t *ms, ULONG_T key);
extern Proc_t *findProcOnList(MUDState_t *ms, ULONG_T key,UINT_T **pLockCount);
extern ULONG_T writeProcNew(MUDState_t *ms, Proc_t *pr);
extern void defineSymbol(MUDState_t *ms, CHAR_T *name, UINT_T nameLen,
			 ULONG_T valu, ULONG_T table);
extern void deleteSymbol(MUDState_t *ms, CHAR_T *name, UINT_T nameLen,
			 ULONG_T key, ULONG_T table);
extern void replaceProc(MUDState_t *ms, ULONG_T key, Proc_t *pr);

/* memory.c */

/* memory.c - interfaces exported to private clients */

extern void memFreeE(MUDState_t *ms, void *p, ULONG_T len);
extern void *memAllocE(MUDState_t *ms, ULONG_T len);
extern void *memAllocET(MUDState_t *ms, ULONG_T len, char *what);

/* memory.c - interfaces exported within the library only */

extern void memFreeBlocks(MUDState_t *ms);
extern void memFree(MUDState_t *ms, void *p, ULONG_T len);
extern void *memAlloc(MUDState_t *ms, ULONG_T len);
extern void *memAllocT(MUDState_t *ms, ULONG_T len, char *what);
extern void memInit(MUDState_t *ms);
extern void memTerm(MUDState_t *ms);

/* symbol.c */

/* symbol.c - interfaces exported within the library only */

extern void symCreate(MUDState_t *ms, Table_t *tb, UINT_T shift);
extern void symDestroy(MUDState_t *ms, Table_t *tb);
extern Entry_t *symEnter(MUDState_t *ms, Table_t *tb, CHAR_T *name,
			 UINT_T nameLen);
extern Entry_t *symLookup(Table_t *tb, CHAR_T *name, UINT_T nameLen);
extern void symDelete(MUDState_t *ms, Table_t *tb,CHAR_T *name,UINT_T nameLen);
extern void symKeyDelete(MUDState_t *ms, Table_t *tb, ULONG_T key);

/* symbol.c - interfaces exported to private clients */

extern void symCreateE(MUDState_t *ms, Table_t *tb, ULONG_T shift);
extern void symDestroyE(MUDState_t *ms, Table_t *tb);
extern Entry_t *symEnterE(MUDState_t *ms, Table_t *tb,
			  CHAR_T *name, ULONG_T nameLen);
extern Entry_t *symLookupE(Table_t *tb, CHAR_T *name, ULONG_T nameLen);
extern void symDeleteE(MUDState_t *ms, Table_t *tb,
		       CHAR_T *name, ULONG_T nameLen);
extern void *symReadE(MUDState_t *ms, Table_t *tb, void *b);
extern void symWriteE(MUDState_t *ms, Table_t *tb);

/* byteIO.c */

/* byteIO.c - interfaces exported within the library only */

extern void writeResult(MUDState_t *ms, BYTE_T **pBytes, UINT_T *pLen);
extern void writeByte(MUDState_t *ms, BYTE_T b);
extern void writeWord(MUDState_t *ms, UINT_T w);
extern void writeLong(MUDState_t *ms, ULONG_T l);
extern void writeBytes(MUDState_t *ms, void *b, UINT_T len);
extern void writeString(MUDState_t *ms, CHAR_T *st);
extern UINT_T readWord(void **pb);
extern ULONG_T readLong(void **pb);

/* byteIO.c - interfaces exported to private clients */

extern void writeResultE(MUDState_t *ms, BYTE_T **pBytes, UINT_T *pLen);
extern void writeLongE(MUDState_t *ms, ULONG_T l);
extern ULONG_T readLongE(void **pb);
extern void writeBytesE(MUDState_t *ms, void *p, UINT_T len);

/* procIO.c */

/* procIO.c - interfaces exported within the library only */

extern void writeProc(MUDState_t *ms, Proc_t *pr);
extern Proc_t *readProc(MUDState_t *ms, void *p);
extern void freeExec(MUDState_t *ms, struct Exec *ex);
extern void freeProc(MUDState_t *ms, Proc_t *pr, ULONG_T myKey,
		     BOOL_T freeRefs);
extern void clearProc(MUDState_t *ms, Proc_t *pr, ULONG_T myKey);

/* procIO.c - interfaces exported to private clients */

extern void writeProcE(MUDState_t *ms, Proc_t *pr);
extern Proc_t *readProcE(MUDState_t *ms, void *p);
extern void freeProcE(MUDState_t *ms, Proc_t *pr,
		      ULONG_T myKey, ULONG_T freeRefs);

/* lex.c */

extern void lexInit(MUDState_t *ms);
extern void lexTerm(MUDState_t *ms);
extern void clearNewLineSignificant(MUDState_t *ms);
extern void setNewLineSignificant(MUDState_t *ms);
extern void lexStart(MUDState_t *ms, CHAR_T current, CHAR_T next);
extern void whiteSpace(MUDState_t *ms);
extern void getToken(MUDState_t *ms);
extern CHAR_T peekChar(MUDState_t *ms);
extern void getNote(MUDState_t *ms);

/* decl.c */

extern UINT_T parseType(MUDState_t *ms);
extern void parseProc(MUDState_t *ms, ULONG_T table, ULONG_T oldKey,
		      CHAR_T *oldName, BOOL_T errorDefine, BOOL_T verbose);
extern void parseAction(MUDState_t *ms, struct Exec **pEx, UINT_T *pType);
extern Proc_t *parseImmediateAction(MUDState_t *ms, BOOL_T allowDir);
extern Proc_t *parseImmediateProc(MUDState_t *ms);
extern RefList_t *makeRef(MUDState_t *ms, ULONG_T key);
extern LocalList_t *localLookup(MUDState_t *ms, CHAR_T *name);

/* pexec.c */

extern BOOL_T libNilType(UINT_T t);
extern void parseBody(MUDState_t *ms, struct Exec **pEx, UINT_T *pType,
		      BOOL_T allowDir);


/* flag bits for MUDInitialize */

#define FLAG_CACHE_PROCS	0x01
#define FLAG_CACHE_SYMBOLS	0x02
#define FLAG_IS_REMOTE		0x04
