#include "MUD.h"
#include "types.h"
#include "MUDInternal.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * util.c - a few utility routines for the MUD library.
 */

/*
 * errorString - output a string to the error/debug/log
 */

void
errorString(MUDState_t *ms, CHAR_T *st)
{

    while (*st != '\0') {
	ms->ms_errorChar(*st++);
    }
}

/*
 * errorHex - output a hex number to the error stuff.
 */

void
errorHex(MUDState_t *ms, ULONG_T n)
{
    CHAR_T *p;
    int cnt;
    BYTE_T b;
    CHAR_T buff[9];

    p = &buff[8];
    *p = '\0';
    for (cnt = 0; cnt < 8; cnt += 1) {
	b = (BYTE_T) n & 0xf;
	n >>= 4;
	*--p = (b > 9) ? (b - 10 + 'a') : (b + '0');
    }
    errorString(ms, p);
}

/*
 * errorDecimal - output a decimal number to the error stuff.
 */

void
errorDecimal(MUDState_t *ms, ULONG_T n)
{
    CHAR_T buffer[11];
    CHAR_T *p;

    p = &buffer[10];
    *p = '\0';
    do {
	*--p = n % 10 + '0';
	n /= 10;
    } while (n != 0);
    errorString(ms, p);
}

/*
 * debugX - utility for use while debugging.
 */

void
debugX(MUDState_t *ms, CHAR_T *m1, ULONG_T n, CHAR_T *m2)
{

    errorString(ms, m1);
    errorHex(ms, n);
    errorString(ms, m2);
}

/*
 * outputString - output a string to the normal output.
 */

void
outputString(MUDState_t *ms, CHAR_T *st)
{

    while (*st != '\0') {
	ms->ms_outputChar(*st++);
    }
}

/*
 * outputDecimal - output a decimal number to the output stuff.
 */

void
outputDecimal(MUDState_t *ms, ULONG_T n)
{
    CHAR_T buffer[11];
    CHAR_T *p;

    p = &buffer[10];
    *p = '\0';
    do {
	*--p = n % 10 + '0';
	n /= 10;
    } while (n != 0);
    outputString(ms, p);
}

/*
 * charsLen1 - return the length of a string, including the trailing \0
 */

UINT_T
charsLen1(CHAR_T *st)
{
    UINT_T len;

    len = 1;
    while (*st != '\0') {
	++st;
	len += 1;
    }
    return len;
}

/*
 * charsCopy - copy a \0 terminated string.
 */

void
charsCopy(CHAR_T *dest, CHAR_T *src)
{
    CHAR_T ch;

    do {
	ch = *src++;
	*dest++ = ch;
    } while (ch != '\0');
}

/*
 * charsDup - allocate and return a copy of a string.
 */

CHAR_T *
charsDup(MUDState_t *ms, CHAR_T *st)
{
    CHAR_T *copy;

    copy = MEM_ALLOC(ms, charsLen1(st), "STNG");
    charsCopy(copy, st);
    return copy;
}

/*
 * charsFree - free a string.
 */

void
charsFree(MUDState_t *ms, CHAR_T *st)
{

    memFree(ms, st, charsLen1(st));
}

/*
 * charsEqual - compare two strings.
 */

BOOL_T
charsEqual(CHAR_T *p1, CHAR_T *p2)
{

    while (*p1 == *p2 && *p1 != '\0') {
	++p1;
	++p2;
    }
    return (BOOL_T) (*p1 == *p2);
}

/*
 * blockCopy - copy a block of memory from one place to another.
 */

void
blockCopy(void *dest, void *src, UINT_T len)
{
    BYTE_T *dp, *sp;

    dp = (BYTE_T *) dest;
    sp = (BYTE_T *) src;
    while (len != 0) {
	len -= 1;
	*dp++ = *sp++;
    }
}
