/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * ClientStuff.h - stuff used by main2.c and its helpers to handle clients,
 *	machines, and the Public stuff.
 */

#define CLIENTSTUFF_H

#define AGENT_MAP_SIZE	1024

#ifndef REQUEST_H
struct Request;
#endif

struct TelnetClient;

typedef struct TableList {
    struct TableList *tl_next;		/* next in global table list */
    struct TableList *tl_prev;		/* make deletion easier */
    Table_t *tl_this;			/* the table */
    NameMap_t *tl_nameMap;		/* the namemap for the table */
    ULONG_T tl_key;			/* the key of the table */
    UINT_T tl_useCount; 		/* number of people using it now */
} TableList_t;

typedef struct UseList {
    struct UseList *ul_next;		/* next in list of used tables */
    Table_t *ul_this;			/* the table */
    TableList_t *ul_entry;		/* pointer to global entry */
} UseList_t;

typedef struct KnownIconList {
    struct KnownIconList *kil_next;
    ULONG_T kil_key;
    BOOL_T kil_temporary;		/* make client forget it on reset */
} KnownIconList_t;

typedef struct KnownEffectList {
    struct KnownEffectList *kel_next;
    struct KnownEffectList *kel_prev;
    ULONG_T kel_key;
} KnownEffectList_t;

typedef struct TextComponentList {
    struct TextComponentList *tcl_next;
    ULONG_T tcl_id;
    UINT_T tcl_width;			/* in characters */
    UINT_T tcl_height;
} TextComponentList_t;

typedef struct DrawComponentList {
    struct DrawComponentList *dcl_next;
    ULONG_T dcl_id;
    UINT_T dcl_rows;			/* in pixels */
    UINT_T dcl_cols;
} DrawComponentList_t;

typedef struct GUIFrame {
    TextComponentList_t *guf_textComponents;
    DrawComponentList_t *guf_drawComponents;
} GUIFrame_t;

typedef struct Client {
    struct Client *cl_next;		/* next in main list of clients */
    struct Client *cl_prev;		/* prev in main list of clients */
    struct Client *cl_nextKHash;	/* next in list hashed on thing key */
    struct Client *cl_nextPHash;	/* next in list hashed on player key */
    struct Client *cl_nextLHash;	/* next in diff location, same hash */
    struct Client *cl_prevLHash;
    struct Client *cl_nextLoc;		/* next in same location */
    struct Client *cl_prevLoc;
    struct TelnetClient *cl_telnet;	/* additional stuff */
    int cl_socket;			/* this client's TCP/IP socket */
    struct Request *cl_outGoing;	/* unsent outgoing messages */
    ULONG_T cl_outGoingPos;		/* sent position in first outgoing */
    struct Request *cl_inComing;	/* any partial incoming message */
    ULONG_T cl_inComingPos;		/* position in incoming message */
    UseList_t *cl_useList;		/* the tables in use */
    KnownIconList_t *cl_knownIcons;	/* which ones the client has seen */
    KnownEffectList_t *cl_knownEffects; /* client has these */
    GUIFrame_t *cl_activeFrame;		/* the one in valid use now */
    GUIFrame_t *cl_newFrame;		/* next one being constructed */
    ULONG_T cl_sessionKey;		/* crypt key for session */
    ULONG_T cl_player;			/* the player key for this client */
    ULONG_T cl_thing;			/* the thing for this player */
    ULONG_T cl_location;		/* his current location */
    ULONG_T cl_inputAction;		/* action to do on input lines */
    ULONG_T cl_rawKeyAction;		/* action to do on special keys */
    ULONG_T cl_mouseDownAction; 	/* action to do on mouse down events */
    ULONG_T cl_buttonAction;		/* action to do on "button" hits */
    ULONG_T cl_effectDoneAction;	/* action to do on effect complete */
    ULONG_T cl_privateTableKey; 	/* key for his private table */
    ULONG_T cl_editStringAction;	/* action to do on editStringDone */
    ULONG_T cl_getStringAction; 	/* action to do on getStringDone */
    ULONG_T cl_queryFileAction; 	/* action to do on queryFileDone */
    ULONG_T cl_defaultTextComponent;	/* default component for text output */
    ULONG_T cl_defaultDrawComponent;	/* default component for graphics */
    ULONG_T cl_lastUsedDrawComponent;	/* last one sent to */
    Table_t *cl_privateSymbols; 	/* his private symbol table */
    NameMap_t *cl_privateMap;		/* his private name-map */
    CHAR_T *cl_name;			/* cached copy of his p_pName */
    CHAR_T *cl_prompt;			/* the current prompt */
    ULONG_T cl_onAt;			/* connect time */
    ULONG_T cl_idleAction;		/* action to do on disconnect */
    ULONG_T cl_activeAction;		/* action to do on restart */
    UINT_T cl_textWidth;		/* width of text display in chars */
    UINT_T cl_textHeight;		/* height of text display in chars */
    UINT_T cl_textColumn;		/* for output formatting */
    UINT_T cl_fontAscent;		/* reported by client */
    UINT_T cl_fontDescent;
    UINT_T cl_fontLeading;
    UINT_T cl_fontHeight;		/* sum of previous 3 */
    UINT_T cl_fontWidth;		/* we want it non-proportional */
    UINT_T cl_passwordCount;		/* times password been tried */
    BOOL_T cl_inserted;			/* done initial fullyInsertClient */
    BOOL_T cl_ended;			/* endClient started */
    BOOL_T cl_inputSync;		/* reply to rt_inputLine */
    BOOL_T cl_wizardMode;		/* currently in wizard mode */
    BOOL_T cl_newPlayer;		/* player has just been created */
    BOOL_T cl_gotTextSize;		/* got text sizes from client */
    PlayerStatus_t cl_status;		/* copy of pl_status */
    BOOL_T cl_shuttingDown;		/* already shutting down */
    BOOL_T cl_destroy;			/* destroy player on exit */
    BOOL_T cl_deleted;
    BOOL_T cl_expectingOldPassword;	/* need old one first on change */
    BOOL_T cl_playerPassword;		/* currently changing player pw */
    BOOL_T cl_editing;			/* currently editing proc or string */
    BOOL_T cl_getting;			/* currently getting a string */
    BOOL_T cl_querying; 		/* currently querying a file */
    BOOL_T cl_noAts;			/* do not print '@'s for non-wiz out */
    BOOL_T cl_fullPaths;		/* SysAdmin: full path all symbols */
    BOOL_T cl_isLocal;			/* is a local connection */

    BOOL_T cl_canEdit;
    BOOL_T cl_canGetString;
    BOOL_T cl_canQueryFile;
    BOOL_T cl_canWizard;
    BOOL_T cl_graphicsOn;
    BOOL_T cl_graphicsPalette;
    BOOL_T cl_voiceOn;
    BOOL_T cl_soundOn;
    BOOL_T cl_musicOn;
    UINT_T cl_graphicsRows, cl_graphicsCols;
    UINT_T cl_graphicsColours;
    CHAR_T *cl_graphicsType;		/* string describing graphics */
    UINT_T cl_clientVersion;
} Client_t;

typedef struct MachineList {
    struct MachineList *ml_next;
    struct MachineList *ml_prev;
    struct MachineList *ml_nextKHash;
    struct MachineList *ml_nextLHash;
    struct MachineList *ml_prevLHash;
    struct MachineList *ml_nextLoc;
    struct MachineList *ml_prevLoc;
    ULONG_T ml_owner;			/* owning player of this machine */
    ULONG_T ml_machine; 		/* key of associated machine */
    ULONG_T ml_thing;			/* key of associated thing */
    ULONG_T ml_location;		/* same as for player */
    ULONG_T ml_sayProc; 		/* when someone talks */
    ULONG_T ml_whisperMeProc;		/* someone whispers to machine */
    ULONG_T ml_whisperOtherProc;	/* overhears some other whisper */
    ULONG_T ml_pageProc;		/* someone pages the machine */
    ULONG_T ml_poseProc;		/* someone does a pose */
    ULONG_T ml_otherProc;		/* an 'OPrint'/'ABPrint' is done */
    ULONG_T ml_idleProc;		/* dungeon goes idle */
    ULONG_T ml_activeProc;		/* dungeon back from idle */
    CHAR_T *ml_name;			/* cached copy of its p_pName */
    BOOL_T ml_dirty;
    BOOL_T ml_deleted;			/* not really here - ignore machine */
} MachineList_t;

typedef struct CallList {
    struct CallList *cll_next;
    Client_t *cll_client;
    MachineList_t *cll_machine;
    ULONG_T cll_action;
    CHAR_T *cll_arg;
} CallList_t;

#ifndef EXTERN
#define EXTERN extern
#endif

EXTERN CHAR_T Path[150];		/* path to access files */
EXTERN ULONG_T CacheSize, MemoryLimit, RequestLimit;
EXTERN BOOL_T AlreadyAborting;

EXTERN Client_t *Clients;		/* master list of clients */
EXTERN Client_t *TrueClient, *ActiveClient;

EXTERN MachineList_t *Machines;
EXTERN MachineList_t *TrueMachine, *ActiveMachine;
EXTERN ULONG_T MachinesCreated, MachinesDeleted;

EXTERN Client_t *(ClientLocMap[AGENT_MAP_SIZE]);
EXTERN MachineList_t *(MachineLocMap[AGENT_MAP_SIZE]);
EXTERN Client_t *(ClientKeyMap[AGENT_MAP_SIZE]);
EXTERN MachineList_t *(MachineKeyMap[AGENT_MAP_SIZE]);
EXTERN Client_t *(ClientPKeyMap[AGENT_MAP_SIZE]);

EXTERN ULONG_T PublicSymbolTableKey;
EXTERN Table_t *PublicSymbolTable;
EXTERN NameMap_t *PublicNameMap;
EXTERN TableList_t *TableList;
EXTERN TableList_t *PlayerTable, *BuiltinTable;

#define PREFIX_MAX	20

EXTERN CHAR_T Prefix[PREFIX_MAX + 1];
EXTERN UINT_T PrefixLength;

EXTERN UINT_T CurrentIndent;
EXTERN BOOL_T ForceWizardMode, MachineDeleted, ClientDeleted;

/* PLAYER_NAME_LEN is too small for some machine names. */
#define ME_STRING_LEN	100
EXTERN CHAR_T MeString[ME_STRING_LEN];

#define DEFAULT_TEXT_WIDTH	78
#define DEFAULT_TEXT_HEIGHT	23
