/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

#define TRACING 	B_TRUE
#define DEBUG_PRESENT	B_TRUE
#define CHECKS_PRESENT	B_TRUE	/* some checks for nil string pointers */

#if TAG_ALLOCS
#define MEM_ALLOC_E(ms, len, what)	memAllocET(ms, len, what)
#else
#define MEM_ALLOC_E(ms, len, what)	memAllocE(ms, len)
#endif

/*
 * MUDServ.h - declare the functions exported from the various source files.
 */

struct TraceBack;
#ifndef REQUEST_H
struct Request;
#endif
#ifndef EXEC_H
struct Exec;
#endif
#ifndef CLIENTSTUFF_H
struct TableList;
struct TelnetClient;
struct GUIFrame;
struct Client;
struct MachineList;
struct CallList;
#endif

/* main1.c */

extern void messChar(char ch);
extern void mess(char *message);
extern void mHex(ULONG_T n);
extern void mNum(ULONG_T n);
extern BOOL_T checkForRetry(char *message);
extern void finalExit(void);

/* main2.c */

extern void addTelnetPort(int port, int flags);
extern void myAbort(char *message);
extern void myAbortNoWrite(void);
extern void addBuiltinProc(CHAR_T *name, ULONG_T key);
extern void subBuiltinProc(ULONG_T key);
extern struct CallList *addCall(struct CallList *oldList, struct Client *cl,
				struct MachineList *ml, ULONG_T action,
				CHAR_T *arg);
extern void runCallList(struct CallList *calls, BOOL_T switchToAgent);
extern void clearClient(struct Client *cl);
extern void removeFromWriteWaiters(struct Client *cl);
extern void addToWriteWaiters(struct Client *cl);
extern LONG_T doSend(struct Client *cl, void *buf, LONG_T len);
extern void putMessageCarefully(struct Client *cl, struct Request *rq);
extern void endClient2(struct Client *cl);
extern void sendChangePassword(struct Client *cl, ULONG_T key);
extern void startChangePassword(void);
extern void setNewPrompt(CHAR_T *newPrompt);
extern void changeClientMode(struct Client *cl, BOOL_T wizard);
extern void bootClient(struct Client *cl);
extern void nukeClient(struct Client *cl);
extern void deleteFromClients(ULONG_T key);
extern void machineShutDown(struct MachineList *ml);
extern void startTimeout(void);
extern void stopTimeout(void);
extern ULONG_T getTimeout(void);
extern void restoreTimeout(ULONG_T N);
extern LONG_T doRecv(struct Client *cl, void *buf, LONG_T len);
extern void setNeedRecover(void);
extern void requestCacheCompact(void);
extern void runServer(void);

/* telnet.c */

extern struct TelnetClient *telnetAllocClient(struct Client *cl, int flags);
extern void telnetStopClient(struct TelnetClient *tncl, BOOL_T doShutdown);
extern void telnetFreeClient(struct TelnetClient *tncl);
extern void telnetChangePassword(struct Client *cl);
extern void telnetSetPrompt(struct Client *cl, CHAR_T *newPrompt);
extern void telnetRead(struct Client *cl);
extern void telnetWrite(struct Client *cl);
extern void telnetPutText(struct Client *cl, struct Request *rq, BOOL_T flush);

/* messages.c */

extern void messageInit(void);
extern void setEditProcActive(BOOL_T state, ULONG_T procKey);
extern ULONG_T GetIndent(void);
extern void SetIndent(ULONG_T newIndent);
extern void sendText(struct Client *cl, CHAR_T *msg, UINT_T len);
extern void messageFlush(void);
extern void setSpecific(ULONG_T agent);
extern void setOthers(BOOL_T flag);
extern void setLogging(void);
extern void setToError(BOOL_T flag);
extern void putChar(CHAR_T ch);
extern void putString(CHAR_T *p);
extern CHAR_T *toggleProcToString(void);

/* clientFuncs.c */

extern struct Client *getClientByThing(ULONG_T thKey);
extern struct MachineList *getMachineByThing(ULONG_T thKey);
extern struct Client *getClientByPlayer(ULONG_T plKey);
extern struct Client *findClientsAt(ULONG_T location);
extern struct MachineList *findMachinesAt(ULONG_T location);
extern void locUnlinkClient(struct Client *cl, ULONG_T oldLoc);
extern void locInsertClient(struct Client *cl, ULONG_T newLoc);
extern void locUnlinkMachine(struct MachineList *ml, ULONG_T oldLoc);
extern void locInsertMachine(struct MachineList *ml, ULONG_T newLoc);
extern void fullyUnlinkClient(struct Client *cl);
extern void fullyInsertClient(struct Client *cl);
extern void fullyInsertMachine(struct MachineList *ml);
extern ULONG_T defineXXX(ULONG_T table, CHAR_T *name, CHAR_T *what,
			 ULONG_T key, ULONG_T thisNil);
extern ULONG_T lookupXXX(CHAR_T *name, CHAR_T *routine, UINT_T typ,
			 ULONG_T table, ULONG_T thisNil);
extern void EditProc(ULONG_T key);
extern UINT_T getTextWidth(struct Client *cl);
extern UINT_T getTextHeight(struct Client *cl);

/* machines.c */

extern void sendToMachines(ULONG_T location, ULONG_T agent1, ULONG_T agent2,
			   CHAR_T *message);

/* Characters.c */

extern void doCreateCharacter(Player_t *pl);
extern void ZapCharacter(ULONG_T key);

/* miscFuncs.c */

extern BOOL_T doFindKey(ULONG_T key, BOOL_T oneOnly);
extern CHAR_T *findPlayerThing(ULONG_T key);
extern CHAR_T *findPlayerTable(ULONG_T key);

/* clients.c */

extern UINT_T beginClient(struct Client *cl, struct Request *rq);
extern void nukePlayer(ULONG_T key);
extern void flushSymbol(ULONG_T key);
extern BOOL_T notInUse(ULONG_T key);
extern struct TableList *addGlobalUse(Table_t *tb, ULONG_T key);
extern void removeGlobalUse(struct TableList *tl);
extern void setupForCall(struct Client *cl);
extern void cleanupAfterCall(void);
extern void freeGUIFrame(struct GUIFrame *guf);
extern void endClient(struct Client *cl);
extern void askPlayer(struct Client *cl, struct Request *rq);
extern void creationCheck(struct Client *cl, struct Request *rq);
extern void createPlayer(struct Client *cl, struct Request *rq);
extern void loginPlayer(struct Client *cl, struct Request *rq);
extern ULONG_T lookup(CHAR_T *name, UINT_T nameLen);
extern CHAR_T *findAnyName(ULONG_T key);
extern void copyToMachine(Machine_t *mch, struct MachineList *ml);
extern BOOL_T doChangePassword(struct Client *cl, CHAR_T *pw);
extern void flushOneMachine(struct MachineList *ml);
extern void flushMachine(ULONG_T key);
extern void flushPlayer(ULONG_T key);
extern void flushClients(void);
extern void doNewPlayer(struct Client *cl);
extern void doStartPlayerStuff(struct Client *cl);
extern void setStatus(ULONG_T key, PlayerStatus_t status, ULONG_T doMode);
extern void SetIt(ULONG_T key);
extern void handleOneRequest(struct Client *cl, struct Request *rq);

/* fileIO.c */

extern void log(char *message);
extern void logX(char *m1, ULONG_T n, char *m2);
extern void logN(char *m1, ULONG_T n, char *m2);
extern void logXNL(char *m1, ULONG_T n);
extern void log2NL(char *m1, char *m2);
extern void ioShutDown(char *message);
extern BOOL_T ioInitialize(char *path, ULONG_T cacheSize);
extern void *ioRead(ULONG_T key);
extern ULONG_T ioCreate(UINT_T typ, void *p, UINT_T len);
extern void ioDirty(ULONG_T key);
extern void *ioExpand(ULONG_T key, UINT_T newLen);
extern void ioShrink(ULONG_T key, UINT_T newLen);
extern void ioReplace(ULONG_T key, void *p, UINT_T newLen);
extern void ioFlush(BOOL_T doIndex);
extern void ioDelete(ULONG_T key);
extern void ioWrite(ULONG_T key);
extern void ioCacheCompact(void);
extern void ioTerminate(void);
extern void ioDisplay(ULONG_T key);
extern ULONG_T getPublicSymbolTableKey(void);
extern ULONG_T getEntryAction(void);
extern void setEntryAction(ULONG_T key);
extern ULONG_T getCreationPassword(void);
extern ULONG_T getDungeonFlags(void);
extern void setDungeonFlags(ULONG_T newFlags);
extern ULONG_T getMachineList(void);
extern ULONG_T getAbortTimeout(void);
extern void setAbortTimeout(ULONG_T newValue);
extern void fileMaybeConCheck(char *where);

/* names.c */

extern void emptyNameMap(NameMap_t *nm);
extern NameMap_t *buildNameMap(Table_t *tb);
extern CHAR_T *findName(NameMap_t *nm, ULONG_T key);
extern void freeNameMap(NameMap_t *nm);
extern void removeNameFromMap(NameMap_t *nm, ULONG_T key);

/* misc.c */

extern void showSymbols(Table_t *tb, CHAR_T *header, CHAR_T *emptyMessage);
extern void fmtValue(UINT_T typ, ULONG_T valu);
extern void freeValue(UINT_T typ, ULONG_T valu);
extern void describeValue(ULONG_T key, CHAR_T *name);
extern void describeBuiltin(ULONG_T key, CHAR_T *name);
extern void describeSymbol(Table_t *tb, CHAR_T *name);
extern BOOL_T deleteable(Entry_t *en, BOOL_T isPrivate);
extern void fmtThing(ULONG_T key);

/* caches.c */

extern void addProcToCache(Proc_t *pr, ULONG_T key);
extern void removeProcFromCache(Proc_t *pr);
extern Proc_t *findCachedProc(ULONG_T key, UINT_T **pLockCount);
extern Proc_t *findCachedProcNoLock(ULONG_T key);
extern BOOL_T removeProc(ULONG_T key, UINT_T *pUseCount);
extern void flushProcCache(void);
extern void updateProcsInCache(ULONG_T ownerKey, PlayerStatus_t newStatus);
extern void incProcRef(ULONG_T key);
extern void decProcRef(ULONG_T key);
extern void decProcRefAndNil(ULONG_T *pKey);
extern Proc_t *findProcByPC(BYTE_T *pc, ULONG_T *pKey);
extern void incRef(ULONG_T key);
extern void decRef(ULONG_T key);
extern void ZapGrammar(ULONG_T key);
extern Grammar_t *findCachedGrammar(ULONG_T key);
extern Table_t *findCachedTable(ULONG_T key);
extern void flushCachedTable(ULONG_T key);
extern void cacheInit(void);
extern void flushRefKey(ULONG_T key);
extern void flushCaches(void);
extern void cacheTerm(BOOL_T freeMemory, BOOL_T doWrites);
extern void clearRefs(ULONG_T key);

/* util.c */

extern void utilInit(void);
extern void trace(char *where, ULONG_T valu);
extern void traceDump(void);
extern void copy(void *dst, void *src, ULONG_T len);
extern CHAR_T *strTail(CHAR_T *p, CHAR_T *q);
extern int strCmp(CHAR_T *left, CHAR_T *right);
extern int strCmpLC(CHAR_T *left, CHAR_T *right);
extern void strCpy(CHAR_T *dest, CHAR_T *src);
extern void strNCpy(CHAR_T *dest, CHAR_T *src, UINT_T len);
extern CHAR_T *nullString(void);
extern CHAR_T *concatenate(CHAR_T *left, CHAR_T *right);
extern ULONG_T nilValue(UINT_T typ);
extern void freeString(CHAR_T *st);
extern CHAR_T *stringCopy(CHAR_T *st);
extern UINT_T strLen1(CHAR_T *st);
extern void newLine(void);
extern void putError(CHAR_T *m);
extern void putStrError(CHAR_T *s1, CHAR_T *s2, CHAR_T *s3);
extern void putDecimal(ULONG_T n);
extern void putSigned(LONG_T n);
extern void putHex(ULONG_T n);
extern BOOL_T checkAction(ULONG_T key, CHAR_T *function);
extern BOOL_T checkBoolAction(ULONG_T key, CHAR_T *function);
extern BOOL_T checkIntAction(ULONG_T key, CHAR_T *function);
extern BOOL_T checkInt2Action(ULONG_T key, CHAR_T *function);
extern BOOL_T checkInt3Action(ULONG_T key, CHAR_T *function);
extern BOOL_T checkStringAction(ULONG_T key, CHAR_T *function);
extern BOOL_T checkThingAction(ULONG_T key, CHAR_T *function);
extern BOOL_T checkStringBoolAction(ULONG_T key, CHAR_T *function);
extern BOOL_T hasBadRefs(Proc_t *pr);
extern ULONG_T GetCurrentTime(void);
extern void ConvTime(ULONG_T seconds, CHAR_T *buffer);
extern void ConvTimeShort(ULONG_T seconds, CHAR_T *buffer);

/* listprop.c (others declared in builtin.c) */

extern void freeList(ULONG_T list, UINT_T typ);
extern void addProp(ULONG_T list, ULONG_T property, ULONG_T valu, UINT_T typ);
extern void delProp(ULONG_T list, ULONG_T property, UINT_T typ);
extern ULONG_T getProp(ULONG_T list, ULONG_T property, UINT_T typ);
extern void xAddTail(ULONG_T list, ULONG_T valu);
extern void DelElement(ULONG_T list, ULONG_T valu);
extern ULONG_T GetIntElement(ULONG_T list, ULONG_T index);
extern ULONG_T GetFixedElement(ULONG_T list, ULONG_T index);
extern ULONG_T GetThingElement(ULONG_T list, ULONG_T index);
extern ULONG_T GetActionElement(ULONG_T list, ULONG_T index);
extern void PutIntElement(ULONG_T list, ULONG_T index, ULONG_T valu);
extern void PutFixedElement(ULONG_T list, ULONG_T index, ULONG_T valu);
extern void PutThingElement(ULONG_T list, ULONG_T index, ULONG_T valu);
extern void PutActionElement(ULONG_T list, ULONG_T index, ULONG_T valu);
extern void nukeThing(ULONG_T key);
extern void ClearThing(ULONG_T key);

/* builtin.c */

extern void builtInInit(ULONG_T seed);
extern void builtInTerm(void);
extern CHAR_T *SubString(CHAR_T *st, LONG_T start, LONG_T length);
extern ULONG_T Random(ULONG_T rang);
extern Proc_t *findBuiltIn(ULONG_T key);

/* exec.c */

extern void runAbort(void);
extern void checkAbort(void);
extern void traceName(ULONG_T key, char *kind);
extern void tbShowValue(QValue_t *p, LocalList_t *ll);
extern void traceBack(void);
extern void runError(CHAR_T *message);
extern void symbolError(CHAR_T *name, CHAR_T *message);
extern void procError(ULONG_T key, CHAR_T *message);
extern BOOL_T nilType(UINT_T typ);
extern ULONG_T doProc0(ULONG_T key);
extern ULONG_T doProc0Top(ULONG_T key);
extern ULONG_T doProc1(ULONG_T key, void *par1);
extern ULONG_T doProc1Top(ULONG_T key, void *par1);
extern ULONG_T doProc2(ULONG_T key, void *par1, void *par2);
extern ULONG_T doProc2Top(ULONG_T key, void *par1, void *par2);
extern ULONG_T doProc3(ULONG_T key, void *par1, void *par2, void *par3);
extern ULONG_T doProc3Top(ULONG_T key, void *par1, void *par2, void *par3);
extern void runExec(struct Exec *ex, QValue_t *pRes, UINT_T *pType);
extern void runExpr(struct Exec *ex, QValue_t *pRes, UINT_T *pType);
extern LONG_T mulFixed(LONG_T a, LONG_T b);
extern LONG_T divFixed(LONG_T a, LONG_T b);

/* parse.c */

extern void parseInit(void);
extern BOOL_T sentEnd(CHAR_T ch);
extern void writeGrammar(Grammar_t *g, ULONG_T key);
extern void readGrammar(Grammar_t *g, ULONG_T key);
extern void freeGrammar(Grammar_t *g);
extern void releaseGrammar(Grammar_t *g);

/* match.c */

extern void setResult(ULONG_T theThing);
extern void matchInit(void);
extern CHAR_T *stuffTail(CHAR_T *p);
extern LONG_T MatchName(CHAR_T *stored, CHAR_T *parsed);
extern CHAR_T *FormatName(CHAR_T *stored);
extern ULONG_T xFindName(ULONG_T thingList, ULONG_T nameProp, CHAR_T *name);

/* print.c */

extern void fmtStart(void);
extern void fmtChar(CHAR_T ch);
extern void fmtString(CHAR_T *st);
extern void fmtLine(void);
extern void fmtCheck(CHAR_T *st);
extern void fmtCheck3(CHAR_T *st1, CHAR_T *st2, CHAR_T *st3);
extern void getTypeName(CHAR_T (*pBuffer)[30], UINT_T t);
extern void fmtNil(UINT_T t);
extern void fmtType(UINT_T t);
extern void fmtComma(CHAR_T *st);
extern void fmtInteger(LONG_T n);
extern void fmtHex(ULONG_T n);
extern CHAR_T *displayFixed(LONG_T n, CHAR_T (*buff)[20]);
extern void fmtFixed(LONG_T n);
extern void fmtBoolean(ULONG_T n);
extern void fmtStatus(ULONG_T n);
extern void fmtThingStatus(ULONG_T n);
extern void fmtAnyName(ULONG_T key, UINT_T typ);
extern void fmtActionName(ULONG_T key);
extern void fmtThingName(ULONG_T key);
extern void fmtPlayerName(ULONG_T key);
extern void fmtStringConst(CHAR_T *st);
extern void fmtProc(Proc_t *pr, CHAR_T *procName);

/* effects.c */

extern void effectsInit(void);
extern void effectsTerm(void);
extern void flushEffects(void);

/* bcComp.d */

extern void Compile(ULONG_T key);
extern void StripBody(ULONG_T key);

/* bcRun.d */

extern void bcInit(void);
extern void bcTerm(BOOL_T freeMemory);
extern void bcTraceBack(struct TraceBack *tb);
extern void doBcRun(ULONG_T key, Proc_t *pr, QValue_t *pArgs, QValue_t *pRes,
		    UINT_T argCount);

/* bcDis.d */
extern void bcByteSwap(BYTE_T *p, UINT_T len);

/*
 * dungeon flag values
 */

#define FLAG_SINGLE_USER		0x00000001
#define FLAG_REMOTE_SYSADMIN_OK 	0x00000002
#define FLAG_MACHINES_ACTIVE		0x00000004
#define FLAG_STARTUP_MACHINES		0x00000008
#define FLAG_DATABASE_STABLE		0x00000010

/* These few are shared by exec.c and bcRun.c */

/* note for future consideration: instead of allocating 'stackframes' for
   parameters and locals during execution, can just allocate a big 'stack
   page' on startup and have a 'stack pointer' into it move around during
   execution. This would be faster, but would require a tradeoff between
   using space for the 'stack' and limiting the call depth/size. */

/* larger values avoid more storage allocator calls, but consume more
   stack space per ANY call */
#define QUICK_LOCAL_COUNT	5 + 2

/* larger value might be wanted, but beware of stack usage! */
#define MAX_CALL_DEPTH		50

typedef struct TraceBack {
    struct TraceBack *tb_next;
    Proc_t *tb_proc;
    ULONG_T tb_key;
    QValue_t *tb_locals;
    BYTE_T *tb_PC;
    ULONG_T *tb_SP, *tb_FP;
    BOOL_T tb_byteCode;
} TraceBack_t;

/*
 * I hate to use global variables, but these are set by both main2.c and
 * exec.c, and referenced (a lot) by listprop.c, so I judge that using
 * accessing/setting functions for them is too expensive. Ick.
 */

#ifndef EXTERN
#define EXTERN extern
#endif

/*
 * This one is used ALL OVER, to call the routines in the MUD library.
 */

EXTERN MUDState_t MS;

EXTERN ULONG_T RealPlayer;		/* key of current player Player_t */
EXTERN ULONG_T EffectivePlayer; 	/* as set by running a proc */
EXTERN PlayerStatus_t RealStatus;	/* need to know wizard/apprentice */
EXTERN PlayerStatus_t EffectiveStatus;

EXTERN ULONG_T ThingsCreated, ThingsDestroyed, RequestsProcessed, MessagesSent;

EXTERN short int BinaryPort;

EXTERN ULONG_T LastRefKey;
EXTERN void *LastRefPtr;

/* These are shared by exec.c and bcRun.c */

EXTERN QValue_t *RunLocals;
EXTERN TraceBack_t *RunTraceBack;
EXTERN UINT_T RunCallDepth;
EXTERN BOOL_T RunExiting, RunAborting, BcAbort;

/* This is for verbose output to help beginning admins */

EXTERN BOOL_T Verbose;

/* use this to turn on any debugging output selectively */

EXTERN BOOL_T WantDebug;

/* these control the activity of machines */

EXTERN BOOL_T MachinesActive, NoMachines;

/* this is for the magic TCP_NODELAY setsockopt call */

EXTERN BOOL_T ForceNoDelay;

#if TRACING
#define TRACE(str, n)	trace(str, (ULONG_T) (n))
#else
#define TRACE(str, n)
#endif
