#include <ctype.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * builtin.c - code to set up all of the builtin routines. Also some direct
 *	builtins.
 */

#define BUILTIN_COUNT	400	/* make DAMN sure this is large enough! */
#define FIRST_BUILTIN	INDEX_MASK

/* functions defined in listprop.c */

extern ULONG_T CreateBoolProp(void);
extern ULONG_T CreateIntProp(void);
extern ULONG_T CreateFixedProp(void);
extern ULONG_T CreateStringProp(void);
extern ULONG_T CreateThingProp(void);
extern ULONG_T CreateActionProp(void);
extern ULONG_T CreateTableProp(void);
extern ULONG_T CreateGrammarProp(void);
extern ULONG_T CreateIntListProp(void);
extern ULONG_T CreateFixedListProp(void);
extern ULONG_T CreateThingListProp(void);
extern ULONG_T CreateActionListProp(void);
extern ULONG_T CreateThing(ULONG_T parent);
extern void SetParent(ULONG_T key, ULONG_T newParent);
extern ULONG_T Parent(ULONG_T th);
extern ULONG_T IsAncestor(ULONG_T myKey, ULONG_T parentKey);
extern ULONG_T SetThingStatus(ULONG_T key, ULONG_T status);
extern ULONG_T GetThingStatus(ULONG_T key);
extern ULONG_T Mine(ULONG_T key);
extern ULONG_T Owner(ULONG_T key);
extern ULONG_T CreateIntList(void);
extern ULONG_T CreateFixedList(void);
extern ULONG_T CreateThingList(void);
extern ULONG_T CreateActionList(void);
extern ULONG_T CreateIntArray(ULONG_T size);
extern ULONG_T Count(ULONG_T list);
extern void AddHead(ULONG_T list, ULONG_T valu);
extern void RemHead(ULONG_T list);
extern void RemTail(ULONG_T list);
extern LONG_T FindElement(ULONG_T list, ULONG_T valu);

extern void CheckList(ULONG_T list, CHAR_T *where);

/* functions defined in main.c */

extern void SetContinue(ULONG_T state);
extern ULONG_T ShutDown(ULONG_T flag);
extern ULONG_T RunLimit(ULONG_T newValue);

/* functions defined in clients.c */

extern ULONG_T It(void);
extern ULONG_T PublicTable(void);
extern ULONG_T PrivateTable(void);
extern ULONG_T ClientsActive(void);
extern ULONG_T UseTable(ULONG_T key);
extern ULONG_T UnUseTable(ULONG_T key);
extern void EditString(CHAR_T *string, ULONG_T handler, ULONG_T mode,
		       CHAR_T *prompt);
extern void GetString(CHAR_T *initial, ULONG_T handler, CHAR_T *prompt);
extern void QueryFile(CHAR_T *path, ULONG_T handler);

/* functions defined in messages.c */

extern LONG_T TextWidth(LONG_T newWidth);
extern LONG_T TextHeight(LONG_T newHeight);
extern void NPrint(CHAR_T *message);
extern void SetPrefix(CHAR_T *newPrefix);
extern void APrint(CHAR_T *message);
extern void OPrint(CHAR_T *message);
extern void SPrint(ULONG_T agent, CHAR_T *st);
extern void ABPrint(ULONG_T location,ULONG_T agent1,ULONG_T agent2,CHAR_T *st);

/* functions in clientFuncs.c */

extern ULONG_T FindAgentAt(ULONG_T location, CHAR_T *name);
extern ULONG_T FindAgent(CHAR_T *name);
extern void DescribeKey(ULONG_T key);
extern void SetAgentLocation(ULONG_T agent, ULONG_T where);
extern void SetSingleUser(ULONG_T state);
extern void SetRemoteSysAdminOK(ULONG_T state);
extern void SetMachinesActive(ULONG_T state);
extern void ForEachAgent(ULONG_T location, ULONG_T action);
extern void ForEachClient(ULONG_T action);
extern void Say(CHAR_T *altName, CHAR_T *what);
extern void SayTo(ULONG_T who, CHAR_T *altName, CHAR_T *what);
extern void Pose(CHAR_T *altName, CHAR_T *what);
extern void PoseTo(ULONG_T who, CHAR_T *altName, CHAR_T *what);
extern ULONG_T Whisper(CHAR_T *altName, CHAR_T *what, ULONG_T who,
		       LONG_T percent);
extern void After(LONG_T seconds, ULONG_T action);
extern void Quit(void);
extern void NewCreationPassword(void);
extern void Normal(void);
extern ULONG_T WizardMode(void);
extern void BootClient(ULONG_T key);
extern void NukeClient(ULONG_T key);
extern ULONG_T AgentLocation(ULONG_T agent);
extern ULONG_T DeleteSymbol(ULONG_T key, CHAR_T *name);
extern ULONG_T MoveSymbol(ULONG_T fromKey, ULONG_T toKey, CHAR_T *name);
extern ULONG_T RenameSymbol(ULONG_T table, CHAR_T *oldName, CHAR_T *newName);
extern ULONG_T IsDefined(ULONG_T table, CHAR_T *name);
extern ULONG_T IsProgrammer(void);
extern ULONG_T IsNormal(void);
extern ULONG_T IsApprentice(void);
extern ULONG_T IsWizard(void);
extern ULONG_T ForceAction(ULONG_T agent, ULONG_T action);
extern ULONG_T CanEdit(void);
extern ULONG_T Editing(void);
extern ULONG_T GettingString(void);
extern ULONG_T PrintNoAts(ULONG_T flag);
extern void FindKey(ULONG_T key);
extern void DumpThing(ULONG_T key);

/* functions defined in machines.c */

extern ULONG_T FindMachineIndexed(CHAR_T *name, ULONG_T index);
extern void CreateMachine(CHAR_T *name, ULONG_T key, ULONG_T where,
			  ULONG_T startup);
extern ULONG_T SetMachineSay(ULONG_T machine, ULONG_T action);
extern ULONG_T SetMachineWhisperMe(ULONG_T machine, ULONG_T action);
extern ULONG_T SetMachineWhisperOther(ULONG_T machine, ULONG_T action);
extern ULONG_T SetMachinePose(ULONG_T machine, ULONG_T action);
extern ULONG_T SetMachineOther(ULONG_T machine, ULONG_T action);
extern ULONG_T SetMachineIdle(ULONG_T machine, ULONG_T action);
extern ULONG_T SetMachineActive(ULONG_T machine, ULONG_T action);
extern void DestroyMachine(ULONG_T key);

/* functions defined in characters.c */

extern ULONG_T Character(CHAR_T *name);
extern ULONG_T CharacterThing(ULONG_T plKey);
extern ULONG_T CharacterTable(ULONG_T plKey);
extern ULONG_T ThingCharacter(ULONG_T thKey);
extern ULONG_T CharacterLocation(ULONG_T plKey);
extern void SetCharacterLocation(ULONG_T plKey, ULONG_T locKey);
extern ULONG_T Me(void);
extern ULONG_T TrueMe(void);
extern CHAR_T *GetMeString(void);
extern void SetMeString(CHAR_T *p);
extern ULONG_T MeCharacter(void);
extern void SetLocation(ULONG_T where);
extern ULONG_T Here(void);
extern ULONG_T SetNewCharacterAction(ULONG_T key);
extern ULONG_T SetCharacterInputAction(ULONG_T key);
extern ULONG_T SetCharacterRawKeyAction(ULONG_T key);
extern ULONG_T SetCharacterMouseDownAction(ULONG_T key);
extern ULONG_T SetCharacterButtonAction(ULONG_T key);
extern ULONG_T SetCharacterIdleAction(ULONG_T key);
extern ULONG_T SetCharacterActiveAction(ULONG_T key);
extern ULONG_T SetCharacterEffectDoneAction(ULONG_T key);
extern void CreateCharacter(CHAR_T *name, CHAR_T *password);
extern void DestroyCharacter(CHAR_T *name);
extern CHAR_T *SetPrompt(CHAR_T *newPrompt);
extern void NewCharacterPassword(void);
extern void ChangeName(CHAR_T *newName);

/* functions defined in miscFuncs.c */

extern void GiveThing(ULONG_T thingKey, ULONG_T player);
extern void ShowTable(ULONG_T key);
extern void DescribeSymbol(ULONG_T key, CHAR_T *name);
extern void Flush(void);
extern ULONG_T ShowClients(ULONG_T longForm);
extern ULONG_T ShowCharacters(ULONG_T longForm);
extern void ShowCharacter(ULONG_T key);
extern ULONG_T CreateTable(void);
extern ULONG_T DefineThing(ULONG_T table, CHAR_T *name, ULONG_T key);
extern ULONG_T DefineAction(ULONG_T table, CHAR_T *name, ULONG_T key);
extern ULONG_T DefineFlag(ULONG_T table, CHAR_T *name, ULONG_T key);
extern ULONG_T DefineCounter(ULONG_T table, CHAR_T *name, ULONG_T key);
extern ULONG_T DefineString(ULONG_T table, CHAR_T *name, ULONG_T key);
extern ULONG_T DefineTable(ULONG_T table, CHAR_T *name, ULONG_T key);
extern ULONG_T LookupThing(ULONG_T table, CHAR_T *name);
extern ULONG_T LookupAction(ULONG_T table, CHAR_T *name);
extern ULONG_T LookupFlag(ULONG_T table, CHAR_T *name);
extern ULONG_T LookupCounter(ULONG_T table, CHAR_T *name);
extern ULONG_T LookupString(ULONG_T table, CHAR_T *name);
extern ULONG_T LookupTable(ULONG_T table, CHAR_T *name);
extern CHAR_T *FindXXXSymbol(ULONG_T table, ULONG_T key);
extern void PrintAction(ULONG_T key);
extern void ScanTable(ULONG_T table, ULONG_T func);
extern CHAR_T *SymbolType(ULONG_T table, CHAR_T *name);
extern void ZapAction(ULONG_T key);
extern void ZapTable(ULONG_T key);

/* functions defined in parse.c */

extern CHAR_T *Verb(void);
extern CHAR_T *ItName(void);
extern CHAR_T *WhoName(void);
extern CHAR_T *Punctuation(void);
extern ULONG_T Word(ULONG_T key, CHAR_T *wrd);
extern void Synonym(ULONG_T key, CHAR_T *oldWord, CHAR_T *newWord);
extern void VerbTail(ULONG_T key, CHAR_T *wrd, ULONG_T action);
extern void Verb0(ULONG_T key, CHAR_T *wrd, ULONG_T separator, ULONG_T action);
extern void Verb1(ULONG_T key, CHAR_T *wrd, ULONG_T separator, ULONG_T action);
extern void Verb2(ULONG_T key, CHAR_T *wrd, ULONG_T separator, ULONG_T action);
extern ULONG_T FindWord(ULONG_T key, CHAR_T *wrd);
extern ULONG_T FindAnyWord(ULONG_T key, CHAR_T *wrd);
extern void ShowWord(ULONG_T key, CHAR_T *wrd);
extern void ShowWords(ULONG_T key);
extern void DeleteWord(ULONG_T key, CHAR_T *wrd);
extern CHAR_T *GetNounPhrase(ULONG_T key, CHAR_T *st, ULONG_T separator);
extern ULONG_T CreateGrammar(void);
extern ULONG_T Parse(ULONG_T key, CHAR_T *sentence);

/* functions defined in match.c */

extern CHAR_T *GetTail(void);
extern CHAR_T *GetWord(void);
extern void SetTail(CHAR_T *p);
extern CHAR_T *SetSay(CHAR_T *what);
extern CHAR_T *SetWhisperMe(CHAR_T *what);
extern CHAR_T *SetWhisperOther(CHAR_T *what);
extern LONG_T HasAdjective(CHAR_T *name, CHAR_T *adjective);
extern CHAR_T *SelectName(CHAR_T *names, ULONG_T which);
extern CHAR_T *SelectWord(CHAR_T *names, ULONG_T which);
extern ULONG_T FindResult(void);
extern CHAR_T *Strip(CHAR_T *st);

/* functions defined in find.c */

extern ULONG_T FindFlagOnList(ULONG_T list, ULONG_T prop);
extern ULONG_T FindChildOnList(ULONG_T list, ULONG_T parent);
extern ULONG_T FindIntOnList(ULONG_T list, ULONG_T prop,
			     ULONG_T valu, ULONG_T compType);
extern ULONG_T FindAgentWithFlag(ULONG_T room, ULONG_T prop);
extern ULONG_T FindAgentAsChild(ULONG_T room, ULONG_T parent);
extern ULONG_T FindAgentAsDescendant(ULONG_T room, ULONG_T parent);
extern ULONG_T FindAgentWithFlagOnList(ULONG_T room, ULONG_T listProp,
				       ULONG_T flagProp);
extern ULONG_T FindAgentWithChildOnList(ULONG_T room, ULONG_T listProp,
					ULONG_T parent);
extern ULONG_T FindAgentWithNameOnList(ULONG_T room, ULONG_T listProp,
				       ULONG_T nameProp, CHAR_T *name);

/* functions in fileIO.c */

extern void ShowCacheStatus(ULONG_T key);
extern void FileDebugOn(void);
extern void FileDebugOff(void);
extern void FileConCheck(CHAR_T *where);
extern void FileTraceOn(void);
extern void FileTraceOff(void);
extern void FileSetPeek(ULONG_T key);
extern void FileSetWatch(ULONG_T key);
extern void FileSetRange(ULONG_T key);
extern void ShowDBStatus(void);

/* functions in util.c */

extern ULONG_T FileOpenForRead(CHAR_T *fileName);
extern ULONG_T FileOpenForWrite(CHAR_T *fileName);
extern ULONG_T FileOpenForUpdate(CHAR_T *fileName);
extern void FileClose(ULONG_T id);
extern CHAR_T *FileRead(ULONG_T id);
extern void FileWrite(ULONG_T id, CHAR_T *theLine);
extern ULONG_T FileReadBinary(ULONG_T id, ULONG_T key);
extern void FileWriteBinary(ULONG_T id, ULONG_T key);
extern void FileSeek(ULONG_T id, ULONG_T pos);
extern void myExecute(CHAR_T *command);
extern CHAR_T *Trim(CHAR_T *s);

/* functions in effects.c */

extern ULONG_T ClientVersion(ULONG_T who);
extern ULONG_T TextComponent(ULONG_T who, ULONG_T newComponent);
extern ULONG_T DrawComponent(ULONG_T who, ULONG_T newComponent);
extern void CreateContainer(ULONG_T encloser, ULONG_T newId,
			    ULONG_T isVertical, ULONG_T alignment);
extern void CreateComponent(ULONG_T encloser, ULONG_T newId,
			    ULONG_T width, ULONG_T height, ULONG_T kind);
extern void MakeFrame(void);
extern ULONG_T GetFontAscent(ULONG_T who);
extern ULONG_T GetFontDescent(ULONG_T who);
extern ULONG_T GetFontLeading(ULONG_T who);
extern ULONG_T GetFontHeight(ULONG_T who);
extern ULONG_T GetFontWidth(ULONG_T who);
extern void IfFound(ULONG_T who);
extern void Else(ULONG_T who);
extern void Fi(ULONG_T who);
extern void FailText(ULONG_T who, CHAR_T *str);
extern void AbortEffect(ULONG_T who, ULONG_T kind, ULONG_T id);
extern void CallEffect(ULONG_T who, ULONG_T key);
extern ULONG_T KnowsEffect(ULONG_T who, ULONG_T key);
extern void DefineEffect(ULONG_T who, ULONG_T key);
extern void EndEffect(void);
extern ULONG_T GOn(ULONG_T who);
extern CHAR_T *GType(ULONG_T who);
extern ULONG_T GRows(ULONG_T who);
extern ULONG_T GCols(ULONG_T who);
extern ULONG_T GColours(ULONG_T who);
extern ULONG_T GPalette(ULONG_T who);
extern void PlaceCursor(LONG_T x, LONG_T y);
extern void PlaceCursorPixels(ULONG_T x, ULONG_T y);
extern void RemoveCursor(void);
extern void SetCursorPen(ULONG_T pen);
extern void SetCursorPattern(ULONG_T newCursor);
extern void AddButton(LONG_T x, LONG_T y, ULONG_T n, CHAR_T *st);
extern void AddButtonPixels(LONG_T x, LONG_T y, ULONG_T n, CHAR_T *st);
extern void EraseButton(ULONG_T n);
extern void ClearButtons(void);
extern void AddRegion(LONG_T x, LONG_T y, ULONG_T width, ULONG_T height,
		      ULONG_T n);
extern void AddRegionPixels(LONG_T x, LONG_T y, ULONG_T width, ULONG_T height,
			    ULONG_T n);
extern void EraseRegion(ULONG_T n);
extern void ClearRegions(void);
extern void SetButtonPen(ULONG_T which, ULONG_T pen);
extern void GSetPen(ULONG_T who, ULONG_T pen);
extern void GSetColour(ULONG_T who, ULONG_T which, ULONG_T colour);
extern void GResetColours(ULONG_T who);
extern void GClear(ULONG_T who);
extern void GPixel(ULONG_T who);
extern void GPixelRow(ULONG_T who, ULONG_T pixels, LONG_T start, LONG_T count);
extern void GAMove(ULONG_T who, LONG_T x, LONG_T y);
extern void GAMovePixels(ULONG_T who, ULONG_T x, ULONG_T y);
extern void GRMove(ULONG_T who, LONG_T deltaX, LONG_T deltaY);
extern void GRMovePixels(ULONG_T who, LONG_T deltaXPixels,LONG_T deltaYPixels);
extern void GADraw(ULONG_T who, LONG_T x, LONG_T y);
extern void GADrawPixels(ULONG_T who, ULONG_T x, ULONG_T y);
extern void GRDraw(ULONG_T who, LONG_T deltaX, LONG_T deltaY);
extern void GRDrawPixels(ULONG_T who, LONG_T deltaXPixels,LONG_T deltaYPixels);
extern void GRectangle(ULONG_T who, LONG_T width, LONG_T height, ULONG_T fill);
extern void GRectanglePixels(ULONG_T who, ULONG_T width, ULONG_T height,
			     ULONG_T fill);
extern void GCirclePixels(ULONG_T who, LONG_T r, ULONG_T fill);
extern void GCircle(ULONG_T who, LONG_T r, ULONG_T fill);
extern void GEllipsePixels(ULONG_T who, LONG_T a, LONG_T b, ULONG_T fill);
extern void GEllipse(ULONG_T who, LONG_T a, LONG_T b, ULONG_T fill);
extern void GPolygonStart(ULONG_T who);
extern void GPolygonEnd(ULONG_T who);
extern void GDefineTile(ULONG_T who, ULONG_T id,
			ULONG_T width, ULONG_T height, ULONG_T data);
extern void GDefineOverlayTile(ULONG_T who, ULONG_T id,
			       ULONG_T transparentColour,
			       ULONG_T width, ULONG_T height, ULONG_T data);
extern void GDisplayTile(ULONG_T who, ULONG_T id);
extern void GSetTextColour(ULONG_T who, ULONG_T which, ULONG_T colour);
extern void GText(ULONG_T who, CHAR_T *str);
extern void GLoadBackGround(ULONG_T who, CHAR_T *name);
extern void GSetImage(ULONG_T who, CHAR_T *name);
extern void GShowImage(ULONG_T who, CHAR_T *name, LONG_T imageX, LONG_T imageY,
		       LONG_T imageWidth, LONG_T imageHeight,
		       LONG_T displayX, LONG_T displayY,
		       LONG_T displayWidth, LONG_T displayHeight);
extern void GShowImagePixels(ULONG_T who, CHAR_T *name,
			     ULONG_T imageX, ULONG_T imageY,
			     ULONG_T imageWidth, ULONG_T imageHeight,
			     ULONG_T displayX, ULONG_T displayY);
extern void GShowBrush(ULONG_T who,CHAR_T *name,LONG_T screenX,LONG_T screenY);
extern void GScrollRectanglePixels(ULONG_T who,
			     LONG_T xDeltaPixels, LONG_T yDeltaPixels,
			     LONG_T xPixels, LONG_T yPixels,
			     LONG_T widthPixels, LONG_T heightPixels);
extern void GScrollRectangle(ULONG_T who,
			     LONG_T xDeltaPixels, LONG_T yDeltaPixels,
			     LONG_T x, LONG_T y,
			     LONG_T width, LONG_T height);
extern void GSetIconPen(ULONG_T theCharacter, ULONG_T colour);
extern void GNewIcon(ULONG_T theCharacter, ULONG_T newIcon);
extern void GShowIcon(ULONG_T who, ULONG_T whoseIcon,
		      ULONG_T isMonster, ULONG_T isTemporary);
extern void GRemoveIcon(ULONG_T who, ULONG_T whoseIcon);
extern void GDeleteIcon(ULONG_T who, ULONG_T whoseIcon);
extern void GResetIcons(ULONG_T who);
#if REDRAW_ICONS
extern void GRedrawIcons(ULONG_T who);
extern void GUndrawIcons(ULONG_T who);
#endif

extern ULONG_T SOn(ULONG_T who);
extern void SVolume(ULONG_T who, ULONG_T volume);
extern void SPlaySound(ULONG_T who, CHAR_T *name, ULONG_T cycles, ULONG_T id);

extern ULONG_T VOn(ULONG_T who);
extern void VVolume(ULONG_T who, ULONG_T volume);
extern void VParams(ULONG_T who, ULONG_T rate, ULONG_T pitch,
		    ULONG_T mode, ULONG_T sex);
extern void VReset(ULONG_T who);
extern void VSpeak(ULONG_T who, CHAR_T *words, ULONG_T id);
extern CHAR_T *VTranslate(CHAR_T *words);
extern void VNarrate(ULONG_T who, CHAR_T *phonemes, ULONG_T id);

extern ULONG_T MOn(ULONG_T who);
extern void MVolume(ULONG_T who, ULONG_T volume);
extern void MPlaySong(ULONG_T who, CHAR_T *name, ULONG_T id);

/* functions defined in bcComp.d */

extern void UnCompile(ULONG_T key);

/* functions defined in bcDis.d */

extern void Disassemble(ULONG_T key);

/* define some short-hand forms */

#define T_E	TYPE_ERROR
#define T_V	TYPE_VOID
#define T_ST	TYPE_STATUS
#define T_P	TYPE_PLAYER
#define T_B	TYPE_BOOLEAN
#define T_I	TYPE_INTEGER
#define T_F	TYPE_FIXED
#define T_T	TYPE_THING
#define T_A	TYPE_ACTION
#define T_S	TYPE_STRING
#define T_TB	TYPE_TABLE
#define T_G	TYPE_GRAMMAR
#define T_L_I	TYPE_LIST_INTEGER
#define T_L_F	TYPE_LIST_FIXED
#define T_L_T	TYPE_LIST_THING
#define T_L_A	TYPE_LIST_ACTION
#define T_P_B	TYPE_PROP_BOOLEAN
#define T_P_I	TYPE_PROP_INTEGER
#define T_P_F	TYPE_PROP_FIXED
#define T_P_T	TYPE_PROP_THING
#define T_P_A	TYPE_PROP_ACTION
#define T_P_S	TYPE_PROP_STRING
#define T_P_TB	TYPE_PROP_TABLE
#define T_P_G	TYPE_PROP_GRAMMAR
#define T_P_L_I TYPE_PROP_LIST_INTEGER
#define T_P_L_F TYPE_PROP_LIST_FIXED
#define T_P_L_T TYPE_PROP_LIST_THING
#define T_P_L_A TYPE_PROP_LIST_ACTION
#define T_LANY	TYPE_LIST_ANY
#define T_ELEM	TYPE_ELEMENT

static ULONG_T NextBuiltinKey;
static ULONG_T Seed;
static Proc_t *LatestProc;

/* table used for direct lookup based on key */

static Proc_t *(BuiltInTable[BUILTIN_COUNT]);

/*
 * builtInTerm - called during shutdown - used to remove all builtins from
 *	the builtin table.
 */

void
builtInTerm(void)
{
    ULONG_T key;

    /* We go to the effort of freeing the proc headers for the builtins so
       that we have freed up the memory chunks they occupy. This allows the
       writing of any large symbol tables (like the public one) to expand the
       byte buffer if needed, with less liklihood that allocation is needed
       to do so. */
    key = NextBuiltinKey;
    while (key != FIRST_BUILTIN) {
	key += 1;
	subBuiltinProc(key | NIL_ACTION);
	freeProcE(&MS, BuiltInTable[FIRST_BUILTIN - (key & INDEX_MASK)], key,
		  BOOLEAN_FALSE);
    }
}

/*
 * routines used to actually set up the builtin routines
 */

static void
bi10(CHAR_T *name, Comp_t body, UINT_T resultType,
     CHAR_T *par1Name, UINT_T par1Type,
     CHAR_T *par2Name, UINT_T par2Type,
     CHAR_T *par3Name, UINT_T par3Type,
     CHAR_T *par4Name, UINT_T par4Type,
     CHAR_T *par5Name, UINT_T par5Type,
     CHAR_T *par6Name, UINT_T par6Type,
     CHAR_T *par7Name, UINT_T par7Type,
     CHAR_T *par8Name, UINT_T par8Type,
     CHAR_T *par9Name, UINT_T par9Type,
     CHAR_T *par10Name, UINT_T par10Type)
{
    Proc_t *pr;
    LocalList_t *ll;
    LocalList_t **pll;
    UINT_T localCount;

    if (NextBuiltinKey == FIRST_BUILTIN - BUILTIN_COUNT) {
	myAbort("too many builtins!");
    }
    pr = MEM_ALLOC_E(&MS, sizeof(Proc_t), "bipr");
    LatestProc = pr;
    pr->pr_locals = P_NULL;
    pr->pr_refs = P_NULL;
    pr->pr_body = P_NULL;
    pr->pr_compiled = body;
    pr->pr_byteCode = P_NULL;
    pr->pr_resultType = resultType;
    pr->pr_owner = SYSADMIN_KEY;
    pr->pr_useCount = 1;
    pr->pr_byteCodeLen = 0;
    pr->pr_utility = B_TRUE;
    pr->pr_wizard = B_FALSE;
    pr->pr_public = B_FALSE;
    pr->pr_isCompiled = B_FALSE;
    pr->pr_stripped = B_FALSE;
    pr->pr_locked = B_FALSE;
    pr->pr_byteSwapped = B_FALSE;
    pr->pr_ownerStatus = ps_wizard;	/* not referenced, however */
    pll = &pr->pr_parameters;
    localCount = 0;
    if (par1Name != P_NULL) {
     localCount += sizeof(QValue_t);
     ll = MEM_ALLOC_E(&MS, sizeof(LocalList_t), "par1");
     ll->ll_name = stringCopy(par1Name);
     ll->ll_type = par1Type;
     *pll = ll;
     pll = &ll->ll_next;
     if (par2Name != P_NULL) {
      localCount += sizeof(QValue_t);
      ll = MEM_ALLOC_E(&MS, sizeof(LocalList_t), "par2");
      ll->ll_name = stringCopy(par2Name);
      ll->ll_type = par2Type;
      *pll = ll;
      pll = &ll->ll_next;
      if (par3Name != P_NULL) {
       localCount += sizeof(QValue_t);
       ll = MEM_ALLOC_E(&MS, sizeof(LocalList_t), "par3");
       ll->ll_name = stringCopy(par3Name);
       ll->ll_type = par3Type;
       *pll = ll;
       pll = &ll->ll_next;
       if (par4Name != P_NULL) {
	localCount += sizeof(QValue_t);
	ll = MEM_ALLOC_E(&MS, sizeof(LocalList_t), "par4");
	ll->ll_name = stringCopy(par4Name);
	ll->ll_type = par4Type;
	*pll = ll;
	pll = &ll->ll_next;
	if (par5Name != P_NULL) {
	 localCount += sizeof(QValue_t);
	 ll = MEM_ALLOC_E(&MS, sizeof(LocalList_t), "par5");
	 ll->ll_name = stringCopy(par5Name);
	 ll->ll_type = par5Type;
	 *pll = ll;
	 pll = &ll->ll_next;
	 if (par6Name != P_NULL) {
	  localCount += sizeof(QValue_t);
	  ll = MEM_ALLOC_E(&MS, sizeof(LocalList_t), "par6");
	  ll->ll_name = stringCopy(par6Name);
	  ll->ll_type = par6Type;
	  *pll = ll;
	  pll = &ll->ll_next;
	  if (par7Name != P_NULL) {
	   localCount += sizeof(QValue_t);
	   ll = MEM_ALLOC_E(&MS, sizeof(LocalList_t), "par7");
	   ll->ll_name = stringCopy(par7Name);
	   ll->ll_type = par7Type;
	   *pll = ll;
	   pll = &ll->ll_next;
	   if (par8Name != P_NULL) {
	    localCount += sizeof(QValue_t);
	    ll = MEM_ALLOC_E(&MS,sizeof(LocalList_t), "par8");
	    ll->ll_name = stringCopy(par8Name);
	    ll->ll_type = par8Type;
	    *pll = ll;
	    pll = &ll->ll_next;
	    if (par9Name != P_NULL) {
	     localCount += sizeof(QValue_t);
	     ll = MEM_ALLOC_E(&MS,sizeof(LocalList_t), "par9");
	     ll->ll_name = stringCopy(par9Name);
	     ll->ll_type = par9Type;
	     *pll = ll;
	     pll = &ll->ll_next;
	     if (par10Name != P_NULL) {
	      localCount += sizeof(QValue_t);
	      ll = MEM_ALLOC_E(&MS,sizeof(LocalList_t), "par10");
	      ll->ll_name = stringCopy(par10Name);
	      ll->ll_type = par10Type;
	      *pll = ll;
	      pll = &ll->ll_next;
	     }
	    }
	   }
	  }
	 }
	}
       }
      }
     }
    }
    *pll = P_NULL;
    pr->pr_localCount = localCount;
    addBuiltinProc(name, NextBuiltinKey | NIL_ACTION);
    BuiltInTable[FIRST_BUILTIN - NextBuiltinKey] = pr;
    NextBuiltinKey -= 1;
}

/*
 * findBuiltIn - used during proc setup to retrieve a pointer to the Proc_t
 */

Proc_t *
findBuiltIn(ULONG_T key)
{

    return BuiltInTable[FIRST_BUILTIN - (key & INDEX_MASK)];
}

static void
bi0(CHAR_T *name, Comp_t body, UINT_T resultType)
{

    bi10(name, body, resultType, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E);
}

static void
bi1(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type)
{

    bi10(name, body, resultType, par1Name, par1Type, P_NULL, T_E, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E);
}

static void
bi2(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E);
}

static void
bi3(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type,
    CHAR_T *par3Name, UINT_T par3Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	 par3Name, par3Type, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E);
}

static void
bi4(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type,
    CHAR_T *par3Name, UINT_T par3Type,
    CHAR_T *par4Name, UINT_T par4Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	 par3Name, par3Type, par4Name, par4Type, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E);
}

static void
bi5(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type,
    CHAR_T *par3Name, UINT_T par3Type,
    CHAR_T *par4Name, UINT_T par4Type,
    CHAR_T *par5Name, UINT_T par5Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	 par3Name, par3Type, par4Name, par4Type, par5Name, par5Type,
	 P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E, P_NULL, T_E);
}

static void
bi6(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type,
    CHAR_T *par3Name, UINT_T par3Type,
    CHAR_T *par4Name, UINT_T par4Type,
    CHAR_T *par5Name, UINT_T par5Type,
    CHAR_T *par6Name, UINT_T par6Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	 par3Name, par3Type, par4Name, par4Type, par5Name, par5Type,
	 par6Name, par6Type, P_NULL, T_E, P_NULL, T_E,
	 P_NULL, T_E, P_NULL, T_E);
}

static void
bi7(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type,
    CHAR_T *par3Name, UINT_T par3Type,
    CHAR_T *par4Name, UINT_T par4Type,
    CHAR_T *par5Name, UINT_T par5Type,
    CHAR_T *par6Name, UINT_T par6Type,
    CHAR_T *par7Name, UINT_T par7Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	par3Name, par3Type, par4Name, par4Type, par5Name, par5Type,
	par6Name, par6Type, par7Name, par7Type, P_NULL, T_E,
	P_NULL, T_E, P_NULL, T_E);
}

static void
bi8(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type,
    CHAR_T *par3Name, UINT_T par3Type,
    CHAR_T *par4Name, UINT_T par4Type,
    CHAR_T *par5Name, UINT_T par5Type,
    CHAR_T *par6Name, UINT_T par6Type,
    CHAR_T *par7Name, UINT_T par7Type,
    CHAR_T *par8Name, UINT_T par8Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	 par3Name, par3Type, par4Name, par4Type, par5Name, par5Type,
	 par6Name, par6Type, par7Name, par7Type, par8Name, par8Type,
	 P_NULL, T_E, P_NULL, T_E);
}

#if 0
static void
bi9(CHAR_T *name, Comp_t body, UINT_T resultType,
    CHAR_T *par1Name, UINT_T par1Type,
    CHAR_T *par2Name, UINT_T par2Type,
    CHAR_T *par3Name, UINT_T par3Type,
    CHAR_T *par4Name, UINT_T par4Type,
    CHAR_T *par5Name, UINT_T par5Type,
    CHAR_T *par6Name, UINT_T par6Type,
    CHAR_T *par7Name, UINT_T par7Type,
    CHAR_T *par8Name, UINT_T par8Type,
    CHAR_T *par9Name, UINT_T par9Type)
{

    bi10(name, body, resultType, par1Name, par1Type, par2Name, par2Type,
	 par3Name, par3Type, par4Name, par4Type, par5Name, par5Type,
	 par6Name, par6Type, par7Name, par7Type, par8Name, par8Type,
	 par9Name, par9Type, P_NULL, T_E);
}
#endif

/*
 * ServerVersion - the version number compiled into this server.
 */
static ULONG_T
ServerVersion(void)
{

    return (VERSION1 - '0') * 10 + (VERSION2 - '0');
}

/*
 * Note - the obvious copyright string.
 */

static void
Note(void)
{
    static CHAR_T string[] = {
	('A'-'\0'-0xc7)&0xff,('m'-'\0'-0xc7)&0xff,
	('i'-'\0'-0xc7)&0xff,('g'-'\0'-0xc7)&0xff,
	('a'-'\0'-0xc7)&0xff,('M'-'\0'-0xc7)&0xff,
	('U'-'\0'-0xc7)&0xff,('D'-'\0'-0xc7)&0xff,
	(' '-'\0'-0xc7)&0xff,('V'-'\0'-0xc7)&0xff,
	(VERSION1-'\0'-0xc7)&0xff,('.'-'\0'-0xc7)&0xff,
	(VERSION2-'\0'-0xc7)&0xff,(' '-'\0'-0xc7)&0xff,
	('C'-'\0'-0xc7)&0xff,('o'-'\0'-0xc7)&0xff,
	('p'-'\0'-0xc7)&0xff,('y'-'\0'-0xc7)&0xff,
	('r'-'\0'-0xc7)&0xff,('i'-'\0'-0xc7)&0xff,
	('g'-'\0'-0xc7)&0xff,('h'-'\0'-0xc7)&0xff,
	('t'-'\0'-0xc7)&0xff,(' '-'\0'-0xc7)&0xff,
	('1'-'\0'-0xc7)&0xff,('9'-'\0'-0xc7)&0xff,
	(YEAR1-'\0'-0xc7)&0xff,(YEAR2-'\0'-0xc7)&0xff,
	(' '-'\0'-0xc7)&0xff,('b'-'\0'-0xc7)&0xff,
	('y'-'\0'-0xc7)&0xff,(' '-'\0'-0xc7)&0xff,
	('C'-'\0'-0xc7)&0xff,('h'-'\0'-0xc7)&0xff,
	('r'-'\0'-0xc7)&0xff,('i'-'\0'-0xc7)&0xff,
	('s'-'\0'-0xc7)&0xff,(' '-'\0'-0xc7)&0xff,
	('G'-'\0'-0xc7)&0xff,('r'-'\0'-0xc7)&0xff,
	('a'-'\0'-0xc7)&0xff,('y'-'\0'-0xc7)&0xff,
	('\n'-'\0'-0xc7)&0xff
    };
    CHAR_T *p;

    p = &string[0];
    while (*p != '\0') {
	putChar(*p + 0xc7);
	++p;
    }
}

static BOOL_T
checkVowel(CHAR_T *p)
{
    CHAR_T ch, ch2;
    BOOL_T vowel;

    ch = *p;
    ch2 = *(p + 1);
    vowel = B_FALSE;
    if (isupper(ch)) {
	ch = tolower(ch);
    }
    if (isupper(ch2)) {
	ch2 = tolower(ch2);
    }
    if (ch == 'a' || ch == 'e' || ch == 'i' || ch == 'o' ||
	(ch == 'u' && (ch2 != 'n' && ch2 != 'k' && ch2 != 's')))
    {
	vowel = B_TRUE;
    }
    return vowel;
}

/*
 * AAn -
 * AAn("You do not have", "apple") => "You do not have an apple"
 * AAn("You do not have", "tree")  => "You do not have a tree"
 * AAn("", "whatzit") => "a whatzit"
 */

static CHAR_T *
AAn(CHAR_T *s1, CHAR_T *s2)
{
    CHAR_T *p, *q;
    ULONG_T len;
    BOOL_T emptyFirst, vowel;

    p = s1;
    len = 1;
    if (*p == '\0') {
	emptyFirst = B_TRUE;
    } else {
	emptyFirst = B_FALSE;
	len = len + 1;
    }
    while (*p != '\0') {
	++p;
	len += 1;
    }
    p = s2;
    vowel = checkVowel(p);
    while (*p != '\0') {
	++p;
	len += 1;
    }
    if (vowel) {
	len += 3;
	p = MEM_ALLOC_E(&MS, len, "aan1");
	q = s1;
	s1 = p;
	while (*q != '\0') {
	    *p++ = *q++;
	}
	q = " an ";
    } else {
	len += 2;
	p = MEM_ALLOC_E(&MS, len, "aan2");
	q = s1;
	s1 = p;
	while (*q != '\0') {
	    *p++ = *q++;
	}
	q = " a ";
    }
    if (emptyFirst) {
	/* no first string - skip the space */
	++q;
    }
    while (*q != '\0') {
	*p++ = *q++;
    }
    q = s2;
    while (*q != '\0') {
	*p++ = *q++;
    }
    *p = '\0';
    return s1;
}

/*
 * IsAre -
 * IsAre("There", "no", "pie", "here.")  => "There is no pie here."
 * IsAre("There", "no", "pies", "here.") => "There are no pies here."
 * IsAre("There", "", "pie", "here.")	 => "There is a pie here."
 * IsAre("There", "", "apple", "here.")  => "There is an apple here."
 * IsAre("There", "", "pies", "here.")	 => "There are some pies here."
 */

static CHAR_T *
IsAre(CHAR_T *s1, CHAR_T *s2, CHAR_T *s3, CHAR_T *s4)
{
    CHAR_T *p, *q;
    ULONG_T len;
    CHAR_T ch;
    BOOL_T vowel;

    /* fixed needs:
	\0 at end;
	space between s3 and s4;
	space between s1 and either s2 or the insert;
	space between the insert and s3.
    */
    len = 4;
    p = s1;
    while (*p != '\0') {
	++p;
	len += 1;
    }
    p = s4;
    while (*p != '\0') {
	++p;
	len += 1;
    }
    p = s3;
    vowel = checkVowel(p);
    while (*p != '\0') {
	++p;
	len += 1;
    }
    ch = *--p;
    if (ch == 'S') {
	ch = 's';
    }
    if (ch == 's') {
	--p;
	if (*p == 's' || *p == 'S') {
	    /* ends in 'ss' - not plural */
	    ch = ' ';
	}
    }
    if (ch == 's') {
	/* plural, use 'are_' */
	len += 4;
    } else {
	/* singular, use 'is_' */
	len += 3;
    }
    p = s2;
    if (*p != '\0') {
	while (*p != '\0') {
	    ++p;
	    len += 1;
	}
    } else {
	if (ch == 's') {
	    /* use 'some' */
	    len = len + 4;
	} else {
	    if (vowel) {
		/* use 'an' */
		len = len + 2;
	    } else {
		/* use 'a' */
		len = len + 1;
	    }
	}
    }
    p = MEM_ALLOC_E(&MS, len, "isar");
    q = s1;
    s1 = p;
    while (*q != '\0') {
	*p++ = *q++;
    }
    if (ch == 's') {
	q = " are ";
    } else {
	q = " is ";
    }
    while (*q != '\0') {
	*p++ = *q++;
    }
    q = s2;
    if (*q != '\0') {
	while (*q != '\0') {
	    *p++ = *q++;
	}
    } else {
	if (ch == 's') {
	    q = "some";
	} else {
	    if (vowel) {
		q = "an";
	    } else {
		q = "a";
	    }
	}
	while (*q != '\0') {
	    *p++ = *q++;
	}
    }
    *p++ = ' ';
    q = s3;
    while (*q != '\0') {
	*p++ = *q++;
    }
    *p++ = ' ';
    q = s4;
    while (*q != '\0') {
	*p++ = *q++;
    }
    *p = '\0';
    return s1;
}

ULONG_T
Random(ULONG_T rang)
{
#define A	16807
#define M	0x7fffffff
#define Q	127773		/* M div A */
#define R	2836		/* M mod A */
    LONG_T lo, hi, test;

    if (rang == 0) {
	return 0;
    } else {
	hi = Seed / Q;
	lo = Seed % Q;
	test = A * lo - R * hi;
	if (test > 0) {
	    Seed = test;
	} else {
	    Seed = test + M;
	}
	return (Seed - 1) % rang;
    }
#undef Q
#undef M
#undef Q
#undef R
}

static void
SetSeed(ULONG_T valu)
{

    Seed = (valu & 0x7fffffff) | 1;
}

ULONG_T
GetSeed(void)
{

    return Seed;
}

CHAR_T *
SubString(CHAR_T *st, LONG_T start, LONG_T length)
{
    CHAR_T *p, *q, *newSt;

    if (start < 0 || length < 0) {
	runError("invalid start/length to SubString");
	return nullString();
    } else {
	while (start > 0 && *st != '\0') {
	    start -= 1;
	    ++st;
	}
	p = st;
	while (length > 0 && *p != '\0') {
	    length -= 1;
	    ++p;
	}
	newSt = MEM_ALLOC_E(&MS, p - st + 1, "sbst");
	q = newSt;
	while (st != p) {
	    *q++ = *st++;
	}
	*q = '\0';
	return newSt;
    }
}

static LONG_T
Length(CHAR_T *st)
{
    CHAR_T *p;

    p = st;
    while (*p != '\0') {
	++p;
    }
    return p - st;
}

static CHAR_T *
Capitalize(CHAR_T *st)
{

    st = stringCopy(st);
    if (islower(*st)) {
	*st = toupper(*st);
    }
    return st;
}

static CHAR_T *
Pluralize(CHAR_T *st)
{
    CHAR_T *p, *q, *result;
    CHAR_T ch;
    UINT_T extra;

    if (*st == '\0') {
	return stringCopy(st);
    } else {
	p = st;
	while (*p != '\0') {
	    ++p;
	}
	ch = *--p;
	if (ch == 'y' && p != st) {
	    ch = *--p;
	    if (ch == 'a' || ch == 'e' || ch == 'i' || ch == 'o' || ch == 'u'){
		++p;
		extra = 3;
	    } else {
		extra = 5;
	    }
	} else if (ch == 'x' || ch == 's' || ch == 'i' || ch == 'o') {
	    extra = 4;
	} else {
	    extra = 3;
	}
	q = MEM_ALLOC_E(&MS, p - st + extra, "plrl");
	result = q;
	while (B_TRUE) {
	    *q++ = *st;
	    if (st == p) {
		break;
	    }
	    ++st;
	}
	if (extra == 5) {
	    *q++ = 'i';
	}
	if (extra >= 4) {
	    *q++ = 'e';
	}
	*q++ = 's';
	*q = '\0';
	return result;
    }
}

static LONG_T
Index(CHAR_T *s1, CHAR_T *s2)
{
    CHAR_T *p, *q;
    CHAR_T ch1, ch2;
    LONG_T n;

    n = 0;
    while (*s1 != '\0') {
	p = s1;
	q = s2;
	do {
	    ch1 = *p++;
	    ch2 = *q++;
	} while (ch1 == ch2 && ch2 != '\0');
	if (ch2 == '\0') {
	    return n;
	}
	++s1;
	n += 1;
    }
    return -1;
}

static CHAR_T *
StringReplace(CHAR_T *s1, LONG_T pos, CHAR_T *s2)
{
    CHAR_T *s, *p;

    s = stringCopy(s1);
    if (pos >= 0 && Length(s1) >= pos + Length(s2)) {
	p = s;
	while (pos != 0) {
	    pos -= 1;
	    ++p;
	}
	while (*s2 != '\0') {
	    *p++ = *s2++;
	}
    } else {
	runError("invalid parameters to StringReplace");
    }
    return s;
}

static void
SetEffectiveToNone(void)
{

    EffectivePlayer = NIL_PLAYER;
    EffectiveStatus = ps_normal;
}

static void
SetEffectiveToReal(void)
{

    if (EffectivePlayer != SYSADMIN_KEY) {
	runError("Only SysAdmin code can use SetEffectiveToReal");
    } else {
	EffectivePlayer = RealPlayer;
	EffectiveStatus = RealStatus;
    }
}

static void
SetEffectiveTo(ULONG_T who)
{

    if (EffectivePlayer != SYSADMIN_KEY) {
	runError("Only SysAdmin code can use SetEffectiveTo");
    } else {
	EffectivePlayer = who;
	EffectiveStatus = ps_apprentice;
    }
}

static void
MakeWizard(ULONG_T key, ULONG_T doMode)
{

    setStatus(key, ps_wizard, doMode);
}

static void
MakeApprentice(ULONG_T key, ULONG_T doMode)
{

    setStatus(key, ps_apprentice, doMode);
}

static void
MakeNormal(ULONG_T key)
{

    setStatus(key, ps_normal, BOOLEAN_TRUE);
}
static LONG_T
IntToFixed(LONG_T n)
{
    return n * (1 << FRAC_BITS);
}

static LONG_T
FixedToInt(LONG_T n)
{
    return n / (1 << FRAC_BITS);
}

static CHAR_T *
IntToString(LONG_T n)
{
    CHAR_T buffer[12];
    CHAR_T *p;
    BOOL_T isNegative;

    if (n < 0) {
	isNegative = B_TRUE;
    } else {
	isNegative = B_FALSE;
	n = -n;
    }
    p = &buffer[11];
    *p = '\0';
    do {
	*--p = (- (n % 10)) + '0';
	n /= 10;
    } while (n != 0);
    if (isNegative) {
	*--p = '-';
    }
    return stringCopy(p);
}

static LONG_T
StringToInt(CHAR_T *p)
{
    ULONG_T n;

    if (isdigit(*p)) {
	n = 0;
	while (isdigit(*p)) {
	    n *= 10;
	    n += (*p - '0');
	    ++p;
	}
	return (LONG_T) n;
    } else {
	return -1;
    }
}

static CHAR_T *
FixedToString(LONG_T n)
{
    CHAR_T buff[20];

    return stringCopy(displayFixed(n, &buff));
}

static LONG_T
StringToFixed(CHAR_T *p)
{
    ULONG_T n, denom, whole;
    int digitCount, bitCount;
    CHAR_T ch;

    ch = *p;
    if (isdigit(ch) || ch == '.') {
	n = 0;
	while (isdigit(ch)) {
	    ++p;
	    n *= 10;
	    n += (ch - '0');
	    ch = *p;
	}
	n <<= FRAC_BITS;
	whole = n;
	if (ch == '.') {
	    ++p;
	    n = 0;
	    digitCount = 0;
	    denom = 1;
	    ch = *p;
	    while (isdigit(ch)) {
		++p;
		if (denom < 1000000) {
		    denom *= 10;
		    n *= 10;
		    n += (ch - '0');
		    digitCount += 1;
		}
		ch = *p;
	    }
	    n += denom >> (FRAC_BITS + 1);
	    for (bitCount = 0; bitCount < FRAC_BITS; bitCount += 1) {
		if (n < (1 << (ULONG_BITS - 1))) {
		    n <<= 1;
		} else {
		    n /= 5;
		    digitCount -= 1;
		}
	    }
	    while (digitCount != 0) {
		digitCount -= 1;
		n /= 10;
	    }
	    whole += n;
	    if (n > whole) {
		whole = ~0;
	    }
	}
	return (LONG_T) whole;
    } else {
	return (-1 * (1 << FRAC_BITS));
    }
}

static ULONG_T
StringToAction(CHAR_T *st)
{

    return parseActionFromString(&MS, st, EffectivePlayer);
}

static ULONG_T
StringToProc(CHAR_T *st)
{

    return parseProcFromString(&MS, st, EffectivePlayer);
}

static CHAR_T *
ProcToString(ULONG_T key)
{
    CHAR_T *res;
    Proc_t *pr;
    UINT_T *pLockCount;

    res = P_NULL;
    if (key == NIL_ACTION) {
	runError("nil action to ProcToString");
    } else {
	pr = findCachedProc(key, &pLockCount);
	if (pr->pr_owner != EffectivePlayer && EffectivePlayer != SYSADMIN_KEY)
	{
	    runError("you don't own that proc");
	} else if (hasBadRefs(pr)) {
	    runError("proc has references to symbols not in any in-use table");
	} else {
	    (void) toggleProcToString();
	    fmtStart();
	    fmtProc(pr, P_NULL);
	    fmtLine();
	    res = toggleProcToString();
	}
	*pLockCount -= 1;
    }
    if (res == P_NULL) {
	res = nullString();
    }
    return res;
}

static CHAR_T *
Date(void)
{
#define DATE_LEN	24
    /* 24 characters plus \0 are put into the buffer */
    CHAR_T dateBuffer[DATE_LEN + 1];

    ConvTime(GetCurrentTime(), &dateBuffer[0]);
    return stringCopy(&dateBuffer[0]);
#undef DATE_LEN
}

static CHAR_T *
DateShort(void)
{
#define DATE_LEN	19
    /* 19 characters plus \0 are put into the buffer */
    CHAR_T dateBuffer[DATE_LEN + 1];

    ConvTimeShort(GetCurrentTime(), &dateBuffer[0]);
    return stringCopy(&dateBuffer[0]);
#undef DATE_LEN
}

static void
TraceEntry(ULONG_T n)
{

    TRACE("user trace", n);
}

static void
DumpBlocks(CHAR_T *where)
{

    dumpBlocksE(&MS, where);
}

static void
ShowBlocks(CHAR_T *where)
{

    showBlocksE(&MS, where);
}

static void
MemValidate(CHAR_T *where)
{

    memValidateE(&MS, where);
}

/*
 * other parts of builtinInit - too many relocations for all in one proc.
 */

static void
effectsBuiltins(void)
{
    void (*r_bi0)(CHAR_T *name, Comp_t body, UINT_T resultType);
    void (*r_bi1)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type);
    void (*r_bi2)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type,
		  CHAR_T *par2Name, UINT_T par2Type);
    void (*r_bi3)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type,
		  CHAR_T *par2Name, UINT_T par2Type,
		  CHAR_T *par3Name, UINT_T par3Type);

    r_bi0 = bi0;
    r_bi1 = bi1;
    r_bi2 = bi2;
    r_bi3 = bi3;

    r_bi0("ServerVersion", (Comp_t) ServerVersion, T_I);
    r_bi1("ClientVersion", (Comp_t) ClientVersion, T_I, "who", T_T);
    r_bi2("TextComponent", (Comp_t) TextComponent, T_I, "who", T_T,
	  "newComponent", T_I);
    r_bi2("DrawComponent", (Comp_t) DrawComponent, T_I, "who", T_T,
	  "newComponent", T_I);
    bi4("CreateContainer", (Comp_t) CreateContainer, T_V,
	"encloser", T_I, "newId", T_I, "isVertical", T_B, "alignment", T_I);
    bi5("CreateComponent", (Comp_t) CreateComponent, T_V,
	"encloser", T_I, "newId", T_I, "width", T_I, "height", T_I,
	"kind", T_I);
    r_bi0("MakeFrame", (Comp_t) MakeFrame, T_V);
    r_bi1("GetFontAscent", (Comp_t) GetFontAscent, T_I, "who", T_T);
    r_bi1("GetFontDescent", (Comp_t) GetFontDescent, T_I, "who", T_T);
    r_bi1("GetFontLeading", (Comp_t) GetFontLeading, T_I, "who", T_T);
    r_bi1("GetFontHeight", (Comp_t) GetFontHeight, T_I, "who", T_T);
    r_bi1("GetFontWidth", (Comp_t) GetFontWidth, T_I, "who", T_T);
    r_bi0("Note", (Comp_t) Note, T_V);
    r_bi1("IfFound", (Comp_t) IfFound, T_V, "who", T_T);
    r_bi1("Else", (Comp_t) Else, T_V, "who", T_T);
    r_bi1("Fi", (Comp_t) Fi, T_V, "who", T_T);
    r_bi2("FailText", (Comp_t) FailText, T_V,
	"who", T_T, "message", T_S);
    r_bi3("AbortEffect", (Comp_t) AbortEffect, T_V,
	"who", T_T, "kind", T_I, "id", T_I);
    r_bi2("CallEffect", (Comp_t) CallEffect, T_V, "who", T_T,
	"whichEffect", T_I);
    r_bi2("KnowsEffect", (Comp_t) KnowsEffect, T_B, "who", T_T,
	"whichEffect", T_I);
    r_bi2("DefineEffect", (Comp_t) DefineEffect, T_V, "who", T_T,
	"whichEffect", T_I);
    r_bi0("EndEffect", (Comp_t) EndEffect, T_V);
    r_bi1("GOn", (Comp_t) GOn, T_B, "who", T_T);
    r_bi1("GType", (Comp_t) GType, T_S, "who", T_T);
    r_bi1("GRows", (Comp_t) GRows, T_I, "who", T_T);
    r_bi1("GCols", (Comp_t) GCols, T_I, "who", T_T);
    r_bi1("GColours", (Comp_t) GColours, T_I, "who", T_T);
    r_bi1("GPalette", (Comp_t) GPalette, T_B, "who", T_T);
    r_bi2("PlaceCursor", (Comp_t) PlaceCursor, T_V,
	"x", T_F, "y", T_F);
    r_bi2("PlaceCursorPixels", (Comp_t) PlaceCursorPixels, T_V,
	"x", T_I, "y", T_I);
    r_bi0("RemoveCursor", (Comp_t) RemoveCursor, T_V);
    r_bi1("SetCursorPen", (Comp_t) SetCursorPen, T_V, "pen", T_I);
    r_bi1("SetCursorPattern", (Comp_t) SetCursorPattern, T_V,
	"newPattern", T_L_I);
    bi4("AddButton", (Comp_t) AddButton, T_V,
	"x", T_F, "y", T_F, "n", T_I, "str", T_S);
    bi4("AddButtonPixels", (Comp_t) AddButtonPixels, T_V,
	"x", T_I, "y", T_I, "n", T_I, "str", T_S);
    r_bi1("EraseButton", (Comp_t) EraseButton, T_V, "n", T_I);
    r_bi0("ClearButtons", (Comp_t) ClearButtons, T_V);
    bi5("AddRegion", (Comp_t) AddRegion, T_V,
	"x", T_F, "y", T_F, "width", T_F, "height", T_F, "n", T_I);
    bi5("AddRegionPixels", (Comp_t) AddRegionPixels, T_V,
	"x", T_I, "y", T_I, "width", T_I, "height", T_I, "n", T_I);
    r_bi1("EraseRegion", (Comp_t) EraseRegion, T_V, "n", T_I);
    r_bi0("ClearRegions", (Comp_t) ClearRegions, T_V);
    r_bi2("SetButtonPen", (Comp_t) SetButtonPen, T_V,
	"which", T_I, "pen", T_I);
    r_bi2("GSetPen", (Comp_t) GSetPen, T_V, "who", T_T, "pen", T_I);
    r_bi3("GSetColour", (Comp_t) GSetColour, T_V, "who", T_T,
	"which", T_I, "colour", T_I);
    r_bi1("GResetColours", (Comp_t) GResetColours, T_V, "who", T_T);
    r_bi1("GClear", (Comp_t) GClear, T_V, "who", T_T);
    r_bi1("GPixel", (Comp_t) GPixel, T_V, "who", T_T);
    bi4("GPixelRow", (Comp_t) GPixelRow, T_V, "who", T_T,
	"pixels", T_L_I, "start", T_I, "count", T_I);
    r_bi3("GAMove", (Comp_t) GAMove, T_V, "who", T_T,
	"x", T_F, "y", T_F);
    r_bi3("GAMovePixels", (Comp_t) GAMovePixels, T_V, "who", T_T,
	"x", T_I, "y", T_I);
    r_bi3("GRMove", (Comp_t) GRMove, T_V, "who", T_T,
	"deltaX", T_F, "deltaY", T_F);
    r_bi3("GRMovePixels", (Comp_t) GRMovePixels, T_V, "who", T_T,
	"deltaXPixels", T_I, "deltaYPixels", T_I);
    r_bi3("GADraw", (Comp_t) GADraw, T_V, "who", T_T,
	"x", T_F, "y", T_F);
    r_bi3("GADrawPixels", (Comp_t) GADrawPixels, T_V, "who", T_T,
	"x", T_I, "y", T_I);
    r_bi3("GRDraw", (Comp_t) GRDraw, T_V, "who", T_T,
	"deltaX", T_F, "deltaY", T_F);
    r_bi3("GRDrawPixels", (Comp_t) GRDrawPixels, T_V, "who", T_T,
	"deltaXPixels", T_I, "deltaYPixels", T_I);
    bi4("GRectangle", (Comp_t) GRectangle, T_V, "who", T_T,
	"width", T_F, "height", T_F, "fill", T_B);
    bi4("GRectanglePixels", (Comp_t) GRectanglePixels, T_V, "who", T_T,
	"width", T_I, "height", T_I, "fill", T_B);
    r_bi3("GCirclePixels", (Comp_t) GCirclePixels, T_V, "who", T_T,
	"r", T_I, "fill", T_B);
    r_bi3("GCircle", (Comp_t) GCircle, T_V, "who", T_T,
	"r", T_F, "fill", T_B);
    bi4("GEllipsePixels", (Comp_t) GEllipsePixels, T_V, "who", T_T,
	"a", T_I, "B", T_I, "fill", T_B);
    bi4("GEllipse", (Comp_t) GEllipse, T_V, "who", T_T,
	"a", T_F, "b", T_F, "fill", T_B);
    r_bi1("GPolygonStart", (Comp_t) GPolygonStart, T_V, "who", T_T);
    r_bi1("GPolygonEnd", (Comp_t) GPolygonEnd, T_V, "who", T_T);
    bi5("GDefineTile", (Comp_t) GDefineTile, T_V, "who", T_T,
	"id", T_I, "pixelWidth", T_I, "pixelHeight", T_I, "data", T_L_I);
    bi6("GDefineOverlayTile", (Comp_t) GDefineOverlayTile, T_V, "who", T_T,
	"id", T_I, "transparentColour", T_I, "pixelWidth", T_I,
	"pixelHeight", T_I, "data", T_L_I);
    r_bi2("GDisplayTile", (Comp_t) GDisplayTile, T_V,
	"who", T_T, "id", T_I);
    r_bi3("GSetTextColour", (Comp_t) GSetTextColour, T_V,
	"who", T_T, "which", T_I, "colour", T_I);
    r_bi2("GText", (Comp_t) GText, T_V, "who", T_T, "text", T_S);
    r_bi2("GLoadBackGround", (Comp_t) GLoadBackGround, T_V,
	"who", T_T, "name", T_S);
    r_bi2("GSetImage", (Comp_t) GSetImage, T_V,
	"who", T_T, "name", T_S);
    bi10("GShowImage", (Comp_t) GShowImage, T_V,
	"who", T_T, "name", T_S,
	"imageX", T_F, "imageY", T_F,
	"imageWidth", T_F, "imageHeight", T_F,
	"displayX", T_F, "displayY", T_F,
	"displayWidth", T_F, "displayHeight", T_F);
    bi8("GShowImagePixels", (Comp_t) GShowImagePixels, T_V,
	"who", T_T, "name", T_S,
	"imageX", T_I, "imageY", T_I,
	"imageWidth", T_I, "imageHeight", T_I,
	"displayX", T_I, "displayY", T_I);
    bi4("GShowBrush", (Comp_t) GShowBrush, T_V, "who", T_T,
	"name", T_S, "displayX", T_F, "displayY", T_F);
    bi7("GScrollRectanglePixels", (Comp_t) GScrollRectanglePixels, T_V,
	"who", T_T, "xDeltaPixels", T_I, "yDeltaPixels", T_I,
	"xPixels", T_I, "yPixels", T_I,
	"widthPixels", T_I, "heightPixels", T_I);
    bi7("GScrollRectangle", (Comp_t) GScrollRectangle, T_V, "who", T_T,
	"xDeltaPixels", T_I, "yDeltaPixels", T_I,
	"x", T_F, "y", T_F, "width", T_F, "height", T_F);
    r_bi2("GSetIconPen", (Comp_t) GSetIconPen, T_V, "who", T_T,
	"colour", T_I);
    r_bi2("GNewIcon", (Comp_t) GNewIcon, T_V, "theCharacter", T_T,
	"newIcon", T_L_I);
    bi4("GShowIcon", (Comp_t) GShowIcon, T_V, "who", T_T,
	"whoseIcon", T_T, "isMonster", T_B, "isTemporary", T_B);
    r_bi2("GRemoveIcon", (Comp_t) GRemoveIcon, T_V, "who", T_T,
	"whoseIcon", T_T);
    r_bi2("GDeleteIcon", (Comp_t) GDeleteIcon, T_V, "who", T_T,
	"whoseIcon", T_T);
    r_bi1("GResetIcons", (Comp_t) GResetIcons, T_V, "who", T_T);
#if REDRAW_ICONS
    r_bi1("GRedrawIcons", (Comp_t) GRedrawIcons, T_V, "who", T_T);
    r_bi1("GUndrawIcons", (Comp_t) GUndrawIcons, T_V, "who", T_T);
#endif

    r_bi1("SOn", (Comp_t) SOn, T_B, "who", T_T);
    r_bi2("SVolume", (Comp_t) SVolume, T_V,
	"who", T_T, "volume", T_I);
    bi4("SPlaySound", (Comp_t) SPlaySound, T_V,
	"who", T_T, "name", T_S, "cycles", T_I, "id", T_I);

    r_bi1("VOn", (Comp_t) VOn, T_B, "who", T_T);
    bi6("VParams", (Comp_t) VParams, T_V, "who", T_T,
	"rate", T_I, "pitch", T_I, "mode", T_I, "sex", T_I, "volume", T_I);
    r_bi1("VReset", (Comp_t) VReset, T_V, "who", T_T);
    r_bi2("VVolume", (Comp_t) VVolume, T_V,
	"who", T_T, "volume", T_I);
    r_bi3("VSpeak", (Comp_t) VSpeak, T_V,
	"who", T_T, "words", T_S, "id", T_I);
    r_bi1("VTranslate", (Comp_t) VTranslate, T_S, "words", T_S);
    r_bi3("VNarrate", (Comp_t) VNarrate, T_V,
	"who", T_T, "phonemes", T_S, "id", T_I);

    r_bi1("MOn", (Comp_t) MOn, T_B, "who", T_T);
    r_bi2("MVolume", (Comp_t) MVolume, T_V,
	"who", T_T, "volume", T_I);
    r_bi3("MPlaySong", (Comp_t) MPlaySong, T_V, "who", T_T,
	"name", T_S, "id", T_I);

    /* these are here only temporarily - they will be moved the next time I
       want to make the database format incompatible */


#if 0
    r_bi1("StartSMUS", (Comp_t) StartSMUS, T_I, "file", T_S);
    r_bi1("StopSMUS", (Comp_t) StopSMUS, T_V, "id", T_I);
    r_bi2("AfterSMUS", (Comp_t) AfterSMUS, T_V,
	"id", T_I, "theAction", T_A);
#endif

#if 0
    r_bi1("ShowCacheStatus", (Comp_t) ShowCacheStatus, T_V,
	"key", T_T);
    r_bi0("FileDebugOn", (Comp_t) FileDebugOn, T_V);
    r_bi0("FileDebugOff", (Comp_t) FileDebugOff, T_V);
    r_bi1("FileConCheck", (Comp_t) FileConCheck, T_V, "where", T_S);
    r_bi0("FileTraceOn", (Comp_t) FileTraceOn, T_V);
    r_bi0("FileTraceOff", (Comp_t) FileTraceOff, T_V);
    r_bi1("FileSetPeek", (Comp_t) FileSetPeek, T_V, "key", T_T);
    r_bi1("FileSetWatchString", (Comp_t) FileSetWatchString, T_V,
	  "key" ,T_S);
    r_bi1("FileSetRangeString", (Comp_t) FileSetRangeString, T_V,
	  "key" ,T_S);
    r_bi1("FileSetWatch", (Comp_t) FileSetWatch, T_V, "key", T_I);
    r_bi1("FileSetRange", (Comp_t) FileSetRange, T_V, "key", T_I);
    r_bi2("CheckList", (Comp_t) CheckList, T_V, "list", T_L_T,
	"where", T_S);
#endif
}

static void
symbolMachineFileBuiltins(void)
{
    void (*r_bi0)(CHAR_T *name, Comp_t body, UINT_T resultType);
    void (*r_bi1)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type);
    void (*r_bi2)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type,
		  CHAR_T *par2Name, UINT_T par2Type);
    void (*r_bi3)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type,
		  CHAR_T *par2Name, UINT_T par2Type,
		  CHAR_T *par3Name, UINT_T par3Type);

    r_bi0 = bi0;
    r_bi1 = bi1;
    r_bi2 = bi2;
    r_bi3 = bi3;

    r_bi0("PublicTable", (Comp_t) PublicTable, T_TB);
    r_bi0("PrivateTable", (Comp_t) PrivateTable, T_TB);
    r_bi0("CreateTable", (Comp_t) CreateTable, T_TB);
    r_bi1("UseTable", (Comp_t) UseTable, T_B, "theTable", T_TB);
    r_bi1("UnUseTable", (Comp_t) UnUseTable, T_B, "theTable", T_TB);
    r_bi1("ShowTable", (Comp_t) ShowTable, T_V, "theTable", T_TB);
    r_bi2("DescribeSymbol", (Comp_t) DescribeSymbol, T_V,
	"theTable", T_TB, "name", T_S);
    r_bi2("DeleteSymbol", (Comp_t) DeleteSymbol, T_B,
	"theTable", T_TB, "name", T_S);
    r_bi1("DescribeThing", (Comp_t) fmtThing, T_V, "theThing", T_T);
    r_bi3("MoveSymbol", (Comp_t) MoveSymbol, T_B,
	"fromTable", T_TB, "toTable", T_TB, "name", T_S);
    r_bi3("RenameSymbol", (Comp_t) RenameSymbol, T_B,
	"theTable", T_TB, "oldName", T_S, "newName", T_S);
    r_bi3("DefineThing", (Comp_t) DefineThing, T_B,
	"theTable", T_TB, "name", T_S, "theThing", T_T);
    r_bi3("DefineAction", (Comp_t) DefineAction, T_B,
	"theTable", T_TB, "name", T_S, "theAction", T_A);
    r_bi3("DefineFlag", (Comp_t) DefineFlag, T_B,
	"theTable", T_TB, "name", T_S, "theFlag", T_P_B);
    r_bi3("DefineCounter", (Comp_t) DefineCounter, T_B,
	"theTable", T_TB, "name", T_S, "theCounter", T_P_I);
    r_bi3("DefineString", (Comp_t) DefineString, T_B,
	"theTable", T_TB, "name", T_S, "theString", T_P_S);
    r_bi3("DefineTable", (Comp_t) DefineTable, T_B,
	"theTable", T_TB, "name", T_S, "newTable", T_TB);
    r_bi2("LookupThing", (Comp_t) LookupThing, T_T,
	"theTable", T_TB, "name", T_S);
    r_bi2("LookupAction", (Comp_t) LookupAction, T_A,
	"theTable", T_TB, "name", T_S);
    r_bi2("LookupFlag", (Comp_t) LookupFlag, T_P_B,
	"theTable", T_TB, "name", T_S);
    r_bi2("LookupCounter", (Comp_t) LookupCounter, T_P_I,
	"theTable", T_TB, "name", T_S);
    r_bi2("LookupString", (Comp_t) LookupString, T_P_S,
	"theTable", T_TB, "name", T_S);
    r_bi2("LookupTable", (Comp_t) LookupTable, T_TB,
	"theTable", T_TB, "name", T_S);
    r_bi2("IsDefined", (Comp_t) IsDefined, T_B,
	"theTable", T_TB, "name", T_S);
    r_bi2("FindThingSymbol", (Comp_t) FindXXXSymbol, T_S,
	"theTable", T_TB, "theThing", T_T);
    r_bi2("FindActionSymbol", (Comp_t) FindXXXSymbol, T_S,
	"theTable", T_TB, "theAction", T_A);
    r_bi2("ScanTable", (Comp_t) ScanTable, T_V,
	"theTable", T_TB, "theAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi2("SymbolType", (Comp_t) SymbolType, T_S,
	"theTable", T_TB, "name", T_S);
    r_bi1("ZapAction", (Comp_t) ZapAction, T_V, "theAction", T_A);
    r_bi1("ZapTable", (Comp_t) ZapTable, T_V, "theTable", T_TB);
    r_bi1("ZapGrammar", (Comp_t) ZapGrammar, T_V, "theGrammar", T_G);
    r_bi1("ZapCharacter", (Comp_t) ZapCharacter, T_V,
	"theCharacter", T_P);

    bi4("CreateMachine", (Comp_t) CreateMachine, T_V,
	"name", T_S, "theThing", T_T, "location", T_T, "startup", T_A);
    r_bi2("SetMachineSay", (Comp_t) SetMachineSay, T_A,
	"machine", T_T, "theAction", T_A);
    r_bi2("SetMachineWhisperMe", (Comp_t) SetMachineWhisperMe, T_A,
	"machine", T_T, "theAction", T_A);
    r_bi2("SetMachineWhisperOther", (Comp_t) SetMachineWhisperOther,T_A,
	"machine", T_T, "theAction", T_A);
    r_bi2("SetMachinePose", (Comp_t) SetMachinePose, T_A,
	"machine", T_T, "theAction", T_A);
    r_bi2("SetMachineOther", (Comp_t) SetMachineOther, T_A,
	"machine", T_T, "theAction", T_A);
    r_bi2("SetMachineIdle", (Comp_t) SetMachineIdle, T_A,
	"machine", T_T, "theAction", T_A);
    r_bi2("SetMachineActive", (Comp_t) SetMachineActive, T_A,
	"machine", T_T, "theAction", T_A);
    r_bi1("DestroyMachine", (Comp_t) DestroyMachine, T_V,
	"machine", T_T);

    r_bi1("FileOpenForRead", (Comp_t) FileOpenForRead, T_I,
	"fileName", T_S);
    r_bi1("FileOpenForWrite", (Comp_t) FileOpenForWrite, T_I,
	"fileName", T_S);
    r_bi1("FileOpenForUpdate", (Comp_t) FileOpenForUpdate, T_I,
	"fileName", T_S);
    r_bi1("FileClose", (Comp_t) FileClose, T_V, "fileId", T_I);
    r_bi1("FileRead", (Comp_t) FileRead, T_S, "fileId", T_I);
    r_bi2("FileWrite", (Comp_t) FileWrite, T_V,
	"fileId", T_I, "theString", T_S);
    r_bi2("FileReadBinary", (Comp_t) FileReadBinary, T_I,
	"fileId", T_I, "theList", T_L_I);
    r_bi2("FileWriteBinary", (Comp_t) FileWriteBinary, T_V,
	"fileId", T_I, "theList", T_L_I);
    r_bi2("FileSeek", (Comp_t) FileSeek, T_V,
	"fileId", T_I, "pos", T_I);
    r_bi1("Execute", (Comp_t) myExecute, T_V, "command", T_S);
}

void
builtInInit(ULONG_T seed)
{
    void (*r_bi0)(CHAR_T *name, Comp_t body, UINT_T resultType);
    void (*r_bi1)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type);
    void (*r_bi2)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type,
		  CHAR_T *par2Name, UINT_T par2Type);
    void (*r_bi3)(CHAR_T *name, Comp_t body, UINT_T resultType,
		  CHAR_T *par1Name, UINT_T par1Type,
		  CHAR_T *par2Name, UINT_T par2Type,
		  CHAR_T *par3Name, UINT_T par3Type);

    NextBuiltinKey = FIRST_BUILTIN;
    Seed = seed;
    r_bi0 = bi0;
    r_bi1 = bi1;
    r_bi2 = bi2;
    r_bi3 = bi3;

    r_bi0("CreateBoolProp", (Comp_t) CreateBoolProp, T_P_B);
    r_bi0("CreateIntProp", (Comp_t) CreateIntProp, T_P_I);
    r_bi0("CreateFixedProp", (Comp_t) CreateFixedProp, T_P_F);
    r_bi0("CreateStringProp", (Comp_t) CreateStringProp, T_P_S);
    r_bi0("CreateThingProp", (Comp_t) CreateThingProp, T_P_T);
    r_bi0("CreateActionProp", (Comp_t) CreateActionProp, T_P_A);
    r_bi0("CreateTableProp", (Comp_t) CreateTableProp, T_P_TB);
    r_bi0("CreateGrammarProp", (Comp_t) CreateGrammarProp, T_P_G);
    r_bi0("CreateIntListProp", (Comp_t) CreateIntListProp, T_P_L_I);
    r_bi0("CreateFixedListProp", (Comp_t) CreateFixedListProp,
	T_P_L_F);
    r_bi0("CreateThingListProp", (Comp_t) CreateThingListProp,T_P_L_T);
    r_bi0("CreateActionListProp",(Comp_t) CreateActionListProp,T_P_L_A);

    r_bi1("CreateThing", (Comp_t) CreateThing, T_T, "parent", T_T);
    r_bi2("SetParent", (Comp_t) SetParent, T_V,
	"theThing", T_T, "newParent", T_T);
    r_bi1("Parent", (Comp_t) Parent, T_T, "theThing", T_T);
    r_bi2("IsAncestor", (Comp_t) IsAncestor, T_B,
	"myKey", T_T, "parentKey", T_T);
    r_bi2("SetThingStatus", (Comp_t) SetThingStatus, T_V,
	"theThing", T_T, "status", TYPE_THING_STATUS);
    r_bi1("GetThingStatus", (Comp_t) GetThingStatus, TYPE_THING_STATUS,
	"theThing", T_T);
    r_bi1("Mine", (Comp_t) Mine, T_B, "theThing", T_T);
    r_bi1("ClearThing", (Comp_t) ClearThing, T_V, "theThing", T_T);
    r_bi2("GiveThing", (Comp_t) GiveThing, T_V,
	"theThing", T_T, "theCharacter", T_P);
    r_bi1("Owner", (Comp_t) Owner, T_P, "theThing", T_T);

    r_bi0("CreateIntList", (Comp_t) CreateIntList, T_L_I);
    r_bi0("CreateFixedList", (Comp_t) CreateFixedList, T_L_F);
    r_bi0("CreateThingList", (Comp_t) CreateThingList, T_L_T);
    r_bi0("CreateActionList", (Comp_t) CreateActionList, T_L_A);
    r_bi1("CreateIntArray", (Comp_t) CreateIntArray, T_L_I,
	"size", T_I);

    r_bi1("Count", (Comp_t) Count, T_I, "theList", T_LANY);
    r_bi2("AddHead", (Comp_t) AddHead, T_V,
	"theList", T_LANY, "theElement", T_ELEM);
    r_bi1("RemHead", (Comp_t) RemHead, T_V, "theList", T_LANY);
    r_bi2("AddTail", (Comp_t) xAddTail, T_V,
	"theList", T_LANY, "theElement", T_ELEM);
    r_bi1("RemTail", (Comp_t) RemTail, T_V, "theList", T_LANY);
    r_bi2("FindElement", (Comp_t) FindElement, T_I,
	"theList", T_LANY, "valu", T_ELEM);
    r_bi2("DelElement", (Comp_t) DelElement, T_V,
	"theList", T_LANY, "valu", T_ELEM);

    r_bi0("CreateGrammar", (Comp_t) CreateGrammar, T_G);
    r_bi2("Word", (Comp_t) Word, T_I, "theGrammar", T_G,
	"theWord", T_S);
    r_bi3("Synonym", (Comp_t) Synonym, T_V, "theGrammar", T_G,
	"oldWord", T_S, "newWord", T_S);
    r_bi3("VerbTail", (Comp_t) VerbTail, T_V, "theGrammar", T_G,
	"theVerb", T_S, "theAction", T_A);
    bi4("Verb0", (Comp_t) Verb0, T_V, "theGrammar", T_G,
	"theVerb", T_S, "separatorCode", T_I, "theAction", T_A);
    bi4("Verb1", (Comp_t) Verb1, T_V, "theGrammar", T_G,
	"theVerb", T_S, "separatorCode", T_I, "theAction", T_A);
    bi4("Verb2", (Comp_t) Verb2, T_V, "theGrammar", T_G,
	"theVerb", T_S, "separatorCode", T_I, "theAction", T_A);
    r_bi2("FindWord", (Comp_t) FindWord, T_I, "theGrammar", T_G,
	"theWord", T_S);
    r_bi2("FindAnyWord", (Comp_t) FindAnyWord, T_I,
	"theGrammar", T_G, "theWord", T_S);
    r_bi2("ShowWord", (Comp_t) ShowWord, T_V, "theGrammar", T_G,
	"theWord", T_S);
    r_bi1("ShowWords", (Comp_t) ShowWords, T_V, "theGrammar", T_G);
    r_bi2("DeleteWord", (Comp_t) DeleteWord, T_V, "theGrammar", T_G,
	"theWord", T_S);
    r_bi3("GetNounPhrase", (Comp_t) GetNounPhrase, T_S,
	"theGrammar", T_G, "theString", T_S, "separator", T_I);
    r_bi2("Parse", (Comp_t) Parse, T_I, "theGrammar", T_G,
	"sentence", T_S);
    r_bi0("Verb", (Comp_t) Verb, T_S);
    r_bi0("ItName", (Comp_t) ItName, T_S);
    r_bi0("WhoName", (Comp_t) WhoName, T_S);
    r_bi0("Punctuation", (Comp_t) Punctuation, T_S);
    r_bi0("GetTail", (Comp_t) GetTail, T_S);
    r_bi0("GetWord", (Comp_t) GetWord, T_S);
    r_bi1("SetTail", (Comp_t) SetTail, T_V, "str", T_S);
    r_bi1("SetSay", (Comp_t) SetSay, T_S, "what", T_S);
    r_bi1("SetWhisperMe", (Comp_t) SetWhisperMe, T_S, "what", T_S);
    r_bi1("SetWhisperOther", (Comp_t) SetWhisperOther, T_S, "what",T_S);
    r_bi2("HasAdjective", (Comp_t) HasAdjective, T_B,
	"name", T_S, "adjective", T_S);
    r_bi2("MatchName", (Comp_t) MatchName, T_I,
	"stored", T_S, "parsed", T_S);
    r_bi2("SelectName", (Comp_t) SelectName, T_S,
	"names", T_S, "which", T_I);
    r_bi2("SelectWord", (Comp_t) SelectWord, T_S,
	"words", T_S, "which", T_I);
    r_bi1("FormatName", (Comp_t) FormatName, T_S, "theName", T_S);

    r_bi2("FindAgentAt", (Comp_t) FindAgentAt, T_T,
	"location", T_T, "name", T_S);
    r_bi1("FindAgent", (Comp_t) FindAgent, T_T, "name", T_S);
    r_bi2("FindMachineIndexed", (Comp_t) FindMachineIndexed, T_T,
	"name", T_S, "index", T_I);
    r_bi3("FindName", (Comp_t) xFindName, T_ST,
	"theList", T_L_T, "nameProp", T_P_S, "theName", T_S);
    r_bi0("FindResult", (Comp_t) FindResult, T_T);
    r_bi2("FindFlagOnList", (Comp_t) FindFlagOnList, T_B,
	"theList", T_L_T, "flagProp", T_P_B);
    r_bi2("FindChildOnList", (Comp_t) FindChildOnList, T_B,
	"theList", T_L_T, "parent", T_T);
    bi4("FindIntOnList", (Comp_t) FindIntOnList, T_T,
	"theList", T_L_T, "intProp", T_P_I, "valu", T_I, "compType", T_I);
    r_bi2("FindAgentWithFlag", (Comp_t) FindAgentWithFlag, T_T,
	"room", T_T, "flagProp", T_P_B);
    r_bi2("FindAgentAsChild", (Comp_t) FindAgentAsChild, T_T,
	"room", T_T, "parent", T_T);
    r_bi2("FindAgentAsDescendant", (Comp_t) FindAgentAsDescendant, T_T,
	"room", T_T, "parent", T_T);
    r_bi3("FindAgentWithFlagOnList", (Comp_t) FindAgentWithFlagOnList,
	T_T, "room", T_T, "listProp", T_P_L_T, "flagProp", T_P_B);
    r_bi3("FindAgentWithChildOnList", (Comp_t) FindAgentWithChildOnList,
	T_T, "room", T_T, "listProp", T_P_L_T, "parent", T_T);
    bi4("FindAgentWithNameOnList", (Comp_t) FindAgentWithNameOnList,T_T,
	"room", T_T, "listProp", T_P_L_T, "nameProp", T_P_S, "name", T_S);

    r_bi1("Print", (Comp_t) putString, T_V, "str", T_S);
    r_bi1("IPrint", (Comp_t) putSigned, T_V, "number", T_I);
    r_bi1("OPrint", (Comp_t) OPrint, T_V, "str", T_S);
    r_bi2("SPrint", (Comp_t) SPrint, T_V,
	"agent", T_T, "str", T_S);
    bi4("ABPrint", (Comp_t) ABPrint, T_V,
	"location", T_T, "agent1", T_T, "agent2", T_T, "str", T_S);
    r_bi1("NPrint", (Comp_t) NPrint, T_V, "str", T_S);
    r_bi1("SetPrefix", (Comp_t) SetPrefix, T_V, "newPrefix", T_S);
    r_bi1("PrintNoAts", (Comp_t) PrintNoAts, T_B, "flag", T_B);
    r_bi3("SubString", (Comp_t) SubString, T_S,
	"str", T_S, "start", T_I, "length", T_I);
    r_bi1("Length", (Comp_t) Length, T_I, "str", T_S);
    r_bi1("Capitalize", (Comp_t) Capitalize, T_S, "str", T_S);
    r_bi1("Pluralize", (Comp_t) Pluralize, T_S, "str", T_S);
    r_bi2("Index", (Comp_t) Index, T_I, "s1", T_S, "s2", T_S);
    r_bi3("StringReplace", (Comp_t) StringReplace, T_S,
	"s1", T_S, "pos", T_I, "s2", T_S);
    r_bi1("Trim", (Comp_t) Trim, T_S, "str", T_S);
    r_bi1("IntToFixed", (Comp_t) IntToFixed, T_F, "n", T_I);
    r_bi1("FixedToInt", (Comp_t) FixedToInt, T_I, "n", T_F);
    r_bi1("IntToString", (Comp_t) IntToString, T_S, "n", T_I);
    r_bi1("StringToInt", (Comp_t) StringToInt, T_I, "str", T_S);
    r_bi1("FixedToString", (Comp_t) FixedToString, T_S, "n",T_F);
    r_bi1("StringToFixed", (Comp_t) StringToFixed, T_F,
	"str", T_S);
    r_bi1("Random", (Comp_t) Random, T_I, "range", T_I);
    r_bi1("SetSeed", (Comp_t) SetSeed, T_V, "newSeed", T_I);
    r_bi0("GetSeed", (Comp_t) GetSeed, T_I);
    r_bi0("Quit", (Comp_t) Quit, T_V);
    r_bi2("AAn", (Comp_t) AAn, T_S, "s1", T_S, "s2", T_S);
    bi4("IsAre", (Comp_t) IsAre, T_S,
	"s1", T_S, "s2", T_S, "s3", T_S, "s4", T_S);
    r_bi1("Strip", (Comp_t) Strip, T_S, "st", T_S);
    r_bi2("After", (Comp_t) After, T_V,
	"seconds", T_F, "theAction", T_A);
    r_bi1("SetPrompt", (Comp_t) SetPrompt, T_S, "prompt", T_S);
    LatestProc->pr_wizard = B_TRUE;
    r_bi0("NewCharacterPassword", (Comp_t) NewCharacterPassword, T_V);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("ChangeName", (Comp_t) ChangeName, T_V, "newName", T_S);
    LatestProc->pr_wizard = B_TRUE;
    r_bi0("Date", (Comp_t) Date, T_S);
    r_bi0("DateShort", (Comp_t) DateShort, T_S);
    r_bi0("Time", (Comp_t) GetCurrentTime, T_I);
    r_bi1("TextWidth", (Comp_t) TextWidth, T_I, "newWidth", T_I);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("TextHeight", (Comp_t) TextHeight, T_I, "newHeight", T_I);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetIndent", (Comp_t) SetIndent, T_V, "newIndent", T_I);
    LatestProc->pr_wizard = B_TRUE;
    r_bi0("GetIndent", (Comp_t) GetIndent, T_I);

    r_bi1("SetLocation", (Comp_t) SetLocation, T_V, "where", T_T);
    LatestProc->pr_wizard = B_TRUE;
    r_bi2("SetAgentLocation", (Comp_t) SetAgentLocation, T_V,
	"agent", T_T, "where", T_T);
    LatestProc->pr_wizard = B_TRUE;
    r_bi0("Here", (Comp_t) Here, T_T);
    r_bi2("Say", (Comp_t) Say, T_V, "alternateName", T_S, "what", T_S);
    r_bi3("SayTo", (Comp_t) SayTo, T_V, "who", T_T,
	"alternativeName", T_S, "what", T_S);
    r_bi2("Pose", (Comp_t) Pose, T_V, "alternateName", T_S, "what", T_S);
    r_bi3("PoseTo", (Comp_t) PoseTo, T_V, "who", T_T,
	 "alternativeName", T_S, "what", T_S);
    bi4("Whisper", (Comp_t) Whisper, T_B, "alternateName", T_S,
	"what", T_S, "who", T_T, "percent", T_I);
    r_bi0("Me", (Comp_t) Me, T_T);
    r_bi0("TrueMe", (Comp_t) TrueMe, T_T);
    r_bi0("MeString", (Comp_t) GetMeString, T_S);
    r_bi1("SetMeString", (Comp_t) SetMeString, T_V, "name", T_S);
    r_bi0("MeCharacter", (Comp_t) MeCharacter, T_P);
    r_bi1("SetIt", (Comp_t) SetIt, T_V, "theThing", T_T);
    r_bi0("It", (Comp_t) It, T_T);
    r_bi1("Character", (Comp_t) Character, T_P, "name", T_S);
    r_bi1("CharacterThing", (Comp_t) CharacterThing, T_T,
	"theCharacter", T_P);
    r_bi1("CharacterTable", (Comp_t) CharacterTable, T_TB,
	"theCharacter", T_P);
    r_bi1("ThingCharacter", (Comp_t) ThingCharacter, T_P,
	"theThing", T_T);
    r_bi1("CharacterLocation", (Comp_t) CharacterLocation, T_T,
	"theCharacter", T_P);
    r_bi2("SetCharacterLocation", (Comp_t) SetCharacterLocation, T_V,
	"theCharacter", T_P, "newLocation", T_T);
    r_bi2("ForEachAgent", (Comp_t) ForEachAgent, T_V,
	"location", T_T, "theAction", T_A);
    r_bi1("ForEachClient", (Comp_t) ForEachClient, T_V,
	"theAction", T_A);
    r_bi1("AgentLocation", (Comp_t) AgentLocation, T_T,
	"theAgent", T_T);
    r_bi2("ForceAction", (Comp_t) ForceAction, T_ST,
	"theAgent", T_T, "theAction", T_A);
    r_bi1("EditProc", (Comp_t) EditProc, T_V, "theProc", T_A);
    r_bi0("CanEdit", (Comp_t) CanEdit, T_B);
    r_bi0("Editing", (Comp_t) Editing, T_B);
    r_bi0("GettingString", (Comp_t) GettingString, T_B);

    r_bi0("Normal", (Comp_t) Normal, T_V);
    r_bi0("WizardMode", (Comp_t) WizardMode, T_B);
    r_bi0("IsProgrammer", (Comp_t) IsProgrammer, T_B);
    r_bi0("IsNormal", (Comp_t) IsNormal, T_B);
    r_bi0("IsApprentice", (Comp_t) IsApprentice, T_B);
    r_bi0("IsWizard", (Comp_t) IsWizard, T_B);
    r_bi1("PrintAction", (Comp_t) PrintAction, T_V, "theAction", T_A);

    r_bi0("Flush", (Comp_t) Flush, T_V);
    r_bi1("SetContinue", (Comp_t) SetContinue, T_V, "state", T_B);
    r_bi1("SetSingleUser", (Comp_t) SetSingleUser, T_V, "state", T_B);
    r_bi1("SetRemoteSysAdminOK", (Comp_t) SetRemoteSysAdminOK, T_V,
	"state", T_B);
    r_bi1("RunLimit", (Comp_t) RunLimit, T_I, "newValue", T_I);
    r_bi1("SetNewCharacterAction", (Comp_t) SetNewCharacterAction, T_A,
	"newAction", T_A);
    r_bi0("NewCreationPassword", (Comp_t) NewCreationPassword, T_V);
    r_bi1("ShutDown", (Comp_t) ShutDown, T_B, "yesNo", T_B);
    r_bi2("CreateCharacter", (Comp_t) CreateCharacter, T_V,
	"name", T_S, "password", T_S);
    r_bi1("DestroyCharacter", (Comp_t) DestroyCharacter, T_V,
	"theCharacter", T_S);
    r_bi1("APrint", (Comp_t) APrint, T_V, "message", T_S);

    r_bi1("SetCharacterInputAction", (Comp_t) SetCharacterInputAction,
	T_A, "newAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetCharacterRawKeyAction", (Comp_t) SetCharacterRawKeyAction,
	T_A, "newAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetCharacterMouseDownAction",
	(Comp_t) SetCharacterMouseDownAction, T_A, "newAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetCharacterButtonAction", (Comp_t) SetCharacterButtonAction,
	T_A, "newAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetCharacterIdleAction", (Comp_t) SetCharacterIdleAction,
	T_A, "newAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetCharacterActiveAction", (Comp_t) SetCharacterActiveAction,
	T_A, "newAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetCharacterEffectDoneAction",
	  (Comp_t) SetCharacterEffectDoneAction,
	  T_A, "newAction", T_A);
    LatestProc->pr_wizard = B_TRUE;
    r_bi2("MakeWizard", (Comp_t) MakeWizard, T_V,
	"who", T_P, "doNow", T_B);
    r_bi2("MakeApprentice", (Comp_t) MakeApprentice, T_V,
	"who", T_P, "doNow", T_B);
    r_bi1("MakeNormal", (Comp_t) MakeNormal, T_V, "who", T_P);
    r_bi0("SetEffectiveToNone", (Comp_t) SetEffectiveToNone, T_V);
    r_bi0("SetEffectiveToReal", (Comp_t) SetEffectiveToReal, T_V);
    r_bi1("SetEffectiveTo", (Comp_t) SetEffectiveTo, T_V,
	"thePlayer", T_P);
    r_bi0("ClientsActive", (Comp_t) ClientsActive, T_B);
    r_bi1("ShowClients", (Comp_t) ShowClients, T_I, "longForm", T_B);
    r_bi1("ShowCharacters", (Comp_t) ShowCharacters, T_I,
	"longForm", T_B);
    r_bi1("ShowCharacter", (Comp_t) ShowCharacter, T_V, "who", T_P);
    r_bi1("BootClient", (Comp_t) BootClient, T_V, "who", T_P);
    r_bi1("NukeClient", (Comp_t) NukeClient, T_V, "who", T_P);
    r_bi1("Log", (Comp_t) log, T_V, "message", T_S);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("StringToAction", (Comp_t) StringToAction, T_A, "str", T_S);
    r_bi1("StringToProc", (Comp_t) StringToProc, T_A, "str", T_S);
    r_bi1("ProcToString", (Comp_t) ProcToString, T_S, "theProc", T_A);
    bi4("EditString", (Comp_t) EditString, T_V,
	"str", T_S, "handler", T_A, "mode", T_I, "prompt", T_S);
    bi3("GetString", (Comp_t) GetString, T_V,
	"initial", T_S, "handler", T_A, "prompt", T_S);
    r_bi2("QueryFile", (Comp_t) QueryFile, T_V,
	"path", T_S, "handler", T_A);
    r_bi1("FindKey", (Comp_t) FindKey, T_V, "key", T_I);
    r_bi1("DumpThing", (Comp_t) DumpThing, T_V, "key", T_T);
    r_bi1("Trace", (Comp_t) TraceEntry, T_V, "n", T_I);
    LatestProc->pr_wizard = B_TRUE;
    r_bi1("SetMachinesActive", (Comp_t) SetMachinesActive, T_V,
	"state", T_B);
    r_bi1("DescribeKey", (Comp_t) DescribeKey, T_V, "key", T_I);
    r_bi1("Compile", (Comp_t) Compile, T_V, "theAction", T_A);
    r_bi1("UnCompile", (Comp_t) UnCompile, T_V, "theAction", T_A);
    r_bi1("StripBody", (Comp_t) StripBody, T_V, "theAction", T_A);
    r_bi1("Disassemble", (Comp_t) Disassemble, T_V, "theAction", T_A);

    symbolMachineFileBuiltins();
    effectsBuiltins();

    r_bi1("DumpBlocks", (Comp_t) DumpBlocks, T_V, "where", T_S);
    r_bi1("ShowBlocks", (Comp_t) ShowBlocks, T_V, "where", T_S);
    r_bi1("MemValidate", (Comp_t) MemValidate, T_V, "where", T_S);
    r_bi0("ShowDBStatus", (Comp_t) ShowDBStatus, T_V);
}
