#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * Characters.c - some builtins dealing with characters.
 */

/*
 * Character - find the indicated character, if any.
 */

ULONG_T
Character(CHAR_T *name)
{
    Entry_t *en;

    TRACE("Character", name);
    en = symLookupE(PlayerTable->tl_this, name, strLen1(name));
    if (en == P_NULL || (en->en_key >> TYPE_SHIFT) != TYPE_PLAYER) {
	return NIL_PLAYER;
    } else {
	return en->en_key;
    }
}

/*
 * CharacterThing - return the thing associated with the given character.
 *	The character does NOT have to be active.
 */

ULONG_T
CharacterThing(ULONG_T plKey)
{
    Player_t *pl;

    TRACE("CharacterThing", plKey);
    if (plKey == NIL_PLAYER) {
	runError("nil character to CharacterThing");
	return NIL_THING;
    } else {
	pl = ioRead(plKey);
	return pl->pl_key;
    }
}

/*
 * CharacterTable - return the private table for the given character.
 */

ULONG_T
CharacterTable(ULONG_T plKey)
{
    Player_t *pl;

    TRACE("CharacterTable", plKey);
    if (plKey == NIL_PLAYER) {
	runError("nil character to CharacterTable");
	return NIL_TABLE;
    } else {
	pl = ioRead(plKey);
	return pl->pl_symbolTableKey;
    }
}

/*
 * ThingCharacter - return the character (or nil) associated with
 *	the passed thing. The character does not have to be active.
 */

ULONG_T
ThingCharacter(ULONG_T thKey)
{
    Client_t *cl;
    Player_t *pl;
    Entry_t *en;
    UINT_T count;

    if (thKey == NIL_THING) {
	runError("nil thing to ThingCharacter");
	return NIL_PLAYER;
    }

    cl = getClientByThing(thKey);
    if (cl != P_NULL) {
	return cl->cl_player;
    }

    /* have to search the hard way */
    en = PlayerTable->tl_this->tb_entries;
    count = PlayerTable->tl_this->tb_entryCount;
    while (count != 0) {
	count -= 1;
	if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE &&
	    (en->en_key >> TYPE_SHIFT) == TYPE_PLAYER)
	{
	    pl = ioRead(en->en_key);
	    if (pl->pl_key == thKey) {
		return en->en_key;
	    }
	}
	++en;
    }
    return NIL_PLAYER;
}

/*
 * CharacterLocation - return the location of the given character - the
 *	character does NOT need to be active.
 */

ULONG_T
CharacterLocation(ULONG_T plKey)
{
    Client_t *cl;
    Player_t *pl;

    TRACE("CharacterLocation", plKey);
    if (plKey == NIL_PLAYER) {
	runError("nil character to CharacterLocation");
	return NIL_THING;
    }

    cl = getClientByPlayer(plKey);
    if (cl != P_NULL) {
	/* character is active */
	return cl->cl_location;
    }

    /* character is not active */
    pl = ioRead(plKey);
    return pl->pl_location;
}

/*
 * SetCharacterLocation - set the current location of the given character.
 *	The character does NOT need to be active.
 */

void
SetCharacterLocation(ULONG_T plKey, ULONG_T locKey)
{
    Client_t *cl;
    Player_t *pl;

    TRACE("SetCharacterLocation", plKey);
    if (plKey == NIL_PLAYER) {
	runError("nil character to SetCharacterLocation");
	return;
    }

    incRef(locKey);

    cl = getClientByPlayer(plKey);
    if (cl != P_NULL) {
	/* character is active */
	decRef(cl->cl_location);
	locUnlinkClient(cl, cl->cl_location);
	locInsertClient(cl, locKey);
	cl->cl_location = locKey;
	return;
    }

    /* character is not active */
    pl = ioRead(plKey);
    decRef(pl->pl_location);
    pl->pl_location = locKey;
    ioDirty(plKey);
}

/*
 * Me - the builtin to return the character's thing.
 */

ULONG_T
Me(void)
{

    if (ActiveClient != P_NULL && ActiveClient->cl_player != NIL_PLAYER) {
	return ActiveClient->cl_thing;
    } else if (ActiveMachine != P_NULL) {
	return ActiveMachine->ml_thing;
    } else {
	return NIL_THING;
    }
}

/*
 * TrueMe - the builtin to return the thing of the active character or
 *	machine if any.
 *	This is not effected by routines which change ActiveClient.
 */

ULONG_T
TrueMe(void)
{

    if (TrueClient != P_NULL && TrueClient->cl_player != NIL_PLAYER) {
	return TrueClient->cl_thing;
    } else if (TrueMachine != P_NULL) {
	return TrueMachine->ml_thing;
    } else {
	return NIL_THING;
    }
}

/*
 * GetMeString - return the current MeString value.
 */

CHAR_T *
GetMeString(void)
{
    return stringCopy(&MeString[0]);
}

/*
 * SetMeString - set a string that FindAgent will substitute for 'me', etc.
 */

void
SetMeString(CHAR_T *p)
{

    strNCpy(&MeString[0], p, ME_STRING_LEN);
}

/*
 * MeCharacter - builtin to return the current character (if a machine is
 *	running, then the owner of that machine is returned).
 */

ULONG_T
MeCharacter(void)
{
    Client_t *cl;

    cl = ActiveClient;
    if (cl != P_NULL) {
	return cl->cl_player;
    } else if (ActiveMachine != P_NULL) {
	return ActiveMachine->ml_owner;
    } else {
	return NIL_PLAYER;
    }
}

/*
 * SetLocation - builtin to set the current character's location.
 */

void
SetLocation(ULONG_T where)
{
    Client_t *cl;
    MachineList_t *ml;

    TRACE("SetLocation", where);
    cl = ActiveClient;
    if (cl != P_NULL) {
	if (cl->cl_player != NIL_PLAYER) {
	    incRef(where);
	    decRef(cl->cl_location);
	    locUnlinkClient(cl, cl->cl_location);
	    cl->cl_location = where;
	    locInsertClient(cl, where);
	}
    } else {
	ml = ActiveMachine;
	if (ml != P_NULL) {
	    incRef(where);
	    decRef(ml->ml_location);
	    locUnlinkMachine(ml, ml->ml_location);
	    ml->ml_location = where;
	    locInsertMachine(ml, where);
	    ml->ml_dirty = B_TRUE;
	}
    }
}

/*
 * Here - the builtin to return the character's location.
 */

ULONG_T
Here(void)
{

    if (ActiveClient != P_NULL && ActiveClient->cl_player != NIL_PLAYER) {
	return ActiveClient->cl_location;
    } else if (ActiveMachine != P_NULL) {
	return ActiveMachine->ml_location;
    } else {
	return NIL_THING;
    }
}

/*
 * SetNewCharacterAction - set the action to be used on character creation.
 */

ULONG_T
SetNewCharacterAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetNewCharacterAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can SetNewCharacterAction");
	return NIL_ACTION;
    } else {
	oldKey = getEntryAction();
	if (key == NIL_ACTION || checkAction(key, "SetNewCharacterAction")) {
	    incProcRef(key);
	    setEntryAction(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * SetCharacterInputAction - set the input action for the current character.
 */

ULONG_T
SetCharacterInputAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetCharacterInputAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	return NIL_ACTION;
    } else {
	oldKey = ActiveClient->cl_inputAction;
	if (key == NIL_ACTION ||
	    checkStringAction(key, "SetCharacterInputAction"))
	{
	    ActiveClient->cl_inputAction = key;
	    incProcRef(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * SetCharacterRawKeyAction - set the raw key action for the current character.
 */

ULONG_T
SetCharacterRawKeyAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetCharacterRawKeyAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	return NIL_ACTION;
    } else {
	oldKey = ActiveClient->cl_rawKeyAction;
	if (key == NIL_ACTION ||
	    checkIntAction(key, "SetCharacterRawKeyAction"))
	{
	    ActiveClient->cl_rawKeyAction = key;
	    incProcRef(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * SetCharacterMouseDownAction - set mouse action for the current character.
 */

ULONG_T
SetCharacterMouseDownAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetCharacterMouseDownAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	return NIL_ACTION;
    } else {
	oldKey = ActiveClient->cl_mouseDownAction;
	if (key == NIL_ACTION ||
	    checkInt3Action(key, "SetCharacterMouseDownAction"))
	{
	    ActiveClient->cl_mouseDownAction = key;
	    incProcRef(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * SetCharacterButtonAction - set the button action for the current character.
 */

ULONG_T
SetCharacterButtonAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetCharacterButtonAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	return NIL_ACTION;
    } else {
	oldKey = ActiveClient->cl_buttonAction;
	if (key == NIL_ACTION ||
	    checkIntAction(key, "SetCharacterButtonAction"))
	{
	    ActiveClient->cl_buttonAction = key;
	    incProcRef(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * SetCharacterIdleAction - set the handler that is called when the current
 *	character leaves the game.
 */

ULONG_T
SetCharacterIdleAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetCharacterIdleAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	return NIL_ACTION;
    } else {
	oldKey = ActiveClient->cl_idleAction;
	if (key == NIL_ACTION || checkAction(key, "SetCharacterIdleAction")) {
	    ActiveClient->cl_idleAction = key;
	    incProcRef(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * SetCharacterActiveAction - set the handler that is called when this
 *	character re-enters the game.
 */

ULONG_T
SetCharacterActiveAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetCharacterActiveAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	return NIL_ACTION;
    } else {
	oldKey = ActiveClient->cl_activeAction;
	if (key == NIL_ACTION || checkAction(key, "SetCharacterActiveAction"))
	{
	    ActiveClient->cl_activeAction = key;
	    incProcRef(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * SetCharacterEffectDoneAction - set the handler that is called when a
 *	non-immediate effect completes.
 */

ULONG_T
SetCharacterEffectDoneAction(ULONG_T key)
{
    ULONG_T oldKey;

    TRACE("SetCharacterEffectDoneAction", key);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	return NIL_ACTION;
    } else {
	oldKey = ActiveClient->cl_effectDoneAction;
	if (key == NIL_ACTION ||
	    checkInt2Action(key, "SetCharacterEffectDoneAction"))
	{
	    ActiveClient->cl_effectDoneAction = key;
	    incProcRef(key);
	    decProcRefAndNil(&oldKey);
	}
	return oldKey;
    }
}

/*
 * doCreateCharacter - common code for CreateCharacter and buildPlayer.
 */

void
doCreateCharacter(Player_t *pl)
{
    Thing_t th;
    Table_t tb;
    BYTE_T *ptr;
    UINT_T len;

    th.th_parent = NIL_THING;
    th.th_owner = SYSADMIN_KEY;
    th.th_useCount = 1;
    th.th_propCount = 0;
    th.th_status = ts_public;
    pl->pl_key = ioCreate(TYPE_THING, &th, sizeof(Thing_t));
    ThingsCreated += 1;
    symCreateE(&MS, &tb, 2);
    symWriteE(&MS, &tb);
    writeResultE(&MS, &ptr, &len);
    pl->pl_symbolTableKey = ioCreate(TYPE_TABLE, ptr, len);
    symDestroyE(&MS, &tb);
    pl->pl_sponsor = SYSADMIN_KEY;
    pl->pl_location = NIL_THING;
    pl->pl_inputAction = NIL_ACTION;
    pl->pl_rawKeyAction = NIL_ACTION;
    pl->pl_mouseDownAction = NIL_ACTION;
    pl->pl_buttonAction = NIL_ACTION;
    pl->pl_promptKey = ioCreate(TYPE_STRING, "input> ", 8);
    pl->pl_lastOn = GetCurrentTime();
    pl->pl_logonCount = 0;
    pl->pl_totalSeconds = 0;
    pl->pl_idleAction = NIL_ACTION;
    pl->pl_activeAction = NIL_ACTION;
    pl->pl_effectDoneAction = NIL_ACTION;
    pl->pl_textWidth = DEFAULT_TEXT_WIDTH;
    pl->pl_textHeight = DEFAULT_TEXT_HEIGHT;
    pl->pl_flags = 0x0000;
    pl->pl_status = ps_normal;
    pl->pl_wizardMode = B_FALSE;
    pl->pl_newPlayer = B_TRUE;
}

/*
 * CreateCharacter - try to satisfy a new character request.
 */

void
CreateCharacter(CHAR_T *name, CHAR_T *password)
{
    CHAR_T *p;
    Entry_t *en;
    CHAR_T ch;
    Player_t pl;
    Table_t tb;
    ULONG_T key;
    BYTE_T *ptr;
    UINT_T nameLen, len, len2;

    TRACE("CreateCharacter", name);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can create characters");
    } else if (*name == '\0') {
	runError("empty name to CreateCharacter");
    } else {
	p = name;
	do {
	    ch = *p++;
	} while (ch != '\0' && ch != ' ' && ch != '\t' &&
	    ch != '.' && ch != ';' && ch != ',' && ch != '\"');
	if (ch != '\0') {
	    symbolError(name,
		"no spaces or punctuation allowed in character name");
	} else {
	    nameLen = strLen1(name);
	    if (nameLen > PLAYER_NAME_LEN) {
		symbolError(name, "character name too long");
	    } else {
		en = symLookupE(PlayerTable->tl_this, name, nameLen);
		if (en != P_NULL) {
		    /* already exists */
		    symbolError(name, "character name already in use");
		} else {
		    len2 = strLen1(password);
		    if (len2 > PASSWORD_LEN) {
			len2 = PASSWORD_LEN;
		    }
		    copy(&pl.pl_password[0], password, len2);
		    doCreateCharacter(&pl);
		    key = ioCreate(TYPE_PLAYER, &pl, sizeof(Player_t));
		    en = symEnterE(&MS, PlayerTable->tl_this, name, nameLen);
		    en->en_key = key;
		    (void) symReadE(&MS, &tb, ioRead(pl.pl_symbolTableKey));
		    tb.tb_owner = key;
		    tb.tb_useCount = 1;
		    tb.tb_dirty = B_TRUE;
		    symWriteE(&MS, &tb);
		    writeResultE(&MS, &ptr, &len);
		    ioReplace(pl.pl_symbolTableKey, ptr, len);
		    symDestroyE(&MS, &tb);
		    key = EffectivePlayer;
		    EffectivePlayer = SYSADMIN_KEY;
		    addProp(pl.pl_key, NAME_KEY, (ULONG_T) name, TYPE_STRING);
		    EffectivePlayer = key;
		}
	    }
	}
    }
}

/*
 * DestroyCharacter.
 */

void
DestroyCharacter(CHAR_T *name)
{
    Entry_t *en;
    Client_t *cl;
    ULONG_T key;
    UINT_T nameLen;

    TRACE("DestroyCharacter", name);
    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can destroy characters");
    } else {
	nameLen = strLen1(name);
	en = symLookupE(PlayerTable->tl_this, name, nameLen);
	if (en == P_NULL || en->en_key >> TYPE_SHIFT != TYPE_PLAYER) {
	    symbolError(name, "is not a character");
	} else {
	    key = en->en_key;
	    if (key == SYSADMIN_KEY) {
		runError("cannot destroy SysAdmin");
	    } else {
		cl = getClientByPlayer(key);
		if (cl != P_NULL) {
		    cl->cl_destroy = B_TRUE;
		    bootClient(cl);
		} else {
		    nukePlayer(key);
		    removeNameFromMap(PlayerTable->tl_nameMap, key);
		    symDeleteE(&MS, PlayerTable->tl_this, name, nameLen);
		}
	    }
	}
    }
}

/*
 * SetPrompt - set a new prompt for a character - return the old one.
 */

CHAR_T *
SetPrompt(CHAR_T *newPrompt)
{
    CHAR_T *oldPrompt;
    UINT_T len;

    if (ActiveClient != P_NULL && ActiveClient->cl_player != NIL_PLAYER) {
	len = strLen1(newPrompt);
	if (len > 30) {
	    runError("prompt too long");
	    return nullString();
	} else {
	    oldPrompt = ActiveClient->cl_prompt;
	    ActiveClient->cl_prompt = stringCopy(newPrompt);
	    setNewPrompt(newPrompt);
	    return oldPrompt;
	}
    } else {
	return nullString();
    }
}

/*
 * NewCharacterPassword
 */

void
NewCharacterPassword(void)
{

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER) {
	runError("machines cannot NewCharacterPassword");
    } else {
	ActiveClient->cl_playerPassword = B_TRUE;
	startChangePassword();
    }
}

/*
 * ChangeName - attempt to change the name of the current player.
 */

void
ChangeName(CHAR_T *newName)
{
    CHAR_T *p;
    Entry_t *en;
    Client_t *cl;
    UINT_T nameLen;
    CHAR_T ch;

    p = newName;
    do {
	ch = *p++;
    } while (ch != '\0' && ch != ' ' && ch != '\t' &&
	ch != '.' && ch != ';' && ch != ',' && ch != '\"');
    cl = ActiveClient;
    if (*newName == '\0') {
	runError("empty character name not allowed");
    } else if (ch != '\0') {
	symbolError(newName,
	    "no spaces or punctuation allowed in character name");
    } else if (cl == P_NULL || cl->cl_player == NIL_PLAYER) {
	runError("only active characters can change their name");
    } else {
	nameLen = strLen1(newName);
	if (nameLen > PLAYER_NAME_LEN) {
	    symbolError(newName, "character name too long");
	} else {
	    en = symLookupE(PlayerTable->tl_this, newName, nameLen);
	    if (en != P_NULL) {
		/* already exists */
		symbolError(newName, "character name already in use");
	    } else {
		symDeleteE(&MS, PlayerTable->tl_this, cl->cl_name,
			   strLen1(cl->cl_name));
		en = symEnterE(&MS, PlayerTable->tl_this, newName, nameLen);
		en->en_key = cl->cl_player;
		removeNameFromMap(PlayerTable->tl_nameMap, cl->cl_player);
		addProp(cl->cl_thing, NAME_KEY, (ULONG_T) newName,
			TYPE_STRING);
		log(cl->cl_name);
		log(" changed name to ");
		log(newName);
		log("\n");
		freeString(cl->cl_name);
		cl->cl_name = stringCopy(newName);
	    }
	}
    }
}

/*
 * ZapCharacter - the builtin.
 */

void
ZapCharacter(ULONG_T plKey)
{
    Client_t *cl;
    Player_t *pl;
    Player_t plCopy;

    if (plKey == NIL_PLAYER) {
	runError("nil character to ZapCharacter");
    } else {
	cl = getClientByPlayer(plKey);
	if (cl != P_NULL) {
	    /* character is active */
	    if (cl->cl_inserted) {
		fullyUnlinkClient(cl);
	    }
	    decRef(cl->cl_location);
	    cl->cl_location = NIL_THING;
	    decProcRef(cl->cl_inputAction);
	    cl->cl_inputAction = NIL_ACTION;
	    decProcRef(cl->cl_rawKeyAction);
	    cl->cl_rawKeyAction = NIL_ACTION;
	    decProcRef(cl->cl_mouseDownAction);
	    cl->cl_mouseDownAction = NIL_ACTION;
	    decProcRef(cl->cl_buttonAction);
	    cl->cl_buttonAction = NIL_ACTION;
	    decProcRef(cl->cl_idleAction);
	    cl->cl_idleAction = NIL_ACTION;
	    decProcRef(cl->cl_activeAction);
	    cl->cl_activeAction = NIL_ACTION;
	    decProcRef(cl->cl_effectDoneAction);
	    cl->cl_effectDoneAction = NIL_ACTION;
	    cl->cl_newPlayer = B_TRUE;
	    ClearThing(cl->cl_thing);
	} else {
	    /* character is not active */
	    pl = ioRead(plKey);
	    /* do this so that we don't have to worry about the decRef's
	       taking the Player_t out of the cache */
	    plCopy = *pl;
	    decRef(plCopy.pl_location);
	    plCopy.pl_location = NIL_THING;
	    decProcRef(plCopy.pl_inputAction);
	    plCopy.pl_inputAction = NIL_ACTION;
	    decProcRef(plCopy.pl_rawKeyAction);
	    plCopy.pl_rawKeyAction = NIL_ACTION;
	    decProcRef(plCopy.pl_mouseDownAction);
	    plCopy.pl_mouseDownAction = NIL_ACTION;
	    decProcRef(plCopy.pl_buttonAction);
	    plCopy.pl_buttonAction = NIL_ACTION;
	    decProcRef(plCopy.pl_idleAction);
	    plCopy.pl_idleAction = NIL_ACTION;
	    decProcRef(plCopy.pl_activeAction);
	    plCopy.pl_activeAction = NIL_ACTION;
	    decProcRef(plCopy.pl_effectDoneAction);
	    plCopy.pl_effectDoneAction = NIL_ACTION;
	    plCopy.pl_newPlayer = B_TRUE;
	    ClearThing(plCopy.pl_key);
	    ioReplace(plKey, &plCopy, sizeof(Player_t));
	}
    }
}
