#include <ctype.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * clientFuncs.c - builtins for clients
 */

#define LOC_CHECKS	1

/*
 * getClientByThing - given a key, find any Client_t structure.
 */

Client_t *
getClientByThing(ULONG_T key)
{
    Client_t *cl;

    cl = ClientKeyMap[key % AGENT_MAP_SIZE];
    while (cl != P_NULL && cl->cl_thing != key) {
	cl = cl->cl_nextKHash;
    }
    return cl;
}

/*
 * getMachineByThing - given a key, find any MachineList_t structure.
 */

MachineList_t *
getMachineByThing(ULONG_T key)
{
    MachineList_t *ml;

    ml = MachineKeyMap[key % AGENT_MAP_SIZE];
    while (ml != P_NULL && ml->ml_thing != key) {
	ml = ml->ml_nextKHash;
    }
    return ml;
}

/*
 * getClientByPlayer - find a Client_t given the key of the player
 */

Client_t *
getClientByPlayer(ULONG_T plKey)
{
    Client_t *cl;

    cl = ClientPKeyMap[plKey % AGENT_MAP_SIZE];
    while (cl != P_NULL && cl->cl_player != plKey) {
	cl = cl->cl_nextPHash;
    }
    return cl;
}

/*
 * findClientsAt - get a pointer to any ring of Client_t's at the given loc.
 */

Client_t *
findClientsAt(ULONG_T location)
{
    Client_t *cl, *clFirst;

    if (location == NIL_THING) {
	return P_NULL;
    }
    cl = ClientLocMap[location % AGENT_MAP_SIZE];
    if (cl == P_NULL) {
	return P_NULL;
    }
    clFirst = cl;
    do {
	if (cl->cl_location == location) {
	    return cl;
	}
	cl = cl->cl_nextLHash;
    } while (cl != clFirst);
    return P_NULL;
}

/*
 * findMachinesAt - get a pointer to any ring of MachineList_t's at given loc.
 */

MachineList_t *
findMachinesAt(ULONG_T location)
{
    MachineList_t *ml, *mlFirst;

    if (location == NIL_THING) {
	return P_NULL;
    }
    ml = MachineLocMap[location % AGENT_MAP_SIZE];
    if (ml == P_NULL) {
	return P_NULL;
    }
    mlFirst = ml;
    do {
	if (ml->ml_location == location) {
	    return ml;
	}
	ml = ml->ml_nextLHash;
    } while (ml != mlFirst);
    return P_NULL;
}

/*
 * FindAgentAt - find an active agent with the given name, in the given
 *	location.
 */

ULONG_T
FindAgentAt(ULONG_T location, CHAR_T *name)
{
    Entry_t *en;
    Client_t *cl, *cl2;
    MachineList_t *ml1, *ml;
    ULONG_T key;

    TRACE("FindAgentAt", location);
#if 1
    /* Try player characters first. */
    cl = findClientsAt(location);
    if (cl != P_NULL) {
	/* At least one client is at that location. */
	en = symLookupE(PlayerTable->tl_this, name, strLen1(name));
	if (en != P_NULL) {
	    key = en->en_key;
	    if ((key >> TYPE_SHIFT) == TYPE_PLAYER) {
		/* The name given is a valid player character name */
		cl2 = cl;
		do {
		    if (cl2->cl_thing == key) {
			/* That player is here. */
			if (cl2->cl_deleted) {
			    return NIL_THING;
			}
			return key;
		    }
		    cl2 = cl2->cl_nextLoc;
		} while (cl2 != cl);
	    }
	}
	/* No luck with name lookup. Try name matching. */
	cl2 = cl;
	do {
	    if (MatchName(cl2->cl_name, name) != -1) {
		if (cl2->cl_deleted) {
		    return NIL_THING;
		}
		return cl2->cl_thing;
	    }
	    cl2 = cl2->cl_nextLoc;
	} while (cl2 != cl);
    }

    /* No client matches. Try machines. */
    ml = findMachinesAt(location);
    if (ml != P_NULL) {
	/* At least one machine there. */
	ml1 = ml;
	do {
	    if (! ml->ml_deleted && ml->ml_name != P_NULL &&
		MatchName(ml->ml_name, name) != -1)
	    {
		return ml->ml_thing;
	    }
	    ml = ml->ml_nextLoc;
	} while (ml != ml1);
    }

    return NIL_THING;
#else
    en = symLookupE(PlayerTable->tl_this, name, strLen1(name));
    if (en != P_NULL) {
	key = en->en_key;
	if ((key >> TYPE_SHIFT) == TYPE_PLAYER) {
	    cl = Clients;
	    while (cl != P_NULL && cl->cl_player != key) {
		cl = cl->cl_next;
	    }
	    if (cl != P_NULL && cl->cl_location == location &&
		! cl->cl_deleted)
	    {
		key = cl->cl_thing;
	    } else {
		key = NIL_THING;
	    }
	} else {
	    key = NIL_THING;
	}
    } else {
	/* not found the easy way - have to scan the machine list */
	ml = Machines;
	while (ml != P_NULL &&
	    (ml->ml_deleted || ml->ml_location != location ||
	     ml->ml_name == P_NULL || MatchName(ml->ml_name, name) == -1))
	{
	    ml = ml->ml_next;
	}
	if (ml != P_NULL) {
	    key = ml->ml_thing;
	}
    }
    if (key == NIL_THING) {
	/* not found there either - check the whole client list */
	cl = Clients;
	while (cl != P_NULL && (cl->cl_location != location ||
	    strCmpLC(cl->cl_name, name) != 0))
	{
	    cl = cl->cl_next;
	}
	if (cl != P_NULL && cl->cl_player != NIL_PLAYER) {
	    key = cl->cl_thing;
	}
    }
    return key;
#endif
}

/*
 * FindAgent - find an active agent with the given name, in the same
 *	location as the current agent.
 */

ULONG_T
FindAgent(CHAR_T *name)
{

    if (strCmpLC(name, "me") == 0 || strCmpLC(name, "myself") == 0 ||
	strCmpLC(name, "yourself") == 0 || strCmpLC(name, "self") == 0)
    {
	if (MeString[0] != '\0') {
	    name = &MeString[0];
	} else if (ActiveClient != P_NULL &&
	    ActiveClient->cl_player != NIL_PLAYER)
	{
	    return ActiveClient->cl_thing;
	} else if (ActiveMachine != P_NULL) {
	    return ActiveMachine->ml_thing;
	} else {
	    return NIL_THING;
	}
    }

    if (ActiveClient != P_NULL && ActiveClient->cl_player != NIL_PLAYER) {
	return FindAgentAt(ActiveClient->cl_location, name);
    } else if (ActiveMachine != P_NULL) {
	return FindAgentAt(ActiveMachine->ml_location, name);
    } else {
	return NIL_THING;
    }
}

/*
 * DescribeKey - the dangerous SysAdmin only peeker.
 */

void
DescribeKey(ULONG_T key)
{

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can DescribeKey");
    } else {
	describeValue(key, findAnyName(key));
    }
}

#if LOC_CHECKS

#if 0

static void
clientLocChecks(char *where)
{
    Client_t *cl, *clFirst, *cl2;
    ULONG_T loc;
    unsigned int i;

    for (i = 0; i != AGENT_MAP_SIZE; i += 1) {
	cl = ClientLocMap[i];
	if (cl != P_NULL) {
	    clFirst = cl;
	    do {
		loc = cl->cl_location;
		if (loc % AGENT_MAP_SIZE != i) {
		    logX("loc ", loc, " index ");
		    logN(P_NULL, loc % AGENT_MAP_SIZE, " not ");
		    logN(P_NULL, i, "\n");
		    myAbort(where);
		}
		cl2 = cl->cl_nextLoc;
		do {
		    if (cl2->cl_location != loc) {
			logX("loc ", loc, " not ");
			logX(P_NULL, loc, "\n");
			myAbort(where);
		    }
		    cl2 = cl2->cl_nextLoc;
		} while (cl2 != cl);
		cl = cl->cl_nextLHash;
		if (cl == P_NULL) {
		    log("cl nil\n");
		    myAbort(where);
		}
	    } while (cl != clFirst);
	}
    }
}

#endif

static void
machineLocChecks(char *where)
{
    MachineList_t *ml, *mlFirst, *ml2;
    ULONG_T loc;
    unsigned int i, count1, count2;

    for (i = 0; i != AGENT_MAP_SIZE; i += 1) {
	ml = MachineLocMap[i];
	if (ml != P_NULL) {
	    mlFirst = ml;
	    count1 = 0;
	    do {
		loc = ml->ml_location;
		if (loc % AGENT_MAP_SIZE != i) {
		    logX("loc ", loc, " index ");
		    logN(P_NULL, loc % AGENT_MAP_SIZE, " not ");
		    logN(P_NULL, i, "\n");
		    myAbort(where);
		}
		ml2 = ml->ml_nextLoc;
		count2 = 0;
		do {
		    if (ml2->ml_location != loc) {
			logX("loc ", loc, " not ");
			logX(P_NULL, loc, "\n");
			myAbort(where);
		    }
		    ml2 = ml2->ml_nextLoc;
		    count2 += 1;
		    if (count2 >= 10000) {
			myAbort(where);
		    }
		} while (ml2 != ml);
		ml = ml->ml_nextLHash;
		if (ml == P_NULL) {
		    log("ml nil\n");
		    myAbort(where);
		}
		count1 += 1;
		if (count1 >= 10000) {
		    myAbort(where);
		}
	    } while (ml != mlFirst);
	}
    }
}

#endif

/*
 * locUnlinkClient - unlink a Client_t from location rings.
 */

void
locUnlinkClient(Client_t *cl, ULONG_T oldLoc)
{
    Client_t *nextL, *nextH, *prevH;

    if (oldLoc == NIL_THING) {
	return;
    }

    oldLoc %= AGENT_MAP_SIZE;
    nextL = cl->cl_nextLoc;
    nextH = cl->cl_nextLHash;
    if (nextL != cl) {
	/* There is more than one client at this loc. */
	nextL->cl_prevLoc = cl->cl_prevLoc;
	cl->cl_prevLoc->cl_nextLoc = nextL;
	if (nextH != P_NULL) {
	    /* This one is the one here on the hash chain. */
	    if (nextH != cl) {
		/* Its not the only one on the hash chain. */
		prevH = cl->cl_prevLHash;
		nextH->cl_prevLHash = nextL;
		prevH->cl_nextLHash = nextL;
		nextL->cl_prevLHash = prevH;
		nextL->cl_nextLHash = nextH;
		if (ClientLocMap[oldLoc] == cl) {
		    ClientLocMap[oldLoc] = nextL;
		}
	    } else {
		nextL->cl_nextLHash = nextL;
		nextL->cl_prevLHash = nextL;
		ClientLocMap[oldLoc] = nextL;
	    }
	}
    } else {
	/* This was the only client at this location. */
	if (nextH != cl) {
	    prevH = cl->cl_prevLHash;
	    nextH->cl_prevLHash = prevH;
	    prevH->cl_nextLHash = nextH;
	    if (ClientLocMap[oldLoc] == cl) {
		ClientLocMap[oldLoc] = nextH;
	    }
	} else {
	    ClientLocMap[oldLoc] = P_NULL;
	}
    }
}

/*
 * locInsertClient - do all of the pointer fiddling for when a Client_t
 *	is inserted into location rings.
 */

void
locInsertClient(Client_t *cl, ULONG_T newLoc)
{
    Client_t *firstH, *thisH, *nextH, *prevH, *nextL;

    if (newLoc == NIL_THING) {
	return;
    }

    firstH = ClientLocMap[newLoc % AGENT_MAP_SIZE];
    if (firstH == P_NULL) {
	cl->cl_nextLoc = cl;
	cl->cl_prevLoc = cl;
	cl->cl_nextLHash = cl;
	cl->cl_prevLHash = cl;
    } else {
	thisH = firstH;
	do {
	    if (thisH->cl_location == newLoc) {
		break;
	    }
	    thisH = thisH->cl_nextLHash;
	} while (thisH != firstH);
	if (thisH->cl_location == newLoc) {
	    /* There is at least one Client_t at this location. It turns
	       out that linking in to the loc ring works the same whether
	       there is more than one Client_t in it or not. */
	    nextL = thisH->cl_nextLoc;
	    nextL->cl_prevLoc = cl;
	    cl->cl_nextLoc = nextL;
	    thisH->cl_nextLoc = cl;
	    cl->cl_prevLoc = thisH;
	    nextH = thisH->cl_nextLHash;
	    if (nextH == thisH) {
		/* The existing Client_t in the hash ring was the only
		   one there. Make it be beside us, and put us there. */
		cl->cl_nextLHash = cl;
		cl->cl_prevLHash = cl;
	    } else {
		/* There was more than one Client_t in the hash ring. */
		prevH = thisH->cl_prevLHash;
		nextH->cl_prevLHash = cl;
		prevH->cl_nextLHash = cl;
		cl->cl_nextLHash = nextH;
		cl->cl_prevLHash = prevH;
	    }
	    thisH->cl_nextLHash = P_NULL;
	    thisH->cl_prevLHash = P_NULL;
	} else {
	    nextH = firstH->cl_nextLHash;
	    cl->cl_nextLoc = cl;
	    cl->cl_prevLoc = cl;
	    cl->cl_nextLHash = nextH;
	    cl->cl_prevLHash = firstH;
	    nextH->cl_prevLHash = cl;
	    firstH->cl_nextLHash = cl;
	}
    }
    ClientLocMap[newLoc % AGENT_MAP_SIZE] = cl;
}

/*
 * locUnlinkMachine - unlink a MachineList_t from location rings.
 */

void
locUnlinkMachine(MachineList_t *ml, ULONG_T oldLoc)
{
    MachineList_t *nextL, *nextH, *prevH;

    if (oldLoc == NIL_THING) {
	return;
    }

#if LOC_CHECKS
    machineLocChecks("locUnlinkMachine start");
#endif
    oldLoc %= AGENT_MAP_SIZE;
    nextL = ml->ml_nextLoc;
    nextH = ml->ml_nextLHash;
    if (nextL != ml) {
	/* There is more than one machine at this loc. */
	nextL->ml_prevLoc = ml->ml_prevLoc;
	ml->ml_prevLoc->ml_nextLoc = nextL;
	if (nextH != P_NULL) {
	    /* This one is the one here on the hash chain. */
	    if (nextH != ml) {
		/* Its not the only one on the hash chain. */
		prevH = ml->ml_prevLHash;
		nextH->ml_prevLHash = nextL;
		prevH->ml_nextLHash = nextL;
		nextL->ml_prevLHash = prevH;
		nextL->ml_nextLHash = nextH;
		if (MachineLocMap[oldLoc] == ml) {
		    MachineLocMap[oldLoc] = nextL;
		}
	    } else {
		nextL->ml_nextLHash = nextL;
		nextL->ml_prevLHash = nextL;
		MachineLocMap[oldLoc] = nextL;
	    }
	}
    } else {
	/* This was the only machine at this location. */
	if (nextH != ml) {
	    prevH = ml->ml_prevLHash;
	    nextH->ml_prevLHash = prevH;
	    prevH->ml_nextLHash = nextH;
	    if (MachineLocMap[oldLoc] == ml) {
		MachineLocMap[oldLoc] = nextH;
	    }
	} else {
	    MachineLocMap[oldLoc] = P_NULL;
	}
    }
#if LOC_CHECKS
    machineLocChecks("locUnlinkMachine done");
#endif
}

/*
 * locInsertMachine - do all of the pointer fiddling for when a MachineList_t
 *	is inserted into location rings.
 */

void
locInsertMachine(MachineList_t *ml, ULONG_T newLoc)
{
    MachineList_t *firstH, *thisH, *nextH, *prevH, *nextL;

#if LOC_CHECKS
    machineLocChecks("locInsertMachine start");
#endif
    if (newLoc == NIL_THING) {
	return;
    }

    firstH = MachineLocMap[newLoc % AGENT_MAP_SIZE];
    if (firstH == P_NULL) {
	ml->ml_nextLoc = ml;
	ml->ml_prevLoc = ml;
	ml->ml_nextLHash = ml;
	ml->ml_prevLHash = ml;
    MachineLocMap[newLoc % AGENT_MAP_SIZE] = ml;
#if LOC_CHECKS
    machineLocChecks("locInsertMachine done 1");
#endif
    } else {
	thisH = firstH;
	do {
	    if (thisH->ml_location == newLoc) {
		break;
	    }
	    thisH = thisH->ml_nextLHash;
	} while (thisH != firstH);
	if (thisH->ml_location == newLoc) {
	    /* There is at least one MachineList_t at this location. It turns
	       out that linking in to the loc ring works the same whether
	       there is more than one MachineList_t in it or not. */
	    nextL = thisH->ml_nextLoc;
	    nextL->ml_prevLoc = ml;
	    ml->ml_nextLoc = nextL;
	    thisH->ml_nextLoc = ml;
	    ml->ml_prevLoc = thisH;
	    nextH = thisH->ml_nextLHash;
	    if (nextH == thisH) {
		/* The existing MachineList_t in the hash ring was the only
		   one there. Make it be beside us, and put us there. */
		ml->ml_nextLHash = ml;
		ml->ml_prevLHash = ml;
	    thisH->ml_nextLHash = P_NULL;
	    thisH->ml_prevLHash = P_NULL;
    MachineLocMap[newLoc % AGENT_MAP_SIZE] = ml;
#if LOC_CHECKS
    machineLocChecks("locInsertMachine done 2");
#endif
	    } else {
		/* There was more than one MachineList_t in the hash ring. */
		prevH = thisH->ml_prevLHash;
		nextH->ml_prevLHash = ml;
		prevH->ml_nextLHash = ml;
		ml->ml_nextLHash = nextH;
		ml->ml_prevLHash = prevH;
	    thisH->ml_nextLHash = P_NULL;
	    thisH->ml_prevLHash = P_NULL;
    MachineLocMap[newLoc % AGENT_MAP_SIZE] = ml;
#if LOC_CHECKS
    machineLocChecks("locInsertMachine done 3");
#endif
	    }
	    thisH->ml_nextLHash = P_NULL;
	    thisH->ml_prevLHash = P_NULL;
    MachineLocMap[newLoc % AGENT_MAP_SIZE] = ml;
#if LOC_CHECKS
    machineLocChecks("locInsertMachine done 4");
#endif
	} else {
	    nextH = firstH->ml_nextLHash;
	    ml->ml_nextLoc = ml;
	    ml->ml_prevLoc = ml;
	    ml->ml_nextLHash = nextH;
	    ml->ml_prevLHash = firstH;
	    nextH->ml_prevLHash = ml;
	    firstH->ml_nextLHash = ml;
    MachineLocMap[newLoc % AGENT_MAP_SIZE] = ml;
#if LOC_CHECKS
    machineLocChecks("locInsertMachine done 5");
#endif
	}
    }
}

/*
 * fullyUnlinkClient - unlink a client from all chains except the main one.
 */

void
fullyUnlinkClient(Client_t *cl)
{
    Client_t **pCl;

    if (cl->cl_location != NIL_THING) {
	locUnlinkClient(cl, cl->cl_location);
    }
    if (cl->cl_thing != NIL_THING) {
	pCl = &ClientKeyMap[cl->cl_thing % AGENT_MAP_SIZE];
	while (*pCl != cl) {
	    pCl = &(*pCl)->cl_nextKHash;
	}
	*pCl = cl->cl_nextKHash;
    }
    if (cl->cl_player != NIL_PLAYER) {
	pCl = &ClientPKeyMap[cl->cl_player % AGENT_MAP_SIZE];
	while (*pCl != cl) {
	    pCl = &(*pCl)->cl_nextPHash;
	}
	*pCl = cl->cl_nextPHash;
    }
}

/*
 * fullyInsertClient - insert client into all needed chains.
 */

void fullyInsertClient(Client_t *cl)
{
    Client_t **pCl;

    locInsertClient(cl, cl->cl_location);
    pCl = &ClientKeyMap[cl->cl_thing % AGENT_MAP_SIZE];
    cl->cl_nextKHash = *pCl;
    *pCl = cl;
    pCl = &ClientPKeyMap[cl->cl_player % AGENT_MAP_SIZE];
    cl->cl_nextPHash = *pCl;
    *pCl = cl;
    cl->cl_inserted = B_TRUE;
}

/*
 * fullyInsertMachine - insert client into all needed chains.
 */

void fullyInsertMachine(MachineList_t *ml)
{
    MachineList_t **pMl;

    locInsertMachine(ml, ml->ml_location);
    pMl = &MachineKeyMap[ml->ml_thing % AGENT_MAP_SIZE];
    ml->ml_nextKHash = *pMl;
    *pMl = ml;
#if LOC_CHECKS
    machineLocChecks("end of fullyUnlinkMachine");
#endif
}

/*
 * SetAgentLocation - move an active agent.
 */

void
SetAgentLocation(ULONG_T agent, ULONG_T where)
{
    Client_t *cl;
    MachineList_t *ml;

    TRACE("SetAgentLocation", agent);
    if (agent == NIL_THING) {
	runError("nil agent passed to SetAgentLocation");
    } else {
	cl = getClientByThing(agent);
	if (cl != P_NULL) {
	    if (cl->cl_player != NIL_PLAYER) {
		incRef(where);
		decRef(cl->cl_location);
		locUnlinkClient(cl, cl->cl_location);
		cl->cl_location = where;
		locInsertClient(cl, where);
	    }
	} else {
	    ml = getMachineByThing(agent);
	    if (ml != P_NULL && ! ml->ml_deleted) {
		incRef(where);
		decRef(ml->ml_location);
		locUnlinkMachine(ml, ml->ml_location);
		ml->ml_location = where;
		locInsertMachine(ml, where);
		ml->ml_dirty = B_TRUE;
	    } else {
		runError("agent not found in SetAgentLocation");
	    }
	}
    }
}

/*
 * SetSingleUser - set the FLAG_SINGLE_USER bit for this world.
 */

void
SetSingleUser(ULONG_T flag)
{
    ULONG_T oldFlags;

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can SetSingleUser");
    } else {
	oldFlags = getDungeonFlags();
	if (flag != BOOLEAN_FALSE) {
	    oldFlags |= FLAG_SINGLE_USER;
	} else {
	    oldFlags &= ~FLAG_SINGLE_USER;
	}
	setDungeonFlags(oldFlags);
    }
}

/*
 * SetRemoteSysAdminOK - set the FLAG_REMOTE_SYSADMIN_OK bit for this world.
 */

void
SetRemoteSysAdminOK(ULONG_T flag)
{
    ULONG_T oldFlags;

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can SetRemoteSysAdminOK");
    } else {
	oldFlags = getDungeonFlags();
	if (flag != BOOLEAN_FALSE) {
	    oldFlags |= FLAG_REMOTE_SYSADMIN_OK;
	} else {
	    oldFlags &= ~FLAG_REMOTE_SYSADMIN_OK;
	}
	setDungeonFlags(oldFlags);
    }
}

/*
 * SetMachinesActive - set the FLAG_MACHINES_ACTIVE bit for this world.
 */

void
SetMachinesActive(ULONG_T flag)
{
    ULONG_T oldFlags;

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can SetMachinesActive");
    } else {
	oldFlags = getDungeonFlags();
	if (flag != BOOLEAN_FALSE) {
	    MachinesActive = B_TRUE;
	    oldFlags |= FLAG_MACHINES_ACTIVE;
	} else {
	    MachinesActive = B_FALSE;
	    oldFlags &= ~FLAG_MACHINES_ACTIVE;
	}
	setDungeonFlags(oldFlags);
    }
}

/*
 * Small agent list helpers for ForEachAgent and ForEachClient.
 */

static CallList_t *
addClientCall(CallList_t *list, Client_t *cl, ULONG_T action)
{

    return addCall(list, cl, P_NULL, action, (CHAR_T *) cl->cl_thing);
}

static CallList_t *
addMachineCall(CallList_t *list, MachineList_t *ml, ULONG_T action)
{

    return addCall(list, P_NULL, ml, action, (CHAR_T *) ml->ml_thing);
}

/*
 * ForEachAgent - call the given proc for each agent in the given location.
 *	If the passed location is nil, then do the action for each agent
 *	currently connected. Note: the action is NOT done for the current
 *	agent. The thing of the client or machine is passed to the action.
 *	NOTE2: the action is done as the current agent - we do not switch
 *	to the agents as we scan them.
 */

void
ForEachAgent(ULONG_T location, ULONG_T action)
{
    Client_t *cl, *me;
    MachineList_t *ml, *moi;
    CallList_t *calls;

    TRACE("ForEachAgent", action);
    if (! checkThingAction(action, "ForEachAgent")) {
	return;
    }

    me = ActiveClient;
    if (me != P_NULL) {
	moi = P_NULL;
    } else {
	moi = ActiveMachine;
	if (moi == P_NULL) {
	    return;
	}
    }

    calls = P_NULL;
    if (location == NIL_THING) {
	/* Just do everyone - expensive! */
	cl = Clients;
	while (cl != P_NULL) {
	    if (cl != me && cl->cl_player != NIL_PLAYER) {
		calls = addClientCall(calls, cl, action);
	    }
	    cl = cl->cl_next;
	}
	ml = Machines;
	while (ml != P_NULL) {
	    if (ml != moi && ! ml->ml_deleted) {
		calls = addMachineCall(calls, ml, action);
	    }
	    ml = ml->ml_next;
	}
    } else {
	/* Just do Client_t's and MachineList_t's at the location. */
	if (me != P_NULL && me->cl_location == location) {
	    /* I am a Client_t, and I am at location. */
	    if (me->cl_location != NIL_THING) {
		cl = me->cl_nextLoc;
		while (cl != me) {
		    if (cl->cl_player != NIL_PLAYER) {
			calls = addClientCall(calls, cl, action);
		    }
		    cl = cl->cl_nextLoc;
		}
	    }
	    ml = findMachinesAt(location);
	    if (ml != P_NULL) {
		moi = ml;
		do {
		    if (! ml->ml_deleted) {
			calls = addMachineCall(calls, ml, action);
		    }
		    ml = ml->ml_nextLoc;
		} while (ml != moi);
	    }
	} else if (moi != P_NULL && moi->ml_location == location) {
	    /* I am a MachineList_t, and I am at location */
	    cl = findClientsAt(location);
	    if (cl != P_NULL) {
		me = cl;
		do {
		    if (cl->cl_player != NIL_PLAYER) {
			calls = addClientCall(calls, cl, action);
		    }
		    cl = cl->cl_nextLoc;
		} while (cl != me);
	    }
	    if (moi->ml_location != NIL_THING) {
		ml = moi->ml_nextLoc;
		while (ml != moi) {
		    if (! ml->ml_deleted) {
			calls = addMachineCall(calls, ml, action);
		    }
		    ml = ml->ml_nextLoc;
		}
	    }
	} else {
	    /* OK, no quick shortcut. */
	    cl = findClientsAt(location);
	    if (cl != P_NULL) {
		me = cl;
		do {
		    if (cl->cl_player != NIL_PLAYER) {
			calls = addClientCall(calls, cl, action);
		    }
		    cl = cl->cl_nextLoc;
		} while (cl != me);
	    }
	    ml = findMachinesAt(location);
	    if (ml != P_NULL) {
		moi = ml;
		do {
		    if (! ml->ml_deleted) {
			calls = addMachineCall(calls, ml, action);
		    }
		    ml = ml->ml_nextLoc;
		} while (ml != moi);
	    }
	}
    }
    /* do NOT change real/effective character stuff */
    runCallList(calls, B_FALSE);
}

/*
 * ForEachClient - call the given proc for each active client in the system.
 *	This WILL include the active agent if that agent is a regular client
 *	and not a machine.
 */

void
ForEachClient(ULONG_T action)
{
    Client_t *cl;
    CallList_t *calls;

    TRACE("ForEachClient", action);
    if (checkThingAction(action, "ForEachClient")) {
	calls = P_NULL;
	cl = Clients;
	while (cl != P_NULL) {
	    if (cl->cl_player != NIL_PLAYER) {
		/* do NOT change real/effective character stuff */
		calls = addClientCall(calls, cl, action);
	    }
	    cl = cl->cl_next;
	}
	runCallList(calls, B_FALSE);
    }
}

/*
 * getName - local utility to get the printable form of a name of a
 *	speaker. The result must be freed by the caller.
 */

static CHAR_T *
getName(CHAR_T *altName, ULONG_T key)
{
    CHAR_T *name, *res;
    BOOL_T needFree;

    needFree = B_FALSE;
    if (*altName != '\0') {
	name = altName;
    } else {
	name = (CHAR_T *) getProp(key, NAME_KEY, TYPE_STRING);
	if (name == P_NULL) {
	    /* if nothing else, ZapCharacter will ClearThing the character,
	       which will remove the NAME_KEY */
	    name = "<someone>";
	} else if (*name == '\0') {
	    freeString(name);
	    name = "<someone>";
	} else {
	    needFree = B_TRUE;
	    if (islower(*name)) {
		*name = toupper(*name);
	    }
	}
    }
    res = FormatName(name);
    if (needFree) {
	freeString(name);
    }
    return res;
}

/*
 * sayAll - common code for Say and Pose.
 *	Send the given string to all clients at this location, and also trigger
 *	the given proc on each machine here.
 *	The active agent's name is prefixed.
 */

static void
sayAll(CHAR_T *altName, CHAR_T *separator, CHAR_T *what, BOOL_T isPose)
{
    CHAR_T *name;
    MachineList_t *ml, *moi, *mlFirst;
    Client_t *me;
    CallList_t *calls;
    ULONG_T key, here, action;
    BOOL_T doneCopy;

    me = ActiveClient;
    if (me != P_NULL) {
	if (me->cl_player == NIL_PLAYER) {
	    return;
	}
	here = me->cl_location;
	key = me->cl_thing;
	moi = P_NULL;	/* keep gcc quiet */
    } else {
	moi = ActiveMachine;
	if (moi == P_NULL) {
	    return;
	}
	here = moi->ml_location;
	key = moi->ml_thing;
    }
    name = getName(altName, key);
    messageFlush();
    flushEffects();
    setOthers(B_TRUE);
    putString(name);
    putString(separator);
    putString(what);
    newLine();
    messageFlush();
    setOthers(B_FALSE);

    doneCopy = B_FALSE;
    ml = findMachinesAt(here);
    if (ml != P_NULL) {
	calls = P_NULL;
	mlFirst = ml;
	do {
	    if (isPose) {
		action = ml->ml_poseProc;
	    } else {
		action = ml->ml_sayProc;
	    }
	    if (ml != moi && ! ml->ml_deleted && action != NIL_ACTION) {
		if (! doneCopy) {
		    doneCopy = B_TRUE;
		    separator = concatenate(name, stringCopy(separator));
		    what = concatenate(separator, stringCopy(what));
		}
		calls = addCall(calls, P_NULL, ml, action, stringCopy(what));
	    }
	    ml = ml->ml_nextLoc;
	} while (ml != mlFirst);
	runCallList(calls, B_TRUE);
    }
    if (doneCopy) {
	freeString(what);
    } else {
	freeString(name);
    }
}

/*
 * Say - an agent is saying something out loud.
 */

void
Say(CHAR_T *altName, CHAR_T *what)
{

    TRACE("Say", what);
    sayAll(altName, " says: ", what, B_FALSE);
}

/*
 * Pose - an agent wants to do something like waving.
 */

void
Pose(CHAR_T *altName, CHAR_T *what)
{

    TRACE("Pose", what);
    sayAll(altName, " ", what, B_TRUE);
}

/*
 * sayOne - common code for SayTo and PoseTo.
 */

static void
sayOne(ULONG_T who, CHAR_T *altName, CHAR_T *separator, CHAR_T *what,
       BOOL_T isPose)
{
    Client_t *cl, *saveClient;
    MachineList_t *ml, *saveMachine;
    char *name;
    ULONG_T key, saveRealPlayer, saveEffectivePlayer;
    PlayerStatus_t saveRealStatus, saveEffectiveStatus;

    if (who == NIL_THING) {
	runError("nil who to SayTo/PoseTo");
	return;
    }

    cl = ActiveClient;
    if (cl != P_NULL) {
	if (cl->cl_player == NIL_PLAYER) {
	    return;
	}
	key = cl->cl_thing;
    } else {
	ml = ActiveMachine;
	if (ml == P_NULL) {
	    return;
	}
	key = ml->ml_thing;
    }

    name = getName(altName, key);
    cl = getClientByThing(who);
    if (cl != P_NULL) {
	/* Speaking to another player. */
	setSpecific(who);
	putString(name);
	freeString(name);
	putString(separator);
	putString(what);
	newLine();
	setSpecific(NIL_THING);
    } else {
	ml = getMachineByThing(who);
	if (ml != P_NULL && ! ml->ml_deleted) {
	    /* Speaking to a machine. */
	    if (isPose) {
		key = ml->ml_poseProc;
	    } else {
		key = ml->ml_sayProc;
	    }
	    if (key != NIL_ACTION) {
		what = concatenate(concatenate(name, stringCopy(separator)),
				   stringCopy(what));
		saveClient = ActiveClient;
		saveMachine = ActiveMachine;
		saveRealPlayer = RealPlayer;
		saveRealStatus = RealStatus;
		saveEffectivePlayer = EffectivePlayer;
		saveEffectiveStatus = EffectiveStatus;
		ActiveClient = P_NULL;
		ActiveMachine = ml;
		RealPlayer = ml->ml_owner;
		RealStatus = ps_apprentice;
		EffectivePlayer = ml->ml_owner;
		EffectiveStatus = ps_apprentice;
		doProc1(key, what);
		ActiveClient = saveClient;
		ActiveMachine = saveMachine;
		RealPlayer = saveRealPlayer;
		RealStatus = saveRealStatus;
		EffectivePlayer = saveEffectivePlayer;
		EffectiveStatus = saveEffectiveStatus;
	    } else {
		freeString(name);
	    }
	} else {
	    freeString(name);
	    runError("agent not found for SayTo/PoseTo");
	}
    }
}

/*
 * SayTo - say to a specific agent.
 */

void
SayTo(ULONG_T who, CHAR_T *altName, CHAR_T *what)
{

    sayOne(who, altName, " says: ", what, B_FALSE);
}

/*
 * PoseTo - pose to a specific agent.
 */

void
PoseTo(ULONG_T who, CHAR_T *altName, CHAR_T *what)
{

    sayOne(who, altName, " ", what, B_TRUE);
}

/*
 * Whisper - this one is more complicated, since one specific target is
 *	guaranteed to get the message, but the others might not.
 */

ULONG_T
Whisper(CHAR_T *altName, CHAR_T *what, ULONG_T who, LONG_T percent)
{
    CHAR_T *name, *name2;
    MachineList_t *ml, *moi, *ml2;
    Client_t *cl, *me, *cl2;
    CallList_t *calls;
    UINT_T len;
    ULONG_T key, here;

    TRACE("Whisper", percent);

    if (who == NIL_THING) {
	runError("nil who to Whisper");
	return BOOLEAN_FALSE;
    }

    messageFlush();
    flushEffects();

    me = ActiveClient;
    if (me != P_NULL) {
	if (me->cl_player == NIL_PLAYER) {
	    return BOOLEAN_FALSE;
	}
	here = me->cl_location;
	key = me->cl_thing;
	moi = P_NULL;	/* keep gcc quiet */
    } else {
	moi = ActiveMachine;
	if (moi == P_NULL) {
	    return BOOLEAN_FALSE;
	}
	here = moi->ml_location;
	key = moi->ml_thing;
    }
    if (*altName != '\0') {
	name = stringCopy(altName);
    } else {
	name = (CHAR_T *) getProp(key, NAME_KEY, TYPE_STRING);
	if (name == P_NULL || *name == '\0') {
	    name = stringCopy("<someone>");
	} else {
	    if (islower(*name)) {
		*name = toupper(*name);
	    }
	}
    }
    name2 = concatenate(FormatName(name), stringCopy(" whispers"));
    freeString(name);

    name = concatenate(stringCopy(name2), stringCopy(": "));
    name = concatenate(name, stringCopy(what));
    cl = getClientByThing(who);
    if (cl != P_NULL) {
	if (cl->cl_location != here) {
	    freeString(name);
	    return BOOLEAN_FALSE;
	}
	len = strLen1(name);
	name = concatenate(name, stringCopy("\n"));
	sendText(cl, name, len);
	memFreeE(&MS, name, len);
	ml = P_NULL;	/* keep gcc quiet */
    } else {
	ml = getMachineByThing(who);
	if (ml == P_NULL || ml->ml_location != here) {
	    freeString(name);
	    return BOOLEAN_FALSE;
	}
	if (ml->ml_whisperMeProc != NIL_ACTION) {
	    runCallList(addCall(P_NULL, P_NULL, ml, ml->ml_whisperMeProc,
				name), B_TRUE);
	}
    }

    name2 = concatenate(name2, stringCopy(" to "));
    name = (CHAR_T *) getProp(who, NAME_KEY, TYPE_STRING);
    if (name == P_NULL || *name == '\0') {
	name = stringCopy("<someone>");
    }
    name2 = concatenate(name2, FormatName(name));
    freeString(name);
    name2 = concatenate(name2, stringCopy(": "));
    what = concatenate(name2, stringCopy(what));
    what = concatenate(what, stringCopy("\n"));
    len = strLen1(what);

    calls = P_NULL;
    if (cl != P_NULL) {
	/* Whisper target is a Client_t. */
	cl2 = cl->cl_nextLoc;
	while (cl2 != cl) {
	    if (cl2 != me && percent > 0 && (LONG_T) Random(100) < percent) {
		sendText(cl2, what, len - 1);
	    }
	    cl2 = cl2->cl_nextLoc;
	}
	*(what + len - 2) = '\0';
	ml = findMachinesAt(here);
	if (ml != P_NULL) {
	    ml2 = ml;
	    do {
		if (ml != moi && ! ml->ml_deleted &&
		    ml->ml_whisperOtherProc != NIL_ACTION &&
		    percent > 0 && (LONG_T) Random(100) < percent)
		{
		    calls = addCall(calls, P_NULL, ml, ml->ml_whisperOtherProc,
				    stringCopy(what));
		}
		ml = ml->ml_nextLoc;
	    } while (ml != ml2);
	}
    } else {
	/* Whisper target is a MachineList_t. */
	cl = findClientsAt(here);
	if (cl != P_NULL) {
	    cl2 = cl;
	    do {
		if (cl != me && percent > 0 && (LONG_T) Random(100) < percent){
		    sendText(cl, what, len - 1);
		}
		cl = cl->cl_nextLoc;
	    } while (cl != cl2);
	}
	*(what + len - 2) = '\0';
	ml2 = ml->ml_nextLoc;
	while (ml2 != ml) {
	    if (ml2 != moi && ! ml2->ml_deleted &&
		ml2->ml_whisperOtherProc != NIL_ACTION &&
		percent > 0 && (LONG_T) Random(100) < percent)
	    {
		calls = addCall(calls, P_NULL, ml, ml->ml_whisperOtherProc,
				stringCopy(what));
	    }
	    ml2 = ml2->ml_next;
	}
    }

    *(what + len - 2) = '\n';
    freeString(what);
    runCallList(calls, B_TRUE);
    return BOOLEAN_TRUE;
}

/*
 * Quit - the scenario wants to boot out the character.
 */

void
Quit(void)
{

    if (ActiveClient != P_NULL) {
	bootClient(ActiveClient);
    }
}

/*
 * NewCreationPassword
 */

void
NewCreationPassword(void)
{

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can NewCreationPassword");
    } else {
	ActiveClient->cl_playerPassword = B_FALSE;
	startChangePassword();
    }
}

/*
 * Normal - used by a wizard to get out of wizard mode.
 */

void
Normal(void)
{
    Client_t *cl;

    cl = ActiveClient;
    if (cl != P_NULL && ! cl->cl_deleted && cl->cl_wizardMode) {
	if (cl->cl_newPlayer) {
	    doNewPlayer(cl);
	}
	if (cl->cl_inputAction == NIL_ACTION) {
	    runError("cannot do Normal when no handler proc specified");
	} else {
	    changeClientMode(cl, B_FALSE);
	}
    }
}

/*
 * WizardMode - can be used to attempt to get into wizard mode. Only works
 *	for apprentices and wizards. Returns true if worked.
 */

ULONG_T
WizardMode(void)
{
    Client_t *cl;

    cl = ActiveClient;
    if (cl != P_NULL && ! cl->cl_deleted && ! cl->cl_wizardMode &&
	cl->cl_status != ps_normal && cl->cl_canWizard)
    {
	changeClientMode(cl, B_TRUE);
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

/*
 * BootClient - boot the client for the given character off.
 */

void
BootClient(ULONG_T key)
{
    Client_t *cl;

    TRACE("BootClient", key);
    if (key == NIL_PLAYER) {
	runError("nil character passed to BootClient");
    } else {
	cl = Clients;
	while (cl != P_NULL && cl->cl_player != key) {
	    cl = cl->cl_next;
	}
	if (cl == P_NULL || cl->cl_player == NIL_PLAYER) {
	    putString("That character is not currently connected.\n");
	} else {
	    bootClient(cl);
	}
    }
}

/*
 * NukeClient - nuke the client for the given character off, right NOW!
 */

void
NukeClient(ULONG_T key)
{
    Client_t *cl;

    TRACE("NukeClient", key);
    if (key == NIL_PLAYER) {
	runError("nil character passed to NukeClient");
    } else {
	cl = Clients;
	while (cl != P_NULL && cl->cl_player != key) {
	    cl = cl->cl_next;
	}
	if (cl == P_NULL || cl->cl_player == NIL_PLAYER) {
	    putString("That character is not currently connected.\n");
	} else {
	    nukeClient(cl);
	}
    }
}

/*
 * AgentLocation - return the location of the given active agent.
 */

ULONG_T
AgentLocation(ULONG_T agent)
{
    Client_t *cl;
    MachineList_t *ml;

    TRACE("AgentLocation", agent);
    if (agent == NIL_THING) {
	runError("nil thing passed to AgentLocation");
	return NIL_THING;
    } else {
	cl = getClientByThing(agent);
	if (cl != P_NULL) {
	    if (cl->cl_player == NIL_PLAYER) {
		return NIL_THING;
	    }
	    return cl->cl_location;
	}
	ml = getMachineByThing(agent);
	if (ml != P_NULL && ! ml->ml_deleted) {
	    return ml->ml_location;
	}
	return NIL_THING;
    }
}

/*
 * DeleteSymbol - interface for the builtin.
 */

ULONG_T
DeleteSymbol(ULONG_T tableKey, CHAR_T *name)
{
    Entry_t *en;
    TableList_t *tl;
    Table_t *tb;
    NameMap_t *nm;
    ULONG_T key;
    UINT_T len;

    tb = findCachedTable(tableKey);
    if (tb->tb_owner != EffectivePlayer && tableKey != PublicSymbolTableKey &&
	EffectivePlayer != SYSADMIN_KEY)
    {
	putError("you don't own that table");
	return BOOLEAN_FALSE;
    } else if (tableKey == BuiltinTable->tl_key) {
	putError("can't delete from Builtin table");
	return BOOLEAN_FALSE;
    } else if (tableKey == PlayerTable->tl_key) {
	putError("can't delete from Characters table");
	return BOOLEAN_FALSE;
    } else if (*name == '\0') {
	putError("empty name to DeleteSymbol");
	return BOOLEAN_FALSE;
    } else {
	len = strLen1(name);
	en = symLookupE(tb, name, len);
	if (en == P_NULL) {
	    putStrError("symbol \"", name, "\" not defined in table");
	    return BOOLEAN_FALSE;
	} else {
	    if (deleteable(en, tableKey != PublicSymbolTableKey)) {
		key = en->en_key;
		if (tableKey == PublicSymbolTableKey) {
		    nm = PublicNameMap;
		} else if (ActiveClient != P_NULL &&
		    tableKey == ActiveClient->cl_privateTableKey)
		{
		    nm = ActiveClient->cl_privateMap;
		} else {
		    tl = TableList;
		    while (tl != P_NULL && tl->tl_this != tb) {
			tl = tl->tl_next;
		    }
		    /* if no-one is 'use'ing table, it will not be in list */
		    if (tl != P_NULL) {
			nm = tl->tl_nameMap;
		    } else {
			nm = P_NULL;
		    }
		}
		if (nm != P_NULL) {
		    removeNameFromMap(nm, key);
		}
		symDeleteE(&MS, tb, name, len);
		clearRefs(key);
		flushSymbol(key);
		return BOOLEAN_TRUE;
	    } else {
		return BOOLEAN_FALSE;
	    }
	}
    }
}

/*
 * MoveSymbol - the builtin.
 */

ULONG_T
MoveSymbol(ULONG_T fromTable, ULONG_T toTable, CHAR_T *name)
{
    Client_t *cl;
    Entry_t *en1;
    TableList_t *tl;
    Table_t *tb1, *tb2;
    Entry_t *en2;
    ULONG_T key;
    UINT_T len;

    cl = ActiveClient;
    if (cl == P_NULL) {
	runError("machines cannot use MoveSymbol");
	return BOOLEAN_FALSE;
    } else if (fromTable == NIL_TABLE) {
	runError("fromTable is nil in MoveSymbol");
	return BOOLEAN_FALSE;
    } else if (fromTable == BuiltinTable->tl_key) {
	runError("fromTable is Builtin table");
	return BOOLEAN_FALSE;
    } else if (fromTable == PlayerTable->tl_key) {
	runError("fromTable is Character table");
	return BOOLEAN_FALSE;
    } else if (toTable == NIL_TABLE) {
	runError("toTable is nil in MoveSymbol");
	return BOOLEAN_FALSE;
    } else if (toTable == BuiltinTable->tl_key) {
	runError("toTable is Builtin table");
	return BOOLEAN_FALSE;
    } else if (toTable == PlayerTable->tl_key) {
	runError("toTable is Character table");
	return BOOLEAN_FALSE;
    } else if (*name == '\0') {
	runError("empty name to MoveSymbol");
	return BOOLEAN_FALSE;
    } else {
	tb1 = findCachedTable(fromTable);
	tb2 = findCachedTable(toTable);
	len = strLen1(name);
	en1 = symLookupE(tb1, name, len);
	en2 = symLookupE(tb2, name, len);
	if (tb1->tb_owner != EffectivePlayer && tb1 != PublicSymbolTable &&
	    EffectivePlayer != SYSADMIN_KEY)
	{
	    putError("you don't own fromTable");
	    return BOOLEAN_FALSE;
	} else if (tb2->tb_owner != EffectivePlayer &&
	    tb2 != PublicSymbolTable && EffectivePlayer != SYSADMIN_KEY)
	{
	    putError("you don't own toTable");
	    return BOOLEAN_FALSE;
	} else if (en1 == P_NULL) {
	    putStrError("symbol \"", name, "\" not defined in fromTable");
	    return BOOLEAN_FALSE;
	} else if (en2 != P_NULL) {
	    putStrError("symbol \"", name, "\" already defined in toTable");
	    return BOOLEAN_FALSE;
	} else {
	    /* It would be nice to use defineXXX and DeleteSymbol, but they
	       do not quite work out right. */
	    key = en1->en_key;
	    tl = TableList;
	    while (tl != P_NULL && tl->tl_this != tb1) {
		tl = tl->tl_next;
	    }
	    /* if no-one is 'use'ing the table, it will not be in list */
	    if (tl != P_NULL) {
		removeNameFromMap(tl->tl_nameMap, key);
	    } else if (tb1 == PublicSymbolTable) {
		removeNameFromMap(PublicNameMap, key);
	    } else if (cl != P_NULL && tb1 == cl->cl_privateSymbols) {
		removeNameFromMap(cl->cl_privateMap, key);
	    }
	    symDeleteE(&MS, tb1, name, len);
	    clearRefs(key);
	    en2 = symEnterE(&MS, tb2, name, len);
	    en2->en_key = key;
	    return BOOLEAN_TRUE;
	}
    }
}

/*
 * RenameSymbol - the builtin.
 */

ULONG_T
RenameSymbol(ULONG_T theTable, CHAR_T *oldName, CHAR_T *newName)
{
    Client_t *cl;
    Entry_t *en1, *en2;
    TableList_t *tl;
    Table_t *tb;
    ULONG_T key;
    UINT_T len1, len2;

    cl = ActiveClient;
    if (cl == P_NULL) {
	runError("machines cannot use RenameSymbol");
	return BOOLEAN_FALSE;
    } else if (theTable == NIL_TABLE) {
	runError("theTable is nil in RenameSymbol");
	return BOOLEAN_FALSE;
    } else if (theTable == PlayerTable->tl_key) {
	runError("theTable is Character table");
	return BOOLEAN_FALSE;
    } else if (theTable == BuiltinTable->tl_key) {
	runError("theTable is Builtin table");
	return BOOLEAN_FALSE;
    } else if (*oldName == '\0') {
	runError("empty oldName to MoveSymbol");
	return BOOLEAN_FALSE;
    } else if (*newName == '\0') {
	runError("empty newName to MoveSymbol");
	return BOOLEAN_FALSE;
    } else {
	tb = findCachedTable(theTable);
	if (tb->tb_owner != EffectivePlayer && tb != PublicSymbolTable &&
	    EffectivePlayer != SYSADMIN_KEY)
	{
	    putError("you don't own theTable");
	    return BOOLEAN_FALSE;
	} else {
	    len1 = strLen1(oldName);
	    en1 = symLookupE(tb, oldName, len1);
	    if (en1 == P_NULL) {
		putStrError("symbol \"", oldName,"\" not defined in theTable");
		return BOOLEAN_FALSE;
	    } else {
		len2 = strLen1(newName);
		en2 = symLookupE(tb, newName, len2);
		if (en2 != P_NULL) {
		    putStrError("symbol \"", newName, "\" is already defined");
		    return BOOLEAN_FALSE;
		} else {
		    key = en1->en_key;
		    tl = TableList;
		    while (tl != P_NULL && tl->tl_this != tb) {
			tl = tl->tl_next;
		    }
		    if (tl != P_NULL) {
			removeNameFromMap(tl->tl_nameMap, key);
		    } else if (tb == PublicSymbolTable) {
			removeNameFromMap(PublicNameMap, key);
		    } else if (cl != P_NULL && tb == cl->cl_privateSymbols) {
			removeNameFromMap(cl->cl_privateMap, key);
		    }
		    symDeleteE(&MS, tb, oldName, len1);
		    clearRefs(key);
		    en2 = symEnterE(&MS, tb, newName, len2);
		    en2->en_key = key;
		    flushSymbol(key);
		    return BOOLEAN_TRUE;
		}
	    }
	}
    }
}

/*
 * defineXXX - common code for defining symbols from a scenario.
 */

ULONG_T
defineXXX(ULONG_T table, CHAR_T *name,CHAR_T *what,ULONG_T key,ULONG_T thisNil)
{
    CHAR_T buf[50];
    Entry_t *en;
    Table_t *tb;
    Client_t *cl;
    UINT_T nameLen;

    cl = ActiveClient;
    if (cl == P_NULL) {
	*strTail(strTail(&buf[0], "machines cannot use Define"), what) = '\0';
	runError(&buf[0]);
	return BOOLEAN_FALSE;
    } else if (table == NIL_TABLE) {
	*strTail(strTail(&buf[0], "nil table to Define"), what) = '\0';
	runError(&buf[0]);
	return BOOLEAN_FALSE;
    } else if (table == BuiltinTable->tl_key) {
	*strTail(strTail(&buf[0], "Builtin table to Define"), what) = '\0';
	runError(&buf[0]);
	return BOOLEAN_FALSE;
    } else if (table == PlayerTable->tl_key) {
	*strTail(strTail(&buf[0], "Character table to Define"), what) = '\0';
	runError(&buf[0]);
	return BOOLEAN_FALSE;
    } else if (key == thisNil) {
	*strTail(strTail(strTail(strTail(&buf[0], "nil "), what),
		 " to Define"), what) = '\0';
	runError(&buf[0]);
	return BOOLEAN_FALSE;
    } else if (*name == '\0') {
	*strTail(strTail(&buf[0], "empty name to Define"), what) ='\0';
	runError(&buf[0]);
	return BOOLEAN_FALSE;
    } else {
	tb = findCachedTable(table);
	if (tb->tb_owner != EffectivePlayer && tb != PublicSymbolTable &&
	    EffectivePlayer != SYSADMIN_KEY)
	{
	    putError("you don't own that table");
	    return BOOLEAN_FALSE;
	} else {
	    nameLen = strLen1(name);
	    en = symLookupE(tb, name, nameLen);
	    if (en != P_NULL) {
		putStrError("symbol \"", name, "\" is already defined");
		return BOOLEAN_FALSE;
	    } else {
		en = symEnterE(&MS, tb, name, nameLen);
		en->en_key = key;
		incRef(key);
		return BOOLEAN_TRUE;
	    }
	}
    }
}

/*
 * lookupXXX - common routine for lookup up a symbol in the private table.
 */

ULONG_T
lookupXXX(CHAR_T *name, CHAR_T *routine, UINT_T typ, ULONG_T table,
	  ULONG_T thisNil)
{
    CHAR_T buf[50];
    Client_t *cl;
    Entry_t *en;
    ULONG_T key;

    cl = ActiveClient;
    if (cl == P_NULL) {
	*strTail(strTail(&buf[0], "machines cannot use Lookup"),routine)='\0';
	runError(&buf[0]);
	return thisNil;
    } else if (*name == '\0') {
	*strTail(strTail(&buf[0], "empty name to Lookup"), routine) = '\0';
	runError(&buf[0]);
	return thisNil;
    } else {
	if (table == NIL_TABLE) {
	    key = lookup(name, strLen1(name));
	    if (key == 0 || (key >> TYPE_SHIFT) != typ) {
		return thisNil;
	    } else {
		return key;
	    }
	} else {
	    en = symLookupE(findCachedTable(table), name, strLen1(name));
	    if (en == P_NULL || (en->en_key >> TYPE_SHIFT) != typ) {
		return thisNil;
	    } else {
		return en->en_key;
	    }
	}
    }
}

/*
 * IsDefined - allow a check for a symbol being defined.
 */

ULONG_T
IsDefined(ULONG_T table, CHAR_T *name)
{

    if (table == NIL_TABLE) {
	if ((lookup(name, strLen1(name)) & INDEX_MASK) != 0) {
	    return BOOLEAN_TRUE;
	} else {
	    return BOOLEAN_FALSE;
	}
    } else {
	if (symLookupE(findCachedTable(table), name, strLen1(name)) == P_NULL){
	    return BOOLEAN_FALSE;
	} else {
	    return BOOLEAN_TRUE;
	}
    }
}

/* Routines to allow testing of the status of the current character. Mostly of
   use for the 'build.m' code. */

ULONG_T
IsProgrammer(void)
{

    if (ActiveClient == P_NULL || ActiveClient->cl_status == ps_normal) {
	return BOOLEAN_FALSE;
    } else {
	return BOOLEAN_TRUE;
    }
}

ULONG_T
IsNormal(void)
{

    if (ActiveClient == P_NULL || ActiveClient->cl_status == ps_normal) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

ULONG_T
IsApprentice(void)
{

    if (ActiveClient != P_NULL && ActiveClient->cl_status == ps_apprentice) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

ULONG_T
IsWizard(void)
{

    if (ActiveClient != P_NULL && ActiveClient->cl_status == ps_wizard) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

/*
 * ForceAction - do an action as another agent.
 *	Return the status of the attempt.
 *	Do NOT change TrueClient or TrueMachine.
 */

ULONG_T
ForceAction(ULONG_T agent, ULONG_T action)
{
    Client_t *cl, *saveClient;
    MachineList_t *ml, *saveMachine;
    Proc_t *pr;
    ULONG_T saveRealPlayer, saveEffectivePlayer, result;
    PlayerStatus_t saveRealStatus, saveEffectiveStatus;

    TRACE("ForceAction", agent);
    if (agent == NIL_THING) {
	runError("nil agent passed to ForceAction");
	return STATUS_FAIL;
    }
    pr = findCachedProcNoLock(action);
    if (pr->pr_parameters != P_NULL || pr->pr_resultType != TYPE_STATUS) {
	runError("action for ForceAction should have "
		 "no parameters and status result");
	return STATUS_FAIL;
    }

    saveClient = ActiveClient;
    saveMachine = ActiveMachine;
    saveRealPlayer = RealPlayer;
    saveEffectivePlayer = EffectivePlayer;
    saveRealStatus = RealStatus;
    saveEffectiveStatus = EffectiveStatus;
    messageFlush();
    flushEffects();
    cl = getClientByThing(agent);
    if (cl != P_NULL) {
	if (cl->cl_player == NIL_PLAYER) {
	    runError("player not active in ForceAction");
	    result = STATUS_FAIL;
	} else {
	    ActiveClient = cl;
	    ActiveMachine = P_NULL;
	    RealPlayer = cl->cl_player;
	    EffectivePlayer = cl->cl_player;
	    RealStatus = cl->cl_status;
	    EffectiveStatus = cl->cl_status;
	    result = doProc0(action);
	    messageFlush();
	    flushEffects();
	}
    } else {
	ml = getMachineByThing(agent);
	if (ml != P_NULL) {
	    ActiveClient = P_NULL;
	    ActiveMachine = ml;
	    RealPlayer = ActiveMachine->ml_owner;
	    EffectivePlayer = RealPlayer;
	    RealStatus = ps_apprentice;
	    EffectiveStatus = ps_apprentice;
	    result = doProc0(action);
	    messageFlush();
	} else {
	    runError("agent not found in ForceAction");
	    result = STATUS_FAIL;
	}
    }
    CurrentIndent = 0;
    ActiveClient = saveClient;
    ActiveMachine = saveMachine;
    RealPlayer = saveRealPlayer;
    EffectivePlayer = saveEffectivePlayer;
    RealStatus = saveRealStatus;
    EffectiveStatus = saveEffectiveStatus;
    return result;
}

/*
 * CanEdit - the builtin
 */

ULONG_T
CanEdit(void)
{

    if (ActiveClient == P_NULL || ! ActiveClient->cl_canEdit) {
	return BOOLEAN_FALSE;
    } else {
	return BOOLEAN_TRUE;
    }
}

/*
 * Editing - return true if the client is currently editing something.
 */

ULONG_T
Editing(void)
{

    if (ActiveClient == P_NULL || ! ActiveClient->cl_editing) {
	return BOOLEAN_FALSE;
    } else {
	return BOOLEAN_TRUE;
    }
}

/*
 * GettingString - return true if the client is currently getting a string.
 */

ULONG_T
GettingString(void)
{

    if (ActiveClient == P_NULL || ! ActiveClient->cl_getting) {
	return BOOLEAN_FALSE;
    } else {
	return BOOLEAN_TRUE;
    }
}

/*
 * EditProc - the builtin
 */

void
EditProc(ULONG_T key)
{
    Client_t *cl;
    UINT_T *pLockCount;
    Proc_t *pr;

    cl = ActiveClient;
    if (cl == P_NULL || ! cl->cl_canEdit) {
	runError("editing is not available");
    } else if (cl->cl_editing) {
	runError("already editing");
    } else if (key == NIL_ACTION) {
	runError("nil proc to EditProc");
    } else if ((key & INDEX_MASK) >= SMALLEST_BUILTIN) {
	runError("can't edit a builtin");
    } else {
	pr = findCachedProc(key, &pLockCount);
	if (pr->pr_owner != EffectivePlayer && EffectivePlayer != SYSADMIN_KEY)
	{
	    runError("you don't own that proc");
	} else {
	    if (hasBadRefs(pr)) {
		runError(
		    "proc has references to symbols not in any in-use table");
	    } else if (pr->pr_stripped) {
		runError("proc is stripped and cannot be edited");
	    } else if (pr->pr_locked) {
		runError("proc is byte-code locked and cannot be edited");
	    } else {
		cl->cl_editing = B_TRUE;
		setEditProcActive(B_TRUE, key);
		fmtProc(pr, findAnyName(key));
		fmtLine();
		setEditProcActive(B_FALSE, key);
	    }
	}
	*pLockCount -= 1;
    }
}

/*
 * PrintNoAts - set client so that it is not shown the '@'s for output from
 *	non-wizard code. Old flag is returned.
 */

ULONG_T
PrintNoAts(ULONG_T flag)
{
    BOOL_T oldValue;

    if (ActiveClient == P_NULL) {
	return BOOLEAN_FALSE;
    } else {
	oldValue = ActiveClient->cl_noAts;
	ActiveClient->cl_noAts = (flag != BOOLEAN_FALSE);
	return (oldValue ? BOOLEAN_TRUE : BOOLEAN_FALSE);
    }
}

/*
 * FindKey - the builtin. Search EVERYWHERE for a name for the given key.
 */

void
FindKey(ULONG_T key)
{

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can FindKey");
    } else if ((key & INDEX_MASK) == 0) {
	runError("nil key passed to FindKey");
    } else {
	if (! doFindKey(key, B_FALSE)) {
	    putString("No name found for that key.\n");
	}
    }
}

void
DumpThing(ULONG_T key)
{
    CHAR_T *p;

    if (ActiveClient == P_NULL || ActiveClient->cl_player == NIL_PLAYER ||
	EffectivePlayer != SYSADMIN_KEY || RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can DumpThing");
    } else if (key == NIL_THING) {
	runError("nil thing passed to DumpThing");
    } else {
	ActiveClient->cl_fullPaths = B_TRUE;
	p = findPlayerThing(key);
	if (p != P_NULL) {
	    fmtCheck("<Characters/");
	    fmtCheck(p);
	    fmtChar('>');
	} else {
	    fmtThingName(key);
	}
	fmtCheck(": thing, ");
	fmtThing(key);
	ActiveClient->cl_fullPaths = B_FALSE;
    }
}

void
ShowDBStatus(void)
{
    extern void ioShowDBStatus(void);

    if (ActiveClient == P_NULL || EffectivePlayer != SYSADMIN_KEY) {
	runError("only SysAdmin can use ShowDBStatus");
    } else {
	ioShowDBStatus();
    }
}

UINT_T
getTextWidth(Client_t *cl)
{
    TextComponentList_t *tcl;

    if (cl == P_NULL) {
	return TEXT_COLUMNS;
    }
    if (cl->cl_activeFrame == P_NULL) {
	return cl->cl_textWidth;
    }
    tcl = cl->cl_activeFrame->guf_textComponents;
    if (tcl == P_NULL) {
	return cl->cl_textWidth;
    }
    if (cl->cl_defaultTextComponent == -1) {
	return tcl->tcl_width;
    }
    while (tcl != P_NULL) {
	if (tcl->tcl_id == cl->cl_defaultTextComponent) {
	    return tcl->tcl_width;
	}
	tcl = tcl->tcl_next;
    }
    return cl->cl_textWidth;
}

UINT_T
getTextHeight(Client_t *cl)
{
    TextComponentList_t *tcl;

    if (cl == P_NULL) {
	return TEXT_COLUMNS;
    }
    if (cl->cl_activeFrame == P_NULL) {
	return cl->cl_textHeight;
    }
    tcl = cl->cl_activeFrame->guf_textComponents;
    if (tcl == P_NULL) {
	return cl->cl_textHeight;
    }
    if (cl->cl_defaultTextComponent == -1) {
	return tcl->tcl_height;
    }
    while (tcl != P_NULL) {
	if (tcl->tcl_id == cl->cl_defaultTextComponent) {
	    return tcl->tcl_height;
	}
	tcl = tcl->tcl_next;
    }
    return cl->cl_textHeight;
}
