#include <time.h>
#include <netinet/in.h>

#include "MUD.h"
#include "types.h"
#include "Request.h"
#include "Effects.h"
#include "exec.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * clients.c - handle client create/delete stuff.
 */

#define MESSAGE_DEBUG	B_FALSE

#if MESSAGE_DEBUG
#define MDEBUG(m)	log(m)
#define MDEBUG2(m1, m2) log2NL(m1, m2)
#define MDEBUGX(m, n)	logXNL(m, n)
#else
#define MDEBUG(m)
#define MDEBUG2(m1, m2)
#define MDEBUGX(m, n)
#endif

static ULONG_T ActiveObject;

/*
 * copyToPlayer - copy from Client_t to Player_t.
 */

static void
copyToPlayer(Player_t *pl, Client_t *cl)
{

    pl->pl_key = cl->cl_thing;
    pl->pl_location = cl->cl_location;
    pl->pl_inputAction = cl->cl_inputAction;
    pl->pl_rawKeyAction = cl->cl_rawKeyAction;
    pl->pl_mouseDownAction = cl->cl_mouseDownAction;
    pl->pl_buttonAction = cl->cl_buttonAction;
    pl->pl_idleAction = cl->cl_idleAction;
    pl->pl_activeAction = cl->cl_activeAction;
    pl->pl_effectDoneAction = cl->cl_effectDoneAction;
    pl->pl_wizardMode = cl->cl_wizardMode;
    pl->pl_status = cl->cl_status;
    pl->pl_newPlayer = cl->cl_newPlayer;
}

/*
 * copyToClient - copy from Player_t to Client_t.
 */

static void
copyToClient(Client_t *cl, Player_t *pl)
{

    cl->cl_thing = pl->pl_key;
    cl->cl_location = pl->pl_location;
    cl->cl_inputAction = pl->pl_inputAction;
    cl->cl_rawKeyAction = pl->pl_rawKeyAction;
    cl->cl_mouseDownAction = pl->pl_mouseDownAction;
    cl->cl_buttonAction = pl->pl_buttonAction;
    cl->cl_privateTableKey = pl->pl_symbolTableKey;
    cl->cl_idleAction = pl->pl_idleAction;
    cl->cl_activeAction = pl->pl_activeAction;
    cl->cl_effectDoneAction = pl->pl_effectDoneAction;
    cl->cl_wizardMode = pl->pl_wizardMode;
    cl->cl_status = pl->pl_status;
    cl->cl_newPlayer = pl->pl_newPlayer;
    if (! cl->cl_gotTextSize) {
	cl->cl_textWidth = pl->pl_textWidth;
	cl->cl_textHeight = pl->pl_textHeight;
    }
}

/*
 * nukePlayer - nuke the given player.
 *	Note: we actually do very little here. This is because there are
 *	so many things that can point at the Player_t that we can never
 *	throw it away. All 'owner' fields point at it! The same is true
 *	of the symbol table - SysAdmin can get at it via something like
 *	PlayerTable(Owner(th)). Also, if the Player_t stays, then its
 *	thing has to stay, and so we shouldn't ClearThing it, since that
 *	might mess up a scenario that makes assumptions. Sigh.
 */

void
nukePlayer(ULONG_T key)
{
    Player_t *pl;
    Player_t plCopy;

    TRACE("nukePlayer", key);
    pl = ioRead(key);
    /* do this so that we don't have to worry about the decRef's
       taking the Player_t out of the cache */
    plCopy = *pl;
    decRef(plCopy.pl_location);
    plCopy.pl_location = NIL_THING;
    decProcRef(plCopy.pl_inputAction);
    plCopy.pl_inputAction = NIL_ACTION;
    decProcRef(plCopy.pl_rawKeyAction);
    plCopy.pl_rawKeyAction = NIL_ACTION;
    decProcRef(plCopy.pl_mouseDownAction);
    plCopy.pl_mouseDownAction = NIL_ACTION;
    decProcRef(plCopy.pl_buttonAction);
    plCopy.pl_buttonAction = NIL_ACTION;
    decProcRef(plCopy.pl_idleAction);
    plCopy.pl_idleAction = NIL_ACTION;
    decProcRef(plCopy.pl_activeAction);
    plCopy.pl_activeAction = NIL_ACTION;
    decProcRef(plCopy.pl_effectDoneAction);
    plCopy.pl_effectDoneAction = NIL_ACTION;
    ioReplace(key, &plCopy, sizeof(Player_t));
    /* Cannot even delete promptKey! */
}

/*
 * flushSymbol - remove symbol from everyone's cache.
 */

void
flushSymbol(ULONG_T key)
{
    Client_t *cl;
    Request_t *rq;

    TRACE("flushSymbol", key);
    cl = Clients;
    while (cl != P_NULL) {
	if (! cl->cl_deleted) {
	    if (cl->cl_telnet == P_NULL) {
		rq = allocRequestE(&MS, 0);
		rq->rq_type = rt_flushSymbol;
		rq->rq_key = key;
		putMessageCarefully(cl, rq);
	    }
	}
	cl = cl->cl_next;
    }
}

/*
 * notInUse - return B_TRUE if given table is not currently in the global
 *	table-in-use list.
 */

BOOL_T
notInUse(ULONG_T key)
{
    TableList_t *tl;

    tl = TableList;
    while (tl != P_NULL && tl->tl_key != key) {
	tl = tl->tl_next;
    }
    return (BOOL_T) (tl == P_NULL);
}

/*
 * addGlobalUse - add/find a use in the global table list.
 */

TableList_t *
addGlobalUse(Table_t *tb, ULONG_T key)
{
    TableList_t *tl;

    TRACE("addGlobalUse", key);
    tl = TableList;
    while (tl != P_NULL && tl->tl_this != tb) {
	tl = tl->tl_next;
    }
    if (tl == P_NULL) {
	tl = MEM_ALLOC_E(&MS, sizeof(TableList_t), "tbls");
	tl->tl_prev = P_NULL;
	tl->tl_next = TableList;
	tl->tl_this = tb;
	tl->tl_nameMap = buildNameMap(tb);
	tl->tl_key = key;
	tl->tl_useCount = 1;
	if (TableList != P_NULL) {
	    TableList->tl_prev = tl;
	}
	TableList = tl;
    } else {
	tl->tl_useCount += 1;
    }
    return tl;
}

/*
 * removeGlobalUse - utility to remove a use of a table from the global list.
 */

void
removeGlobalUse(TableList_t *tl)
{
    TableList_t *prev, *next;

    TRACE("removeGlobalUse", tl->tl_key);
    tl->tl_useCount -= 1;
    if (tl->tl_useCount == 0) {
	prev = tl->tl_prev;
	next = tl->tl_next;
	if (prev == P_NULL) {
	    TableList = next;
	} else {
	    prev->tl_next = next;
	}
	if (next != P_NULL) {
	    next->tl_prev = prev;
	}
	freeNameMap(tl->tl_nameMap);
	/* This is safe to do, since only explicitly created tables can be in
	   the global table list. It is not quite certain that we WANT to,
	   since the client can still enter symbols into the table, which
	   will require re-reading it from the database, which, if the table
	   is large enough, will even require re-reading it from disk. */
	flushCachedTable(tl->tl_key);
	memFreeE(&MS, tl, sizeof(TableList_t));
    }
}

/*
 * setupForCall - common setup before calling an interpreted routine.
 */

void
setupForCall(Client_t *cl)
{
    extern ULONG_T FindResult(void);

    RealPlayer = cl->cl_player;
    EffectivePlayer = cl->cl_player;
    RealStatus = cl->cl_status;
    EffectiveStatus = cl->cl_status;
    strCpy(&MeString[0], cl->cl_name);
    ActiveMachine = P_NULL;
    TrueMachine = P_NULL;
    (void) FindResult();
}

/*
 * cleanupAfterCall - common cleanup after calling an interpreted routine.
 */

void
cleanupAfterCall(void)
{

    CurrentIndent = 0;
    Prefix[0] = '\0';
    PrefixLength = 0;
}

/*
 * doActionForClient - do a parameterless action for the given client.
 */

static void
doActionForClient(Client_t *cl, ULONG_T key)
{

    TRACE("doActionForClient", key);
    if (key != NIL_ACTION) {
	setupForCall(cl);
	(void) doProc0Top(key);
	cleanupAfterCall();
    }
}

/*
 * doNewPlayer - non-wizard-mode stuff for a new player.
 */

void
doNewPlayer(Client_t *cl)
{
    ULONG_T key;

    TRACE("doNewPlayer", cl);
    key = getEntryAction();
    if (key != NIL_ACTION) {
	cl->cl_newPlayer = B_FALSE;	     /* only do it once */
	/* Note that the player's location is not yet set here - it will
	   normally be set by the entry action. */
	doActionForClient(cl, key);
    } else {
	log("New player when no entry action!\n");
	sendText(cl,
	    "World not ready for use yet - contact its administrator\n", 56);
	if (Verbose) {
	    mess(
"A non-wizard client just attempted to connect, but there is no complete\n"
"scenario set up in this database yet.\n"
	    );
	}
    }
}

/*
 * doStartPlayerStuff - common code for binary and telnet player startup.
 */

void
doStartPlayerStuff(Client_t *cl)
{

    if (cl->cl_newPlayer) {
	doNewPlayer(cl);
    } else {
	doActionForClient(cl, cl->cl_activeAction);
    }
}

/*
 * runClient - when we are ready for the client to start up the session,
 *	we do this stuff, then send an rt_runClient to let him proceed.
 */

static void
runClient(Client_t *cl)
{
    Request_t *rq;

    log2NL("runClient for character ", cl->cl_name);
    fullyInsertClient(cl);	/* BEFORE the doStartPlayer */
    setNewPrompt(cl->cl_prompt);
    if (ForceWizardMode && cl->cl_player == SYSADMIN_KEY &&
	cl->cl_canWizard)
    {
	cl->cl_wizardMode = B_TRUE;
    }
    if (cl->cl_wizardMode) {
	changeClientMode(cl, B_TRUE);
    } else {
	doStartPlayerStuff(cl);
    }
    rq = allocRequestE(&MS, 0);
    rq->rq_type = rt_runClient;
    putMessageCarefully(cl, rq);
}

/*
 * beginClient - the client now wants to run a game. Return the res
 *	result, so that a telnet client can see it. Note that when
 *	we are called from telnet code, 'rq' is nil.
 */

UINT_T
beginClient(Client_t *cl, Request_t *rq)
{
    Player_t *pl;
    EffectsInfo_t *ei;
    GUIFrame_t *guf;
    TextComponentList_t *tcl;
    UINT_T res;

    TRACE("beginClient", cl);

    if (rq != P_NULL) {
	if (rq->rq_usedLen != sizeof(struct EffectsInfo)) {
	    logN("beginClient with incorrect length ", rq->rq_usedLen, P_NULL);
	    logN(" not ", sizeof(struct EffectsInfo), "\n");
	    endClient(cl);
	    freeRequestE(&MS, rq);
	    return 0;
	}
	for (res = 0; res != BEGIN_KEY_LEN; res += 1) {
	    if (rq->rq_u.ru_bytes[res] != (BYTE_T) (res * res)) {
		log("beginClient with incorrect header\n");
		endClient(cl);
		freeRequestE(&MS, rq);
		return 0;
	    }
	}
    }

    res = Random(0xffff);
    cl->cl_sessionKey = (ULONG_T) cl ^ ((ULONG_T) res << 16 | res);
    if (rq != P_NULL) {
	rq->rq_key = cl->cl_sessionKey;
	rq->rq_usedLen = 0;
    }
    if ((getDungeonFlags() & FLAG_SINGLE_USER) != 0) {
	/* single-player dungeon (an adventure game) */
	if (Clients != cl || cl->cl_next != P_NULL) {
	    /* Someone already playing - disallow connection */
	    res = 0;
	    if (rq != P_NULL) {
		rq->rq_uint = 0;
		putMessageCarefully(cl, rq);
	    }
	} else {
	    /* allow connection, but special setup, since no true character */
	    res = 1;
	    cl->cl_textColumn = 0;
	    cl->cl_player = SYSADMIN_KEY;
	    pl = ioRead(SYSADMIN_KEY);
	    copyToClient(cl, pl);
	    cl->cl_wizardMode = B_FALSE; /* override, for next connect */
	    cl->cl_prompt = stringCopy(ioRead(pl->pl_promptKey));
	    cl->cl_name = stringCopy("<SINGLE>");
	    log("beginClient in single player mode\n");
	    if (cl->cl_inputAction == NIL_ACTION) {
		cl->cl_newPlayer = B_TRUE;
	    } else {
		cl->cl_newPlayer = B_FALSE;
	    }
	    cl->cl_privateSymbols = findCachedTable(pl->pl_symbolTableKey);
	    cl->cl_privateMap = buildNameMap(cl->cl_privateSymbols);
	    if (rq != P_NULL) {
		rq->rq_uint = 1;
		putMessageCarefully(cl, rq);
	    }
	    runClient(cl);
	}
    } else {
	res = 2;
	if (rq != P_NULL) {
	    rq->rq_uint = 2;
	    putMessageCarefully(cl, rq);
	    log("beginClient for binary client\n");
	} else {
	    log("beginClient for telnet client\n");
	}
    }
    if (cl->cl_telnet == P_NULL) {
	/* note that ei WILL be word aligned */
	ei = (EffectsInfo_t *) &rq->rq_u.ru_bytes[0];
	cl->cl_graphicsType = stringCopy(&ei->ei_graphicsType[0]);
	cl->cl_graphicsRows = ntohs(ei->ei_graphicsRows);
	cl->cl_graphicsCols = ntohs(ei->ei_graphicsCols);
	cl->cl_graphicsColours = ntohs(ei->ei_graphicsColours);
	if (ei->ei_textWidth != 0) {
	    cl->cl_gotTextSize = B_TRUE;
	    cl->cl_textWidth = ntohs(ei->ei_textWidth);
	    cl->cl_textHeight = ntohs(ei->ei_textHeight);
	    guf = MEM_ALLOC_E(&MS, sizeof(GUIFrame_t), "GUIF");
	    tcl = MEM_ALLOC_E(&MS, sizeof(TextComponentList_t), "TCL_");
	    tcl->tcl_next = P_NULL;
	    tcl->tcl_width = cl->cl_textWidth;
	    tcl->tcl_height = cl->cl_textHeight;
	    guf->guf_textComponents = tcl;
	    guf->guf_drawComponents = P_NULL;
	    cl->cl_activeFrame = guf;
	}
	if (ei->ei_fontAscent != 0) {
	    cl->cl_fontAscent = ntohs(ei->ei_fontAscent);
	    cl->cl_fontDescent = ntohs(ei->ei_fontDescent);
	    cl->cl_fontLeading = ntohs(ei->ei_fontLeading);
	    cl->cl_fontHeight = ntohs(ei->ei_fontHeight);
	    cl->cl_fontWidth = ntohs(ei->ei_fontWidth);
	}
	cl->cl_clientVersion = ntohs(ei->ei_version);
	cl->cl_canEdit = ei->ei_canEdit;
	cl->cl_canGetString = ei->ei_canGetString;
	cl->cl_canQueryFile = ei->ei_canQueryFile;
	cl->cl_canWizard = ei->ei_canWizard;
	cl->cl_graphicsOn = ei->ei_graphicsOn;
	cl->cl_graphicsPalette = ei->ei_graphicsPalette;
	cl->cl_voiceOn = ei->ei_voiceOn;
	cl->cl_soundOn = ei->ei_soundOn;
	cl->cl_musicOn = ei->ei_musicOn;
    }
    cl->cl_onAt = GetCurrentTime();
    return res;
}

/*
 * freeGUIFrame - free a GUI frame description, as setup for Java client.
 */

void
freeGUIFrame(GUIFrame_t *guf)
{
    TextComponentList_t *tcl;
    DrawComponentList_t *dcl;

    if (guf == P_NULL) {
	return;
    }
    while (B_TRUE) {
	tcl = guf->guf_textComponents;
	if (tcl == P_NULL) {
	    break;
	}
	guf->guf_textComponents = tcl->tcl_next;
	memFreeE(&MS, tcl, sizeof(TextComponentList_t));
    }
    while (B_TRUE) {
	dcl = guf->guf_drawComponents;
	if (dcl == P_NULL) {
	    break;
	}
	guf->guf_drawComponents = dcl->dcl_next;
	memFreeE(&MS, dcl, sizeof(DrawComponentList_t));
    }
    memFreeE(&MS, guf, sizeof(GUIFrame_t));
}

/*
 * endClient - a client player is leaving - get rid of his info, etc.
 */

void
endClient(Client_t *cl)
{
    Player_t *pl;
    KnownIconList_t *kil;
    KnownEffectList_t *kel;
    UseList_t *ul;
    ULONG_T key, promptKey;

    TRACE("endClient", cl);
    if (cl->cl_ended) {
	return;
    }
    cl->cl_ended = B_TRUE;
    /* note: this is also called when a client is finally going away, so it
       must be callable twice in a row */
    key = cl->cl_player;
    if (key != NIL_PLAYER && cl->cl_inserted) {
	log2NL("endClient for character ",
	       (cl->cl_name == P_NULL ? (CHAR_T *) "<UNKNOWN>" : cl->cl_name));
	if (Verbose) {
	    mess("Connection for character ");
	    if (cl->cl_name != P_NULL) {
		mess(cl->cl_name);
	    } else {
		mess("<UNKNOWN>");
	    }
	    mess(" is shutting down.\n");
	}
	if (cl->cl_inserted) {
	    ActiveClient = cl;
	    TrueClient = cl;
	    doActionForClient(cl, cl->cl_idleAction);
	    ActiveClient = P_NULL;
	    TrueClient = P_NULL;
	    flushEffects();
	    messageFlush();
	}
	if (cl->cl_editStringAction != NIL_ACTION) {
	    decProcRef(cl->cl_editStringAction);
	    cl->cl_editStringAction = NIL_ACTION;
	}
	if (cl->cl_getStringAction != NIL_ACTION) {
	    decProcRef(cl->cl_getStringAction);
	    cl->cl_getStringAction = NIL_ACTION;
	}
	if (cl->cl_destroy) {
	    /* needed to call nukePlayer */
	    EffectivePlayer = SYSADMIN_KEY;
	    RealPlayer = SYSADMIN_KEY;
	    nukePlayer(cl->cl_player);
	    removeNameFromMap(PlayerTable->tl_nameMap, cl->cl_player);
	    symDeleteE(&MS, PlayerTable->tl_this,
		       cl->cl_name, strLen1(cl->cl_name));
	} else {
	    pl = ioRead(cl->cl_player);
	    if (cl->cl_prompt != P_NULL &&
		strCmp(cl->cl_prompt, ioRead(pl->pl_promptKey)) != 0)
	    {
		ioDelete(pl->pl_promptKey);
		promptKey =
		    ioCreate(TYPE_STRING,cl->cl_prompt,strLen1(cl->cl_prompt));
		/* just in case - theoretically, we could have nuked 'pl' */
		pl = ioRead(cl->cl_player);
		pl->pl_promptKey = promptKey;
	    }
	    copyToPlayer(pl, cl);
	    pl->pl_lastOn = GetCurrentTime();
	    pl->pl_logonCount += 1;
	    pl->pl_totalSeconds += (pl->pl_lastOn - cl->cl_onAt);
	    ioDirty(cl->cl_player);
	    flushCachedTable(cl->cl_privateTableKey);
	}
	fullyUnlinkClient(cl);
	cl->cl_location = NIL_THING;
	cl->cl_thing = NIL_THING;
	cl->cl_player = NIL_PLAYER;
    } else {
	log("endClient for non-started connection\n");
    }

    if (cl->cl_privateMap != P_NULL) {
	freeNameMap(cl->cl_privateMap);
	cl->cl_privateMap = P_NULL;
    }
    if (cl->cl_prompt != P_NULL) {
	freeString(cl->cl_prompt);
	cl->cl_prompt = P_NULL;
    }
    if (cl->cl_name != P_NULL) {
	freeString(cl->cl_name);
	cl->cl_name = P_NULL;
    }
    if (cl->cl_graphicsType != P_NULL) {
	freeString(cl->cl_graphicsType);
	cl->cl_graphicsType = P_NULL;
    }
    while (B_TRUE) {
	kil = cl->cl_knownIcons;
	if (kil == P_NULL) {
	    break;
	}
	cl->cl_knownIcons = kil->kil_next;
	memFreeE(&MS, kil, sizeof(KnownIconList_t));
    }
    while (B_TRUE) {
	kel = cl->cl_knownEffects;
	if (kel == P_NULL) {
	    break;
	}
	cl->cl_knownEffects = kel->kel_next;
	memFreeE(&MS, kel, sizeof(KnownEffectList_t));
    }
    while (B_TRUE) {
	ul = cl->cl_useList;
	if (ul == P_NULL) {
	    break;
	}
	removeGlobalUse(ul->ul_entry);
	cl->cl_useList = ul->ul_next;
	memFreeE(&MS, ul, sizeof(UseList_t));
    }
    freeGUIFrame(cl->cl_activeFrame);
    cl->cl_activeFrame = P_NULL;
    freeGUIFrame(cl->cl_newFrame);
    cl->cl_newFrame = P_NULL;

    endClient2(cl);
    cl->cl_deleted = B_TRUE;
}

/*
 * askPlayer - client is asking to start up the indicated player.
 */

void
askPlayer(Client_t *cl, Request_t *rq)
{
    Entry_t *en;
    Client_t *cl2;
    CHAR_T *pw;
    Player_t *pl;
    ULONG_T key;

    TRACE("askPlayer", cl);
    log2NL("askPlayer for character ", &rq->rq_u.ru_text[0]);
    if (Verbose) {
	mess("Handling a request to run client for character '");
	mess(&rq->rq_u.ru_text[0]);
	mess("'.\n");
    }

    en = symLookupE(PlayerTable->tl_this,
		    &rq->rq_u.ru_text[0], rq->rq_usedLen);
    rq->rq_usedLen = 0;
    if (en == P_NULL) {
	/* no such player. Perhaps create it */
	pw = ioRead(getCreationPassword());
	if (*pw == '*') {
	    /* automatic creation disallowed - set key to 0 */
	    rq->rq_key = 0;
	} else if (*pw == '\0') {
	    /* Empty CreationPassword - none needed. */
	    rq->rq_key = 1;
	} else {
	    /* else set key to 2 */
	    rq->rq_key = 2;
	    cl->cl_passwordCount = 0;
	}
    } else if ((en->en_key >> TYPE_SHIFT) != TYPE_PLAYER) {
	/* the given name is in the table, but not of type player. Hmmm. */
	rq->rq_key = 3;
    } else {
	key = en->en_key;
	if (key == SYSADMIN_KEY && rq->rq_key != 0 &&
	    (getDungeonFlags() & FLAG_REMOTE_SYSADMIN_OK) == 0)
	{
	    /* remote SysAdmin login when not allowed - disallow it */
	    rq->rq_key = 4;
	} else {
	    cl2 = getClientByPlayer(key);
	    if (cl2 == P_NULL) {
		cl->cl_textColumn = 0;
		cl->cl_player = key;
		pl = ioRead(key);
		cl->cl_name = stringCopy(&rq->rq_u.ru_text[0]);
		cl->cl_prompt = stringCopy(ioRead(pl->pl_promptKey));
		copyToClient(cl, pl);
		/* set key to 4 - OK to proceed normally */
		rq->rq_key = 5;
		/* Do this *after* the last reference via 'pl' */
		cl->cl_privateSymbols = findCachedTable(pl->pl_symbolTableKey);
		cl->cl_privateMap = buildNameMap(cl->cl_privateSymbols);
		cl->cl_passwordCount = 0;
	    } else {
		/* player already active - set key to 5 */
		rq->rq_key = 6;
	    }
	}
    }
}

/*
 * creationCheck - handle incoming rt_creationCheck
 */

void
creationCheck(Client_t *cl, Request_t *rq)
{
    char *pw;

    TRACE("creationCheck", cl);
    MUDCrypt(&rq->rq_u.ru_text[0], B_TRUE, cl->cl_sessionKey);
    pw = ioRead(getCreationPassword());
    rq->rq_usedLen = 0;
    if (strCmp(&rq->rq_u.ru_text[0], pw) == 0) {
	rq->rq_key = 0;
    } else {
	cl->cl_passwordCount += 1;
	if (cl->cl_passwordCount == 3) {
	    rq->rq_key = 2;
	} else {
	    rq->rq_key = 1;
	}
    }
}

/*
 * createPlayer - try to satisfy a new player request.
 */

void
createPlayer(Client_t *cl, Request_t *rq)
{
    Entry_t *en;
    CHAR_T *name;
    Player_t pl;
    UINT_T nameLen;

    TRACE("createPlayer", cl);
    rq->rq_usedLen = 0;
    name = &rq->rq_u.ru_text[PASSWORD_LEN];
    log2NL("createPlayer for character ", name);
    nameLen = strLen1(name);
    en = symLookupE(PlayerTable->tl_this, name, nameLen);
    if (en != P_NULL) {
	/* already exists */
	rq->rq_key = 0;
	if (Verbose) {
	    mess("Denying creation of character '");
	    mess(name);
	    mess("' since character already exists.\n");
	}
	if (cl->cl_telnet == P_NULL) {
	    putMessageCarefully(cl, rq);
	}
    } else {
	if (Verbose) {
	    mess("Creating new character '");
	    mess(name);
	    mess("'.\n");
	}
	copy(&pl.pl_password[0], &rq->rq_u.ru_text[0], PASSWORD_LEN);
	MUDCrypt(&pl.pl_password[0], B_TRUE, cl->cl_sessionKey);
	doCreateCharacter(&pl);
	cl->cl_privateSymbols = findCachedTable(pl.pl_symbolTableKey);
	/* delay doing this until here - we want the map to point at the
	   cached table structure - not the temporary one on the stack
	   in doCreateCharacter */
	cl->cl_privateMap = buildNameMap(cl->cl_privateSymbols);
	cl->cl_name = stringCopy(name);
	cl->cl_prompt = stringCopy("input> ");
	cl->cl_player = ioCreate(TYPE_PLAYER, &pl, sizeof(Player_t));
	cl->cl_privateSymbols->tb_owner = cl->cl_player;
	cl->cl_privateSymbols->tb_useCount = 1;
	cl->cl_privateSymbols->tb_dirty = B_TRUE;
	copyToClient(cl, &pl);
	en = symEnterE(&MS, PlayerTable->tl_this, name, nameLen);
	en->en_key = cl->cl_player;
	rq->rq_key = 1;
	EffectivePlayer = SYSADMIN_KEY;
	RealPlayer = SYSADMIN_KEY;
	addProp(pl.pl_key, NAME_KEY, (ULONG_T) name, TYPE_STRING);
	if (cl->cl_telnet == P_NULL) {
	    putMessageCarefully(cl, rq);
	    runClient(cl);
	}
    }
}

/*
 * loginPlayer - handle an rt_loginPlayer.
 */

void
loginPlayer(Client_t *cl, Request_t *rq)
{
    Player_t *pl;

    TRACE("loginPlayer", cl);
    log2NL("loginPlayer for character ", cl->cl_name);
    pl = ioRead(cl->cl_player);
    MUDCrypt(&rq->rq_u.ru_text[0], B_TRUE, cl->cl_sessionKey);
    rq->rq_usedLen = 0;
    if (strCmp(&rq->rq_u.ru_text[0], &pl->pl_password[0]) == 0) {
	rq->rq_key = 0;
	if (cl->cl_telnet == P_NULL) {
	    putMessageCarefully(cl, rq);
	    runClient(cl);
	}
    } else {
	cl->cl_passwordCount += 1;
	if (cl->cl_passwordCount == 3) {
	    rq->rq_key = 2;
	} else {
	    rq->rq_key = 1;
	}
	if (cl->cl_telnet == P_NULL) {
	    putMessageCarefully(cl, rq);
	}
    }
}

/*
 * lookupForClient - look up a symbol in the appropriate tables. Return the
 *	key for the found entry, else 0.
 */

ULONG_T
lookupForClient(Client_t *cl, CHAR_T *name, UINT_T nameLen)
{
    Entry_t *en;
    UseList_t *ul;

    TRACE("lookupForClient", name);
    en = symLookupE(BuiltinTable->tl_this, name, nameLen);
    if (en == P_NULL && cl != P_NULL && cl->cl_privateSymbols != P_NULL) {
	en = symLookupE(cl->cl_privateSymbols, name, nameLen);
	if (en == P_NULL) {
	    ul = cl->cl_useList;
	    while (ul != P_NULL && en == P_NULL) {
		en = symLookupE(ul->ul_this, name, nameLen);
		ul = ul->ul_next;
	    }
	}
    }
    if (en == P_NULL) {
	en = symLookupE(PublicSymbolTable, name, nameLen);
    }
    if (en != P_NULL) {
	return en->en_key;
    } else {
	return 0;
    }
}

/*
 * lookup - externally callable lookup - use the current player if any.
 */

ULONG_T
lookup(CHAR_T *name, UINT_T nameLen)
{

    return lookupForClient(ActiveClient, name, nameLen);
}

/*
 * findAnyName1 - find any (private or public) name for the given key.
 */

static CHAR_T *
findAnyName1(Client_t *cl, ULONG_T key)
{
    CHAR_T *name;
    UseList_t *ul;

    TRACE("findAnyName1", key);
    name = findName(BuiltinTable->tl_nameMap, key);
    if (name == P_NULL && cl != P_NULL && ! cl->cl_deleted) {
	name = findName(cl->cl_privateMap, key);
	if (name == P_NULL) {
	    ul = cl->cl_useList;
	    while (ul != P_NULL && name == P_NULL) {
		name = findName(ul->ul_entry->tl_nameMap, key);
		ul = ul->ul_next;
	    }
	}
    }
    if (name == P_NULL) {
	name = findName(PublicNameMap, key);
    }
    return name;
}

/*
 * findAnyName - external interface - use 'ActiveClient'.
 */

CHAR_T *
findAnyName(ULONG_T key)
{

    if (ActiveClient != P_NULL && ActiveClient->cl_fullPaths) {
	if (doFindKey(key, B_TRUE)) {
	    return "";
	} else {
	    return P_NULL;
	}
    } else {
	return findAnyName1(ActiveClient, key);
    }
}

/*
 * copyToMachine - copy stuff from MachineList_t to Machine_t.
 */

void
copyToMachine(Machine_t *mch, MachineList_t *ml)
{

    mch->mch_location = ml->ml_location;
    mch->mch_sayProc = ml->ml_sayProc;
    mch->mch_whisperMeProc = ml->ml_whisperMeProc;
    mch->mch_whisperOtherProc = ml->ml_whisperOtherProc;
    mch->mch_pageProc = ml->ml_pageProc;
    mch->mch_poseProc = ml->ml_poseProc;
    mch->mch_otherProc = ml->ml_otherProc;
    mch->mch_idleProc = ml->ml_idleProc;
    mch->mch_activeProc = ml->ml_activeProc;
}

/*
 * doChangePassword - common for changePassword and telnet stuff.
 */

BOOL_T
doChangePassword(Client_t *cl, CHAR_T *pw)
{
    Player_t *pl;
    CHAR_T *pw2;

    MUDCrypt(pw, B_TRUE, cl->cl_sessionKey);
    if (cl->cl_expectingOldPassword) {
	cl->cl_expectingOldPassword = B_FALSE;
	if (cl->cl_playerPassword) {
	    pl = ioRead(cl->cl_player);
	    pw2 = &pl->pl_password[0];
	} else {
	    pw2 = ioRead(getCreationPassword());
	}
	if (strCmp(pw, pw2) == 0) {
	    if (cl->cl_telnet == P_NULL) {
		sendChangePassword(cl, 1);
	    }
	} else {
	    if (cl->cl_telnet == P_NULL) {
		sendChangePassword(cl, 2);
	    }
	    return B_FALSE;
	}
    } else {
	if (cl->cl_playerPassword) {
	    pl = ioRead(cl->cl_player);
	    copy(&pl->pl_password[0], pw, PASSWORD_LEN);
	    ioDirty(cl->cl_player);
	} else {
	    ioReplace(getCreationPassword(), pw, strLen1(pw));
	}
    }
    return B_TRUE;
}

/*
 * changePassword - handle a rt_changePassword reply.
 */

static void
changePassword(Client_t *cl, Request_t *rq)
{

    TRACE("handlePassword", rq);
    if (cl->cl_expectingOldPassword) {
	if (doChangePassword(cl, &rq->rq_u.ru_text[0])) {
	    /* Do not want to re-set prompt here. */
	    return;
	}
    } else {
	if (rq->rq_usedLen != 0) {
	    (void) doChangePassword(cl, &rq->rq_u.ru_text[0]);
	}
    }
    if (! cl->cl_wizardMode) {
	setNewPrompt(cl->cl_prompt);
    }
}

/*
 * flushOneMachine - shared by flushMachine and flushClients.
 */

void
flushOneMachine(MachineList_t *ml)
{
    Machine_t *mch;

    TRACE("flushOneMachine", ml->ml_machine);
    ml->ml_dirty = B_FALSE;
    mch = ioRead(ml->ml_machine);
    copyToMachine(mch, ml);
    ioDirty(ml->ml_machine);
}

/*
 * flushMachine - called by 'describeSymbol'.
 */

void
flushMachine(ULONG_T key)
{
    MachineList_t *ml;

    TRACE("flushMachine", key);
    ml = Machines;
    while (ml != P_NULL && (ml->ml_deleted || ml->ml_machine != key)) {
	ml = ml->ml_next;
    }
    if (ml != P_NULL && ml->ml_dirty) {
	flushOneMachine(ml);
    }
}

/*
 * flushOnePlayer - shared by flushPlayer and flushClients.
 */

static void
flushOnePlayer(Client_t *cl)
{
    Player_t *pl;

    TRACE("flushOnePlayer", cl->cl_player);
    pl = ioRead(cl->cl_player);
    copyToPlayer(pl, cl);
    if (cl->cl_prompt != P_NULL &&
	strCmp(cl->cl_prompt, ioRead(pl->pl_promptKey)) != 0)
    {
	ioDelete(pl->pl_promptKey);
	pl->pl_promptKey =
	    ioCreate(TYPE_STRING, cl->cl_prompt, strLen1(cl->cl_prompt));
    }
    ioDirty(cl->cl_player);
}

/*
 * flushPlayer - ditto.
 */

void
flushPlayer(ULONG_T key)
{
    Client_t *cl;

    TRACE("flushPlayer", key);
    cl = getClientByPlayer(key);
    if (cl != P_NULL) {
	flushOnePlayer(cl);
    }
}

/*
 * flushClients - flush the current state of all machines and players.
 */

void
flushClients(void)
{
    Client_t *cl;
    MachineList_t *ml;

    TRACE("flushClients", 0);
    cl = Clients;
    while (cl != P_NULL) {
	if (! cl->cl_deleted && cl->cl_player != NIL_PLAYER) {
	    flushOnePlayer(cl);
	}
	cl = cl->cl_next;
    }
    ml = Machines;
    while (ml != P_NULL) {
	if (! ml->ml_deleted && ml->ml_dirty) {
	    flushOneMachine(ml);
	}
	ml = ml->ml_next;
    }
}

/*
 * setStatus - change the status of the indicated player.
 */

void
setStatus(ULONG_T key, PlayerStatus_t status, ULONG_T doMode)
{
    Client_t *cl;
    Player_t *pl;

    TRACE("setStatus", key);
    if (key == NIL_PLAYER) {
	runError("nil player passed to MakeXXX");
    } else {
	cl = ActiveClient;
	if (key == SYSADMIN_KEY) {
	    runError("cannot change status of \"SysAdmin\"");
	} else if (cl == P_NULL) {
	    runError("only players can change player status");
	} else if (cl->cl_status != ps_wizard) {
	    runError("only full wizards can change player status");
	} else if (cl->cl_player == key) {
	    runError("cannot change own status");
	} else {
	    pl = ioRead(key);
	    if ((cl == P_NULL || EffectivePlayer != SYSADMIN_KEY) &&
		((status == ps_normal && pl->pl_status != ps_normal) ||
		 (status == ps_apprentice && pl->pl_status == ps_wizard)))
	    {
		runError("only SysAdmin can demote players");
	    } else {
		pl->pl_status = status;
		pl->pl_sponsor = cl->cl_player;
		if (status != ps_normal && pl->pl_newPlayer) {
		    pl->pl_wizardMode = B_TRUE;
		}
		ioDirty(key);
		cl = getClientByPlayer(key);
		if (cl != P_NULL && ! cl->cl_deleted && cl->cl_canWizard) {
		    cl->cl_status = status;
		    if (status == ps_normal) {
			if (cl->cl_wizardMode) {
			    changeClientMode(cl, B_FALSE);
			}
		    } else {
			if (! cl->cl_wizardMode && doMode == BOOLEAN_TRUE) {
			    changeClientMode(cl, B_TRUE);
			}
		    }
		}
		updateProcsInCache(key, status);
	    }
	}
    }
}

/*
 * routines for handling 'It'
 */

void
SetIt(ULONG_T key)
{

    ActiveObject = key;
}

ULONG_T
It(void)
{

    return ActiveObject;
}

/*
 * routines for some symbol tables.
 */

ULONG_T
PublicTable(void)
{

    return PublicSymbolTableKey;
}

ULONG_T
PrivateTable(void)
{

    if (ActiveClient == P_NULL) {
	return NIL_TABLE;
    } else {
	return ActiveClient->cl_privateTableKey;
    }
}

/*
 * ClientsActive - tell a machine whether or not anyone is logged on.
 */

ULONG_T
ClientsActive(void)
{

    if (Clients == P_NULL) {
	return BOOLEAN_FALSE;
    } else {
	return BOOLEAN_TRUE;
    }
}

/*
 * flushEffect - handle an rt_flushEffect request.
 */

static BOOL_T
flushEffect(Client_t *cl, ULONG_T key)
{
    KnownEffectList_t *kel, *next, *prev;

    kel = cl->cl_knownEffects;
    if (key == 0) {
	while (kel != P_NULL) {
	    next = kel;
	    kel = kel->kel_next;
	    memFreeE(&MS, next, sizeof(KnownEffectList_t));
	}
	cl->cl_knownEffects = P_NULL;
	return B_TRUE;
    } else {
	while (kel != P_NULL && kel->kel_key != key) {
	    kel = kel->kel_next;
	}
	if (kel != P_NULL) {
	    next = kel->kel_next;
	    prev = kel->kel_prev;
	    if (next != P_NULL) {
		next->kel_prev = prev;
	    }
	    if (prev != P_NULL) {
		prev->kel_next = next;
	    } else {
		cl->cl_knownEffects = next;
	    }
	    memFreeE(&MS, kel, sizeof(KnownEffectList_t));
	    return B_TRUE;
	} else {
	    return B_FALSE;
	}
    }
}

/*
 * symbolEnter - handle client request to define a symbol.
 */

static void
symbolEnter(Client_t *cl, Request_t *rq)
{
    Table_t *tb;
    Entry_t *en;
    ULONG_T table;

    /* it would be desireable to disallow synonyms for a given key, so
       that it is possible to delete one of them (see the checks in
       'deleteable' in 'misc.c'), but with the multiple symbol table
       stuff, that is not as reasonable. */
    table = rq->rq_u.ru_otherKey;
    TRACE("symbolEnter", table);
    tb = P_NULL;
    if (table == TABLE_PUBLIC) {
	tb = PublicSymbolTable;
    } else if (table == TABLE_PRIVATE) {
	tb = cl->cl_privateSymbols;
    } else if (table == BuiltinTable->tl_key) {
	putError("cannot modify Builtin table");
    } else if (table == PlayerTable->tl_key) {
	putError("cannot modify Character table");
    } else {
	tb = findCachedTable(table);
	if (tb->tb_owner != RealPlayer && EffectivePlayer != SYSADMIN_KEY) {
	    putError("you don't own that table");
	    tb = P_NULL;
	}
    }
    if (tb != P_NULL) {
	en = symEnterE(&MS, tb, &rq->rq_u.ru_text[sizeof(ULONG_T)],
			rq->rq_usedLen - sizeof(ULONG_T));
	en->en_key = rq->rq_key;
	incRef(rq->rq_key);
    }
}

/*
 * symbolDelete - handle a client request to delete a symbol.
 */

static void
symbolDelete(Client_t *cl, Request_t *rq)
{
    Table_t *tb;
    NameMap_t *nm;
    TableList_t *tl;
    Entry_t *en;
    ULONG_T key, table;
    UINT_T nameLen;

    key = rq->rq_key;
    table = rq->rq_u.ru_otherKey;
    TRACE("symbolDelete", key);
    tb = P_NULL;
    nm = P_NULL;
    if (table == TABLE_PUBLIC) {
	tb = PublicSymbolTable;
	nm = PublicNameMap;
    } else if (table == TABLE_PRIVATE) {
	tb = cl->cl_privateSymbols;
	nm = cl->cl_privateMap;
    } else if (table == BuiltinTable->tl_key) {
	putError("cannot modify Builtin table");
    } else if (table == PlayerTable->tl_key) {
	putError("cannot modify Characters table");
    } else {
	tb = findCachedTable(table);
	if (tb->tb_owner != RealPlayer && EffectivePlayer != SYSADMIN_KEY) {
	    putError("you don't own that table");
	    tb = P_NULL;
	} else {
	    tl = TableList;
	    while (tl != P_NULL && tl->tl_this != tb) {
		tl = tl->tl_next;
	    }
	    if (tl != P_NULL) {
		nm = tl->tl_nameMap;
	    }
	}
    }
    if (tb != P_NULL) {
	RealPlayer = cl->cl_player;    /* tested in 'deleteable' */
	EffectivePlayer = cl->cl_player;
	nameLen = rq->rq_usedLen - sizeof(ULONG_T);
	en = symLookupE(tb, &rq->rq_u.ru_text[sizeof(ULONG_T)], nameLen);
	if (en == P_NULL) {
	    putError("symbol not defined in that table");
	} else if (deleteable(en, table != TABLE_PUBLIC)) {
	    if (key == 0) {
		key = en->en_key;
	    }
	    if (nm != P_NULL) {
		removeNameFromMap(nm, key);
	    }
	    symDeleteE(&MS, tb, &rq->rq_u.ru_text[sizeof(ULONG_T)], nameLen);
	    clearRefs(key);
	    flushSymbol(key);
	}
    }
}

/*
 * addUse - add a use table to the given client.
 */

static void
addUse(Client_t *cl, ULONG_T key)
{
    UseList_t *ul;
    Table_t *tb;

    TRACE("addUse", key);
    tb = findCachedTable(key);
    ul = cl->cl_useList;
    while (ul != P_NULL && ul->ul_this != tb) {
	ul = ul->ul_next;
    }
    if (ul != P_NULL || tb == cl->cl_privateSymbols ||
	key == BuiltinTable->tl_key || key == PublicSymbolTableKey)
    {
	putError("that table is already in use");
    } else {
	ul = MEM_ALLOC_E(&MS, sizeof(UseList_t), "ul_t");
	ul->ul_this = tb;
	ul->ul_entry = addGlobalUse(tb, key);
	ul->ul_next = cl->cl_useList;
	cl->cl_useList = ul;
    }
}

/*
 * UseTable - the builtin
 */

ULONG_T
UseTable(ULONG_T key)
{

    if (ActiveClient != P_NULL) {
	if (key == NIL_TABLE) {
	    runError("nil table to 'UseTable'");
	    return BOOLEAN_FALSE;
	} else {
	    addUse(ActiveClient, key);
	    return BOOLEAN_TRUE;
	}
    } else {
	return BOOLEAN_FALSE;
    }
}

/*
 * removeUse - remove a use of a table from a client.
 */

static void
removeUse(Client_t *cl, ULONG_T key)
{
    UseList_t *ul;
    UseList_t **ulp;
    Entry_t *en;
    Request_t *rq;
    Table_t *tb;
    UINT_T count;

    TRACE("removeUse", key);
    tb = findCachedTable(key);
    ulp = &cl->cl_useList;
    while (B_TRUE) {
	ul = *ulp;
	if (ul == P_NULL || ul->ul_this == tb) {
	    break;
	}
	ulp = &ul->ul_next;
    }
    if (tb == cl->cl_privateSymbols) {
	putError("cannot unuse your private table");
    } else if (key == PublicSymbolTableKey) {
	putError("cannot unuse the public table");
    } else if (key == BuiltinTable->tl_key) {
	putError("cannot unuse the Builtin table");
    } else if (ul == P_NULL) {
	putError("that table is not currently in use");
    } else {
	*ulp = ul->ul_next;
	if (cl->cl_telnet == P_NULL) {
	    en = tb->tb_entries;
	    count = tb->tb_entryCount;
	    while (count != 0) {
		if (en->en_nameLen != EN_EMPTY && en->en_nameLen != EN_FREE) {
		    rq = allocRequestE(&MS, 0);
		    rq->rq_type = rt_flushSymbol;
		    rq->rq_key = en->en_key;
		    putMessageCarefully(cl, rq);
		}
		count -= 1;
		++en;
	    }
	}
	removeGlobalUse(ul->ul_entry);
	memFreeE(&MS, ul, sizeof(UseList_t));
    }
}

/*
 * UnUseTable - the builtin
 */

ULONG_T
UnUseTable(ULONG_T key)
{

    if (ActiveClient != P_NULL) {
	if (key == NIL_TABLE) {
	    runError("nil table to 'UnUseTable'");
	    return BOOLEAN_FALSE;
	} else {
	    removeUse(ActiveClient, key);
	    return BOOLEAN_TRUE;
	}
    } else {
	return BOOLEAN_FALSE;
    }
}

/*
 * EditString - the builtin
 */

void
EditString(CHAR_T *str, ULONG_T handler, ULONG_T mode, CHAR_T *prompt)
{
    Client_t *cl;
    Request_t *rq;
    UINT_T promptLen;

    TRACE("EditString", handler);
    promptLen = strLen1(prompt);
    cl = ActiveClient;
    if (cl == P_NULL || cl->cl_player == NIL_PLAYER) {
	runError("only players can edit strings");
    } else if (! cl->cl_canEdit) {
	runError("editing not available");
    } else if (cl->cl_editing) {
	runError("already editing");
    } else if (promptLen > 80) {
	runError("prompt unreasonably long");
    } else if (mode > EDIT_CODE) {
	runError("invalid mode value");
    } else if (checkStringBoolAction(handler, "EditString")) {
	cl->cl_editing = B_TRUE;
	cl->cl_editStringAction = handler;
	incProcRef(handler);
	rq = allocRequestE(&MS, strLen1(str) + promptLen);
	rq->rq_type = rt_editString;
	rq->rq_key = mode;
	strCpy(&rq->rq_u.ru_text[0], prompt);
	strCpy(&rq->rq_u.ru_text[promptLen], str);
	putMessageCarefully(cl, rq);
    }
}

/*
 * GetString - the builtin
 */

void
GetString(CHAR_T *str, ULONG_T handler, CHAR_T *prompt)
{
    Client_t *cl;
    Request_t *rq;
    UINT_T promptLen;

    TRACE("GetString", handler);
    promptLen = strLen1(prompt);
    cl = ActiveClient;
    if (cl == P_NULL || cl->cl_player == NIL_PLAYER) {
	runError("only players can get strings");
    } else if (! cl->cl_canGetString) {
	runError("GetString not available");
    } else if (cl->cl_getting) {
	runError("already GetStringing");
    } else if (promptLen > 80) {
	runError("prompt unreasonably long");
    } else if (checkStringBoolAction(handler, "GetString")) {
	cl->cl_getting = B_TRUE;
	cl->cl_getStringAction = handler;
	incProcRef(handler);
	rq = allocRequestE(&MS, strLen1(str) + promptLen);
	rq->rq_type = rt_getString;
	strCpy(&rq->rq_u.ru_text[0], prompt);
	strCpy(&rq->rq_u.ru_text[promptLen], str);
	putMessageCarefully(cl, rq);
    }
}

/*
 * QueryFile - the builtin
 */

void
QueryFile(CHAR_T *name, ULONG_T handler)
{
    Client_t *cl;
    Request_t *rq;

    TRACE("QueryFile", handler);
    cl = ActiveClient;
    if (cl == P_NULL || cl->cl_player == NIL_PLAYER) {
	runError("only players can query files");
    } else if (! cl->cl_canQueryFile) {
	runError("QueryFile not available");
    } else if (cl->cl_querying) {
	runError("already QueryFileing");
    } else if (checkBoolAction(handler, "QueryFile")) {
	cl->cl_querying = B_TRUE;
	cl->cl_queryFileAction = handler;
	incProcRef(handler);
	rq = allocRequestE(&MS, strLen1(name));
	rq->rq_type = rt_queryFile;
	strCpy(&rq->rq_u.ru_text[0], name);
	putMessageCarefully(cl, rq);
    }
}

/*
 * handleOneRequest - handle a single request from a client.
 */

void
handleOneRequest(Client_t *cl, Request_t *rq)
{
    CHAR_T *name;
    Proc_t *pr;
    BYTE_T *bytes;
    Exec_t *saveBody;
    UINT_T *pLockCount;
    ULONG_T key;
    UINT_T len, mLen, resultType;
    RequestType_t rt;
    QValue_t qv;
    Exec_t fakeBody;
    Integer_t intValue;

    TRACE("handleOneRequest", rq->rq_type - rt_first);
    rt = rq->rq_type;
    if (rt >= rt_illegal) {
	logN("request type code ", rt - rt_first, "\n");
	myAbort("invalid request type");
    }
    RequestsProcessed += 1;
    ActiveObject = NIL_THING;
    ActiveClient = cl;
    TrueClient = cl;

    /* actual request processing */

    /* any messages that require an immediate reply are ReplyMsg'd right
       away, then rq is niled. After the 'switch', if rq is not nil, the
       request is freed. */

    switch (rt) {

    /* message types that require an immediate reply */

    case rt_beginClient:
	MDEBUG("BEGINCLIENT\n");
	(void) beginClient(cl, rq);
	rq = P_NULL;
	break;
    case rt_endClient:
	MDEBUG("ENDCLIENT\n");
	rq->rq_usedLen = 0;
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	endClient(cl);
	ActiveClient = P_NULL;
	break;
    case rt_askPlayer:
	MDEBUG2("ASKPLAYER ", &rq->rq_u.ru_text[0]);
	askPlayer(cl, rq);
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	break;
    case rt_creationCheck:
	MDEBUG("CREATIONCHECK\n");
	creationCheck(cl, rq);
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	break;
    case rt_createPlayer:
	MDEBUG2("CREATEPLAYER ", &rq->rq_u.ru_text[PASSWORD_LEN]);
	createPlayer(cl, rq);
	rq = P_NULL;
	break;
    case rt_loginPlayer:
	MDEBUG2("LOGINPLAYER ", cl->cl_name);
	loginPlayer(cl, rq);
	rq = P_NULL;
	break;
    case rt_lookup:
	MDEBUG2("LOOKUP ", &rq->rq_u.ru_text[0]);
	rq->rq_key = lookupForClient(cl, &rq->rq_u.ru_text[0], rq->rq_usedLen);
	rq->rq_usedLen = 0;
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	break;
    case rt_findName:
	MDEBUGX("FINDNAME ", rq->rq_key);
	name = findAnyName1(cl, rq->rq_key);
	if (name == P_NULL) {
	    rq->rq_u.ru_text[0] = '\0';
	    rq->rq_usedLen = 1;
	} else {
	    len = strLen1(name);
	    if (len > rq->rq_availLen) {
		freeRequestE(&MS, rq);
		rq = allocRequestE(&MS, len);
		rq->rq_type = rt_findName;
	    }
	    strCpy(&rq->rq_u.ru_text[0], name);
	    rq->rq_usedLen = len;
	}
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	break;
    case rt_readProc:
	MDEBUGX("READPROC ", rq->rq_key);
	key = rq->rq_key;
	pr = findCachedProc(key, &pLockCount);
	saveBody = pr->pr_body;
	fakeBody.ex_kind = ex_error;
	pr->pr_body = &fakeBody;
	writeProcE(&MS, pr);
	writeResultE(&MS, &bytes, &mLen);
	len = mLen;
	pr->pr_body = saveBody;
	/* Must keep the lock until after the writeProcE call. */
	*pLockCount -= 1;
	if (len > rq->rq_availLen) {
	    /* must expand and replace the client request */
	    freeRequestE(&MS, rq);
	    rq = allocRequestE(&MS, len);
	    rq->rq_type = rt_readProc;
	}
	copy(&rq->rq_u.ru_bytes[0], bytes, len);
	rq->rq_usedLen = len;
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	break;
    case rt_writeProcNew:
	rq->rq_key =
	    ioCreate(TYPE_ACTION, &rq->rq_u.ru_bytes[0], rq->rq_usedLen);
	rq->rq_usedLen = 0;
	MDEBUGX("WRITEPROCNEW => ", rq->rq_key);
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	break;
    case rt_runProc:
	MDEBUG("RUNPROC\n");
	pr = readProcE(&MS, &rq->rq_u.ru_bytes[0]);
	if (cl->cl_wizardMode) {
	    setupForCall(cl);
	    runExpr(pr->pr_body, &qv, &resultType);
	    messageFlush();		    /* make indent work! */
	    cleanupAfterCall(); 	    /* in case someone forgot */
	    EffectiveStatus = ps_wizard;    /* get rid of the '@' on output */
	    switch (rq->rq_key) {
	    case 4:
		/* client wants to edit the proc */
		if (resultType == TYPE_ACTION) {
		    EditProc(qv.qv_integer);
		} else {
		    putError("value is not a proc");
		    if (resultType == TYPE_STRING) {
			freeString(qv.qv_string);
		    }
		}
		break;
	    case 3:
		/* client wants to see the key */
		putString("Raw hex value: ");
		putHex(qv.qv_integer);
		newLine();
		if (resultType == TYPE_STRING || resultType == TYPE_THING ||
		    (resultType >= TYPE_LIST_INTEGER &&
			resultType <= TYPE_LIST_ACTION))
		{
		    freeValue(resultType, qv.qv_integer);
		}
		break;
	    case 2:
		/* client wants it described */
		fmtStart();
		/* There is special describe output for some types. */
		switch (resultType) {
		case TYPE_PLAYER:
		case TYPE_THING:
		case TYPE_ACTION:
		case TYPE_TABLE:
		case TYPE_GRAMMAR:
		case TYPE_PROP_BOOLEAN:
		case TYPE_PROP_INTEGER:
		case TYPE_PROP_FIXED:
		case TYPE_PROP_THING:
		case TYPE_PROP_ACTION:
		case TYPE_PROP_STRING:
		case TYPE_PROP_TABLE:
		case TYPE_PROP_GRAMMAR:
		case TYPE_PROP_LIST_INTEGER:
		case TYPE_PROP_LIST_FIXED:
		case TYPE_PROP_LIST_THING:
		case TYPE_PROP_LIST_ACTION:
		    describeValue(qv.qv_integer,
				  findAnyName1(cl, qv.qv_integer));
		    break;
		default:
		    fmtValue(resultType, qv.qv_integer);
		    fmtLine();
		    break;
		}
		freeValue(resultType, qv.qv_integer);
		break;
	    case 1:
		/* client wants the result printed */
		if (resultType != TYPE_VOID && resultType != TYPE_ERROR) {
		    fmtStart();
		    fmtString("==> ");
		    fmtValue(resultType, qv.qv_integer);
		    fmtLine();
		    freeValue(resultType, qv.qv_integer);
		}
		break;
	    case 0:
		/* client will be putting the value into a symbol table. */
		rq->rq_key = qv.qv_integer;
		if (resultType == TYPE_INTEGER || resultType == TYPE_FIXED) {
		    /* Why is this limitation here? It is because when we
		       are executing code, we do not want to have to use
		       ioRead to fetch the value of an int or fixed
		       constant whenever we want to use it. So, we use the
		       spare 3 bytes in the Exec_t to cache the actual
		       value. That field is only ever referenced by the
		       code in exec.c, print.c and procIO.c . This sort
		       of thing is the reason for the ms_getIntConstValue
		       routine - it is used by readExec to get the proper
		       value from the database for insertion into the
		       Exec_t. In the future, we should be able to come up
		       with a less limited scheme. E.g. have the expanded
		       Exec_t point at a structure containing both the
		       full 32 bit value and its constant key. This can
		       easily ride along with the Exec_t and be freed as
		       needed, just like the expanded structures for
		       binary operators, etc. */
		    if (qv.qv_integer > (LONG_T) 0x00ffffff ||
			qv.qv_integer < (LONG_T) 0xff000000)
		    {
			if (resultType == TYPE_INTEGER) {
			    putError(
				"integer value out of range for definition");
			} else {
			    putError(
				"fixed value out of range for definition");
			}
			rq->rq_key = 0;
		    } else {
			intValue.int_value = qv.qv_integer;
			intValue.int_owner = RealPlayer;
			intValue.int_useCount = 0;
			rq->rq_key = ioCreate(resultType, &intValue,
					      sizeof(Integer_t));
		    }
		} else if (resultType == TYPE_STRING) {
		    freeString(qv.qv_string);
		    putError("cannot define a symbol of that type");
		    rq->rq_key = 0;
		} else if (resultType == TYPE_VOID) {
		    putError("cannot define a symbol of that type");
		    /* The value is already 0 */
		}
		break;
	    }
	} else {
	    putError("You are not in wizard mode!\n");
	    rq->rq_key = 0;
	}
	/* NOTE: we pass 'B_TRUE' here so that references will be
	   decremented. The corresponding call in the client passes
	   'B_FALSE', so that we don't get multiple decrements. */
	/* NOTE: if the proc we ran consisted only of a direct proc (e.g.
	   "proc Print(Me()@p_pName); corp"), or a direct proc whose
	   contents is only another direct proc, etc. then we have a
	   problem if we let the following 'freeProcE' call delete refs
	   in 'pr' (the only ref will be to the direct proc). Since the
	   direct proc has no other refs to it, it will be freed. The
	   problem is that we will be yielding the pointer to it back to
	   the client, who can try to use it, etc. No good. So, we have
	   to detect that odd situation, and *not* free refs in 'pr' in
	   that case. Ugh!!!! What I have actually done is to simply
	   not let the case happen, in the parser, so nothing is done
	   here. Further note: Only in the 'define symbol' case is this
	   a problem. In the other cases, having the proc freed is just
	   what we want. */
	freeProcE(&MS, pr, NIL_ACTION, BOOLEAN_TRUE);
	rq->rq_usedLen = 0;
	putMessageCarefully(cl, rq);
	rq = P_NULL;
	break;
    case rt_inputLine:
	MDEBUG2("INPUTLINE ", &rq->rq_u.ru_text[0]);
	if (cl->cl_inputAction != NIL_ACTION) {
	    setupForCall(cl);
	    name = stringCopy(&rq->rq_u.ru_text[0]);
	    (void) doProc1Top(cl->cl_inputAction, name);
	    cleanupAfterCall();
	}
	if (cl->cl_inputSync) {
	    flushEffects();
	    messageFlush();
	    rq->rq_usedLen = 0;
	    putMessageCarefully(cl, rq);
	    rq = P_NULL;
	}
	break;
    case rt_setSync:
	MDEBUGX("SETSYNC ", rq->rq_key);
	cl->cl_inputSync = (BOOL_T) (rq->rq_key != 0);
	break;
    case rt_rawKey:
	MDEBUGX("RAWKEY ", rq->rq_key);
	if (cl->cl_rawKeyAction != NIL_ACTION) {
	    setupForCall(cl);
	    (void) doProc1Top(cl->cl_rawKeyAction, (CHAR_T *) rq->rq_key);
	    cleanupAfterCall();
	}
	break;
    case rt_regionSelect:
	MDEBUGX("REGIONSELECT ", rq->rq_key);
	if (cl->cl_mouseDownAction != NIL_ACTION) {
	    setupForCall(cl);
	    (void) doProc3Top(cl->cl_mouseDownAction,
			      (CHAR_T *) rq->rq_key,
			      (CHAR_T *) (rq->rq_u.ru_otherKey >> 16),
			      (CHAR_T *) (rq->rq_u.ru_otherKey & 0xffff));
	    cleanupAfterCall();
	}
	break;
    case rt_buttonHit:
	MDEBUGX("BUTTONHIT ", rq->rq_key);
	if (cl->cl_buttonAction != NIL_ACTION) {
	    setupForCall(cl);
	    (void) doProc1Top(cl->cl_buttonAction, (CHAR_T *) rq->rq_key);
	    cleanupAfterCall();
	}
	break;
    case rt_flushEffect:
	MDEBUGX("FLUSHEFFECT ", rq->rq_key);
	if (flushEffect(cl, rq->rq_key)) {
	    putMessageCarefully(cl, rq);
	    rq = P_NULL;
	}
	break;
    case rt_textResize:
	MDEBUGX("TEXTRESIZE ", rq->rq_key);
	cl->cl_textHeight = rq->rq_key >> 16;
	cl->cl_textWidth = rq->rq_key & 0xffff;
	cl->cl_gotTextSize = B_TRUE;
	break;
    case rt_incRef:
	MDEBUGX("INCREF ", rq->rq_key);
	incRef(rq->rq_key);
	break;
    case rt_decRef:
	MDEBUGX("DECREF ", rq->rq_key);
	decRef(rq->rq_key);
	break;
    case rt_symbolEnter:
	MDEBUG2("SYMBOLENTER ", &rq->rq_u.ru_text[sizeof(ULONG_T)]);
	symbolEnter(cl, rq);
	break;
    case rt_symbolDelete:
	MDEBUG2("SYMBOLDELETE ", &rq->rq_u.ru_text[sizeof(ULONG_T)]);
	symbolDelete(cl, rq);
	break;
    case rt_useTable:
	MDEBUGX("USETABLE ", rq->rq_key);
	addUse(cl, rq->rq_key);
	break;
    case rt_unuseTable:
	MDEBUGX("UNUSETABLE ", rq->rq_key);
	removeUse(cl, rq->rq_key);
	break;
    case rt_describeBuiltin:
	MDEBUGX("DESCRIBEBUILTIN ", rq->rq_key);
	messageFlush(); 		/* make indent work! */
	cleanupAfterCall();		/* in case someone forgot */
	EffectiveStatus = ps_wizard;	/* get rid of '@' on output */
	describeBuiltin(rq->rq_key, &rq->rq_u.ru_text[0]);
	break;
    case rt_describe:
	MDEBUGX("DESCRIBE ", rq->rq_key);
	if (cl->cl_player == SYSADMIN_KEY) {
	    putHex(rq->rq_key);
	    putString(" => ");
	    name = findAnyName(rq->rq_key);
	    if (name != P_NULL) {
		putString(name);
	    } else {
		ioDisplay(rq->rq_key);
	    }
	    newLine();
	} else {
	    putError("only SysAdmin can do that");
	}
	break;
    case rt_changePassword:
	MDEBUG("CHANGEPASSWORD\n");
	changePassword(cl, rq);
	break;
    case rt_graphicsFlip:
	MDEBUG("GRAPHICSFLIP\n");
	if (rq->rq_key == 0) {
	    cl->cl_graphicsOn = B_FALSE;
	} else {
	    cl->cl_graphicsOn = B_TRUE;
	}
	break;
    case rt_voiceFlip:
	MDEBUG("VOICEFLIP\n");
	if (rq->rq_key == 0) {
	    cl->cl_voiceOn = B_FALSE;
	} else {
	    cl->cl_voiceOn = B_TRUE;
	}
	break;
    case rt_soundFlip:
	MDEBUG("SOUNDFLIP\n");
	if (rq->rq_key == 0) {
	    cl->cl_soundOn = B_FALSE;
	} else {
	    cl->cl_soundOn = B_TRUE;
	}
	break;
    case rt_musicFlip:
	MDEBUG("MUSICFLIP\n");
	if (rq->rq_key == 0) {
	    cl->cl_musicOn = B_FALSE;
	} else {
	    cl->cl_musicOn = B_TRUE;
	}
	break;
    case rt_doEditProc:
	MDEBUGX("DOEDITPROC ", rq->rq_key);
	key = rq->rq_key;
	if ((key & TYPE_MASK) == (TYPE_ACTION << TYPE_SHIFT)) {
	    if (key == NIL_ACTION) {
		/* client is already doing it */
		if (cl->cl_editing) {
		    myAbort("client already editing");
		}
		cl->cl_editing = B_TRUE;
	    } else {
		EditProc(rq->rq_key);
	    }
	} else {
	    putError("value is not a proc");
	}
	break;
    case rt_editStringDone:
	MDEBUG("EDITSTRINGDONE\n");
	key = cl->cl_editStringAction;
	if (key != NIL_ACTION) {
	    cl->cl_editing = B_FALSE;	/* Before the call! */
	    setupForCall(cl);
	    name = stringCopy(&rq->rq_u.ru_text[0]);
	    /* do these here so we can start another EditString from inside the
	       handler for this one */
	    cl->cl_editStringAction = NIL_ACTION;
	    (void) doProc2Top(key, name,
		(CHAR_T *) (rq->rq_flag ? BOOLEAN_TRUE : BOOLEAN_FALSE));
	    decProcRef(key);
	    cleanupAfterCall();
	}
	break;
    case rt_replaceProc:
	MDEBUGX("REPLACEPROC ", rq->rq_key);
	/* The complexity here is because we want, in the database, the
	   code from the replacement, but the usecount from the original.
	   We *know* (see /library/procIO.c) that the useCount field is
	   the first in the bytestream, so we can poke it directly. */
	key = rq->rq_key;
	if (removeProc(key, &len)) {
	    /* Replace the pr_useCount directly. */
	    rq->rq_u.ru_bytes[0] = len >> 8;	/* big endian! */
	    rq->rq_u.ru_bytes[1] = len;
	    ioReplace(key, &rq->rq_u.ru_bytes[0], rq->rq_usedLen);
	} else {
	    name = "*** edits discarded - proc is byte-code locked ***\n";
	    sendText(cl, name, strLen1(name));
	}
	break;
    case rt_editProcDone:
	MDEBUG("EDITPROCDONE\n");
	cl->cl_editing = B_FALSE;
	break;
    case rt_getStringDone:
	MDEBUG("GETSTRINGDONE\n");
	key = cl->cl_getStringAction;
	if (key != NIL_ACTION) {
	    cl->cl_getting = B_FALSE;	/* Before the call! */
	    setupForCall(cl);
	    /* do these here so we can start another GetString from inside the
	       handler for this one */
	    cl->cl_getStringAction = NIL_ACTION;
	    name = stringCopy(&rq->rq_u.ru_text[0]);
	    (void) doProc2Top(key, name,
		(CHAR_T *) (rq->rq_flag ? BOOLEAN_TRUE : BOOLEAN_FALSE));
	    decProcRef(key);	/* match the inc done on the request */
	    cleanupAfterCall();
	}
	break;
    case rt_effectDone:
	MDEBUGX("EFFECTDONE ", rq->rq_key);
	if (cl->cl_effectDoneAction != NIL_ACTION) {
	    setupForCall(cl);
	    (void) doProc2Top(cl->cl_effectDoneAction,
			      (CHAR_T *) rq->rq_key,
			      (CHAR_T *) rq->rq_u.ru_otherKey);
	    cleanupAfterCall();
	}
	break;
    case rt_queryFileDone:
	MDEBUGX("FILEQUERYDONE ", rq->rq_flag - B_FALSE);
	key = cl->cl_queryFileAction;
	if (key != NIL_ACTION) {
	    setupForCall(cl);
	    cl->cl_queryFileAction = NIL_ACTION;
	    (void) doProc1Top(key,
		(CHAR_T *) (rq->rq_flag ? BOOLEAN_TRUE : BOOLEAN_FALSE));
	    decProcRef(key);
	    cleanupAfterCall();
	}
	cl->cl_querying = B_FALSE;
	break;
    default:
	logN("unknown request type ", rt - rt_first, "\n");
	break;
    }
    if (rq != P_NULL) {
	freeRequestE(&MS, rq);
    }
    flushEffects();
    messageFlush();
    ActiveClient = P_NULL;
    TrueClient = P_NULL;
}
