#include <netinet/in.h>

#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "Request.h"
#include "Effects.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * effects.c - code to do server part of graphics/sound/voice, etc.
 */

/* There used to be a comment here that said that EFFECTS_BUFFER_SIZE must
   be less than TEXT_LENGTH. I can't now think of a reason why. Perhaps it
   was because way back when all requests were the same size?

   Later: it probably has to do with the fact that TEXT_LENGTH is the
   declared buffer size in a message. Sending a message with a buffer
   longer than that involves some arithmetic with negative numbers, which
   *should* work, but... */

#define EFFECTS_BUFFER_SIZE	4000
#define EFFECT_CALLS_MAX	100

typedef struct DefineStack {
    struct DefineStack *ds_next;
    BYTE_T *ds_data;
    ULONG_T *ds_calls;
    ULONG_T ds_key;
    UINT_T ds_dataLength;
    UINT_T ds_callsCount;
} DefineStack_t;

static ULONG_T EffectsKey, DefiningKey;
static Client_t *EffectsClient;
static DefineStack_t *DefineStack;
static UINT_T EffectsPos, CallsCount;

static BYTE_T EffectsBuffer[EFFECTS_BUFFER_SIZE];
static ULONG_T CallsBuffer[EFFECT_CALLS_MAX];

static BOOL_T Defining;

/*
 * effectsInit - initialize the stuff here.
 */

void
effectsInit(void)
{

    EffectsKey = NIL_THING;
    EffectsClient = P_NULL;
    EffectsPos = 0;
    Defining = B_FALSE;
    DefineStack = P_NULL;
}

/*
 * freeDefineStack - free up any nested DefineEffect info.
 */

void
freeDefineStack(void)
{
    DefineStack_t *ds;

    while (B_TRUE) {
	ds = DefineStack;
	if (ds == P_NULL) {
	    break;
	}
	DefineStack = ds->ds_next;
	if (ds->ds_dataLength != 0) {
	    memFreeE(&MS, ds->ds_data, ds->ds_dataLength);
	}
	if (ds->ds_callsCount != 0) {
	    memFreeE(&MS, ds->ds_calls, ds->ds_callsCount * sizeof(ULONG_T));
	}
	memFreeE(&MS, ds, sizeof(DefineStack_t));
    }
    Defining = B_FALSE;
    EffectsPos = 0;
    CallsCount = 0;
}

/*
 * effectsTerm - terminate the stuff here.
 */

void
effectsTerm(void)
{

    freeDefineStack();
}

/*
 * flushEffects1 - internal form of flushEffects.
 */

static void
flushEffects1(void)
{
    Request_t *rq;

    if (EffectsPos != 0 && EffectsClient->cl_player != NIL_PLAYER &&
	EffectsClient->cl_telnet == P_NULL && ! EffectsClient->cl_deleted)
    {
	rq = allocRequestE(&MS, EffectsPos);
	rq->rq_type = rt_effects;
	rq->rq_key = EffectsClient->cl_lastUsedDrawComponent;
	copy(&rq->rq_u.ru_bytes[0], &EffectsBuffer[0], EffectsPos);
	putMessageCarefully(EffectsClient, rq);
    }
    EffectsPos = 0;
}

/*
 * flushEffects - flush the effects buffer to the destined client.
 *	External entry point.
 */

void
flushEffects(void)
{

    TRACE("flushEffects", EffectsClient);
    if (EffectsPos != 0) {
	flushEffects1();
    }
    EffectsClient = P_NULL;
    EffectsKey = NIL_THING;
}

/*
 * effectsStart - setup for a single effects activity.
 */

static Client_t *
effectsStart(ULONG_T key, UINT_T len)
{
    Client_t *cl;

    if (len <= EFFECTS_BUFFER_SIZE) {
	if (Defining) {
	    if (key != NIL_THING) {
		runError("non-nil 'who' while defining an effect");
		freeDefineStack();
		return P_NULL;
	    } else {
		if (EffectsPos + len > EFFECTS_BUFFER_SIZE) {
		    runError("buffer overflow while defining an effect");
		    freeDefineStack();
		    return P_NULL;
		} else {
		    return EffectsClient;
		}
	    }
	} else {
	    if (key == NIL_THING) {
		cl = ActiveClient;
		if (cl == P_NULL) {
		    return P_NULL;
		}
		key = cl->cl_thing;
		if (key != EffectsKey) {
		    if (EffectsPos != 0) {
			flushEffects1();
		    }
		} else {
		    if (cl->cl_defaultDrawComponent !=
			cl->cl_lastUsedDrawComponent)
		    {
			if (EffectsPos != 0) {
			    flushEffects1();
			}
		    }
		}
		EffectsKey = key;
		EffectsClient = cl;
	    } else if (key == EffectsKey) {
		/* more effects stuff for the same client */
		cl = EffectsClient;
		if (cl->cl_defaultDrawComponent !=
		    cl->cl_lastUsedDrawComponent)
		{
		    if (EffectsPos != 0) {
			flushEffects1();
		    }
		}
	    } else {
		cl = getClientByThing(key);
		if (cl == P_NULL) {
		    return P_NULL;
		}
		if (EffectsPos != 0) {
		    flushEffects1();
		}
		EffectsKey = key;
		EffectsClient = cl;
	    }
	    cl->cl_lastUsedDrawComponent = cl->cl_defaultDrawComponent;
	    if (EffectsPos + len > EFFECTS_BUFFER_SIZE) {
		flushEffects1();
	    }
	    return cl;
	}
    } else {
	runError("buffer overflow while defining an effect");
	freeDefineStack();
	return P_NULL;
    }
}

/*
 * effectsPtr - handy shortcut routine.
 */

static Client_t *
effectsPtr(ULONG_T key)
{

    if (key == NIL_THING) {
	return ActiveClient;
    }
    if (key != EffectsKey) {
	return getClientByThing(key);
    }
    return EffectsClient;
}

/*
 * effectsByte - add a byte to the effects buffer.
 */

static void
effectsByte(BYTE_T b)
{

    EffectsBuffer[EffectsPos] = b;
    EffectsPos += 1;
}

/*
 * effectsWord - add a word to the effects buffer. Big endian!
 */

static void
effectsWord(UINT_T w)
{

    EffectsBuffer[EffectsPos] = w >> 8;
    EffectsBuffer[EffectsPos + 1] = w;
    EffectsPos += 2;
}

/*
 * effectsLong - add a long to the effects buffer. Big endian!
 */

static void
effectsLong(ULONG_T u)
{

    effectsWord(u >> 16);
    effectsWord(u);
}

/*
 * effectsString - add a string to the effects buffer.
 */

static void
effectsString(CHAR_T *p)
{
    BYTE_T *ep;
    UINT_T pos;
    CHAR_T ch;

    pos = EffectsPos;
    ep = &EffectsBuffer[pos];
    do {
	ch = *p++;
	*ep++ = ch;
	pos += 1;
    } while (ch != '\0');
    EffectsPos = pos;
}

/*
 * effectsLongs - add a chunk of data to the effects buffer.
 *	The data is in the form of a 'list int', and so the normal
 *	byteswapping is needed. However, only the number of bytes
 *	indicated should be sent.
 */

static void
effectsLongs(ULONG_T *p, UINT_T len)
{
    ULONG_T extra;
    UINT_T cnt;

    cnt = len / sizeof(ULONG_T);
    while (cnt != 0) {
	effectsLong(*p);
	++p;
	cnt -= 1;
    }
    len &= (sizeof(ULONG_T) - 1);
    extra = *p;
    if (len >= 1) {
	effectsByte(extra >> 24);
	if (len >= 2) {
	    effectsByte(extra >> 16);
	    if (len == 3) {
		effectsByte(extra >> 8);
	    }
	}
    }
}

/* first, some somewhat generic effects-like routines */

ULONG_T
ClientVersion(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	return (ULONG_T) cl->cl_clientVersion;
    } else {
	return 0;
    }
}

ULONG_T
TextComponent(ULONG_T who, ULONG_T new)
{
    Client_t *cl;
    ULONG_T old;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	old = cl->cl_defaultTextComponent;
	if (new != old) {
	    messageFlush();
	}
	cl->cl_defaultTextComponent = new;
	return old;
    } else {
	return -1;
    }
}

ULONG_T
DrawComponent(ULONG_T who, ULONG_T new)
{
    Client_t *cl;
    ULONG_T old;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	old = cl->cl_defaultDrawComponent;
	cl->cl_defaultDrawComponent = new;
	return old;
    } else {
	return -1;
    }
}

void
CreateContainer(ULONG_T encloser, ULONG_T newId, ULONG_T isVertical,
		ULONG_T alignment)
{
    Client_t *cl;
    Request_t *rq;

    if ((LONG_T) alignment < -1 || (LONG_T) alignment > +1) {
	runError("Invalid alignment to CreateContainer");
	return;
    }
    cl = ActiveClient;
    if (cl == P_NULL) {
	return;
    }
    rq = allocRequestE(&MS, sizeof(ULONG_T));
    rq->rq_type = rt_createContainer;
    rq->rq_key = newId;
    rq->rq_flag = (BOOL_T) (isVertical != BOOLEAN_FALSE);
    rq->rq_uint = (UINT_T) alignment;
    rq->rq_u.ru_otherKey = encloser;
    putMessageCarefully(cl, rq);
}

void
CreateComponent(ULONG_T encloser, ULONG_T newId,
		ULONG_T width, ULONG_T height, ULONG_T kind)
{
    Client_t *cl;
    Request_t *rq;
    GUIFrame_t *guf;
    TextComponentList_t *tcl;
    DrawComponentList_t *dcl;

    /*
      1 - PROMPTED_TEXT
      2 - GENERAL_TEXT
      3 - ICON_LIST
      4 - ICONED_CANVAS
      5 - BUTTON_PANEL
     */
    if (kind <= 0 || kind > 5) {
	runError("Invalid kind to CreateComponent");
	return;
    }
    cl = ActiveClient;
    if (cl == P_NULL) {
	return;
    }
    rq = allocRequestE(&MS, sizeof(ULONG_T) + sizeof(UINT_T) * 2);
    rq->rq_type = rt_createComponent;
    rq->rq_key = newId;
    rq->rq_uint = (UINT_T) kind;
    rq->rq_u.ru_otherKey = encloser;
    rq->rq_u.ru_ints[2] = (UINT_T) width;
    rq->rq_u.ru_ints[3] = (UINT_T) height;
    putMessageCarefully(cl, rq);
    guf = cl->cl_newFrame;
    if (guf == P_NULL) {
	guf = MEM_ALLOC_E(&MS, sizeof(GUIFrame_t), "guf_");
	guf->guf_textComponents = P_NULL;
	guf->guf_drawComponents = P_NULL;
	cl->cl_newFrame = guf;
    }
    if (kind == 2) {
	tcl = MEM_ALLOC_E(&MS, sizeof(TextComponentList_t), "tcl_");
	tcl->tcl_next = guf->guf_textComponents;
	tcl->tcl_id = newId;
	tcl->tcl_width = width;
	tcl->tcl_height = height;
	guf->guf_textComponents = tcl;
    } else if (kind == 4 || kind == 5) {
	dcl = MEM_ALLOC_E(&MS, sizeof(DrawComponentList_t), "dcl_");
	dcl->dcl_next = guf->guf_drawComponents;
	dcl->dcl_id = newId;
	dcl->dcl_cols = width;
	dcl->dcl_rows = height;
	guf->guf_drawComponents = dcl;
    }
}

void
MakeFrame(void)
{
    Client_t *cl;
    Request_t *rq;

    cl = ActiveClient;
    if (cl == P_NULL) {
	return;
    }
    flushEffects1();
    rq = allocRequestE(&MS, 0);
    rq->rq_type = rt_makeFrame;
    putMessageCarefully(cl, rq);
    freeGUIFrame(cl->cl_activeFrame);
    cl->cl_activeFrame = cl->cl_newFrame;
    cl->cl_newFrame = P_NULL;
    setNewPrompt(cl->cl_prompt);
}

ULONG_T
GetFontAscent(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	return cl->cl_fontAscent;
    }
    return 0;
}

ULONG_T
GetFontDescent(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	return cl->cl_fontDescent;
    }
    return 0;
}

ULONG_T
GetFontLeading(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	return cl->cl_fontLeading;
    }
    return 0;
}

ULONG_T
GetFontHeight(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	return cl->cl_fontHeight;
    }
    return 0;
}

ULONG_T
GetFontWidth(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL) {
	return cl->cl_fontWidth;
    }
    return 0;
}

void
IfFound(ULONG_T who)
{

    if (effectsStart(who, 1) != P_NULL) {
	effectsByte(ef_IfFound - ef_null);
    }
}

void
Else(ULONG_T who)
{

    if (effectsStart(who, 1) != P_NULL) {
	effectsByte(ef_Else - ef_null);
    }
}

void
Fi(ULONG_T who)
{

    if (effectsStart(who, 1) != P_NULL) {
	effectsByte(ef_Fi - ef_null);
    }
}

void
FailText(ULONG_T who, CHAR_T *message)
{

    if (effectsStart(who, 1 + strLen1(message)) != P_NULL) {
	effectsByte(ef_FailText - ef_null);
	effectsString(message);
    }
}

void
AbortEffect(ULONG_T who, ULONG_T kind, ULONG_T id)
{

    if (effectsStart(who, 1 + 1 + 4) != P_NULL) {
	effectsByte(ef_Abort - ef_null);
	effectsByte(kind);
	effectsLong(id);
    }
}

void
CallEffect(ULONG_T who, ULONG_T key)
{
    DefineStack_t *ds;

    if (effectsStart(who, 1 + 4) != P_NULL) {
	if (Defining) {
	    ds = DefineStack;
	    while (ds != P_NULL && ds->ds_key != key) {
		ds = ds->ds_next;
	    }
	    if (ds != P_NULL || key == DefiningKey) {
		runError("effect cannot call itself");
		freeDefineStack();
		return;
	    }
	    if (CallsCount == EFFECT_CALLS_MAX) {
		runError("too many calls in effect");
		freeDefineStack();
		return;
	    }
	    CallsBuffer[CallsCount] = htonl(key);
	    CallsCount += 1;
	}
	effectsByte(ef_Call - ef_null);
	effectsLong(key);
    }
}

ULONG_T
KnowsEffect(ULONG_T who, ULONG_T key)
{
    KnownEffectList_t *kel, *prev, *next;
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl == P_NULL) {
	return BOOLEAN_FALSE;
    } else {
	kel = cl->cl_knownEffects;
	while (kel != P_NULL && kel->kel_key != key) {
	    kel = kel->kel_next;
	}
	if (kel == P_NULL) {
	    return BOOLEAN_FALSE;
	} else {
	    prev = kel->kel_prev;
	    if (prev != P_NULL) {
		/* move it to the head of the queue */
		next = kel->kel_next;
		prev->kel_next = next;
		if (next != P_NULL) {
		    next->kel_prev = prev;
		}
		next = cl->cl_knownEffects;
		kel->kel_next = next;
		next->kel_prev = kel;
		kel->kel_prev = P_NULL;
		cl->cl_knownEffects = kel;
	    }
	    return BOOLEAN_TRUE;
	}
    }
}

void
DefineEffect(ULONG_T who, ULONG_T key)
{
    DefineStack_t *ds;
    Client_t *cl;

    if (Defining) {
	if (who != EffectsKey) {
	    runError("nested DefineEffect must be for same client");
	    freeDefineStack();
	    return;
	}
	ds = MEM_ALLOC_E(&MS, sizeof(DefineStack_t), "ds_t");
	ds->ds_next = DefineStack;
	ds->ds_key = DefiningKey;
	ds->ds_dataLength = EffectsPos;
	ds->ds_callsCount = CallsCount;
	if (EffectsPos != 0) {
	    ds->ds_data = MEM_ALLOC_E(&MS, EffectsPos, "dsef");
	    copy(ds->ds_data, &EffectsBuffer[0], EffectsPos);
	    EffectsPos = 0;
	}
	if (CallsCount != 0) {
	    ds->ds_calls =
		MEM_ALLOC_E(&MS, CallsCount * sizeof(ULONG_T), "dscl");
	    copy(ds->ds_calls, &CallsBuffer[0], CallsCount * sizeof(ULONG_T));
	    CallsCount = 0;
	}
	DefineStack = ds;
    }
    if (who == NIL_THING) {
	cl = ActiveClient;
	if (cl == P_NULL) {
	    runError("DefineEffect called when no client is active");
	    freeDefineStack();
	    return;
	}
    } else {
	cl = getClientByThing(who);
	if (cl == P_NULL) {
	    runError("given client is not active");
	    freeDefineStack();
	    return;
	}
    }
    if (EffectsPos != 0) {
	flushEffects1();
    }
    EffectsKey = who;
    EffectsClient = cl;
    DefiningKey = key;
    Defining = B_TRUE;
}

void
EndEffect(void)
{
    Client_t *cl;
    Request_t *rq;
    KnownEffectList_t *kel;
    DefineStack_t *ds;
    ULONG_T len, total;

    cl = EffectsClient;
    if (cl == P_NULL) {
	Defining = B_FALSE;
    }
    if (! Defining) {
	runError("EndEffect called when not defining an effect");
	return;
    }
    if (EffectsPos != 0 && cl->cl_player != NIL_PLAYER) {
	len = CallsCount * sizeof(ULONG_T);
	total = len + 1 + EffectsPos;
	rq = allocRequestE(&MS, total);
	rq->rq_type = rt_defineEffect;
	rq->rq_key = DefiningKey;
	rq->rq_u.ru_bytes[0] = CallsCount;
	if (len != 0) {
	    copy(&rq->rq_u.ru_bytes[1], &CallsBuffer[0], len);
	}
	copy(&rq->rq_u.ru_bytes[len + 1], &EffectsBuffer[0], EffectsPos);
	rq->rq_usedLen = total;
	putMessageCarefully(cl, rq);
	if (DefiningKey != 0) {
	    /* never remember key 0 effect - it is continually reused */
	    kel = MEM_ALLOC_E(&MS, sizeof(KnownEffectList_t), "kef_");
	    kel->kel_next = cl->cl_knownEffects;
	    kel->kel_prev = P_NULL;
	    kel->kel_key = DefiningKey;
	    cl->cl_knownEffects = kel;
	}
    }
    ds = DefineStack;
    if (ds != P_NULL) {
	len = ds->ds_dataLength;
	EffectsPos = len;
	if (len != 0) {
	    copy(&EffectsBuffer[0], ds->ds_data, len);
	    memFreeE(&MS, ds->ds_data, len);
	}
	len = ds->ds_callsCount;
	CallsCount = len;
	if (len != 0) {
	    len *= sizeof(ULONG_T);
	    copy(&CallsBuffer[0], ds->ds_calls, len);
	    memFreeE(&MS, ds->ds_calls, len);
	}
	DefiningKey = ds->ds_key;
	DefineStack = ds->ds_next;
	memFreeE(&MS, ds, sizeof(DefineStack_t));
    } else {
	EffectsPos = 0;
	CallsCount = 0;
	Defining = B_FALSE;
    }
}

/* now the various graphics effects routines */

ULONG_T
GOn(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_graphicsOn) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

CHAR_T *
GType(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_graphicsOn) {
	return stringCopy(cl->cl_graphicsType);
    } else {
	return nullString();
    }
}

ULONG_T
GRows(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_graphicsOn) {
	return cl->cl_graphicsRows;
    } else {
	return 0;
    }
}

ULONG_T
GCols(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_graphicsOn) {
	return cl->cl_graphicsCols;
    } else {
	return 0;
    }
}

ULONG_T
GColours(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_graphicsOn) {
	return cl->cl_graphicsColours;
    } else {
	return 0;
    }
}

ULONG_T
GPalette(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_graphicsPalette) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

static BOOL_T
graphicsStart(ULONG_T who, UINT_T len)
{
    Client_t *cl;

    cl = effectsStart(who, len);
    return (BOOL_T) (cl != P_NULL && cl->cl_graphicsOn);
}

static BOOL_T
getGraphicsSize(Client_t *cl, UINT_T *pRows, UINT_T *pCols)
{
    DrawComponentList_t *dcl;

    if (cl->cl_graphicsOn && cl->cl_activeFrame != P_NULL) {
	dcl = cl->cl_activeFrame->guf_drawComponents;
	while (dcl != P_NULL) {
	    if (dcl->dcl_id == cl->cl_defaultDrawComponent) {
		*pRows = dcl->dcl_rows;
		*pCols = dcl->dcl_cols;
		return B_TRUE;
	    }
	    dcl = dcl->dcl_next;
	}
    }
    *pRows = 100;
    *pCols = 100;
    return B_FALSE;
}

static void
scaleCoordinates(ULONG_T key, LONG_T *pX, LONG_T *pY)
{
    Client_t *cl;
    UINT_T rows, cols;

    cl = effectsStart(key, 0);
    if (! getGraphicsSize(cl, &rows, &cols)) {
	runError("No active graphics client/component to scale graphics for");
	freeDefineStack();
    }
    /* Specifically round down here - not to nearest, so that we do not get
       any funniness in scaling. */

    *pX = *pX * cols / (1 << FRAC_BITS);
    *pY = *pY * rows / (1 << FRAC_BITS);
}

static void
scaleCoordinatesRound(ULONG_T key, LONG_T *pX, LONG_T *pY)
{
    Client_t *cl;
    ULONG_T x, y;
    UINT_T rows, cols;

    cl = effectsStart(key, 0);
    if (! getGraphicsSize(cl, &rows, &cols)) {
	runError("No active graphics client/component to scale graphics for");
	freeDefineStack();
    }
    /* This version is used to scale distances, rather than coordinates.
       To make things work out, we *want* to round on this one, not
       truncate. */
    x = *pX * cols;
    y = *pY * rows;
    if (x < 0) {
	x -= (1 << (FRAC_BITS - 1));
    } else {
	x += (1 << (FRAC_BITS - 1));
    }
    if (y < 0) {
	y -= (1 << (FRAC_BITS - 1));
    } else {
	y += (1 << (FRAC_BITS - 1));
    }
    *pX = x / (1 << FRAC_BITS);
    *pY = y / (1 << FRAC_BITS);
}

void
PlaceCursorPixels(ULONG_T x, ULONG_T y)
{

    if (x <= 0xff && y <= 0xff) {
	if (graphicsStart(NIL_THING, 1 + 1 + 1)) {
	    effectsByte(ef_PlaceCursorShort - ef_null);
	    effectsByte(x);
	    effectsByte(y);
	}
    } else {
	if (graphicsStart(NIL_THING, 1 + 2 + 2)) {
	    effectsByte(ef_PlaceCursor - ef_null);
	    effectsWord(x);
	    effectsWord(y);
	}
    }
}

void
PlaceCursor(LONG_T x, LONG_T y)
{

    scaleCoordinates(NIL_THING, &x, &y);
    PlaceCursorPixels(x, y);
}

void
RemoveCursor(void)
{

    if (graphicsStart(NIL_THING, 1)) {
	effectsByte(ef_RemoveCursor - ef_null);
    }
}

void
SetCursorPen(ULONG_T pen)
{

    if (graphicsStart(NIL_THING, 1 + 2)) {
	effectsByte(ef_SetCursorPen - ef_null);
	effectsWord(pen);
    }
}

void
SetCursorPattern(ULONG_T newPattern)
{
    ItemList_t *il;

    if (newPattern == NIL_LIST_INTEGER) {
	runError("nil list integer to SetCursorPattern");
	freeDefineStack();
    } else {
	il = ioRead(newPattern);
	if (il->l_elementCount != 32 / sizeof(ULONG_T)) {
	    runError("invalid list integer to SetCursorPattern");
	    freeDefineStack();
	} else {
	    if (effectsStart(NIL_THING, 1 + 32) != P_NULL) {
		effectsByte(ef_SetCursorPattern - ef_null);
		effectsLongs((ULONG_T *) (il + 1), 32);
	    }
	}
    }
}

void
AddButtonPixels(ULONG_T x, ULONG_T y, ULONG_T n, CHAR_T *str)
{

    if (graphicsStart(NIL_THING, 1 + 2 + 2 + 2 + strLen1(str))) {
	effectsByte(ef_AddButton - ef_null);
	effectsWord(x);
	effectsWord(y);
	effectsWord(n);
	effectsString(str);
    }
}

void
AddButton(LONG_T x, LONG_T y, ULONG_T n, CHAR_T *str)
{

    scaleCoordinates(NIL_THING, &x, &y);
    AddButtonPixels(x, y, n, str);
}

void
EraseButton(ULONG_T n)
{

    if (graphicsStart(NIL_THING, 1 + 2)) {
	effectsByte(ef_EraseButton - ef_null);
	effectsWord(n);
    }
}

void
ClearButtons(void)
{

    if (graphicsStart(NIL_THING, 1)) {
	effectsByte(ef_ClearButtons - ef_null);
    }
}

void
AddRegionPixels(LONG_T x, LONG_T y, LONG_T width, LONG_T height, ULONG_T n)
{

    if (graphicsStart(NIL_THING, 1 + 2 + 2 + 2 + 2 + 2)) {
	effectsByte(ef_AddRegion - ef_null);
	effectsWord(x);
	effectsWord(y);
	effectsWord(width);
	effectsWord(height);
	effectsWord(n);
    }
}

void
AddRegion(LONG_T x, LONG_T y, LONG_T width, LONG_T height, ULONG_T n)
{

    scaleCoordinates(NIL_THING, &x, &y);
    scaleCoordinatesRound(NIL_THING, &width, &height);
    AddRegionPixels(x, y, width, height, n);
}

void
EraseRegion(ULONG_T n)
{

    if (graphicsStart(NIL_THING, 1 + 2)) {
	effectsByte(ef_EraseRegion - ef_null);
	effectsWord(n);
    }
}

void
ClearRegions(void) {

    if (graphicsStart(NIL_THING, 1)) {
	effectsByte(ef_ClearRegions - ef_null);
    }
}

void
SetButtonPen(ULONG_T which, ULONG_T pen)
{

    if (graphicsStart(NIL_THING, 2 + 2)) {
	effectsByte(ef_SetButtonPen - ef_null);
	effectsWord(which);
	effectsWord(pen);
    }
}

void
GSetPen(ULONG_T who, ULONG_T pen)
{

    if (graphicsStart(who, 1 + 2)) {
	effectsByte(ef_SetPen - ef_null);
	effectsWord(pen);
    }
}

void
GSetColour(ULONG_T who, ULONG_T which, ULONG_T colour)
{

    if (graphicsStart(who, 1 + 2 + 2)) {
	effectsByte(ef_SetColour - ef_null);
	effectsWord(which);
	effectsLong(colour);
    }
}

void
GResetColours(ULONG_T who)
{

    if (graphicsStart(who, 1)) {
	effectsByte(ef_ResetColours - ef_null);
    }
}

void
GClear(ULONG_T who)
{

    if (graphicsStart(who, 1)) {
	effectsByte(ef_Clear - ef_null);
    }
}

void
GPixel(ULONG_T who)
{

    if (graphicsStart(who, 1)) {
	effectsByte(ef_Pixel - ef_null);
    }
}

void
GPixelRow(ULONG_T who, ULONG_T pixels, LONG_T start, LONG_T count)
{
    ItemList_t *il;
    int i;

    if (pixels == NIL_LIST_INTEGER) {
	runError("nil list integer to GPixelRow");
	freeDefineStack();
    } else if (start < 0) {
	runError("negative start to GPixelRow");
	freeDefineStack();
    } else if (count < 0) {
	runError("negative count to GPixelRow");
	freeDefineStack();
    } else {
	il = ioRead(pixels);
	if (il->l_elementCount < start + count) {
	    runError("too few elements in array to GPixelRow");
	    freeDefineStack();
	} else if (graphicsStart(who, 1 + 2 + count)) {
	    effectsByte(ef_PixelRow - ef_null);
	    effectsWord(count);
	    for (i = start; i != start + count; i += 1) {
		effectsByte(((ULONG_T *) (il + 1))[i]);
	    }
	}
    }
}

void
GAMovePixels(ULONG_T who, ULONG_T x, ULONG_T y)
{

    if (x <= 0xff && y <= 0xff) {
	if (graphicsStart(who, 1 + 1 + 1)) {
	    effectsByte(ef_AMoveShort - ef_null);
	    effectsByte(x);
	    effectsByte(y);
	}
    } else {
	if (graphicsStart(who, 1 + 2 + 2)) {
	    effectsByte(ef_AMove - ef_null);
	    effectsWord(x);
	    effectsWord(y);
	}
    }
}

void
GAMove(ULONG_T who, LONG_T x, LONG_T y)
{

    scaleCoordinates(who, &x, &y);
    GAMovePixels(who, x, y);
}

void
GRMovePixels(ULONG_T who, LONG_T x, LONG_T y)
{

    if (x <= 127 && y <= 127 && x >= -128 && y >= -128) {
	if (graphicsStart(who, 1 + 1 + 1)) {
	    effectsByte(ef_RMoveShort - ef_null);
	    effectsByte(x);
	    effectsByte(y);
	}
    } else {
	if (graphicsStart(who, 1 + 2 + 2)) {
	    effectsByte(ef_RMove - ef_null);
	    effectsWord(x);
	    effectsWord(y);
	}
    }
}

void
GADrawPixels(ULONG_T who, LONG_T x, LONG_T y)
{

    if (x <= 127 && y <= 127 && x >= -128 && y >= -128) {
	if (graphicsStart(who, 1 + 1 + 1)) {
	    effectsByte(ef_ADrawShort - ef_null);
	    effectsByte(x);
	    effectsByte(y);
	}
    } else {
	if (graphicsStart(who, 1 + 2 + 2)) {
	    effectsByte(ef_ADraw - ef_null);
	    effectsWord(x);
	    effectsWord(y);
	}
    }
}

void
GADraw(ULONG_T who, LONG_T x, LONG_T y)
{

    scaleCoordinates(who, &x, &y);
    GADrawPixels(who, x, y);
}

void
GRMove(ULONG_T who, LONG_T x, LONG_T y)
{

    scaleCoordinatesRound(who, &x, &y);
    GRMovePixels(who, x, y);
}

void
GRDrawPixels(ULONG_T who, LONG_T x, LONG_T y)
{

    if (x <= 127 && y <= 127 && x >= -128 && y >= -128) {
	if (graphicsStart(who, 1 + 1 + 1)) {
	    effectsByte(ef_RDrawShort - ef_null);
	    effectsByte(x);
	    effectsByte(y);
	}
    } else {
	if (graphicsStart(who, 1 + 2 + 2)) {
	    effectsByte(ef_RDraw - ef_null);
	    effectsWord(x);
	    effectsWord(y);
	}
    }
}

void
GRDraw(ULONG_T who, LONG_T x, LONG_T y)
{

    scaleCoordinatesRound(who, &x, &y);
    GRDrawPixels(who, x, y);
}

void
GRectanglePixels(ULONG_T who, ULONG_T width, ULONG_T height, ULONG_T fill)
{

    if (graphicsStart(who, 1 + 2 + 2 + 1)) {
	effectsByte(ef_Rectangle - ef_null);
	effectsWord(width);
	effectsWord(height);
	effectsByte(fill == BOOLEAN_FALSE ? 0 : 1);
    }
}

void
GRectangle(ULONG_T who, LONG_T width, LONG_T height, ULONG_T fill)
{

    scaleCoordinatesRound(who, &width, &height);
    GRectanglePixels(who, width, height, fill);
}

void
GCirclePixels(ULONG_T who, LONG_T r, ULONG_T fill)
{

    if (graphicsStart(who, 1 + 2 + 1)) {
	effectsByte(ef_Circle - ef_null);
	effectsWord(r);
	effectsByte(fill == BOOLEAN_FALSE ? 0 : 1);
    }
}

void
GCircle(ULONG_T who, LONG_T r, ULONG_T fill)
{
    LONG_T dummy;

    scaleCoordinatesRound(who, &r, &dummy);
    GCirclePixels(who, r, fill);
}

void
GEllipsePixels(ULONG_T who, LONG_T a, LONG_T b, ULONG_T fill)
{

    if (graphicsStart(who, 1 + 2 + 2 + 1)) {
	effectsByte(ef_Ellipse - ef_null);
	effectsWord(a);
	effectsWord(b);
	effectsByte(fill == BOOLEAN_FALSE ? 0 : 1);
    }
}

void
GEllipse(ULONG_T who, LONG_T a, LONG_T b, ULONG_T fill)
{

    scaleCoordinatesRound(who, &a, &b);
    GEllipsePixels(who, a, b, fill);
}

void
GPolygonStart(ULONG_T who)
{

    if (graphicsStart(who, 1)) {
	effectsByte(ef_PolygonStart - ef_null);
    }
}

void
GPolygonEnd(ULONG_T who)
{

    if (graphicsStart(who, 1)) {
	effectsByte(ef_PolygonEnd - ef_null);
    }
}

void
GDefineTile(ULONG_T who, ULONG_T id,
	    ULONG_T pixelWidth, ULONG_T pixelHeight, ULONG_T data)
{
    ItemList_t *il;
    UINT_T byteSize;

    if (data == NIL_LIST_INTEGER) {
	runError("nil list integer to GDefineTile");
	freeDefineStack();
    } else {
	byteSize = pixelWidth * pixelHeight;
	il = ioRead(data);
	if (il->l_elementCount !=
	    (byteSize + (sizeof(ULONG_T) - 1)) / sizeof(ULONG_T))
	{
	    runError("invalid list integer to GDefineTile");
	    freeDefineStack();
	} else if (graphicsStart(who, 1 + 2 + 1 + 1 + byteSize)) {
	    effectsByte(ef_DefineTile - ef_null);
	    effectsWord(id);
	    effectsByte(pixelWidth);
	    effectsByte(pixelHeight);
	    effectsLongs((ULONG_T *) (il + 1), byteSize);
	}
    }
}

void
GDefineOverlayTile(ULONG_T who, ULONG_T id, ULONG_T transparentColour,
		   ULONG_T pixelWidth, ULONG_T pixelHeight, ULONG_T data)
{
    ItemList_t *il;
    UINT_T byteSize;

    if (data == NIL_LIST_INTEGER) {
	runError("nil list integer to GDefineTile");
	freeDefineStack();
    } else {
	byteSize = pixelWidth * pixelHeight;
	il = ioRead(data);
	if (il->l_elementCount !=
	    (byteSize + (sizeof(ULONG_T) - 1)) / sizeof(ULONG_T))
	{
	    runError("invalid list integer to GDefineTile");
	    freeDefineStack();
	} else if (graphicsStart(who, 1 + 2 + 1 + 1 + byteSize)) {
	    effectsByte(ef_DefineOverlayTile - ef_null);
	    effectsWord(id);
	    effectsWord(transparentColour);
	    effectsByte(pixelWidth);
	    effectsByte(pixelHeight);
	    effectsLongs((ULONG_T *) (il + 1), byteSize);
	}
    }
}

void
GDisplayTile(ULONG_T who, ULONG_T id)
{

    if (graphicsStart(who, 1 + 2)) {
	effectsByte(ef_DisplayTile - ef_null);
	effectsWord(id);
    }
}

void
GSetTextColour(ULONG_T who, ULONG_T which, ULONG_T colour)
{

    if (graphicsStart(who, 1 + 2 + 2)) {
	effectsByte(ef_SetTextColour - ef_null);
	effectsWord(which);
	effectsLong(colour);
    }
}

void
GText(ULONG_T who, CHAR_T *str)
{

    if (graphicsStart(who, 1 + strLen1(str))) {
	effectsByte(ef_Text - ef_null);
	effectsString(str);
    }
}

void
GLoadBackGround(ULONG_T who, CHAR_T *name)
{

    if (graphicsStart(who, 1 + strLen1(name))) {
	effectsByte(ef_LoadBackGround - ef_null);
	effectsString(name);
    }
}

void
GSetImage(ULONG_T who, CHAR_T *name)
{

    if (graphicsStart(who, 1 + strLen1(name))) {
	effectsByte(ef_SetImage - ef_null);
	effectsString(name);
    }
}

void
GShowImagePixels(ULONG_T who, CHAR_T *name, ULONG_T imageX, ULONG_T imageY,
		 ULONG_T imageWidth, ULONG_T imageHeight,
		 ULONG_T displayX, ULONG_T displayY)
{

    if (graphicsStart(who, 1 + strLen1(name) + 2 + 2 + 2 + 2 + 2 + 2)) {
	effectsByte(ef_ShowImagePixels - ef_null);
	effectsString(name);
	effectsWord(imageX);
	effectsWord(imageY);
	effectsWord(imageWidth);
	effectsWord(imageHeight);
	effectsWord(displayX);
	effectsWord(displayY);
    }
}

void
GShowImage(ULONG_T who, CHAR_T *name,
	   LONG_T imageX, LONG_T imageY, LONG_T imageWidth, LONG_T imageHeight,
	   LONG_T displayX, LONG_T displayY,
	   LONG_T displayWidth, LONG_T displayHeight)
{

    if (graphicsStart(who, 1 + strLen1(name) + 4 * 8)) {
	effectsByte(ef_ShowImage - ef_null);
	effectsString(name);
	effectsLong(imageX);
	effectsLong(imageY);
	effectsLong(imageWidth);
	effectsLong(imageHeight);
	effectsLong(displayX);
	effectsLong(displayY);
	effectsLong(displayWidth);
	effectsLong(displayHeight);
    }
}

void
GShowBrush(ULONG_T who, CHAR_T *name, LONG_T displayX, LONG_T displayY)
{

    scaleCoordinates(who, &displayX, &displayY);
    if (graphicsStart(who, 1 + strLen1(name) + 2 + 2)) {
	effectsByte(ef_ShowBrush - ef_null);
	effectsString(name);
	effectsWord(displayX);
	effectsWord(displayY);
    }
}

void
GScrollRectanglePixels(ULONG_T who, LONG_T xDelta, LONG_T yDelta,
		       LONG_T xPixels, LONG_T yPixels,
		       LONG_T widthPixels, LONG_T heightPixels)
{

    if (graphicsStart(who, 1 + 2 + 2 + 2 + 2 + 2 + 2)) {
	effectsByte(ef_ScrollRectangle - ef_null);
	effectsWord(xDelta);
	effectsWord(yDelta);
	effectsWord(xPixels);
	effectsWord(yPixels);
	effectsWord(widthPixels);
	effectsWord(heightPixels);
    }
}

void
GScrollRectangle(ULONG_T who, LONG_T xDelta, LONG_T yDelta,
		 LONG_T x, LONG_T y, LONG_T width, LONG_T height)
{

    /* Do *not* scale the deltas - they are integer pixel values. */
    scaleCoordinates(who, &x, &y);
    scaleCoordinatesRound(who, &width, &height);
    if (graphicsStart(who, 1 + 2 + 2 + 2 + 2 + 2 + 2)) {
	effectsByte(ef_ScrollRectangle - ef_null);
	effectsWord(xDelta);
	effectsWord(yDelta);
	effectsWord(x);
	effectsWord(y);
	effectsWord(width);
	effectsWord(height);
    }
}

void
GSetIconPen(ULONG_T who, ULONG_T colour)
{

    if (graphicsStart(who, 1 + 2)) {
	effectsByte(ef_SetIconPen - ef_null);
	effectsWord(colour);
    }
}

void
GNewIcon(ULONG_T theCharacter, ULONG_T newIcon)
{
    ItemList_t *il;
    Client_t *cl;
    KnownIconList_t *kil;
    KnownIconList_t **pkil;

    if (theCharacter == NIL_THING) {
	runError("nil thing to GNewIcon");
	freeDefineStack();
    } else if (newIcon == NIL_LIST_INTEGER) {
	runError("nil list integer to GNewIcon");
	freeDefineStack();
    } else {
	il = ioRead(newIcon);
	if (il->l_elementCount != 32 / sizeof(ULONG_T)) {
	    runError("invalid list integer to GNewIcon");
	    freeDefineStack();
	} else {
	    addProp(theCharacter, ICON_KEY, newIcon, TYPE_LIST_INTEGER);
	    cl = Clients;
	    while (cl != P_NULL) {
		if (cl->cl_graphicsOn) {
		    /* does this client know about this icon? */
		    kil = cl->cl_knownIcons;
		    while (kil != P_NULL && kil->kil_key != theCharacter) {
			kil = kil->kil_next;
		    }
		    if (kil != P_NULL) {
			/* yes - send him the updated info */
			if (effectsStart(cl->cl_thing, 1 + 4 + 32) != P_NULL) {
			    effectsByte(ef_NewIcon - ef_null);
			    effectsLong(theCharacter);
			    effectsLongs((ULONG_T *) (il + 1), 32);
			}
		    }
		} else {
		    /* This client is not currently using graphics. If we
		       think he knows about this icon, remove it from his
		       known list, so we will have to re-send it. */
		    pkil = &cl->cl_knownIcons;
		    while (B_TRUE) {
			kil = *pkil;
			if (kil == P_NULL || kil->kil_key == theCharacter) {
			    break;
			}
			pkil = &kil->kil_next;
		    }
		    if (kil != P_NULL) {
			*pkil = kil->kil_next;
			memFreeE(&MS, kil, sizeof(KnownIconList_t));
		    }
		}
		cl = cl->cl_next;
	    }
	}
    }
}

void
GShowIcon(ULONG_T who, ULONG_T whoseIcon,
	  ULONG_T isMonster, ULONG_T isTemporary)
{
    Client_t *cl;
    KnownIconList_t *kil;
    ItemList_t *il;
    ULONG_T key;

    if (whoseIcon == NIL_THING) {
	runError("nil whoseIcon to GShowIcon");
	freeDefineStack();
    } else {
	cl = effectsPtr(who);
	if (cl != P_NULL && cl->cl_graphicsOn) {
	    kil = cl->cl_knownIcons;
	    while (kil != P_NULL && kil->kil_key != whoseIcon) {
		kil = kil->kil_next;
	    }
	    if (kil == P_NULL) {
		/* this client does not yet know about the icon - inform it */
		key = getProp(whoseIcon, ICON_KEY, TYPE_LIST_INTEGER);
		if (key != NIL_LIST_INTEGER) {
		    /* the player has an icon */
		    if (effectsStart(who, 1 + 4 + 32) != P_NULL) {
			kil = MEM_ALLOC_E(&MS, sizeof(KnownIconList_t),"kil_");
			kil->kil_key = whoseIcon;
			kil->kil_next = cl->cl_knownIcons;
			kil->kil_temporary = isTemporary != BOOLEAN_FALSE;
			cl->cl_knownIcons = kil;
			il = ioRead(key);
			/* we ignore the possibility of incorrect length */
			effectsByte(ef_NewIcon - ef_null);
			effectsLong(whoseIcon);
			effectsLongs((ULONG_T *) (il + 1), 32);
		    }
		}
	    } else {
		kil->kil_temporary = (isTemporary != BOOLEAN_FALSE);
	    }
	}
	/* go ahead and get the client to display whatever it now has */
	if (graphicsStart(who, 1 + 4 + 1)) {
	    effectsByte(ef_ShowIcon - ef_null);
	    effectsLong(whoseIcon);
	    effectsByte(isMonster);
	}
    }
}

void
GRemoveIcon(ULONG_T who, ULONG_T whoseIcon)
{

    if (whoseIcon != NIL_THING) {
	if (graphicsStart(who, 1 + 4)) {
	    effectsByte(ef_RemoveIcon - ef_null);
	    effectsLong(whoseIcon);
	}
    }
}

void
GDeleteIcon(ULONG_T who, ULONG_T whoseIcon)
{
    Client_t *cl;
    KnownIconList_t **pKil;
    KnownIconList_t *kil;

    if (whoseIcon != NIL_THING) {
	cl = effectsStart(who, 1 + 4);
	if (cl != P_NULL && cl->cl_graphicsOn) {
	    pKil = &cl->cl_knownIcons;
	    while (B_TRUE) {
		kil = *pKil;
		if (kil == P_NULL || kil->kil_key == whoseIcon) {
		    break;
		}
		pKil = &kil->kil_next;
	    }
	    if (kil != P_NULL) {
		*pKil = kil->kil_next;
		memFreeE(&MS, kil, sizeof(KnownIconList_t));
	    }
	    effectsByte(ef_DeleteIcon - ef_null);
	    effectsLong(whoseIcon);
	}
    }
}

void
GResetIcons(ULONG_T who)
{
    Client_t *cl;
    KnownIconList_t **pKil;
    KnownIconList_t *kil;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_graphicsOn) {
	pKil = &cl->cl_knownIcons;
	while (B_TRUE) {
	    kil = *pKil;
	    if (kil == P_NULL) {
		break;
	    }
	    if (kil->kil_temporary) {
		if (effectsStart(who, 1 + 4) != P_NULL) {
		    effectsByte(ef_DeleteIcon - ef_null);
		    effectsLong(kil->kil_key);
		}
		*pKil = kil->kil_next;
		memFreeE(&MS, kil, sizeof(KnownIconList_t));
	    } else {
		pKil = &kil->kil_next;
	    }
	}
	if (effectsStart(who, 1) != P_NULL) {
	    effectsByte(ef_ResetIcons - ef_null);
	}
    }
}

#if REDRAW_ICONS
void
GRedrawIcons(ULONG_T who)
{

    if (graphicsStart(who, 1)) {
	effectsByte(ef_RedrawIcons - ef_null);
    }
}

void
GUndrawIcons(ULONG_T who)
{

    if (graphicsStart(who, 1)) {
	effectsByte(ef_UndrawIcons - ef_null);
    }
}
#endif

/* effects routines for sound */

ULONG_T 
SOn(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_soundOn) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

static BOOL_T
soundStart(ULONG_T who, UINT_T len)
{
    Client_t *cl;

    cl = effectsStart(who, len);
    return (BOOL_T) (cl != P_NULL && cl->cl_soundOn);
}

void
SVolume(ULONG_T who, ULONG_T volume)
{

    if (soundStart(who, 1 + 2)) {
	effectsByte(ef_SoundVolume - ef_null);
	effectsWord(volume);
    }
}

void
SPlaySound(ULONG_T who, CHAR_T *name, ULONG_T cycles, ULONG_T id)
{

    if (soundStart(who, 1 + strLen1(name))) {
	effectsByte(ef_PlaySound - ef_null);
	effectsString(name);
	effectsLong(id);
	if ((cycles & 0xffff) == 0) {
	    /* prevent accidental infinite replay */
	    cycles = 0xffff;
	}
	effectsWord(cycles);
    }
}

/* effects routines for voice */

/* The Amiga had a "translator.library" that translated from simple
   English text to phoneme sequences. This was called here in the server
   for the 'Speak' builtin. Remote machines used the 'narrator.device'
   to speak phoneme sequences. All of that is a bit hard to do generically,
   so many of the routines here are empty. */

ULONG_T 
VOn(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_voiceOn) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

static BOOL_T
voiceStart(ULONG_T who, UINT_T len)
{
    Client_t *cl;

    cl = effectsStart(who, len);
    return (BOOL_T) (cl != P_NULL && cl->cl_voiceOn);
}

void
VVolume(ULONG_T who, ULONG_T volume)
{

    if (voiceStart(who, 1 + 2)) {
	effectsByte(ef_VoiceVolume - ef_null);
	effectsWord(volume);
    }
}

#ifndef AMIGA

#define DEFPITCH	110
#define DEFRATE		150
#define DEFVOL		64
#define DEFFREQ		22200
#define MALE		0
#define FEMALE		1
#define NATURALF0	0
#define ROBOTICF0	1
#define DEFSEX		MALE
#define DEFMODE		NATURALF0

#define MINRATE		40
#define MAXRATE		400
#define MINPITCH	65
#define MAXPITCH	320
#define MINFREQ		5000
#define MAXFREQ		28000
#define MINVOL		0
#define MAXVOL		64

#endif

void
VParams(ULONG_T who, ULONG_T rate, ULONG_T pitch, ULONG_T mode,
	ULONG_T sex, ULONG_T volume)
{

    if (voiceStart(who, 1 + 2 + 2 + 2 + 2)) {
	if (rate < MINRATE) {
	    rate = MINRATE;
	} else if (rate > MAXRATE) {
	    rate = MAXRATE;
	}
	if (pitch < MINPITCH) {
	    pitch = MINPITCH;
	} else if (pitch > MAXPITCH) {
	    pitch = MAXPITCH;
	}
	if (mode != NATURALF0 && mode != ROBOTICF0) {
	    mode = DEFMODE;
	}
	if (sex != MALE && sex != FEMALE) {
	    sex = DEFSEX;
	}
	if (volume < MINVOL) {
	    volume = MINVOL;
	} else if (volume > MAXVOL) {
	    volume = MAXVOL;
	}
	
	effectsByte(ef_Params - ef_null);
	effectsWord(rate);
	effectsWord(pitch);
	effectsByte(mode);
	effectsByte(sex);
	effectsByte(volume);
    }
}

void
VReset(ULONG_T who)
{

    if (voiceStart(who, 1)) {
	effectsByte(ef_VReset - ef_null);
    }
}

void
VSpeak(ULONG_T who, CHAR_T *words, ULONG_T id)
{
#if 0
    LONG_T res;
    UINT_T len;

    if (voiceStart(who, 1)) {
	if (! TranslatorOpen) {
	    TranslatorBase = OpenLibrary("translator.library", 0);
	    if (TranslatorBase != P_NULL) {
		TranslatorOpen = B_TRUE;
	    }
	}
	if (TranslatorOpen) {
	    flushEffects1();
	    res = Translate(words, strLen1(words) - 1,
			    (CHAR_T *) &EffectsBuffer[1],
			    EFFECTS_BUFFER_SIZE - 2 - 4);
	    if (res == 0) {
		len = strLen1((CHAR_T *) &EffectsBuffer[1]);
		EffectsBuffer[0] = ef_Narrate - ef_null;
		EffectsPos = len + 1;
		effectsLong(id);
	    }
	}
    }
#endif
}

CHAR_T *
VTranslate(CHAR_T *words)
{
#if 0
    LONG_T len;

    if (! TranslatorOpen) {
	TranslatorBase = OpenLibrary("translate.library", 0);
	if (TranslatorBase != P_NULL) {
	    TranslatorOpen = B_TRUE;
	}
    }
    if (TranslatorOpen) {
	flushEffects1();
	len = Translate(words, strLen1(words) - 1,
			(CHAR_T *) &EffectsBuffer[0],
			EFFECTS_BUFFER_SIZE - 1);
	if (len == 0) {
	    return stringCopy((CHAR_T *) &EffectsBuffer[0]);
	} else {
	    return nullString();
	}
    } else {
	return nullString();
    }
#else
    return nullString();
#endif
}

void
VNarrate(ULONG_T who, CHAR_T *phonemes, ULONG_T id)
{

    if (voiceStart(who, 1 + strLen1(phonemes) + 4)) {
	effectsByte(ef_Narrate - ef_null);
	effectsString(phonemes);
	effectsLong(id);
    }
}

/* effects routines for music */

/* I never had music implemented in the Amiga client, but the code here
   is pretty generic - it just triggers whatever is in the client. */

ULONG_T
MOn(ULONG_T who)
{
    Client_t *cl;

    cl = effectsPtr(who);
    if (cl != P_NULL && cl->cl_musicOn) {
	return BOOLEAN_TRUE;
    } else {
	return BOOLEAN_FALSE;
    }
}

static BOOL_T
musicStart(ULONG_T who, UINT_T len)
{
    Client_t *cl;

    cl = effectsStart(who, len);
    return (BOOL_T) (cl != P_NULL && cl->cl_musicOn);
}

void
MVolume(ULONG_T who, ULONG_T volume)
{

    if (musicStart(who, 1 + 2)) {
	effectsByte(ef_MusicVolume - ef_null);
	effectsWord(volume);
    }
}

void
MPlaySong(ULONG_T who, CHAR_T *name, ULONG_T id)
{

    if (musicStart(who, 1 + strLen1(name) + 4)) {
	effectsByte(ef_PlaySong - ef_null);
	effectsString(name);
	effectsLong(id);
    }
}
