#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * find.c - builtins for finding flags, etc. various ways.
 */

/*
 * FindFlagOnList - return 'true' if any element in the given list of things
 *	has the given flag set. Note that we do NO permissions/ownership
 *	testing here. I'm not sure just what would be best.
 */

ULONG_T
FindFlagOnList(ULONG_T list, ULONG_T prop)
{
    ItemList_t *l;
    ULONG_T *p;
    Thing_t *th;
    Attr_t *at;
    ULONG_T thing, count, limit, resultThing;
    BYTE_T n;

    if (prop == NIL_PROP_BOOLEAN) {
	runError("nil property to FindFlagOnList");
    } else if (list != NIL_LIST_THING) {
	/* perfectly OK to have a nil list here */
	l = ioRead(list);
	limit = l->l_elementCount;
	count = 0;
	while (count != limit) {
	    l = ioRead(list);
	    p = (ULONG_T *) (l + 1) + count;
	    thing = *p;
	    resultThing = thing;
	    while (thing != NIL_THING) {
		th = ioRead(thing);
		thing = th->th_parent;
		n = th->th_propCount;
		at = (Attr_t *) (th + 1);
		while (n != 0) {
		    if (at->at_prop == prop && at->at_value != BOOLEAN_FALSE) {
			setResult(resultThing);
			return BOOLEAN_TRUE;
		    }
		    n -= 1;
		    ++at;
		}
	    }
	    count += 1;
	}
    }
    return BOOLEAN_FALSE;
}

/*
 * FindChildOnList - return 'true' if any element of the list is a direct
 *	child of the given thing.
 */

ULONG_T
FindChildOnList(ULONG_T list, ULONG_T parent)
{
    ItemList_t *l;
    ULONG_T *p;
    Thing_t *th;
    ULONG_T thing, count, limit;

    if (list != NIL_LIST_THING) {
	l = ioRead(list);
	limit = l->l_elementCount;
	count = 0;
	while (count != limit) {
	    l = ioRead(list);
	    p = (ULONG_T *) (l + 1) + count;
	    thing = *p;
	    th = ioRead(thing);
	    if (th->th_parent == parent) {
		setResult(thing);
		return BOOLEAN_TRUE;
	    }
	    count += 1;
	}
    }
    return BOOLEAN_FALSE;
}

/*
 * FindIntOnList - try to find a thing on a list of things, such that
 *	the thing has a given int property whose value is appropriate.
 *	The test can be any of =, !=, <, <=, >, or >= as chosen by
 *	the compType parameter.
 */

ULONG_T
FindIntOnList(ULONG_T list, ULONG_T prop, LONG_T valu, ULONG_T compType)
{
    ItemList_t *l;
    ULONG_T *p;
    Thing_t *th;
    Attr_t *at;
    ULONG_T thing, count, limit, resultThing;
    LONG_T val;
    BYTE_T n;

    if (prop == NIL_PROP_INTEGER) {
	runError("nil property to FindIntOnList");
    } else if (compType > 5) {
	runError("invalid compType to FindIntOnList");
    } else if (list != NIL_LIST_THING) {
	/* perfectly OK to have a nil list here */
	l = ioRead(list);
	limit = l->l_elementCount;
	count = 0;
	while (count != limit) {
	    l = ioRead(list);
	    p = (ULONG_T *) (l + 1) + count;
	    thing = *p;
	    resultThing = thing;
	    while (thing != NIL_THING) {
		th = ioRead(thing);
		thing = th->th_parent;
		n = th->th_propCount;
		at = (Attr_t *) (th + 1);
		while (n != 0) {
		    if (at->at_prop == prop) {
			val = at->at_value;
			switch (compType) {
			case 0:
			    if (val == valu) {
				return resultThing;
			    }
			    break;
			case 1:
			    if (val != valu) {
				return resultThing;
			    }
			    break;
			case 2:
			    if (val <  valu) {
				return resultThing;
			    }
			    break;
			case 3:
			    if (val <= valu) {
				return resultThing;
			    }
			    break;
			case 4:
			    if (val >  valu) {
				return resultThing;
			    }
			    break;
			case 5:
			    if (val >= valu) {
				return resultThing;
			    }
			    break;
			}
		    }
		    n -= 1;
		    ++at;
		}
	    }
	    count += 1;
	}
    }
    return NIL_THING;
}

/*
 * propSet - utility for FindAgentWithFlag - return B_TRUE if the given
 *	thing has the given property, either directory or via inheritance.
 */

static BOOL_T
propSet(ULONG_T thing, ULONG_T prop)
{
    Thing_t *th;
    Attr_t *at;
    BYTE_T count;

    while (thing != NIL_THING) {
	th = ioRead(thing);
	thing = th->th_parent;
	count = th->th_propCount;
	at = (Attr_t *) (th + 1);
	while (count != 0) {
	    if (at->at_prop == prop &&
		at->at_value != BOOLEAN_FALSE)
	    {
		return B_TRUE;
	    }
	    count -= 1;
	    ++at;
	}
    }
    return B_FALSE;
}

/*
 * FindAgentWithFlag - look for a client/machine in the given location which
 *	has the given flag set. Return the found thing.
 */

ULONG_T
FindAgentWithFlag(ULONG_T room, ULONG_T prop)
{
    Client_t *cl, *clFirst;
    MachineList_t *ml, *mlFirst;

    if (room == NIL_THING) {
	runError("nil room to FindAgentWithFlag");
    } else if (prop == NIL_PROP_BOOLEAN) {
	runError("nil property to FindAgentWithFlag");
    } else {
	cl = findClientsAt(room);
	if (cl != P_NULL) {
	    clFirst = cl;
	    do {
		if (propSet(cl->cl_thing, prop)) {
		    return cl->cl_thing;
		}
		cl = cl->cl_nextLoc;
	    } while (cl != clFirst);
	}
	ml = findMachinesAt(room);
	if (ml != P_NULL) {
	    mlFirst = ml;
	    do {
		if (! ml->ml_deleted) {
		    if (propSet(ml->ml_thing, prop)) {
			return ml->ml_thing;
		    }
		}
		ml = ml->ml_nextLoc;
	    } while (ml != mlFirst);
	}
    }
    return NIL_THING;
}

/*
 * FindAgentAsChild - look in the given room to see if there is a machine
 *	here which is a child of the given thing. Return the first such agent.
 */

ULONG_T
FindAgentAsChild(ULONG_T room, ULONG_T parent)
{
    MachineList_t *ml, *mlFirst;
    Thing_t *th;

    if (room == NIL_THING) {
	runError("nil room to FindAgentAsChild");
    } else {
	ml = findMachinesAt(room);
	if (ml != P_NULL) {
	    mlFirst = ml;
	    do {
		if (! ml->ml_deleted) {
		    th = ioRead(ml->ml_thing);
		    if (th->th_parent == parent) {
			return ml->ml_thing;
		    }
		}
		ml = ml->ml_nextLoc;
	    } while (ml != mlFirst);
	}
    }
    return NIL_THING;
}

/*
 * FindAgentAsDescendant - look in the given room to see if there is a machine
 *	here which is a descendant of the given thing. Return the first such.
 */

ULONG_T
FindAgentAsDescendant(ULONG_T room, ULONG_T parent)
{
    MachineList_t *ml, *mlFirst;
    Thing_t *th;
    ULONG_T key;

    if (room == NIL_THING) {
	runError("nil room to FindAgentAsChild");
    } else {
	ml = findMachinesAt(room);
	if (ml != P_NULL) {
	    mlFirst = ml;
	    do {
		if (! ml->ml_deleted) {
		    th = ioRead(ml->ml_thing);
		    key = ml->ml_thing;
		    while (key != NIL_THING) {
			th = ioRead(key);
			key = th->th_parent;
			if (key == parent) {
			    return ml->ml_thing;
			}
		    }
		}
		ml = ml->ml_nextLoc;
	    } while (ml != mlFirst);
	}
    }
    return NIL_THING;
}

/*
 * flagInList - utility for FindAgentWithFlagOnList. Return B_TRUE if
 *	the thing or an ancestor contains the given thing list property,
 *	and the list of things for that property contains a thing with has
 *	the given boolean property present.
 */

static BOOL_T
flagInList(ULONG_T thing, ULONG_T listProp, ULONG_T flagProp)
{
    Thing_t *th;
    Attr_t *at;
    ULONG_T parent;
    UINT_T count, limit;
    
    while (thing != NIL_THING) {
	th = ioRead(thing);
	parent = th->th_parent;
	limit = th->th_propCount;
	count = 0;
	while (count != limit) {
	    th = ioRead(thing);
	    at = (Attr_t *) (th + 1) + count;
	    if (at->at_prop == listProp &&
		FindFlagOnList(at->at_value, flagProp) != BOOLEAN_FALSE) {
		return B_TRUE;
	    }
	    count += 1;
	}
	thing = parent;
    }
    return B_FALSE;
}

/*
 * FindAgentWithFlagOnList - look for an agent at the given location which has
 *	a thing on the specified list which has the given flag set.
 */

ULONG_T
FindAgentWithFlagOnList(ULONG_T room, ULONG_T listProp, ULONG_T flagProp)
{
    Client_t *cl, *clFirst;
    MachineList_t *ml, *mlFirst;

    if (room == NIL_THING) {
	runError("nil room to FindAgentWithFlagOnList");
    } else if (listProp == NIL_PROP_LIST_THING) {
	runError("nil list property to FindAgentWithFlagOnList");
    } else if (flagProp == NIL_PROP_BOOLEAN) {
	runError("nil flag property to FindAgentWithFlagOnList");
    } else {
	cl = findClientsAt(room);
	if (cl != P_NULL) {
	    clFirst = cl;
	    do {
		if (flagInList(cl->cl_thing, listProp, flagProp)) {
		    return cl->cl_thing;
		}
		cl = cl->cl_nextLoc;
	    } while (cl != clFirst);
	}
	ml = findMachinesAt(room);
	if (ml != P_NULL) {
	    mlFirst = ml;
	    do {
		if (! ml->ml_deleted) {
		    if (flagInList(ml->ml_thing, listProp, flagProp)) {
			return ml->ml_thing;
		    }
		}
		ml = ml->ml_nextLoc;
	    } while (ml != mlFirst);
	}
    }
    return NIL_THING;
}

/*
 * childInList - utility for FindAgentWithChildOnList. Return B_TRUE if
 *	the passed thing, or an ancestor, has the given list property,
 *	and the given targetParent is the parent of any thing on that list.
 */

static BOOL_T
childInList(ULONG_T thing, ULONG_T listProp, ULONG_T targetParent)
{
    Thing_t *th;
    Attr_t *at;
    ULONG_T parent;
    UINT_T count, limit;

    while (thing != NIL_THING) {
	th = ioRead(thing);
	parent = th->th_parent;
	limit = th->th_propCount;
	count = 0;
	while (count != limit) {
	    th = ioRead(thing);
	    at = (Attr_t *) (th + 1) + count;
	    if (at->at_prop == listProp &&
		FindChildOnList(at->at_value, targetParent) != BOOLEAN_FALSE)
	    {
		return B_TRUE;
	    }
	    count += 1;
	}
	thing = parent;
    }
    return B_FALSE;
}

/*
 * FindAgentWithChildOnList - look for an agent at the given location which has
 *	a thing on the specified list which has the given parent.
 */

ULONG_T
FindAgentWithChildOnList(ULONG_T room, ULONG_T listProp, ULONG_T targetParent)
{
    Client_t *cl, *clFirst;
    MachineList_t *ml, *mlFirst;

    if (room == NIL_THING) {
	runError("nil room to FindAgentWithChildOnList");
    } else if (listProp == NIL_PROP_LIST_THING) {
	runError("nil list property to FindAgentWithChildOnList");
    } else {
	cl = findClientsAt(room);
	if (cl != P_NULL) {
	    clFirst = cl;
	    do {
		if (childInList(cl->cl_thing, listProp, targetParent)) {
		    return cl->cl_thing;
		}
		cl = cl->cl_nextLoc;
	    } while (cl != clFirst);
	}
	ml = findMachinesAt(room);
	if (ml != P_NULL) {
	    mlFirst = ml;
	    do {
		if (! ml->ml_deleted) {
		    if (childInList(ml->ml_thing, listProp, targetParent)) {
			return ml->ml_thing;
		    }
		}
		ml = ml->ml_nextLoc;
	    } while (ml != mlFirst);
	}
    }
    return NIL_THING;
}

/*
 * nameInList - utility for FindAgentWithNameOnList. Return B_TRUE if
 *	the passed thing, or any ancestor, has the given thing list property,
 *	and that list contains a thing with a nameProp matching the name.
 */

static BOOL_T
nameInList(ULONG_T thing, ULONG_T listProp, ULONG_T nameProp, char *name)
{
    Thing_t *th;
    Attr_t *at;
    ULONG_T parent;
    UINT_T count, limit;

    while (thing != NIL_THING) {
	th = ioRead(thing);
	parent = th->th_parent;
	limit = th->th_propCount;
	count = 0;
	while (count != limit) {
	    th = ioRead(thing);
	    at = (Attr_t *) (th + 1) + count;
	    if (at->at_prop == listProp &&
		xFindName(at->at_value, nameProp, name) != STATUS_FAIL)
	    {
		return B_TRUE;
	    }
	    count += 1;
	}
	thing = parent;
    }
    return B_FALSE;
}

/*
 * FindAgentWithNameOnList - look for an agent who has the given list, and on
 *	that list is a thing with the given string property matching the
 *	supplied name. Sounds complicated. It is, except that I can just steal
 *	the hard stuff from elsewhere. This routine returns the thing for the
 *	first agent found which satisfied the testing.
 *	Note that this routine directly involves 10 levels of looping!
 */

ULONG_T
FindAgentWithNameOnList(ULONG_T room, ULONG_T listProp, ULONG_T nameProp,
			CHAR_T *name)
{
    Client_t *cl, *clFirst;
    MachineList_t *ml, *mlFirst;

    if (room == NIL_THING) {
	runError("nil room to FindAgentWithNameOnList");
    } else if (listProp == NIL_PROP_LIST_THING) {
	runError("nil list property to FindAgentWithNameOnList");
    } else if (nameProp == NIL_PROP_STRING) {
	runError("nil flag property to FindAgentWithNameOnList");
    } else {
	cl = findClientsAt(room);
	if (cl != P_NULL) {
	    clFirst = cl;
	    do {
		if (nameInList(cl->cl_thing, listProp, nameProp, name)) {
		    return cl->cl_thing;
		}
		cl = cl->cl_nextLoc;
	    } while (cl != clFirst);
	}
	ml = findMachinesAt(room);
	if (ml != P_NULL) {
	    mlFirst = ml;
	    do {
		if (! ml->ml_deleted) {
		    if (nameInList(ml->ml_thing, listProp, nameProp, name)) {
			return ml->ml_thing;
		    }
		}
		ml = ml->ml_nextLoc;
	    } while (ml != mlFirst);
	}
    }
    return NIL_THING;
}
