#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * Machines.c - routines for dealing with machines.
 */

/*
 * FindMachineIndexed - find the N'th machine whose name matches the
 *	given one. This allows checking all such machines.
 */

ULONG_T
FindMachineIndexed(CHAR_T *name, ULONG_T index)
{
    MachineList_t *ml;

    ml = Machines;
    while (ml != P_NULL) {
	if (! ml->ml_deleted && ml->ml_name != P_NULL &&
	    MatchName(ml->ml_name, name) != -1)
	{
	    index -= 1;
	    if (index == 0) {
		return ml->ml_thing;
	    }
	}
	ml = ml->ml_next;
    }
    return NIL_THING;
}

/*
 * sendToMachines - send the given string to the "otherProc" of all
 *	machines in the given room, except the given two.
 */

void
sendToMachines(ULONG_T here, ULONG_T agent1, ULONG_T agent2, CHAR_T *message)
{
    MachineList_t *ml, *mlFirst;
    CHAR_T *p;
    CallList_t *calls;
    ULONG_T action, saveIndent;
    CHAR_T saveChar;

    messageFlush();
    flushEffects();
    saveIndent = GetIndent();
    p = message + strLen1(message) - 2;
    saveChar = *p;
    if (saveChar == '\n') {
	if (p == message) {
	    return;
	}
	*p = '\0';
    }

    ml = findMachinesAt(here);
    if (ml != P_NULL) {
	calls = P_NULL;
	mlFirst = ml;
	do {
	    action = ml->ml_otherProc;
	    if (! ml->ml_deleted && action != NIL_ACTION &&
		ml->ml_thing != agent1 && ml->ml_thing != agent2)
	    {
		calls = addCall(calls, P_NULL, ml, action,stringCopy(message));
	    }
	    ml = ml->ml_nextLoc;
	} while (ml != mlFirst);
	runCallList(calls, B_TRUE);
    }

    *p = saveChar;
    SetIndent(saveIndent);
    messageFlush();
}

/*
 * CreateMachine - allow an agent to create a machine.
 */

void
CreateMachine(CHAR_T *name, ULONG_T key, ULONG_T where, ULONG_T startup)
{
    CHAR_T *p;
    MachineList_t *ml, *prevMachine;
    Client_t *prevClient;
    Thing_t *th;
    Machine_t mch;
    ULONG_T saveEffective;
    CHAR_T ch;

    TRACE("CreateMachine", key);
    if (key == NIL_THING) {
	runError("nil thing passed to CreateMachine");
    } else if (startup == NIL_ACTION ||
	checkAction(startup, "CreateMachine"))
    {
	p = name;
	do {
	    ch = *p++;
	} while (ch != '\0' && ch != ' ' && ch != '\t' && ch != '\"');
	if (ch != '\0') {
	    symbolError(name, "no spaces or newlines allowed in machine name");
	} else {
	    th = ioRead(key);
	    if (th->th_owner != EffectivePlayer) {
		runError("you do not own the thing");
	    } else {
		th->th_owner = SYSADMIN_KEY;
		th->th_status = ts_public;
		ioDirty(key);
		ml = MEM_ALLOC_E(&MS, sizeof(MachineList_t), "mach");
		MachinesCreated += 1;
		ml->ml_owner = EffectivePlayer;
		ml->ml_location = where;
		incRef(where);
		ml->ml_thing = key;
		incRef(key);
		mch.mch_owner = ml->ml_owner;
		mch.mch_key = ml->ml_thing;
		mch.mch_useCount = 0;
		ml->ml_sayProc = NIL_ACTION;
		ml->ml_whisperMeProc = NIL_ACTION;
		ml->ml_whisperOtherProc = NIL_ACTION;
		ml->ml_pageProc = NIL_ACTION;
		ml->ml_poseProc = NIL_ACTION;
		ml->ml_otherProc = NIL_ACTION;
		ml->ml_idleProc = NIL_ACTION;
		ml->ml_activeProc = NIL_ACTION;
		copyToMachine(&mch, ml);
		ml->ml_name = stringCopy(name);
		ml->ml_dirty = B_TRUE;
		ml->ml_deleted = B_FALSE;
		ml->ml_machine =
		    ioCreate(TYPE_MACHINE, &mch, sizeof(Machine_t));
		saveEffective = EffectivePlayer;
		EffectivePlayer = SYSADMIN_KEY;
		addProp(key, NAME_KEY, (ULONG_T) name, TYPE_STRING);
		EffectivePlayer = saveEffective;
		ml->ml_next = Machines;
		Machines = ml;
		xAddTail(getMachineList(), ml->ml_machine);
		fullyInsertMachine(ml);
		/* want these two calls before we change ActiveClient */
		messageFlush();
		flushEffects();
		if (startup != NIL_ACTION) {
		    /* now switch over to the new machine for its startup */
		    prevMachine = ActiveMachine;
		    ActiveMachine = ml;
		    prevClient = ActiveClient;
		    ActiveClient = P_NULL;
		    /* do not change real/effective here */
		    /* machine must be totally set up at this point, since the
		       user-code could destroy the machine */
		    (void) doProc0(startup);
		    messageFlush();
		    flushEffects();
		    /* now switch back to the causing client/machine */
		    ActiveClient = prevClient;
		    ActiveMachine = prevMachine;
		}
	    }
	}
    }
}

/* procs to allow the setting of machine actions */

static MachineList_t *
findMachine(ULONG_T key)
{
    MachineList_t *ml;

    ml = getMachineByThing(key);
    if (ml == P_NULL || ml->ml_deleted) {
	runError("can't find machine to set proc");
    } else if (ActiveClient != P_NULL) {
	if (ActiveClient->cl_player != ml->ml_owner) {
	    ml = P_NULL;
	    runError("not owner of machine");
	}
    } else if (ActiveMachine == P_NULL ||
	ActiveMachine->ml_owner != ml->ml_owner)
    {
	ml = P_NULL;
	runError("not same machine owner");
    }
    return ml;
}

ULONG_T
SetMachineSay(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkStringAction(action, "SetMachineSay")) {
	oldAction = ml->ml_sayProc;
	ml->ml_sayProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

ULONG_T
SetMachineWhisperMe(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkStringAction(action, "SetMachineWhisperMe")) {
	oldAction = ml->ml_whisperMeProc;
	ml->ml_whisperMeProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

ULONG_T
SetMachineWhisperOther(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkStringAction(action, "SetMachineWhisperOther")) {
	oldAction = ml->ml_whisperOtherProc;
	ml->ml_whisperOtherProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

#if 0

/* This isn't needed, since we can use ForceAction to make a machine do
   whatever we want. */

ULONG_T
SetMachinePage(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkStringAction(action, "SetMachinePage")) {
	oldAction = ml->ml_pageProc;
	ml->ml_pageProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

#endif

ULONG_T
SetMachinePose(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkStringAction(action, "SetMachinePose")) {
	oldAction = ml->ml_poseProc;
	ml->ml_poseProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

ULONG_T
SetMachineOther(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkStringAction(action, "SetMachineOther")) {
	oldAction = ml->ml_otherProc;
	ml->ml_otherProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

ULONG_T
SetMachineIdle(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkAction(action, "SetMachineIdle")) {
	oldAction = ml->ml_idleProc;
	ml->ml_idleProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

ULONG_T
SetMachineActive(ULONG_T key, ULONG_T action)
{
    MachineList_t *ml;
    ULONG_T oldAction;

    ml = findMachine(key);
    if (ml != P_NULL && checkAction(action, "SetMachineActive")) {
	oldAction = ml->ml_activeProc;
	ml->ml_activeProc = action;
	incProcRef(action);
	decProcRefAndNil(&oldAction);
	ml->ml_dirty = B_TRUE;
	return oldAction;
    } else {
	return NIL_ACTION;
    }
}

/*
 * DestroyMachine.
 */

void
DestroyMachine(ULONG_T key)
{
    MachineList_t *ml;
    MachineList_t **pMl;
    ULONG_T saveEffective;

    TRACE("DestroyMachine", key);
    if (key == NIL_THING) {
	runError("nil thing to DestroyMachine");
	return;
    }
    ml = getMachineByThing(key);
    if (ml == P_NULL || ml->ml_deleted) {
	runError("can't find machine to destroy");
	return;
    }
    if (EffectivePlayer != ml->ml_owner && EffectivePlayer != SYSADMIN_KEY) {
	runError("not machine owner");
	return;
    }

    /* Do this first, so there is no way we can find it again, such as
       during a recursive decRef cycle. */
    ml->ml_deleted = B_TRUE;
    ml->ml_thing = NIL_THING;
    flushRefKey(key);
    if (ActiveClient == P_NULL && ActiveMachine == ml) {
	ActiveMachine = P_NULL;
    }
    machineShutDown(ml);
    saveEffective = EffectivePlayer;
    EffectivePlayer = SYSADMIN_KEY;
    DelElement(getMachineList(), ml->ml_machine);
    EffectivePlayer = saveEffective;
    decRef(key);		    /* someone might still refer to it */
    decRef(ml->ml_location);
    decProcRef(ml->ml_sayProc);
    decProcRef(ml->ml_whisperMeProc);
    decProcRef(ml->ml_whisperOtherProc);
    decProcRef(ml->ml_pageProc);
    decProcRef(ml->ml_poseProc);
    decProcRef(ml->ml_otherProc);
    decProcRef(ml->ml_idleProc);
    decProcRef(ml->ml_activeProc);
    ioDelete(ml->ml_machine);
    ml->ml_dirty = B_FALSE;
    /* Do this part here, while we still know the key to its thing. The
       rest of the unlinking is done in 'handleRequests'. This should be
       safe since we never call any scenario functions while scanning
       down the key hash list. */
    pMl = &MachineKeyMap[key % AGENT_MAP_SIZE];
    while (*pMl != ml) {
	pMl = &(*pMl)->ml_nextKHash;
    }
    *pMl = ml->ml_nextKHash;
    MachineDeleted = B_TRUE;
}
