#if defined(unix) || defined(AMIGA)

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#if defined(AIX)
#include <sys/select.h>
#endif
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#elif defined(_WIN32)
#include <winsock2.h>

#endif

#if defined(AMIGA)
#include <proto/dos.h>
extern int fcntl(int fd, int cmd, int arg);
#endif

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "MUD.h"
#include "types.h"
#include "Request.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * main2.c - server stuff, aborting, timers.
 */

#define SOCKET_DEBUG	B_FALSE

#if SOCKET_DEBUG
#define SDEBUG(x)	mess(x)
#else
#define SDEBUG(x)
#endif

#define MAX_PORTS	5

/* Note that because we use mud.library for all of our alloc/free other
   than public messages, we do not have to free things on a shutdown - it
   will all go away anyway. */

#define FREE_MEMORY_ON_SHUTDOWN B_TRUE

#define EVENT_MAX_PER_BUNCH	100

typedef struct ActionList {
    struct ActionList *al_next;
    Client_t *al_client;
    MachineList_t *al_machine;
    LONG_T al_delay;
    ULONG_T al_action;
} ActionList_t;

static int BinarySocket;
static int TelnetPorts[MAX_PORTS];
static int TelnetFlags[MAX_PORTS];
static int TelnetSockets[MAX_PORTS];
static int TelnetPortCount = 0;
static fd_set AllClientBits, ClientWriteBits;
static unsigned int MaxFd;
static ULONG_T ActionDelay;
static ActionList_t *Actions;
static ULONG_T LastTime;
static CallList_t *FreeCallLists;

#if defined(AMIGA)
LONG_T MyAddr;
#else
int MyAddr;
#endif

#if defined(_WIN32)
static WSADATA WSAData;
#endif

static ULONG_T ClientCounter;			/* just count them */
static UINT_T ClientCount;			/* number active now */
static BOOL_T CacheDone, QuitRequested, AbortedOnce,
    NeedRecover, NeedCacheCompact;

static Client_t *ShutDowns;
static ULONG_T AbortMax, AbortCount;
static BOOL_T InDoDelayedActions, CodeRunning, ShuttingDown;

/*
 * addTelnetPort - add a request for a telnet port. Called from the
 *	command-line parsing code.
 */

void
addTelnetPort(int port, int flags)
{

    if (TelnetPortCount == MAX_PORTS) {
	mess("Too many telnet ports\n");
	finalExit();
    }
    TelnetPorts[TelnetPortCount] = port;
    TelnetFlags[TelnetPortCount] = flags;
    TelnetSockets[TelnetPortCount] = -1;
    TelnetPortCount += 1;
}

#if 0
    static char *RequestNames[] = {
	"first",
	"killClient",
	"nukeClient",
	"runClient",
	"message",
	"enterWizardMode",
	"leaveWizardMode",
	"setContinue",
	"changePassword",
	"setPrompt",
	"deleteProc",
	"effects",
	"editString",
	"editProc",
	"getString",
	"flushSymbol",
	"defineEffect",
	"queryFile",
	"shutDown",
	"flush",
	"syncShutDown",
	"beginClient",
	"endClient",
	"askPlayer",
	"creationCheck",
	"createPlayer",
	"loginPlayer",
	"lookup",
	"findName",
	"readProc",
	"writeProcNew",
	"runProc",
	"log",
	"inputLine",
	"setSync",
	"rawKey",
	"regionSelect",
	"buttonHit",
	"flushEffect",
	"textResize",
	"incRef",
	"decRef",
	"symbolEnter",
	"symbolDelete",
	"useTable",
	"unuseTable",
	"describeBuiltin",
	"describe",
	"graphicsFlip",
	"voiceFlip",
	"soundFlip",
	"musicFlip",
	"doEditProc",
	"editStringDone",
	"replaceProc",
	"editProcDone",
	"getStringDone",
	"effectDone",
	"queryFileDone",
	"createContainer",
	"createComponent",
	"makeFrame",
	"illegal"
    };
#endif

/*
 * killSocket - get rid of a socket.
 */

static void
killSocket(int s)
{

    TRACE("killSocket", s);
    shutdown(s, 2);
#if defined(_WIN32)
    closesocket(s);
#elif defined(AMIGA)
    CloseSocket(s);
#else
    close(s);
#endif
}

/*
 * freeLocalLists - free our local lists of things.
 */

static void
freeLocalLists(BOOL_T freeMemory)
{
    ActionList_t *al;
    CallList_t *cll;

    while (B_TRUE) {
	al = Actions;
	if (al == P_NULL) {
	    break;
	}
	Actions = al->al_next;
	decProcRef(al->al_action);
	if (freeMemory) {
	    memFreeE(&MS, al, sizeof(ActionList_t));
	}
    }
    while (B_TRUE) {
	cll = FreeCallLists;
	if (cll == P_NULL) {
	    break;
	}
	FreeCallLists = cll->cll_next;
	if (freeMemory) {
	    memFreeE(&MS, cll, sizeof(CallList_t));
	}
    }
}

/*
 * processShutDowns - run through and notify those waiting for shutdown.
 */

static void
processShutDowns(void)
{
    Client_t *cl;
    Request_t rq;

    while (B_TRUE) {
	cl = ShutDowns;
	if (cl == P_NULL) {
	    break;
	}
	ShutDowns = cl->cl_next;
	rq.rq_availLen = 0;
	rq.rq_usedLen = 0;
	rq.rq_type = rt_syncShutDown;
	(void) send(cl->cl_socket, (char *) &rq.rq_usedLen, REQ_LEN, 0);
	killSocket(cl->cl_socket);
	memFreeE(&MS, cl, sizeof(Client_t));
    }
}

/*
 * shutDown2 - second half of shutting the world down.
 */

static void
shutDown2(BOOL_T freeMemory, BOOL_T doWrites)
{
    BOOL_T reallyFreeMemory;

    TRACE("shutDown2", ((freeMemory - B_FALSE) << 8) | (doWrites - B_FALSE));
#if FREE_MEMORY_ON_SHUTDOWN
    reallyFreeMemory = freeMemory;
#else
    reallyFreeMemory = B_FALSE;
#endif
    ShuttingDown = B_TRUE;
    if (CacheDone) {
	freeLocalLists(reallyFreeMemory);
	if (reallyFreeMemory) {
	    builtInTerm();
	    freeNameMap(PublicNameMap);
	    if (doWrites) {
		removeGlobalUse(BuiltinTable);
		removeGlobalUse(PlayerTable);
	    }
	} else {
	    BuiltinTable->tl_this->tb_usedCount = 0;
	    BuiltinTable->tl_this->tb_entryCount = 0;
	}
	effectsTerm();
	cacheTerm(reallyFreeMemory, doWrites);
	bcTerm(reallyFreeMemory);
    }
    /* Do this regardless of FREE_MEMORY_ON_SHUTDOWN,
       since the library uses AllocMem directly */
    if (freeMemory) {
	libTermE(&MS);
    }
    if (doWrites) {
	ioTerminate();
    }
    processShutDowns();
}

/*
 * genericAbort - common aborting code.
 */

static void
genericAbort(char *message, BOOL_T freeMemory, BOOL_T doWrites)
{
    int i;

    TRACE("genericAbort", message);
    MS.ms_ReclaimMem = P_NULL;
    log2NL("genericAbort: ", message);
    setLogging();
    traceBack();
#if TRACING
    traceDump();
#endif
    messageFlush();
    if (! AlreadyAborting) {
	AlreadyAborting = B_TRUE;
	shutDown2(freeMemory, doWrites);
	ioShutDown(message);
    }
    if (BinarySocket != -1) {
	killSocket(BinarySocket);
	BinarySocket = -1;
    }
    for (i = 0; i != TelnetPortCount; i += 1) {
	if (TelnetSockets[i] != -1) {
	    killSocket(TelnetSockets[i]);
	    TelnetSockets[i] = -1;
	}
    }
#if defined(_WIN32)
    (void) WSACleanup();
#endif
    finalExit();
}

/*
 * myAbort - something serious has gone wrong - shut down and exit.
 */

void
myAbort(char *message)
{

    genericAbort(message, B_TRUE, B_TRUE);
}

/*
 * libraryAbort - abort for the MUD library.
 */

static void
libraryAbort(BOOL_T isServerNuke)
{

    genericAbort("lib", B_FALSE, B_TRUE);
}

/*
 * myAbortNoWrite - abort, but do not try to do any 'io' calls.
 */

void
myAbortNoWrite(void)
{

    genericAbort("abortNoWrite", B_TRUE, B_FALSE);
}

/*
 * myPerror - use 'mess' for a perror-type call. Note that winsock
 *	doesn't provide a 'strerror', so we have to fake it.
 *	Note that all of our calls are from socket calls.
 */

static void
myPerror(char *message)
{
    char *reason;
#if defined(_WIN32)
    %%%% use GetModuleHandle("wsock32") and FormatMessage %%%%
    extern char *wserror(int errorCode);

    reason = wserror(WSAGetLastError());
#else
    reason = strerror(errno);
#endif
    mess(message);
    mess(": ");
    if (reason == P_NULL) {
	mess("unknown error ");
	mNum(errno);
    } else {
	mess(reason);
    }
    messChar('\n');
}

/*
 * addBuiltinProc - entry to allow the builtin procs to be added.
 */

void
addBuiltinProc(CHAR_T *name, ULONG_T key)
{
    Entry_t *en;

    en = symEnterE(&MS, BuiltinTable->tl_this, name, strLen1(name));
    en->en_key = key;
}

/*
 * subBuiltinProc - used during shutdown to remove builtins from builtin
 *	table before that table is written out.
 */

void
subBuiltinProc(ULONG_T key)
{
    CHAR_T *name;

    TRACE("subPublicProc", key);
    name = findName(BuiltinTable->tl_nameMap, key);
    if (name == P_NULL) {
	AlreadyAborting = B_TRUE;
	myAbort("can't find name for builtin in subPublicProc");
    }
    removeNameFromMap(BuiltinTable->tl_nameMap, key);
    symDeleteE(&MS, BuiltinTable->tl_this, name, strLen1(name));
}

/*
 * addCall - add an agent to a CallList_t.
 */

CallList_t *
addCall(CallList_t *oldList, Client_t *cl, MachineList_t *ml,
	ULONG_T action, CHAR_T *arg)
{
    CallList_t *cll;

    cll = FreeCallLists;
    if (cll != P_NULL) {
	FreeCallLists = cll->cll_next;
    } else {
	cll = MEM_ALLOC_E(&MS, sizeof(CallList_t), "clls");
    }
    cll->cll_next = oldList;
    cll->cll_client = cl;
    cll->cll_machine = ml;
    cll->cll_action = action;
    cll->cll_arg = arg;
    return cll;
}

/*
 * runCallList - run an action for an agent list.
 */

void
runCallList(CallList_t *calls, BOOL_T switchToAgent)
{
    CallList_t *cll;
    Client_t *cl, *saveClient;
    MachineList_t *ml, *saveMachine;
    ULONG_T saveRealPlayer, saveEffectivePlayer;
    PlayerStatus_t saveRealStatus, saveEffectiveStatus;

    if (calls == P_NULL) {
	return;
    }
    /* Could put these inside 'if (switchToAgent)', but want to shut gcc up. */
    saveClient = ActiveClient;
    saveMachine = ActiveMachine;
    saveRealPlayer = RealPlayer;
    saveRealStatus = RealStatus;
    saveEffectivePlayer = EffectivePlayer;
    saveEffectiveStatus = EffectiveStatus;
    while (calls != P_NULL) {
	cll = calls;
	calls = calls->cll_next;
	cl = cll->cll_client;
	if (cl != P_NULL) {
	    if (! cl->cl_deleted && cl->cl_player != NIL_PLAYER) {
		if (switchToAgent) {
		    ActiveClient = cl;
		    ActiveMachine = NULL;
		    RealPlayer = cl->cl_player;
		    RealStatus = cl->cl_status;
		    EffectivePlayer = cl->cl_player;
		    EffectiveStatus = cl->cl_status;
		}
		doProc1(cll->cll_action, cll->cll_arg);
	    }
	} else {
	    ml = cll->cll_machine;
	    if (! ml->ml_deleted) {
		if (switchToAgent) {
		    ActiveClient = P_NULL;
		    ActiveMachine = ml;
		    RealPlayer = ml->ml_owner;
		    RealStatus = ps_apprentice;
		    EffectivePlayer = ml->ml_owner;
		    EffectiveStatus = ps_apprentice;
		}
		doProc1(cll->cll_action, cll->cll_arg);
	    }
	}
	cll->cll_next = FreeCallLists;
	FreeCallLists = cll;
    }
    if (switchToAgent) {
	ActiveClient = saveClient;
	ActiveMachine = saveMachine;
	RealPlayer = saveRealPlayer;
	RealStatus = saveRealStatus;
	EffectivePlayer = saveEffectivePlayer;
	EffectiveStatus = saveEffectiveStatus;
    }
}

/*
 * unlinkClient - unlink a client from the client list.
 */

static void
unlinkClient(Client_t *cl)
{

    if (cl->cl_prev != P_NULL) {
	cl->cl_prev->cl_next = cl->cl_next;
    } else {
	Clients = cl->cl_next;
    }
    if (cl->cl_next != P_NULL) {
	cl->cl_next->cl_prev = cl->cl_prev;
    }
}

/*
 * freeClient - unlink and free a client structure.
 */

static void
freeClient(Client_t *cl)
{

    TRACE("freeClient", cl);
    unlinkClient(cl);
    if (cl->cl_telnet != P_NULL) {
	telnetFreeClient(cl->cl_telnet);
    }
    memFreeE(&MS, cl, sizeof(Client_t));
}

/*
 * clearClient - clear a client structure for use or reuse.
 */

void
clearClient(Client_t *cl)
{

    TRACE("clearClient", cl);
    cl->cl_outGoing = P_NULL;
    cl->cl_outGoingPos = 0;
    cl->cl_inComing = P_NULL;
    cl->cl_inComingPos = 0;
    cl->cl_useList = P_NULL;
    cl->cl_knownIcons = P_NULL;
    cl->cl_knownEffects = P_NULL;
    cl->cl_activeFrame = P_NULL;
    cl->cl_newFrame = P_NULL;
    cl->cl_player = NIL_PLAYER;
    cl->cl_thing = NIL_THING;
    cl->cl_location = NIL_THING;
    cl->cl_inputAction = NIL_ACTION;
    cl->cl_rawKeyAction = NIL_ACTION;
    cl->cl_mouseDownAction = NIL_ACTION;
    cl->cl_buttonAction = NIL_ACTION;
    cl->cl_effectDoneAction = NIL_ACTION;
    cl->cl_privateTableKey = NIL_TABLE;
    cl->cl_editStringAction = NIL_ACTION;
    cl->cl_getStringAction = NIL_ACTION;
    cl->cl_queryFileAction = NIL_ACTION;
    cl->cl_defaultTextComponent = -1;
    cl->cl_defaultDrawComponent = -1;
    cl->cl_lastUsedDrawComponent = -1;
    cl->cl_privateSymbols = P_NULL;
    cl->cl_privateMap = P_NULL;
    cl->cl_name = P_NULL;
    cl->cl_prompt = P_NULL;
    cl->cl_onAt = 0;
    cl->cl_idleAction = NIL_ACTION;
    cl->cl_activeAction = NIL_ACTION;
    cl->cl_textWidth = DEFAULT_TEXT_WIDTH;
    cl->cl_textHeight = DEFAULT_TEXT_HEIGHT;
    cl->cl_textColumn = 0;
    cl->cl_fontAscent = 8;
    cl->cl_fontDescent = 2;
    cl->cl_fontLeading = 1;
    cl->cl_fontHeight = 11;
    cl->cl_fontWidth = 6;
    cl->cl_passwordCount = 0;
    cl->cl_inserted = B_FALSE;
    cl->cl_ended = B_FALSE;
    cl->cl_inputSync = B_FALSE;
    cl->cl_wizardMode = B_FALSE;
    cl->cl_newPlayer = B_FALSE;
    cl->cl_gotTextSize = B_FALSE;
    cl->cl_status = ps_normal;
    cl->cl_shuttingDown = B_FALSE;
    cl->cl_destroy = B_FALSE;
    cl->cl_deleted = B_FALSE;
    cl->cl_expectingOldPassword = B_FALSE;
    cl->cl_playerPassword = B_FALSE;
    cl->cl_editing = B_FALSE;
    cl->cl_getting = B_FALSE;
    cl->cl_querying = B_FALSE;
    cl->cl_noAts = B_FALSE;
    cl->cl_fullPaths = B_FALSE;
    cl->cl_canEdit = B_FALSE;
    cl->cl_canGetString = B_FALSE;
    cl->cl_canQueryFile = B_FALSE;
    cl->cl_canWizard = B_FALSE;
    cl->cl_graphicsOn = B_FALSE;
    cl->cl_graphicsPalette = B_FALSE;
    cl->cl_voiceOn = B_FALSE;
    cl->cl_soundOn = B_FALSE;
    cl->cl_musicOn = B_FALSE;
    cl->cl_graphicsRows = 100;
    cl->cl_graphicsCols = 100;
    cl->cl_graphicsColours = 8;
    cl->cl_graphicsType = P_NULL;
    cl->cl_clientVersion = 0;
}

/*
 * doSend - helper when sending to a client. This is not 'static' since
 *	it is also used from the telnet code.
 */

LONG_T
doSend(Client_t *cl, void *buf, LONG_T len)
{
    LONG_T res;

    TRACE("doSend", cl->cl_socket);
    res = send(cl->cl_socket, buf, len, 0);
    if (res < 0) {
#if defined(_WIN32)
	if (WSA_GetLastError() != WSA_EWOULDBLOCK)
#else
	if (errno != EWOULDBLOCK)
#endif
	{
	    /* Something wrong - shutdown the client. */
	    SDEBUG("doSend error, calling endClient\n");
	    endClient(cl);
	} else {
	    /* Fake this, so caller will not think we have died. */
	    res = 0;
	}
    }
    return res;
}

/*
 * doRecv - local helper routine for clientRead. Also used for telnet mode.
 */

LONG_T
doRecv(Client_t *cl, void *buf, LONG_T len)
{
    LONG_T res;

    TRACE("doRecv", cl->cl_socket);
    res = recv(cl->cl_socket, buf, len, 0);
    if (res <= 0) {
	if (res < 0) {
#if defined(_WIN32)
	    if (WSAGetLastError() != WSA_EWOULDBLOCK)
#else
	    if (errno != EWOULDBLOCK)
#endif
	    {
		/* Something wrong with the client. */
#if SOCKET_DEBUG
		mess("doRecv problem errno ");
		mNum(errno);
		mess(" - calling endClient\n");
#endif
		endClient(cl);
	    }
	} else {
	    /* On Linux, length 0 means connection gone away, I think. */
	    SDEBUG("doRecv got length 0 - calling endClient\n");
	    endClient(cl);
	}
    } else {
#if SOCKET_DEBUG
	mess("doRecv got ");
	mNum(res);
	mess(" bytes\n");
#endif
    }
    return res;
}

/*
 * addToWriteWaiters - we have filled up an outgoing socket, and must
 *	stop sending to it for a while. Add that socket to the set
 *	we select on for output ready.
 */

void
addToWriteWaiters(Client_t *cl)
{

    FD_SET(cl->cl_socket, &ClientWriteBits);
}

/*
 * removeFromWriteWaiters - this client no longer needs to be woken up
 *	when something can be sent on its socket.
 */

void
removeFromWriteWaiters(Client_t *cl)
{

    FD_CLR(cl->cl_socket, &ClientWriteBits);
}

/*
 * detachSocket - detach the client from the socket.
 */

static void
detachSocket(Client_t *cl)
{
    int s, max, i;

    TRACE("detachSocket", cl);
    s = cl->cl_socket;
    if (s != -1) {
	cl->cl_socket = -1;
	FD_CLR(s, &AllClientBits);
	FD_CLR(s, &ClientWriteBits);
	killSocket(s);
	if (s == MaxFd) {
	    /* This was the biggest, and is now gone. */
	    max = BinarySocket;
	    for (i = 0; i != TelnetPortCount; i += 1) {
		if (TelnetSockets[i] > max) {
		    max = TelnetSockets[i];
		}
	    }
	    cl = Clients;
	    while (cl != P_NULL) {
		if (cl->cl_socket > max) {
		    max = cl->cl_socket;
		}
		cl = cl->cl_next;
	    }
	    MaxFd = max;
	}
    }
}

/*
 * clientWrite - we can write to the client socket.
 */

static void
clientWrite(Client_t *cl)
{
    Request_t *rq;
    LONG_T pos, len, res;

    rq = cl->cl_outGoing;
    if (rq == P_NULL) {
	removeFromWriteWaiters(cl);
    } else {
	pos = cl->cl_outGoingPos;
	len = rq->rq_usedLen + REQ_LEN - pos;
	rq->rq_usedLen = htons(rq->rq_usedLen);
	res = doSend(cl, (BYTE_T *) &rq->rq_usedLen + pos, len);
	if (res >= 0) {
	    if (res == len) {
		while (B_TRUE) {
		    /* all sent */
		    MessagesSent += 1;
		    cl->cl_outGoing = rq->rq_next;
		    freeRequestE(&MS, rq);
		    rq = cl->cl_outGoing;
		    if (rq == P_NULL) {
			/* Nothing left to send right now. */
			removeFromWriteWaiters(cl);
			break;
		    }
		    len = rq->rq_usedLen + REQ_LEN;
		    rq->rq_usedLen = htons(rq->rq_usedLen);
		    res = doSend(cl, &rq->rq_usedLen, len);
		    if (res < 0) {
			break;
		    }
		    if (res != len) {
			rq->rq_usedLen = ntohs(rq->rq_usedLen);
			cl->cl_outGoingPos = res;
			break;
		    }
		}
	    } else {
		pos += res;
		rq->rq_usedLen = ntohs(rq->rq_usedLen);
		cl->cl_outGoingPos = pos;
	    }
	}
    }
}

static char *RqNames[] = {
    "::first::",

    "::killClient::",
    "::nukeClient::",
    "::runClient::",
    "::message::",
    "::enterWizardMode::",
    "::leaveWizardMode::",
    "::setContinue::",
    "::changePassword::",
    "::setPrompt::",
    "::deleteProc::",
    "::effects::",
    "::editString::",
    "::editProc::",
    "::getString::",
    "::flushSymbol::",
    "::defineEffect::",
    "::queryFile::",

    "::shutDown::",
    "::flush::",
    "::syncShutDown::",
    "::beginClient::",
    "::endClient::",
    "::askPlayer::",
    "::creationCheck::",
    "::createPlayer::",
    "::loginPlayer::",
    "::lookup::",
    "::findName::",
    "::readProc::",
    "::writeProcNew::",
    "::runProc::",

    "::log::",
    "::inputLine::",
    "::setSync::",
    "::rawKey::",
    "::regionSelect::",
    "::buttonHit::",
    "::flushEffect::",
    "::textResize::",
    "::incRef::",
    "::decRef::",
    "::symbolEnter::",
    "::symbolDelete::",
    "::useTable::",
    "::unuseTable::",
    "::describeBuiltin::",
    "::describe::",
    "::graphicsFlip::",
    "::voiceFlip::",
    "::soundFlip::",
    "::musicFlip::",
    "::doEditProc::",
    "::editStringDone::",
    "::replaceProc::",
    "::editProcDone::",
    "::getStringDone::",
    "::effectDone::",
    "::queryFileDone::",
    "::createContainer::",
    "::createComponent::",
    "::makeFrame::",

    "::illegal::"
};

/*
 * putMessageCarefully - carefully send a message to a client.
 */

void
putMessageCarefully(Client_t *cl, Request_t *rq)
{
    Request_t **tail;
    int s, count;
    LONG_T len, res;
    RequestType_t rt;

    TRACE("putMessageCarefully", cl);
    s = cl->cl_socket;
    if (s != -1) {
	if (cl->cl_telnet != P_NULL) {
	    mess("Got into putMessageCarefully for telnet client!\n");
	    mess("Type of request is ");
	    mNum(rq->rq_type - rt_first);
	    messChar('\n');
	    freeRequestE(&MS, rq);
	} else {
	    rt = rq->rq_type;
#if TRACING
	    TRACE(RqNames[rt - rt_first], rq->rq_key);
#endif
	    rq->rq_key = htonl(rq->rq_key);
	    rq->rq_uint = htons(rq->rq_uint);
	    if (rt == rt_symbolEnter || rt == rt_symbolDelete ||
		rt == rt_regionSelect || rt == rt_effectDone ||
		rt == rt_createContainer || rt == rt_createComponent)
	    {
		rq->rq_u.ru_otherKey = htonl(rq->rq_u.ru_otherKey);
		if (rt == rt_createComponent) {
		    /* Ack! */
		    rq->rq_u.ru_ints[2] = htons(rq->rq_u.ru_ints[2]);
		    rq->rq_u.ru_ints[3] = htons(rq->rq_u.ru_ints[3]);
		}
	    }
	    if (cl->cl_outGoing == P_NULL) {
		/* Nothing queued - can try it right away. */
		len = REQ_LEN + rq->rq_usedLen;
		rq->rq_usedLen = htons(rq->rq_usedLen);
		res = doSend(cl, &rq->rq_usedLen, len);
		if (res < 0) {
		    freeRequestE(&MS, rq);
		} else {
		    if (res == len) {
			/* Sent it immediately. */
			MessagesSent += 1;
			freeRequestE(&MS, rq);
		    } else {
			/* Didn't send the whole message. */
			rq->rq_usedLen = ntohs(rq->rq_usedLen);
			cl->cl_outGoing = rq;
			cl->cl_outGoingPos = res;
			rq->rq_next = P_NULL;
			addToWriteWaiters(cl);
		    }
		}
	    } else {
		/* Something already started - just queue this one. We will
		   retry sends to this client from the main 'select' call. */
		count = 0;
		tail = &cl->cl_outGoing->rq_next;
		while (*tail != P_NULL) {
		    tail = &(*tail)->rq_next;
		    count += 1;
		}
		*tail = rq;
		rq->rq_next = P_NULL;
		if (count > 5) {
		    /* We have queued up several - try to send right now. */
		    clientWrite(cl);
		}
	    }
	}
    } else {
	freeRequestE(&MS, rq);
    }
}

/*
 * endClient2 - free stuff here for this client, including socket.
 */

void
endClient2(Client_t *cl)
{
    ActionList_t *al;
    ActionList_t **alp;
    Request_t *rq;

    TRACE("endClient2", cl);

    /* Free any timed events associated with this client. */
    alp = &Actions;
    while (B_TRUE) {
	al = *alp;
	if (al == P_NULL) {
	    break;
	}
	if (al->al_client == cl) {
	    decProcRef(al->al_action);
	    *alp = al->al_next;
	    memFreeE(&MS, al, sizeof(ActionList_t));
	} else {
	    alp = &al->al_next;
	}
    }

    /* Free any messages queued to go out to this client. */
    while (B_TRUE) {
	rq = cl->cl_outGoing;
	if (rq == P_NULL) {
	    break;
	}
	cl->cl_outGoing = rq->rq_next;
	freeRequestE(&MS, rq);
    }

    detachSocket(cl);
    ClientDeleted = B_TRUE;
}

/*
 * getTime - return a current time number in 100ths of a second.
 *	Since the value will not fit in a 32 bit quantity, we modulo the
 *	second count by 20,000,000, which makes sure the final value is
 *	an unsigned 32 bit value. Ick. Hey ANSI! How do I do portable
 *	time arithmetic that survives the year 2000?
 */

ULONG_T
getTime(void)
{
#if defined(AMIGA)
    struct DateStamp ds;

    DateStamp(&ds);
    return (ULONG_T)
	(ds.ds_Days * (24 * 60) + ds.ds_Minute) * 60 % 20000000 * 100 +
	 ds.ds_Tick * 2;
#elif defined(UNIX)
    struct timeval tv;

    gettimeofday(&tv, (struct timezone *) P_NULL);
    return tv.tv_sec % 20000000 * 100 + tv.tv_usec / 10000;
#else
%%%%%
#endif
}

/*
 * After - user routine to schedule an action after a given second count.
 */

void
After(LONG_T seconds, ULONG_T theAction)
{
    ActionList_t **alp;
    ActionList_t *al, *al2;
    ULONG_T actualDelta, now;

    TRACE("After", theAction);
    if (ActiveClient == P_NULL && ActiveMachine == P_NULL) {
	runError("After called by destroyed machine");
    } else if (seconds < 0) {
	runError("After called with negative interval");
    } else if (checkAction(theAction, "After")) {
	al = MEM_ALLOC_E(&MS, sizeof(ActionList_t), "actl");
	/* Convert from a fixed representing seconds to a ulong
	   representing 100ths of seconds. All this rounding up stuff is
	   needed to get the right answer for values like 0.1, 0.4, etc. */
	seconds = (seconds + ((1 << 7) - 1)) / (1 << 8);
	seconds = seconds * 100;
	seconds =
	    (seconds + ((1 << (FRAC_BITS - 9)) - 1)) / (1 << (FRAC_BITS - 8));
	al->al_client = ActiveClient;
	al->al_machine = ActiveMachine;
	al->al_delay = seconds;
	al->al_action = theAction;
	incProcRef(theAction);
	if (! InDoDelayedActions) {
	    now = getTime();
	    actualDelta = now - LastTime;
	    LastTime = now;
	    if (actualDelta != 0) {
		al2 = Actions;
		while (al2 != P_NULL) {
		    al2->al_delay = al2->al_delay - actualDelta;
		    al2 = al2->al_next;
		}
	    }
	    al2 = Actions;
	    if (al2 == P_NULL || al2->al_delay > seconds) {
		ActionDelay = seconds;
	    }
	}
	alp = &Actions;
	while (B_TRUE) {
	    al2 = *alp;
	    if (al2 == P_NULL || al2->al_delay > seconds) {
		break;
	    }
	    alp = &al2->al_next;
	}
	al->al_next = al2;
	*alp = al;
    }
}

/*
 * sendChangePassword - send an rt_changePassword to the client.
 */

void
sendChangePassword(Client_t *cl, ULONG_T key)
{
    Request_t *rq;

    rq = allocRequestE(&MS, 0);
    rq->rq_type = rt_changePassword;
    rq->rq_key = key;
    rq->rq_flag = cl->cl_playerPassword;
    putMessageCarefully(cl, rq);
}

/*
 * startChangePassword - get the client to ask for a password.
 */

void
startChangePassword(void)
{
    Client_t *cl;

    TRACE("changePassword", 0);
    cl = ActiveClient;
    if (cl != P_NULL && cl->cl_player != NIL_PLAYER) {
	cl->cl_expectingOldPassword = B_TRUE;
	if (cl->cl_telnet == P_NULL) {
	    sendChangePassword(cl, 0);
	} else {
	    telnetChangePassword(cl);
	}
    }
}

/*
 * setNewPrompt - ask a client to change its prompt.
 */

void
setNewPrompt(CHAR_T *newPrompt)
{
    Client_t *cl;
    Request_t *rq;
    UINT_T len;

    TRACE("setNewPrompt", newPrompt);
    cl = ActiveClient;
    if (cl != P_NULL && cl->cl_player != NIL_PLAYER) {
	if (cl->cl_telnet == P_NULL) {
	    len = strLen1(newPrompt);
	    rq = allocRequestE(&MS, len);
	    rq->rq_type = rt_setPrompt;
	    copy(&rq->rq_u.ru_text[0], newPrompt, len);
	    putMessageCarefully(cl, rq);
	} else {
	    telnetSetPrompt(cl, newPrompt);
	}
    }
}

/*
 * changeClientMode - send a change mode to the given client.
 */

void
changeClientMode(Client_t *cl, BOOL_T wizard)
{
    Request_t *rq;

    TRACE("changeClientMode", cl);
    if (ActiveClient != P_NULL && ActiveClient->cl_player != NIL_PLAYER &&
	ActiveClient->cl_canWizard && ! cl->cl_deleted)
    {
	rq = allocRequestE(&MS, 0);
	if (wizard) {
	    rq->rq_type = rt_enterWizardMode;
	    if (cl->cl_name != NULL && *cl->cl_name != '\0') {
		log2NL("Setting WIZARD mode for ", cl->cl_name);
	    } else {
		logXNL("Setting WIZARD mode for ", cl->cl_player);
	    }
	    cl->cl_wizardMode = B_TRUE;
	} else {
	    rq->rq_type = rt_leaveWizardMode;
	    if (cl->cl_name != NULL && *cl->cl_name != '\0') {
		log2NL("Clearing WIZARD mode for ", cl->cl_name);
	    } else {
		logXNL("Clearing WIZARD mode for ", cl->cl_player);
	    }
	    cl->cl_wizardMode = B_FALSE;
	}
	rq->rq_key = cl->cl_player;
	putMessageCarefully(cl, rq);
	if (! wizard) {
	    setNewPrompt(cl->cl_prompt);
	}
    }
}

/*
 * SetContinue - tell client to enable/disable error defining.
 */

void
SetContinue(ULONG_T flag)
{
    Client_t *cl;
    Request_t *rq;

    TRACE("SetContinue", flag);
    cl = ActiveClient;
    if (cl == P_NULL || EffectivePlayer != SYSADMIN_KEY ||
	RealPlayer != SYSADMIN_KEY)
    {
	runError("only SysAdmin can SetContinue");
    } else if (cl->cl_player != NIL_PLAYER) {
	if (cl->cl_telnet == P_NULL) {
	    rq = allocRequestE(&MS, 0);
	    rq->rq_type = rt_setContinue;
	    rq->rq_key = flag;
	    putMessageCarefully(cl, rq);
	}
    }
}

/*
 * shutdownCarefully - common code for bootClient and nukeClient.
 */

static void
shutdownCarefully(Client_t *cl, RequestType_t rt)
{
    Request_t *rq;
    LONG_T len;
    int s;

    s = cl->cl_socket;
    if (s != -1) {
	if (cl->cl_telnet == P_NULL) {
	    rq = allocRequestE(&MS, 0);
	    rq->rq_type = rt;
	    len = REQ_LEN;
	    if (send(s, (char *) &rq->rq_usedLen, len, 0) == len) {
		MessagesSent += 1;
	    } else {
		detachSocket(cl);
	    }
	    freeRequestE(&MS, rq);
	} else {
	    telnetStopClient(cl->cl_telnet, (BOOL_T) (rt == rt_killClient));
	}
    }
}

/*
 * bootClient - try to boot a client off.
 */

void
bootClient(Client_t *cl)
{

    TRACE("bootClient", cl);
    log2NL("BootClient for player ", cl->cl_name);
    if (ActiveClient != P_NULL && ActiveClient->cl_player != NIL_PLAYER) {
	shutdownCarefully(cl, rt_killClient);
    }
}

/*
 * nukeClient - forcibly remove a client. Dangerous. Emergency use only.
 */

void
nukeClient(Client_t *cl)
{

    TRACE("nukeClient", cl);
    log2NL("NukeClient for player ", cl->cl_name);
    if (ActiveClient != P_NULL && ActiveClient->cl_player != NIL_PLAYER) {
	shutdownCarefully(cl, rt_nukeClient);
	endClient(cl);
    }
}

/*
 * deleteFromClients - ask all clients to remove the given proc from their
 *	proc lists.
 */

void
deleteFromClients(ULONG_T key)
{
    Client_t *cl;
    Request_t *rq;

    TRACE("deleteFromClients", key);
    cl = Clients;
    while (cl != P_NULL) {
	if (cl->cl_player != NIL_PLAYER) {
	    if (cl->cl_telnet == P_NULL) {
		rq = allocRequestE(&MS, 0);
		rq->rq_type = rt_deleteProc;
		rq->rq_key = key;
		putMessageCarefully(cl, rq);
	    }
	}
	cl = cl->cl_next;
    }
}

/*
 * machineShutdown - remove any delayed actions the machine may have.
 */

void
machineShutDown(MachineList_t *ml)
{
    ActionList_t **pal;
    ActionList_t *al;

    TRACE("machineShutDown", ml);
    pal = &Actions;
    while (B_TRUE) {
	al = *pal;
	if (al == P_NULL) {
	    break;
	}
	if (al->al_machine == ml) {
	    decProcRef(al->al_action);
	    *pal = al->al_next;
	    memFreeE(&MS, al, sizeof(ActionList_t));
	} else {
	    pal = &al->al_next;
	}
    }
}

/*
 * ShutDown - builtin routine to set/clear the shutdown requested flag.
 */

ULONG_T
ShutDown(ULONG_T flag)
{
    ULONG_T oldState;

    TRACE("ShutDown", flag);
    oldState = QuitRequested ? BOOLEAN_TRUE : BOOLEAN_FALSE;
    if (ActiveClient != P_NULL && EffectivePlayer == SYSADMIN_KEY &&
	RealPlayer == SYSADMIN_KEY)
    {
	if (flag == BOOLEAN_FALSE) {
	    log("SysAdmin has cleared the shutdown flag.\n");
	    if (Verbose) {
		mess(
"SysAdmin has cleared the 'shutdown' flag using 'ShutDown(false)'.\n"
		);
	    }
	    QuitRequested = B_FALSE;
	} else {
	    log("SysAdmin has set the shutdown flag.\n");
	    if (Verbose) {
		mess(
"SysAdmin has set the 'shutdown' flag using 'ShutDown(true)'.\n"
"MUDServ will shut down when there are no more active clients.\n"
		);
	    }
	    QuitRequested = B_TRUE;
	}
    }
    return oldState;
}

/*
 * doDelayedActions - do any actions on the action list that are scheduled
 *	to happen now or earlier.
 */

static void
doDelayedActions(void)
{
    ActionList_t *al;
    MachineList_t *ml;
    Client_t *cl;
    LONG_T actualDelta;
    ULONG_T eventCount;
    ULONG_T now;

    TRACE("doDelayedActions", 0);
    InDoDelayedActions = B_TRUE;
    now = getTime();
    actualDelta = now - LastTime;
    LastTime = now;

    /* We have to use this double-scan method, along with the signed delays,
       to handle the situation (likely quite common), where an action called
       here will queue up another one. Also, it is quite possible that a new
       action added by a call here will get done on this pass through the
       delayed actions. If there are enough such, we might not ever get out
       of this loop! Hmm. This is not good. Kludge in something that will
       get us out every now and then, so we handle client requests, etc. */

    al = Actions;
    while (al != P_NULL) {
	if (al->al_client != P_NULL || MachinesActive) {
	    al->al_delay = al->al_delay - actualDelta;
	}
	al = al->al_next;
    }
    eventCount = 0;
    while (B_TRUE) {
	al = Actions;
	if (al == P_NULL || al->al_delay > 0 ||
	    eventCount == EVENT_MAX_PER_BUNCH)
	{
	    break;
	}
	eventCount += 1;
	Actions = al->al_next;
	ml = al->al_machine;
	ActiveMachine = ml;
	TrueMachine = ml;
	cl = al->al_client;
	ActiveClient = cl;
	TrueClient = cl;
	if (cl == P_NULL) {
	    if (ml->ml_name == P_NULL) {
		MeString[0] = '\0';
	    } else {
		strNCpy(&MeString[0], ml->ml_name, ME_STRING_LEN);
	    }
	    RealPlayer = ActiveMachine->ml_owner;
	    /* machines run as apprentices! */
	    RealStatus = ps_apprentice;
	} else {
	    strCpy(&MeString[0], cl->cl_name);
	    RealPlayer = cl->cl_player;
	    RealStatus = cl->cl_status;
	}
	EffectivePlayer = RealPlayer;
	EffectiveStatus = RealStatus;
	/* beware: this could start a timer right away! */
	(void) doProc0Top(al->al_action);
	/* WARNING!!! if we are running a machine, the previous call may
	   have destroyed the machine and thus removed any delayed actions
	   for that machine from the list! */
	/* flush for each proc called, since the value of ActiveClient for
	   them can be changing */
	messageFlush();
	flushEffects();
	CurrentIndent = 0;
	/* Do this AFTER running the proc, so that it can keep running even
	   if its owner has lost all references to it */
	decProcRef(al->al_action);
	memFreeE(&MS, al, sizeof(ActionList_t));
    }
    ActiveClient = P_NULL;
    ActiveMachine = P_NULL;
    TrueClient = P_NULL;
    TrueMachine = P_NULL;
    al = Actions;
    if (al != P_NULL && al->al_delay > 0) {
	ActionDelay = al->al_delay;
    } else {
	ActionDelay = 0;
    }
    InDoDelayedActions = B_FALSE;
}

/* NOTE: need to change this. On UNIX, want to use abort signal only when
   user code is executing - otherwise it will interrupt all sorts of
   system calls. What if user code requires flushes to sockets???? */

/*
 * abortEnable - enable one pulse on the abort timer.
 */

static void
abortEnable(void)
{

}

#if 0
/*
 * abortHandler - handle the interrupt for the abort timer.
 */

static void
abortHandler(void)
{

    if (CodeRunning) {
	AbortCount += 1;
	if (AbortCount > AbortMax) {
	    runAbort();
	}
    }
    if (! ShuttingDown) {
	abortEnable();
    }
}
#endif

/*
 * startTimeout - start the abort countdown.
 */

void
startTimeout(void)
{

    AbortedOnce = B_FALSE;
    AbortCount = 0;
    CodeRunning = B_TRUE;
}

/*
 * stopTimeout - stop the abort countdown.
 */

void
stopTimeout(void)
{

    CodeRunning = B_FALSE;
}

/*
 * allow external people (FileIO.c) to have time NOT counted in the timer.
 */

ULONG_T
getTimeout(void)
{

    return AbortCount;
}

void
restoreTimeout(ULONG_T n)
{

    AbortCount = n;
}

/*
 * RunLimit - allow set/get of the execution timer.
 */

ULONG_T
RunLimit(ULONG_T newValue)
{
    ULONG_T oldValue;

    oldValue = getAbortTimeout();
    setAbortTimeout(newValue);
    AbortMax = newValue;
    return oldValue;
}

/*
 * handleMessage - handle a single message to the server.
 */

static void
handleMessage(Request_t *rq, Client_t *cl)
{
    RequestType_t rt;

    rt = rq->rq_type;
    TRACE("handleMessage", rt);
#if TRACING
    if (rt <= rt_illegal) {
	TRACE(RqNames[rt - rt_first], rq->rq_key);
    }
#endif
    if (rt == rt_shutDown) {
	TRACE("rt_shutDown", cl);
	freeClient(cl);
	logN("Server shutdown request via message - currently ",
	     ClientCount, " clients\n");
	if (Verbose) {
	    mess("Server shutdown request via message. There are currently ");
	    mNum(ClientCount);
	    mess(" active clients.\n");
	}
	QuitRequested = B_TRUE;
    } else if (rt == rt_syncShutDown) {
	TRACE("rt_syncShutDown", cl);
	logN("Server shutdown request via message - currently ",
	     ClientCount, " clients\n");
	if (Verbose) {
	    mess("Server shutdown request via message. There are currently ");
	    mNum(ClientCount);
	    mess(" active clients.\n");
	}
	QuitRequested = B_TRUE;
	/* This client is really just a shutdown request. Move it. */
	unlinkClient(cl);
	cl->cl_next = ShutDowns;
	ShutDowns = cl;
	if (Verbose) {
	    mess("Entry added to shutdown notification list.\n");
	}
    } else if (rt == rt_flush) {
	TRACE("rt_flush", cl);
	freeClient(cl);
	log("Server flush request via message\n");
	flushClients();
	flushCaches();
	ioFlush(B_TRUE);
    } else if (rt == rt_log) {
	TRACE("rt_log", cl);
	log(&rq->rq_u.ru_text[0]);
	freeRequestE(&MS, rq);
    } else {
	if (rt == rt_askPlayer) {
	    /* Want to know if this is a local connection, so we can allow
	       a SysAdmin login. */
	    if (cl->cl_isLocal) {
		rq->rq_key = 0;
	    }
	} else if (rt == rt_symbolEnter || rt == rt_symbolDelete ||
	    rt == rt_regionSelect || rt == rt_effectDone)
	{
	    rq->rq_u.ru_otherKey = ntohl(rq->rq_u.ru_otherKey);
	}
	handleOneRequest(cl, rq);
    }
}

/*
 * gotHeader - we have just received all of a request header.
 */

static void
gotHeader(Client_t *cl, Request_t *rq, BOOL_T alloced)
{
    Request_t *rqNew;
    LONG_T res;
    UINT_T len;

    len = ntohs(rq->rq_usedLen);
    rq->rq_usedLen = len;
#if SOCKET_DEBUG
    mess("gotHeader, usedLen ");
    mNum(len);
    mess(" type ");
    mNum(rq->rq_type);
    messChar('\n');
#endif 
    if (len == 0) {
	/* no tail needed */
	if (! alloced) {
	    rqNew = allocRequestE(&MS, 0);
	    rqNew->rq_key = ntohl(rq->rq_key);
	    rqNew->rq_type = rq->rq_type;
	    rqNew->rq_flag = rq->rq_flag;
	    rqNew->rq_uint = ntohs(rq->rq_uint);
	    rq = rqNew;
	}
	cl->cl_inComing = P_NULL;	/* may point to this rq */
	handleMessage(rq, cl);
    } else {
	/* got a tail - get space for it */
	rqNew = allocRequestE(&MS, len);
	rqNew->rq_key = ntohl(rq->rq_key);
	rqNew->rq_type = rq->rq_type;
	rqNew->rq_flag = rq->rq_flag;
	rqNew->rq_uint = ntohs(rq->rq_uint);
	if (alloced) {
	    freeRequestE(&MS, rq);
	}
	res = doRecv(cl, &rqNew->rq_u, len);
	if (res > 0) {
	    if (res == len) {
		/* got the whole tail */
		cl->cl_inComing = P_NULL;
		handleMessage(rqNew, cl);
	    } else {
		/* more to come later */
		rqNew->rq_next = P_NULL;
		cl->cl_inComing = rqNew;
		cl->cl_inComingPos = REQ_LEN + res;
	    }
	}
    }
}

/*
 * clientRead - read from a given client socket.
 */

static void
clientRead(Client_t *cl)
{
    Request_t *rq;
    LONG_T res, pos;
    Request_t initialRq;

    if (cl->cl_inComing == P_NULL) {
	/* a new request coming in */
	res = doRecv(cl, &initialRq.rq_usedLen, REQ_LEN);
	if (res > 0) {
	    RequestsProcessed += 1;
	    if (res == REQ_LEN) {
		/* read the entire header part */
		gotHeader(cl, &initialRq, B_FALSE);
	    } else {
		/* More of the header to come. */
		rq = allocRequestE(&MS, 0);
		memcpy(&rq->rq_usedLen, &initialRq.rq_usedLen, res);
		rq->rq_next = P_NULL;
		cl->cl_inComing = rq;
		cl->cl_inComingPos = res;
	    }
	}
    } else {
	/* There is already a partial request received. */
	rq = cl->cl_inComing;
	pos = cl->cl_inComingPos;
	if (pos < REQ_LEN) {
	    /* Only partial header, so far. */
	    res = doRecv(cl, (BYTE_T *) &rq->rq_usedLen + pos, REQ_LEN - pos);
	    if (res > 0) {
		if (res == REQ_LEN - pos) {
		    /* Got all of the header. */
		    gotHeader(cl, rq, B_TRUE);
		} else {
		    cl->cl_inComingPos = pos + res;
		}
	    }
	} else {
	    /* It was a partial tail. */
	    pos -= REQ_LEN;
	    res = doRecv(cl, &rq->rq_u.ru_bytes[pos], rq->rq_usedLen - pos);
	    if (res > 0) {
		pos += res;
		if (pos == rq->rq_usedLen) {
		    /* Got it all now. */
		    cl->cl_inComing = P_NULL;
		    handleMessage(rq, cl);
		} else {
		    cl->cl_inComingPos = pos + REQ_LEN;
		}
	    }
	}
    }
}

/*
 * scanClients - scan the client sockets for something we can do.
 */

static void
scanClients(fd_set *readFds, fd_set *writeFds, fd_set *exceptFds)
{
    Client_t *cl;

    TRACE("scanClients", 0);
    cl = Clients;
    while (cl != P_NULL) {
	if (cl->cl_socket == -1) {
	    /* Might have been nuked by a client earlier in list. */
	} else if (FD_ISSET(cl->cl_socket, exceptFds)) {
	    SDEBUG("exception on client socket\n");
	    endClient(cl);
	} else {
	    if (cl->cl_socket != -1 && FD_ISSET(cl->cl_socket, readFds)) {
		/* Something to read from the client socket. */
		if (cl->cl_telnet == P_NULL) {
		    SDEBUG("scanClients doing clientRead\n");
		    clientRead(cl);
		} else {
		    SDEBUG("scanClients doing telnetRead\n");
		    ActiveClient = cl;
		    TrueClient = cl;
		    telnetRead(cl);
		    ActiveClient = P_NULL;
		    TrueClient = P_NULL;
		}
	    }
	    if (cl->cl_socket != -1 && FD_ISSET(cl->cl_socket, writeFds)) {
		/* We can write more to the client socket. */
		if (cl->cl_telnet == P_NULL) {
		    SDEBUG("scanClients doing clientWrite\n");
		    clientWrite(cl);
		} else {
		    SDEBUG("scanClients doing telnetWrite\n");
		    telnetWrite(cl);
		}
	    }
	}
	cl = cl->cl_next;
    }
}

/*
 * handleConnection - a new connection has come in on a main socket.
 */

static void
handleConnection(int mySocket, BOOL_T isTelnet, int telnetFlags)
{
    struct sockaddr_in from;
    Client_t *cl;
    int s;
    long int flag;
#if defined(AMIGA)
    LONG fromLen;
#else
    int fromLen;
#endif
    BOOL_T isLocal;

    TRACE("handleConnection", 0);
    fromLen = sizeof(struct sockaddr_in);
    s = accept(mySocket, (struct sockaddr *) &from, &fromLen);
#if defined(_WIN32)
    if (s == INVALID_SOCKET) {
	s = WSAGetLastError();
	if (s != WSAEINTR && s != WSAEWOULDBLOCK) {
	    myPerror("accept");
	    myAbort("client socket error");
	}
    }
#else
    if (s < 0) {
	if (errno != EINTR && errno != EWOULDBLOCK) {
	    myPerror("accept");
	    myAbort("client socket error");
	}
    }
#endif
    else {
	if (from.sin_addr.s_addr == MyAddr) {
	    isLocal = B_TRUE;
	} else {
	    isLocal = B_FALSE;
	}
	flag = 1;
	if (ioctl(s, FIONBIO, (char *) &flag) == -1) {
	    myPerror("ioctl on new socket");
	} else if ((ForceNoDelay || isLocal) &&
		   setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
			      (char *) &flag, sizeof(int)) != 0)
	{
	    myPerror("setsockopt on new socket");
	} else {
	    cl = MEM_ALLOC_E(&MS, sizeof(Client_t), "clnt");
	    TRACE("handleConnection got one OK", cl);
	    cl->cl_next = P_NULL;
	    cl->cl_prev = P_NULL;
	    cl->cl_nextKHash = P_NULL;
	    cl->cl_nextPHash = P_NULL;
	    cl->cl_nextLHash = P_NULL;
	    cl->cl_prevLHash = P_NULL;
	    cl->cl_nextLoc = P_NULL;
	    cl->cl_prevLoc = P_NULL;
	    clearClient(cl);
	    cl->cl_socket = s;
	    cl->cl_isLocal = isLocal;
	    if (isTelnet) {
		SDEBUG("handleConnection got new telnet connection\n");
		log("New telnet connection\n");
		cl->cl_telnet = telnetAllocClient(cl, telnetFlags);
	    } else {
		SDEBUG("handleConnection got new non-telnet connection\n");
		log("New binary connection\n");
		cl->cl_telnet = P_NULL;
	    }
	    FD_SET(s, &AllClientBits);
	    if (s > MaxFd) {
		MaxFd = s;
	    }
	    cl->cl_next = Clients;
	    cl->cl_prev = P_NULL;
	    if (Clients != P_NULL) {
		Clients->cl_prev = cl;
	    }
	    Clients = cl;
	    ClientCounter += 1;
	    ClientCount += 1;
	    if (isTelnet) {
		if (beginClient(cl, NULL) == 0) {
		    sendText(cl, "Single user game already in use\n", 32);
		    SDEBUG("telnet bad beginClient - calling endClient\n");
		    endClient(cl);
		}
	    }
	}
    }
}

/*
 * handleRequests - wait for and handle requests from clients.
 */

static void
handleRequests(void)
{
    MachineList_t *ml;
    MachineList_t **pml;
    Client_t *cl, *clTemp;
    BOOL_T doneDelayedActions;
    fd_set readFds, writeFds, exceptFds;
    struct timeval timeval;
    int res, i;

    TRACE("handleRequests", 0);
    MachinesActive =
	! NoMachines && (getDungeonFlags() & FLAG_MACHINES_ACTIVE) != 0x0;
    MachineDeleted = B_FALSE;

    NeedRecover = B_FALSE;
    NeedCacheCompact = B_FALSE;
    while (! QuitRequested || Clients != P_NULL) {
	if (NeedRecover) {
	    NeedRecover = B_FALSE;
	    memRecoverE(&MS);
	}
	if (NeedCacheCompact) {
	    NeedCacheCompact = B_FALSE;
	    ioCacheCompact();
	}
	if (Actions != P_NULL) {
	    if (Actions->al_delay > 0) {
		/* Nothing to do now - wait the indicated time. */
		timeval.tv_sec = Actions->al_delay / 100;
		timeval.tv_usec = Actions->al_delay % 100 * 10000;
	    } else {
		/* Some actions need doing now - just poll. */
		timeval.tv_sec = 0;
		timeval.tv_usec = 0;
	    }
	} else {
	    /* Nothing queued at all - just wait 10 seconds for input. */
	    timeval.tv_sec = 10;
	    timeval.tv_usec = 0;
	}
	readFds = AllClientBits;
	FD_SET(BinarySocket, &readFds);
	for (i = 0; i != TelnetPortCount; i += 1) {
	    FD_SET(TelnetSockets[i], &readFds);
	}
	writeFds = ClientWriteBits;
	exceptFds = readFds;
	res = select(MaxFd + 1, &readFds, &writeFds, &exceptFds, &timeval);
	doneDelayedActions = B_FALSE;
#if defined(_WIN32)
	if (res == SOCKET_ERROR) {
	    if (WSAGetLastError() == WSAEINTR)
#else
	if (res < 0) {
	    if (errno == EINTR)
#endif
	    {
		logN("Server quit requested via signal - currently ",
		     ClientCount, " clients.\n");
		QuitRequested = B_TRUE;
		if (Verbose) {
		    if (Clients == P_NULL) {
			mess(
		  "Server quit requested via signal - MUDServ shutting down.\n"
			);
			break;
		    } else {
			mess(
		       "Server quit requested via signal. There are currently "
			);
			mNum(ClientCount);
			mess(" active clients.\n"
		    "MUDServ will shut down when those clients are all gone.\n"
			);
		    }
		}
	    } else {
		myPerror("select");
		break;
	    }
	} else if (res > 0) {
	    /* Activity on some socket. */
#if SOCKET_DEBUG
	    mess("select res ");
	    mNum(res);
	    messChar('\n');
#endif
	    if (FD_ISSET(BinarySocket, &exceptFds)) {
		SDEBUG("exception on binary socket - shutting down\n");
		break;
	    }
	    res = 0;
	    for (i = 0; i != TelnetPortCount; i += 1) {
		if (FD_ISSET(TelnetSockets[i], &exceptFds)) {
		    SDEBUG("exception on telnet socket - shutting down\n");
		    res = 1;
		    break;
		}
	    }
	    if (res != 0) {
		break;
	    }
	    if (FD_ISSET(BinarySocket, &readFds)) {
		SDEBUG("new binary connection\n");
		handleConnection(BinarySocket, B_FALSE, 0x0);
	    }
	    for (i = 0; i != TelnetPortCount; i += 1) {
		if (FD_ISSET(TelnetSockets[i], &readFds)) {
		    SDEBUG("new telnet connection\n");
		    handleConnection(TelnetSockets[i], B_TRUE, TelnetFlags[i]);
		}
	    }
	    scanClients(&readFds, &writeFds, &exceptFds);
	} else {
	    /* Nothing on any socket - timer has gone off. */
	    doDelayedActions();
	    doneDelayedActions = B_TRUE;
	}
	if (! doneDelayedActions && Actions != P_NULL &&
	    Actions->al_delay <= 0)
	{
	    /* Have to go through if we didn't recently go through, otherwise
	       the al_delay values never get decremented! Also, we may want
	       to start the timer. */
	    doDelayedActions();
	}
	if (ClientDeleted) {
	    ClientDeleted = B_FALSE;
	    cl = Clients;
	    while (cl != P_NULL) {
		clTemp = cl->cl_next;
		if (cl->cl_deleted) {
		    if (cl->cl_inserted && cl->cl_player != NIL_PLAYER) {
			fullyUnlinkClient(cl);
		    }
		    freeClient(cl);
		    ClientCount -= 1;
		}
		cl = clTemp;
	    }
	}
	if (MachineDeleted) {
	    MachineDeleted = B_FALSE;
	    pml = &Machines;
	    while (B_TRUE) {
		ml = *pml;
		if (ml == P_NULL) {
		    break;
		}
		if (ml->ml_deleted) {
		    *pml = ml->ml_next;
		    if (ml->ml_name != P_NULL) {
			freeString(ml->ml_name);
		    }
		    locUnlinkMachine(ml, ml->ml_location);
		    /* It was already deleted from the key table, back
		       in 'DestroyMachine'. */
		    memFreeE(&MS, ml, sizeof(MachineList_t));
		    MachinesDeleted += 1;
		} else {
		    pml = &ml->ml_next;
		}
	    }
	}
    }

    if (ClientCounter == 2) {
	log("Shutting down, 1 client serviced\n");
    } else {
	logN("Shutting down, ", ClientCounter - 1, " clients serviced\n");
    }
    logN("Machines: ", MachinesCreated, " created, ");
    logN(P_NULL, MachinesDeleted, " deleted\n");
}

/*
 * setNeedRecover - allow memAllocE/reclaimMem to request a memRecoverE.
 */

void
setNeedRecover(void)
{

    NeedRecover = B_TRUE;
}

/*
 * requestCacheCompact - allow database code to request a cache compact
 *	callout. That is only safe to do when no MUD code is executing.
 */

void
requestCacheCompact(void)
{

    NeedCacheCompact = B_TRUE;
}

static void
doStartMsg(void)
{
    /* Grrr. This wouldn't be necessary if C had proper string escapes! */
    static char message[] = {
	'M', 'U', 'D', ' ', 's', 'e', 'r', 'v', 'e', 'r', ' ',
	'V', VERSION1, '.', VERSION2, ' ',
	'C', 'o', 'p', 'y', 'r', 'i', 'g', 'h', 't', ' ',
	'1', '9', YEAR1, YEAR2, ' ', 'b', 'y', ' ',
	'C', 'h', 'r', 'i', 's', ' ', 'G', 'r', 'a', 'y', ' ',
	's', 't', 'a', 'r', 't', 'e', 'd', '\n', '\0'
    };

    log(&message[0]);
}

/*
 * startServing - everything all set - start cranking.
 */

static void
startServing(ULONG_T libStart)
{
    extern int main(int, char **);
    MachineList_t *ml;
    MachineList_t **pMl;
    ItemList_t *l;
    ULONG_T *p;
    Machine_t *mch;
    Entry_t *en;
    UINT_T count;

    TRACE("startServing", 0);
    bcInit();
    cacheInit();
    matchInit();
    parseInit();
    effectsInit();
    TableList = P_NULL;
    PublicSymbolTableKey = getPublicSymbolTableKey();
    PublicSymbolTable = findCachedTable(PublicSymbolTableKey);
    PublicNameMap = buildNameMap(PublicSymbolTable);
    en = symLookupE(PublicSymbolTable, "Characters", 11);
    if (en == P_NULL || (en->en_key >> TYPE_SHIFT) != TYPE_TABLE) {
	myAbort("can't find Characters in public symbol table");
    }
    PlayerTable = addGlobalUse(findCachedTable(en->en_key), en->en_key);
    en = symLookupE(PublicSymbolTable, "Builtin", 8);
    if (en == P_NULL || (en->en_key >> TYPE_SHIFT) != TYPE_TABLE) {
	myAbort("can't find Builtin in public symbol table");
    }
    BuiltinTable = addGlobalUse(findCachedTable(en->en_key), en->en_key);
    builtInInit(((ULONG_T) time((time_t *) P_NULL) & 0x7ffeffff) | 1);
    CacheDone = B_TRUE;
    doStartMsg();

    EffectivePlayer = SYSADMIN_KEY;
    RealPlayer = SYSADMIN_KEY;
    EffectiveStatus = ps_wizard;
    RealStatus = ps_wizard;
    if (! NoMachines && (getDungeonFlags() & FLAG_STARTUP_MACHINES) != 0x0) {
	l = ioRead(getMachineList());
	count = l->l_elementCount;
	p = (ULONG_T *) (l + 1);
	while (count != 0) {
	    count -= 1;
	    mch = ioRead(*p);
	    ml = MEM_ALLOC_E(&MS, sizeof(MachineList_t), "mach");
	    ml->ml_next = Machines;
	    ml->ml_machine = *p;
	    ml->ml_owner = mch->mch_owner;
	    ml->ml_thing = mch->mch_key;
	    ml->ml_location = mch->mch_location;
	    ml->ml_sayProc = mch->mch_sayProc;
	    ml->ml_whisperMeProc = mch->mch_whisperMeProc;
	    ml->ml_whisperOtherProc = mch->mch_whisperOtherProc;
	    ml->ml_pageProc = mch->mch_pageProc;
	    ml->ml_poseProc = mch->mch_poseProc;
	    ml->ml_otherProc = mch->mch_otherProc;
	    ml->ml_idleProc = mch->mch_idleProc;
	    ml->ml_activeProc = mch->mch_activeProc;
	    ml->ml_name =
		(CHAR_T *) getProp(ml->ml_thing, NAME_KEY, TYPE_STRING);
	    if (ml->ml_name == P_NULL) {
		ml->ml_name = stringCopy("<machine>");
	    }
	    ml->ml_dirty = B_FALSE;
	    ml->ml_deleted = B_FALSE;
	    fullyInsertMachine(ml);
	    Machines = ml;
	    ++p;
	}
    }

    LastTime = getTime();
    ShuttingDown = B_FALSE;
    CodeRunning = B_FALSE;
    AbortMax = getAbortTimeout();
    abortEnable();

    if (Verbose) {
	mess("Initialization complete - starting up active machines.\n");
    }

    ml = Machines;
    while (ml != P_NULL) {
	if (! ml->ml_deleted && ml->ml_activeProc != NIL_ACTION) {
	    ActiveMachine = ml;
	    TrueMachine = ml;
	    RealPlayer = ml->ml_owner;
	    EffectivePlayer = RealPlayer;
	    RealStatus = ps_apprentice;
	    EffectiveStatus = ps_apprentice;
	    if (ml->ml_name == P_NULL) {
		MeString[0] = '\0';
	    } else {
		strNCpy(&MeString[0], ml->ml_name, ME_STRING_LEN);
	    }
	    (void) doProc0Top(ml->ml_activeProc);
	    CurrentIndent = 0;
	    ActiveMachine = P_NULL;
	    TrueMachine = P_NULL;
	}
	ml = ml->ml_next;
    }

    if (Verbose) {
	mess("Active machines started - ready for client requests.\n");
    }

    handleRequests();

    if (Verbose) {
	mess("Shutting down active machines.\n");
    }

    ActiveClient = P_NULL;
    TrueClient = P_NULL;
    /* Do not need the careful machine scanning code here - we are just
       destroying them unconditionally */
    while (B_TRUE) {
	ml = Machines;
	if (ml == P_NULL) {
	    break;
	}
	Machines = ml->ml_next;
	if (ml->ml_idleProc != NIL_ACTION) {
	    ActiveMachine = ml;
	    TrueMachine = ml;
	    RealPlayer = ml->ml_owner;
	    EffectivePlayer = RealPlayer;
	    RealStatus = ps_apprentice;
	    EffectiveStatus = ps_apprentice;
	    if (ml->ml_name == P_NULL) {
		MeString[0] = '\0';
	    } else {
		strNCpy(&MeString[0], ml->ml_name, ME_STRING_LEN);
	    }
	    (void) doProc0Top(ml->ml_idleProc);
	    CurrentIndent = 0;
	    ActiveMachine = P_NULL;
	    TrueMachine = P_NULL;
	}
	locUnlinkMachine(ml, ml->ml_location);
	pMl = &MachineKeyMap[ml->ml_thing % AGENT_MAP_SIZE];
	while (*pMl != ml) {
	    pMl = &(*pMl)->ml_nextKHash;
	}
	*pMl = ml->ml_nextKHash;
	if (ml->ml_dirty) {
	    flushOneMachine(ml);
	}
	if (ml->ml_name != P_NULL) {
	    freeString(ml->ml_name);
	}
	memFreeE(&MS, ml, sizeof(MachineList_t));
    }
    ShuttingDown = B_TRUE;

    if (Verbose) {
	mess("Active machines shut down - freeing stuff.\n");
    }

    freeLocalLists(FREE_MEMORY_ON_SHUTDOWN);
#if FREE_MEMORY_ON_SHUTDOWN
    builtInTerm();
#else
    BuiltinTable->tl_this->tb_usedCount = 0;
    BuiltinTable->tl_this->tb_entryCount = 0;
#endif
    removeGlobalUse(BuiltinTable);
    removeGlobalUse(PlayerTable);
#if FREE_MEMORY_ON_SHUTDOWN
    freeNameMap(PublicNameMap);
#endif
    effectsTerm();
    cacheTerm(FREE_MEMORY_ON_SHUTDOWN, B_TRUE);
    bcTerm(FREE_MEMORY_ON_SHUTDOWN);
}

/*
 * writeProcNew - supply the stubs left that we need for the library.
 */

static ULONG_T
writeProcNew(BYTE_T *p, UINT_T len)
{

    return ioCreate(TYPE_ACTION, p, len);
}

/*
 * retryMem - allow library to ask user to free up some memory.
 */

static BOOL_T
retryMem(void)
{

    return checkForRetry("MUDServ: out of memory");
}

/*
 * getIntConstValue - retrieve the value of a defined integer constant.
 */

static ULONG_T
getIntConstValue(ULONG_T key)
{

    return ((Integer_t *) ioRead(key))->int_value;
}

/*
 * Some stubs to convert parameter conventions.
 */

static void *
myAllocMem(ULONG_T len)
{
    return malloc(len);
}

static void
myFreeMem(void *ptr, ULONG_T len)
{
    if (ptr == P_NULL) {
	myAbort("nil ptr to free");
    }
    if (len == 0) {
	myAbort("0 len to free");
    }
    free(ptr);
}

/*
 * sigIntHandler - for UNIX systems, catch SIGINT and request a shutdown.
 */

static void
sigIntHandler(int signum)
{

    QuitRequested = B_TRUE;
}

/*
 * runServer2 - second phase of 'runServer' - library specific stuff.
 */

static void
runServer2(void)
{
    extern void errorChar(char ch);

    TRACE("runServer2", 0);
    QuitRequested = B_FALSE;
    if (signal(SIGINT, sigIntHandler) == SIG_ERR) {
	myPerror("signal SIGINT");
	return;
    }
    /* Got a SIGPIPE on a 'send' using java client on a very busy world. */
    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
	myPerror("signal SIGPIPE");
	return;
    }
    MS.ms_AllocMem = myAllocMem;
    MS.ms_FreeMem = myFreeMem;
    /* we stuff in 'reclaimMem' later, after cacheInit has set things up */
    MS.ms_ReclaimMem = P_NULL;
    MS.ms_RetryMem = retryMem;
    MS.ms_Open = (int (*)(const char *, int, ...)) open;
    MS.ms_Close = close;
    MS.ms_Read = read;

    MS.ms_abort = libraryAbort;
    MS.ms_errorChar = errorChar;
    MS.ms_outputChar = putChar;

    MS.ms_incRef = incRef;
    MS.ms_decRef = decRef;
    MS.ms_lookup = lookup;
    MS.ms_findAnyName = findAnyName;
    MS.ms_findProc = findCachedProc;
    MS.ms_writeProcNew = writeProcNew;
    MS.ms_findBuiltin = findBuiltIn;
    MS.ms_getIntConstValue = getIntConstValue;

    MS.ms_editErrorStart = P_NULL;
    MS.ms_defineEffect = P_NULL;
    MS.ms_flushEffect = P_NULL;

    CacheDone = B_FALSE;
    ThingsCreated = 0;
    ThingsDestroyed = 0;
    RequestsProcessed = 0;
    MessagesSent = 0;

    startServing(libInitE(&MS, MemoryLimit, RequestLimit));

    logN("Things: ", ThingsCreated, " created, ");
    logN(P_NULL, ThingsDestroyed, " destroyed\n");
    logN("Messages: ", RequestsProcessed, " processed, ");
    logN(P_NULL, MessagesSent, " sent\n");
    libTermE(&MS);
}

/*
 * setupSocket - do all of the work needed to set up one of our main sockets.
 */

static BOOL_T
setupSocket(int *pSocket, short int port)
{
    struct sockaddr_in myAddr;
    int s;
    long int flag;
    
    s = socket(AF_INET, SOCK_STREAM, 0);
#if defined(_WIN32)
    if (s == INVALID_SOCKET)
#else
    if (s < 0)
#endif
    {
	myPerror("socket 1");
	return B_FALSE;
    }
    flag = 1;
    if (ioctl(s, FIONBIO, (char *) &flag) == -1) {
	myPerror("ioctl on socket");
	killSocket(s);
	return B_FALSE;
    }
    /* Do this *before* the bind/listen */
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag)) != 0) {
	myPerror("setsockopt(SO_REUSEADDR)");
	killSocket(s);
	return B_FALSE;
    }
    memset(&myAddr, 0, sizeof(struct sockaddr_in));
    myAddr.sin_port = htons(port);
    myAddr.sin_family = AF_INET;
    if (bind(s, (struct sockaddr *) &myAddr,
	sizeof(struct sockaddr_in)) < 0)
    {
	myPerror("bind");
	killSocket(s);
	return B_FALSE;
    }
    if (listen(s, 5) < 0) {
	myPerror("listen");
	killSocket(s);
	return B_FALSE;
    }
    *pSocket = s;
    return B_TRUE;
}

/*
 * runServer - start the MUD-specific stuff.
 */

void
runServer(void)
{
    char nameBuffer[MAXHOSTNAMELEN];
    struct hostent *hp;
    int i;
    BOOL_T ok;

    TRACE("runServer", 0);

#if defined AMIGA
    SetErrnoPtr(&errno, sizeof(int));
#elif defined(_WIN32)
    if (WSAStartup(MAKEWORD(1, 0), &WSAData) < 0) {
	mess("WSAStartup failure");
	return;
    }
#endif

    Actions = P_NULL;
    FreeCallLists = P_NULL;
    ShutDowns = P_NULL;
    Machines = P_NULL;
    ActiveMachine = P_NULL;
    TrueMachine = P_NULL;
    MachinesCreated = 0;
    MachinesDeleted = 0;
    ClientCounter = 1;
    ClientCount = 0;
    Clients = P_NULL;
    ActiveClient = P_NULL;
    TrueClient = P_NULL;
    for (i = 0; i != AGENT_MAP_SIZE; i += 1) {
	ClientKeyMap[i] = P_NULL;
	MachineKeyMap[i] = P_NULL;
	ClientLocMap[i] = P_NULL;
	MachineLocMap[i] = P_NULL;
	ClientPKeyMap[i] = P_NULL;
    }

    FD_ZERO(&AllClientBits);
    FD_ZERO(&ClientWriteBits);
    BinarySocket = -1;

    if (ioInitialize(&Path[0], CacheSize)) {
	if (gethostname(&nameBuffer[0], MAXHOSTNAMELEN) < 0) {
	    myPerror("gethostname");
	} else {
	    hp = gethostbyname(&nameBuffer[0]);
	    if (hp == P_NULL) {
		myPerror("gethostbyname for self");
	    } else {
		MyAddr = (ULONG_T) hp->h_addr;
		if (setupSocket(&BinarySocket, BinaryPort)) {
		    ok = B_TRUE;
		    i = 0;
		    while (ok && i != TelnetPortCount) {
			if (! setupSocket(&TelnetSockets[i], TelnetPorts[i])) {
			    ok = B_FALSE;
			} else {
			    i += 1;
			}
		    }
		    if (ok) {
			MaxFd = BinarySocket;
			for (i = 0; i != TelnetPortCount; i += 1) {
			    if (TelnetSockets[i] > MaxFd) {
				MaxFd = TelnetSockets[i];
			    }
			}
			runServer2();
			for (i = 0; i != TelnetPortCount; i += 1) {
			    killSocket(TelnetSockets[i]);
			}
			killSocket(BinarySocket);
		    } else {
			while (i != 0) {
			    i -= 1;
			    killSocket(TelnetSockets[i]);
			}
		    }
		}
	    }
	}
	ioTerminate();
	processShutDowns();
    }

#if defined(_WIN32)
    (void) WSACleanup();
#endif
}
