#include "MUD.h"
#include "types.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * names.c - name table routines.
 */

/*
 * emptyNameMap - empty out the pointer part of a namemap.
 */

void
emptyNameMap(NameMap_t *nm)
{
    Entry_t **pen;
    UINT_T i;

    pen = &nm->nm_map[0];
    for (i = 0; i < NAME_MAP_SIZE; i += 1) {
	*pen++ = P_NULL;
    }
}

/*
 * buildNameMap - build a namemap structure from the given symbol table,
 *	and return a pointer to it.
 */

NameMap_t *
buildNameMap(Table_t *tb)
{
    NameMap_t *nm;

    nm = MEM_ALLOC_E(&MS, sizeof(NameMap_t), "nmmp");
    emptyNameMap(nm);
    nm->nm_table = tb;
    return nm;
}

/*
 * findName - find the name associated with the given key in the given namemap.
 *	Return 'P_NULL' if none found.
 */

CHAR_T *
findName(NameMap_t *nm, ULONG_T key)
{
    Entry_t **pen;
    Table_t *tb;
    Entry_t *en;
    UINT_T count;

    pen = &nm->nm_map[key % NAME_MAP_SIZE];
    en = *pen;
    if (en == P_NULL || en->en_key != key ||
	en->en_nameLen == EN_FREE || en->en_nameLen == EN_EMPTY)
    {
	tb = nm->nm_table;
	count = tb->tb_entryCount;
	en = tb->tb_entries;
	while (count != 0) {
	    if (en->en_key == key && en->en_nameLen != EN_FREE &&
		en->en_nameLen != EN_EMPTY)
	    {
		*pen = en;
		return en->en_name;
	    }
	    count -= 1;
	    ++en;
	}
	return P_NULL;
    } else {
	return en->en_name;
    }
}

/*
 * freeNameMap - free all of the passed namemap.
 */

void
freeNameMap(NameMap_t *nm)
{

    memFreeE(&MS, nm, sizeof(NameMap_t));
}

/*
 * removeNameFromMap - remove the indicated name from the indicated namemap.
 */

void
removeNameFromMap(NameMap_t *nm, ULONG_T key)
{
    Entry_t **pen;

    pen = &nm->nm_map[key % NAME_MAP_SIZE];
    if (*pen != P_NULL && (*pen)->en_key == key) {
	*pen = P_NULL;
    }
}
