#include "MUD.h"
#include "types.h"
#include "exec.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * print.c - routines for pretty-printing various things.
 *	Note: the routines in this file are fairly expensive to use, so
 *	they are not used in normal game output. They are intended for
 *	wizard/apprentice type output.
 */

static UINT_T IndentLevel, CurrentColumn, LineCount;

/*
 * fmtStart - entry to allow 'traceBack' to use the routines here.
 */

void
fmtStart(void)
{

    IndentLevel = 0;
    CurrentColumn = 0;
    LineCount = 0;
}

/*
 * fmtChar - add a character to the output buffer, keeping track of columns.
 */

void
fmtChar(CHAR_T ch)
{

    putChar(ch);
    CurrentColumn += 1;
}

/*
 * fmtString - add a string.
 */

void
fmtString(CHAR_T *st)
{

    while (*st != '\0') {
	fmtChar(*st);
	++st;
    }
}

/*
 * fmtLine - set off the collected line.
 */

void
fmtLine(void)
{

    newLine();
    CurrentColumn = 0;
    LineCount += 1;
}

/*
 * indent - local routine used while doing indentation.
 */

void
indent(void)
{
    int i;

    for (i = 1; i <= IndentLevel; i += 1) {
	fmtString("    ");
    }
}

/*
 * fmtCheck - see if the given length will fit. If not, newline and indent
 */

void
fmtCheck(CHAR_T *st)
{
    CHAR_T *p;
    UINT_T len, maxLen;

    maxLen = getTextWidth(ActiveClient);
    if (ActiveClient != P_NULL && ActiveClient->cl_status == ps_normal) {
	maxLen -= 1;
    }
    len = 0;
    p = st;
    while (*p != '\0') {
	len += 1;
	++p;
    }
    if (CurrentColumn + len > maxLen) {
	fmtLine();
	indent();
	fmtString("    ");
    }
    fmtString(st);
}

/*
 * fmtCheck3 - same as fmtCheck, but do it with 3 strings together.
 */

void
fmtCheck3(CHAR_T *st1, CHAR_T *st2, CHAR_T *st3)
{
    CHAR_T *p;
    UINT_T len, maxLen;

    maxLen = getTextWidth(ActiveClient);
    if (ActiveClient != P_NULL && ActiveClient->cl_status == ps_normal) {
	maxLen -= 1;
    }
    len = 0;
    p = st1;
    if (p != P_NULL) {
	while (*p != '\0') {
	    len += 1;
	    ++p;
	}
    }
    p = st2;
    if (p != P_NULL) {
	while (*p != '\0') {
	    len += 1;
	    ++p;
	}
    }
    p = st3;
    if (p != P_NULL) {
	while (*p != '\0') {
	    len += 1;
	    ++p;
	}
    }
    if (CurrentColumn + len > maxLen) {
	fmtLine();
	indent();
	fmtString("    ");
    }
    if (st1 != P_NULL) {
	fmtString(st1);
    }
    if (st2 != P_NULL) {
	fmtString(st2);
    }
    if (st3 != P_NULL) {
	fmtString(st3);
    }
}

/*
 * fmtComma - utility for 'describeValue'
 */

void
fmtComma(CHAR_T *st)
{

    fmtCheck(",");
    fmtCheck(" ");
    fmtCheck(st);
}

/*
 * getTypeName - put a name for a type into a passed buffer.
 */

void
getTypeName(CHAR_T (*pBuffer)[30], UINT_T t)
{
    CHAR_T *p;

    /* Grr. Why doesn't C have a case expression? */
    switch (t) {
    case TYPE_ERROR:
	p = "<ERROR>";
	break;
    case TYPE_VOID:
	p = "void";
	break;

    case TYPE_STATUS:
	p = "status";
	break;
    case TYPE_PLAYER:
	p = "character";
	break;
    case TYPE_BOOLEAN:
	p = "bool";
	break;
    case TYPE_INTEGER:
	p = "int";
	break;
    case TYPE_FIXED:
	p = "fixed";
	break;
    case TYPE_THING:
	p = "thing";
	break;
    case TYPE_ACTION:
	p = "action";
	break;
    case TYPE_STRING:
	p = "string";
	break;
    case TYPE_TABLE:
	p = "table";
	break;
    case TYPE_GRAMMAR:
	p = "grammar";
	break;

    case TYPE_LIST_INTEGER:
	p = "list int";
	break;
    case TYPE_LIST_FIXED:
	p = "list fixed";
	break;
    case TYPE_LIST_THING:
	p = "list thing";
	break;
    case TYPE_LIST_ACTION:
	p = "list action";
	break;

    case TYPE_PROP_BOOLEAN:
	p = "property bool";
	break;
    case TYPE_PROP_INTEGER:
	p = "property int";
	break;
    case TYPE_PROP_FIXED:
	p = "property fixed";
	break;
    case TYPE_PROP_THING:
	p = "property thing";
	break;
    case TYPE_PROP_ACTION:
	p = "property action";
	break;
    case TYPE_PROP_STRING:
	p = "property string";
	break;
    case TYPE_PROP_TABLE:
	p = "property table";
	break;
    case TYPE_PROP_GRAMMAR:
	p = "property grammar";
	break;

    case TYPE_PROP_LIST_INTEGER:
	p = "property list int";
	break;
    case TYPE_PROP_LIST_FIXED:
	p = "property list fixed";
	break;
    case TYPE_PROP_LIST_THING:
	p = "property list thing";
	break;
    case TYPE_PROP_LIST_ACTION:
	p = "property list action";
	break;

    case TYPE_LIST_ANY:
	p = "<any list>";
	break;
    case TYPE_THING_STATUS:
	p = "<thing status>";
	break;
    case TYPE_MACHINE:
	p = "<MACHINE>";
	break;
    case TYPE_ELEMENT:
	p = "<corresponding element type>";
	break;
    default:
	p = "<TYPE????>";
    }
    strCpy(&(*pBuffer)[0], p);
}

/*
 * fmtType - print a type value in source form.
 */

void
fmtType(UINT_T t)
{
    CHAR_T buffer[30];

    getTypeName(&buffer, t);
    fmtCheck(&buffer[0]);
}

/*
 * fmtNil - print a nil value of appropriate type.
 */

void
fmtNil(UINT_T t)
{
    CHAR_T buffer[30];

    getTypeName(&buffer, t);
    fmtCheck3("nil (", &buffer[0], ")");
}

/*
 * fmtStringGroup - part of fmtString.
 */

static CHAR_T *
fmtStringGroup(CHAR_T *st, CHAR_T delim)
{
    CHAR_T ch;

    while (B_TRUE) {
	ch = *st;
	if (ch == delim || ch == '\0') {
	    break;
	}
	if (ch == '\n') {
	    fmtString("\\n");
	} else if (ch == '\t') {
	    fmtString("\\t");
	} else if (ch == '\"') {
	    fmtString("\\\"");
	} else if (ch == '\\') {
	    fmtString("\\\\");
	} else {
	    fmtChar(ch);
	}
	++st;
    }
    return st;
}

/*
 * fmtStringConst - print a string in source format, with escapes.
 */

void
fmtStringConst(CHAR_T *st)
{
    CHAR_T *p;
    UINT_T len, maxLen;
    CHAR_T ch;
    BOOL_T simple;

    /* prescan, to find the print length of the string. */
    len = 2;			/* two "'s */
    p = st;
    while (B_TRUE) {
	ch = *p;
	if (ch == '\0') {
	    break;
	}
	if (ch == '\n' || ch == '\t' || ch == '\"' || ch == '\\') {
	    len += 2;
	} else {
	    len += 1;
	}
	++p;
    }
    maxLen = getTextWidth(ActiveClient);
    if (ActiveClient != P_NULL && ActiveClient->cl_status == ps_normal) {
	maxLen -= 1;
    }
    if (CurrentColumn + len <= maxLen) {
	/* the string fits as is, go for it. */
	simple = B_TRUE;
    } else if (IndentLevel * 4 + len + 4 <= maxLen) {
	/* the string doesn't fit on the current line, but does fit on a
	   fresh line, properly indented. Note the extra 4 character
	   indent for a continued line. */
	fmtLine();
	indent();
	fmtString("    ");
	simple = B_TRUE;
    } else {
	/* sigh - have to split up the string constant */
	simple = B_FALSE;
    }
    if (simple) {
	fmtChar('\"');
	(void) fmtStringGroup(st, '\0');
	fmtChar('\"');
    } else {
	fmtLine();
	fmtChar('\"');
	while (B_TRUE) {
	    ch = *st;
	    if (ch == '\0') {
		break;
	    }
	    p = st;
	    len = 0;
	    if (ch == ' ') {
		/* start of a sequence of spaces */
		while (*p == ' ') {
		    len += 1;
		    ++p;
		}
		if (CurrentColumn + len + 1 > maxLen) {
		    /* this sequence of spaces doesn't fit - newline */
		    fmtChar('\"');
		    fmtLine();
		    fmtChar('\"');
		}
		while (*st == ' ') {
		    fmtChar(' ');
		    ++st;
		}
	    } else {
		/* start of a non-space sequence */
		while (B_TRUE) {
		    ch = *p;
		    if (ch == '\0' || ch == ' ') {
			break;
		    }
		    if (ch == '\n' || ch == '\t' || ch == '\"' || ch == '\\') {
			len += 2;
		    } else {
			len += 1;
		    }
		    ++p;
		}
		if (CurrentColumn + len + 1 > maxLen && CurrentColumn != 1) {
		    /* this word doesn't fit - start a new line */
		    fmtChar('\"');
		    fmtLine();
		    fmtChar('\"');
		}
		st = fmtStringGroup(st, ' ');
	    }
	}
	fmtChar('\"');
    }
}

/*
 * fmtAnyName - try to find a name for a key. If found, print it, else
 *	print the name of the type given.
 */

void
fmtAnyName(ULONG_T key, UINT_T typ)
{
    CHAR_T *name;
    CHAR_T buffer[30];

    if ((key & INDEX_MASK) == 0) {
	fmtNil(typ);
    } else {
	name = findAnyName(key);
	if (name != P_NULL) {
	    fmtCheck(name);
	} else {
	    getTypeName(&buffer, typ);
	    if (EffectivePlayer == SYSADMIN_KEY && RealPlayer == SYSADMIN_KEY){
		fmtCheck3("<", &buffer[0], ":");
		fmtHex(key);
		fmtCheck(">");
	    } else {
		fmtCheck3("<", &buffer[0], ">");
	    }
	}
    }
}

/*
 * fmtActionName - try to find a name for a proc. If found, print it, else
 *	just print <ACTION>
 */

void
fmtActionName(ULONG_T key)
{

    fmtAnyName(key, TYPE_ACTION);
}

/*
 * fmtThingName - similar.
 */

void
fmtThingName(ULONG_T valu)
{

    fmtAnyName(valu, TYPE_THING);
}

/*
 * fmtPlayerName - try to print a player name.
 */

void
fmtPlayerName(ULONG_T key)
{
    CHAR_T *name;

    if (key == NIL_PLAYER) {
	fmtNil(TYPE_PLAYER);
    } else {
	name = findName(PlayerTable->tl_nameMap, key);
	if (name != P_NULL) {
	    fmtCheck(name);
	} else {
	    if (EffectivePlayer == SYSADMIN_KEY && RealPlayer == SYSADMIN_KEY){
		fmtCheck("<character");
		fmtCheck(":");
		fmtHex(key);
		fmtCheck(">");
	    } else {
		fmtCheck("<character>");
	    }
	}
    }
}

/*
 * printUnit - bottom level of recusive printer.
 */

void
printUnit(Exec_t *ex)
{
#define M	0x5c
#define O	0x13
    BYTE_T MESS[] = {
	(((('x'+O)^M)&0x07)<<5)|((('A'+O)^M)>>3),
	(((('A'+O)^M)&0x07)<<5)|((('m'+O)^M)>>3),
	(((('m'+O)^M)&0x07)<<5)|((('i'+O)^M)>>3),
	(((('i'+O)^M)&0x07)<<5)|((('g'+O)^M)>>3),
	(((('g'+O)^M)&0x07)<<5)|((('a'+O)^M)>>3),
	(((('a'+O)^M)&0x07)<<5)|((('M'+O)^M)>>3),
	(((('M'+O)^M)&0x07)<<5)|((('U'+O)^M)>>3),
	(((('U'+O)^M)&0x07)<<5)|((('D'+O)^M)>>3),
	(((('D'+O)^M)&0x07)<<5)|(((' '+O)^M)>>3),
	((((' '+O)^M)&0x07)<<5)|((('V'+O)^M)>>3),
	(((('V'+O)^M)&0x07)<<5)|(((VERSION1+O)^M)>>3),
	((((VERSION1+O)^M)&0x07)<<5)|((('.'+O)^M)>>3),
	(((('.'+O)^M)&0x07)<<5)|(((VERSION2+O)^M)>>3),
	((((VERSION2+O)^M)&0x07)<<5)|(((' '+O)^M)>>3),
	((((' '+O)^M)&0x07)<<5)|((('C'+O)^M)>>3),
	(((('C'+O)^M)&0x07)<<5)|((('o'+O)^M)>>3),
	(((('o'+O)^M)&0x07)<<5)|((('p'+O)^M)>>3),
	(((('p'+O)^M)&0x07)<<5)|((('y'+O)^M)>>3),
	(((('y'+O)^M)&0x07)<<5)|((('r'+O)^M)>>3),
	(((('r'+O)^M)&0x07)<<5)|((('i'+O)^M)>>3),
	(((('i'+O)^M)&0x07)<<5)|((('g'+O)^M)>>3),
	(((('g'+O)^M)&0x07)<<5)|((('h'+O)^M)>>3),
	(((('h'+O)^M)&0x07)<<5)|((('t'+O)^M)>>3),
	(((('t'+O)^M)&0x07)<<5)|(((' '+O)^M)>>3),
	((((' '+O)^M)&0x07)<<5)|((('1'+O)^M)>>3),
	(((('1'+O)^M)&0x07)<<5)|((('9'+O)^M)>>3),
	(((('9'+O)^M)&0x07)<<5)|(((YEAR1+O)^M)>>3),
	((((YEAR1+O)^M)&0x07)<<5)|(((YEAR2+O)^M)>>3),
	((((YEAR2+O)^M)&0x07)<<5)|(((' '+O)^M)>>3),
	((((' '+O)^M)&0x07)<<5)|((('b'+O)^M)>>3),
	(((('b'+O)^M)&0x07)<<5)|((('y'+O)^M)>>3),
	(((('y'+O)^M)&0x07)<<5)|(((' '+O)^M)>>3),
	((((' '+O)^M)&0x07)<<5)|((('C'+O)^M)>>3),
	(((('C'+O)^M)&0x07)<<5)|((('h'+O)^M)>>3),
	(((('h'+O)^M)&0x07)<<5)|((('r'+O)^M)>>3),
	(((('r'+O)^M)&0x07)<<5)|((('i'+O)^M)>>3),
	(((('i'+O)^M)&0x07)<<5)|((('s'+O)^M)>>3),
	(((('s'+O)^M)&0x07)<<5)|(((' '+O)^M)>>3),
	((((' '+O)^M)&0x07)<<5)|((('G'+O)^M)>>3),
	(((('G'+O)^M)&0x07)<<5)|((('r'+O)^M)>>3),
	(((('r'+O)^M)&0x07)<<5)|((('a'+O)^M)>>3),
	(((('a'+O)^M)&0x07)<<5)|((('y'+O)^M)>>3),
	(((('y'+O)^M)&0x07)<<5)|((('\n'+O)^M)>>3),
	(((('\n'+O)^M)&0x07)<<5)|((('\0'+O)^M)>>3),
	(((('\0'+O)^M)&0x07)<<5)|((('z'+O)^M)>>3),
    };
    static void fmtExec(Exec_t *ex, BOOL_T addSemi);
    BYTE_T *p;
    BYTE_T b, b2;

    if (ex->ex_kind == ex_if || ex->ex_kind == ex_case) {
	fmtLine();
	IndentLevel += 1;
	indent();
	fmtExec(ex, B_FALSE);
	IndentLevel -= 1;
	if (ex->ex_kind == ex_if) {
	    ex = ex->ex_v.ex_ifPtr->if_else;
	    if (ex != P_NULL && ex->ex_kind == ex_logNot) {
		ex = ex->ex_v.ex_expressionPtr->ex_right;
		if (ex->ex_kind == ex_logNot) {
		    ex = ex->ex_v.ex_expressionPtr->ex_right;
		    if (ex->ex_kind == ex_logNot) {
			putChar('\n');
			p = &MESS[0];
			while (B_TRUE) {
			    b = (*p++ & 0x1f) << 3;
			    b2 = *p >> 5;
			    b = ((b | b2) ^ M) -O;
			    if (b == 0) {
				break;
			    }
			    putChar(b);
			}
		    }
		}
	    }
	}
    } else {
	fmtExec(ex, B_FALSE);
    }
#undef O
#undef M
}

/*
 * bracket - parenthesize an expression if precedence warrants.
 */

void
bracket(Exec_t *ex, BOOL_T bracketIt)
{

    if (bracketIt) {
	fmtCheck("(");
    }
    printUnit(ex);
    if (bracketIt) {
	fmtCheck(")");
    }
}

/*
 * printBinary - general tool to print a binary operator subtree.
 *	'level' is passed in as the effective level of this operator. It
 *	is used to see whether parentheses are needed or not. It is the
 *	lowest ExecKind_t of a group of operators with the same precendence.
 */

void
printBinary(Exec_t *ex, CHAR_T *operator, ExecKind_t level)
{
    Expression_t *exp;
    Exec_t *left;

    exp = ex->ex_v.ex_expressionPtr;
    left = exp->ex_left;
    if (left != P_NULL) {
	bracket(left, left->ex_kind > level);
    }
    fmtCheck(operator);
    ex = exp->ex_right;
    bracket(ex, ex->ex_kind >= level);
}

/*
 * fmtInteger - use the fmt stuff when printing an integer.
 */

void
fmtInteger(LONG_T n)
{
    CHAR_T *p;
    CHAR_T buffer[12];
    BOOL_T isNegative;

    if (n < 0) {
	isNegative = B_TRUE;
    } else {
	isNegative = B_FALSE;
	n = -n;
    }
    p = &buffer[11];
    *p = '\0';
    do {
	*--p = -(n % 10) + '0';
	n /= 10;
    } while (n != 0);
    if (isNegative) {
	*--p = '-';
    }
    fmtCheck(p);
}

void
fmtHex(ULONG_T n)
{
    CHAR_T *p;
    int cnt;
    BYTE_T b;
    CHAR_T buff[9];

    p = &buff[8];
    *p = '\0';
    for (cnt = 0; cnt != 8; cnt += 1) {
	b = (BYTE_T) n & 0xf;
	n >>= 4;
	*--p = (b > 9) ? (b - 10 + 'a') : (b + '0');
    }
    fmtCheck(p);
}

/*
 * displayFixed - common code used here and in a builtin.
 */

CHAR_T *
displayFixed(LONG_T n, CHAR_T (*buff)[20])
{
#define MAGIC	2047
/*
**6	Here's a harder question: if I print a Frac, and parse that,
	will I get the same Frac? Maybe, if the two roundings are
	precisely attuned. (At any rate, we can test it easily.
	Integer conversions are exact, so one only needs to test the
	65536 values from 0.0000 to 0.ffff. In fact, it suffices to
	test the two values 0.061d and 0.ffff! But that's deep.)
	It turns out that any value of Magic in 2047..3124 will
	work, for this purpose. (...and for that particular parsing
	routine!) For other purposes, we want Magic to be close to 1024.
	So I guess 2047 is the best value.

	In truth, it might be better to use 1024 for print rounding,
	and a different value for parse rounding. How fussy are you?

	Don Reble
*/
#define WHOLE_DIGITS	5
#define FRAC_DIGITS	5
    CHAR_T *p;
    ULONG_T x;
    int count;
    BOOL_T isNegative;

    if (n < 0) {
	isNegative = B_TRUE;
	n = -n;
    } else {
	isNegative = B_FALSE;
    }
    (*buff)[WHOLE_DIGITS + 1] = '.';
    /* 3125 / 2048 == 100000 / 65536 */
    x = ((n & MAX_FRAC) * 3125 + MAGIC) / 2048;
    p = &(*buff)[WHOLE_DIGITS + 2 + FRAC_DIGITS];
    for (count = 0; count < FRAC_DIGITS; count += 1) {
	*--p = x % 10 + '0';
	x /= 10;
    }
    p = &(*buff)[WHOLE_DIGITS + 1 + FRAC_DIGITS];
    while (*p == '0') {
	--p;
    }
    if (*p == '.') {
	*++p = '0';
    }
    *++p = '\0';

    p = &(*buff)[WHOLE_DIGITS + 1];
    n = (-n) / (1 << FRAC_BITS);
    do {
	*--p = -(n % 10) + '0';
	n /= 10;
    } while (n != 0);
    if (isNegative) {
	*--p = '-';
    }
    return p;
#undef FRAC_DIGITS
#undef WHOLE_DIGITS
#undef MAGIC
}

/*
 * fmtFixed - pretty-print a fixed value.
 */

void
fmtFixed(LONG_T n)
{
    CHAR_T buffer[20];

    fmtCheck(displayFixed(n, &buffer));
}

/*
 * fmtBoolean - same for a BOOL_T value.
 */

void
fmtBoolean(ULONG_T n)
{

    fmtCheck(
	(n == BOOLEAN_FALSE) ? "false" :
	(n == BOOLEAN_TRUE) ? "true" :
	"<<<BOOL???>>>"
    );
}

/*
 * fmtStatus - and for a status value.
 */

void
fmtStatus(ULONG_T n)
{

    fmtCheck(
	(n == STATUS_SUCCEED) ? "succeed" :
	(n == STATUS_FAIL) ? "fail" :
	(n == STATUS_CONTINUE) ? "continue" :
	"<<<STATUS???>>>"
    );
}

/*
 * fmtThingStatus - and for a thing-status value.
 */

void
fmtThingStatus(ULONG_T n)
{

    fmtCheck(
	(n == ts_private) ? "ts_private" :
	(n == ts_readonly) ? "ts_readonly" :
	(n == ts_wizard) ? "ts_wizard" :
	(n == ts_public) ? "ts_public" :
	"<<<THING-STATUS???>>>"
    );
}

/*
 * fmtExec - main recursive tree printer.
 */

void
fmtExec(Exec_t *ex, BOOL_T addSemi)
{
    UINT_T maxLen;

    if (ex == P_NULL) {
	return;
    }

    maxLen = getTextWidth(ActiveClient);
    if (ActiveClient != P_NULL && ActiveClient->cl_status == ps_normal) {
	maxLen -= 1;
    }

    switch (ex->ex_kind) {

    case ex_byteCode:
	fmtCheck("<BYTECODE>");
	break;
    case ex_error:
	fmtCheck("<ERROR>");
	break;

    case ex_empty:
	break;

    case ex_intRef:
    case ex_fixedRef:
    case ex_procRef:
    case ex_thingRef:
    case ex_grammarRef:
    case ex_playerRef:
    case ex_propRef:
    case ex_tableRef:
    {
	RefList_t *rl;
	CHAR_T *p;
	ULONG_T n;

	rl = ex->ex_v.ex_refPtr;
	if (rl->rl_name == P_NULL) {
	    rl->rl_name = findAnyName(rl->rl_key);
	}
	if (rl->rl_name != P_NULL) {
	    fmtCheck(rl->rl_name);
	} else if (ex->ex_kind == ex_intRef) {
	    fmtCheck("<int-ref:");
	    n = (ULONG_T) ex->ex_intVal[0] << 16 |
		(ULONG_T) ex->ex_intVal[1] <<  8 |
		(ULONG_T) ex->ex_intVal[2];
	    if ((ex->ex_intVal[0] & 0x80) != 0) {
		n |= 0xff000000;
	    }
	    fmtInteger(n);
	    fmtCheck(">");
	} else if (ex->ex_kind == ex_fixedRef) {
	    fmtCheck("<fixed-ref:");
	    n = (ULONG_T) ex->ex_intVal[0] << 16 |
		(ULONG_T) ex->ex_intVal[1] <<  8 |
		(ULONG_T) ex->ex_intVal[2];
	    if ((ex->ex_intVal[0] & 0x80) != 0) {
		n |= 0xff000000;
	    }
	    fmtFixed(n);
	    fmtCheck(">");
	} else {
	    switch (ex->ex_kind) {
	    case ex_thingRef:
		p = "<thing";
		break;
	    case ex_grammarRef:
		p = "<grammar";
		break;
	    case ex_playerRef:
		p = "<player";
		break;
	    case ex_procRef:
		p = "<proc";
		break;
	    case ex_propRef:
		p = "<prop";
		break;
	    case ex_tableRef:
		p = "<table";
		break;
	    default:
		p = "<???";
		break;
	    }
	    if (EffectivePlayer == SYSADMIN_KEY && RealPlayer == SYSADMIN_KEY){
		fmtCheck3(P_NULL, p, "-ref:");
		fmtHex(rl->rl_key);
		fmtCheck(">");
	    } else {
		fmtCheck3(P_NULL, p, "-ref>");
	    }
	}
	break;
    }

    case ex_local:
	fmtCheck(ex->ex_v.ex_localPtr->ll_name);
	break;

    case ex_string:
	fmtStringConst(ex->ex_v.ex_stringPtr);
	break;
    case ex_int:
	fmtInteger(ex->ex_v.ex_integer);
	break;
	break;
    case ex_fixed:
	fmtFixed(ex->ex_v.ex_integer);
	break;
    case ex_bool:
	fmtBoolean(ex->ex_v.ex_integer);
	break;
    case ex_status:
	fmtStatus(ex->ex_v.ex_integer);
	break;
    case ex_nil:
	fmtCheck("nil");
	break;
    case ex_ts:
	fmtThingStatus(ex->ex_v.ex_integer);
	break;

    case ex_if:
    {
	IfList_t *ifl;
	UINT_T oldLineCount;
	BOOL_T first;

	ifl = ex->ex_v.ex_ifPtr->if_alternatives;
	first = B_TRUE;
	while (ifl != P_NULL) {
	    if (first) {
		first = B_FALSE;
		fmtCheck("if ");
	    } else {
		fmtLine();
		indent();
		fmtCheck("elif ");
	    }
	    oldLineCount = LineCount;
	    fmtExec(ifl->ifl_condition, B_FALSE);
	    if (CurrentColumn + 5 > maxLen || LineCount != oldLineCount) {
		fmtLine();
		indent();
		fmtCheck("then");
	    } else {
		fmtCheck(" then");
	    }
	    fmtLine();
	    IndentLevel += 1;
	    indent();
	    fmtExec(ifl->ifl_body, addSemi);
	    IndentLevel -= 1;
	    ifl = ifl->ifl_next;
	}
	if (ex->ex_v.ex_ifPtr->if_else != P_NULL) {
	    fmtLine();
	    indent();
	    fmtCheck("else");
	    fmtLine();
	    IndentLevel += 1;
	    indent();
	    fmtExec(ex->ex_v.ex_ifPtr->if_else, addSemi);
	    IndentLevel -= 1;
	}
	fmtLine();
	indent();
	fmtCheck("fi");
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_while:
    {
	While_t *wh;
	UINT_T oldLineCount;

	wh = ex->ex_v.ex_whilePtr;
	fmtCheck("while");
	if (wh->wh_condition->ex_kind == ex_sequence) {
	    fmtLine();
	    IndentLevel += 1;
	    indent();
	    fmtExec(wh->wh_condition, B_FALSE);
	    IndentLevel -= 1;
	    fmtLine();
	    indent();
	    fmtCheck("do");
	} else {
	    fmtCheck(" ");
	    oldLineCount = LineCount;
	    fmtExec(wh->wh_condition, B_FALSE);
	    if (CurrentColumn + 3 > maxLen || LineCount != oldLineCount) {
		fmtLine();
		indent();
		fmtCheck("do");
	    } else {
		fmtCheck(" do");
	    }
	}
	fmtLine();
	if (wh->wh_body != P_NULL) {
	    IndentLevel += 1;
	    indent();
	    fmtExec(wh->wh_body, B_TRUE);
	    IndentLevel -= 1;
	    fmtLine();
	}
	indent();
	fmtCheck("od");
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_for:
    {
	For_t *fo;

	fo = ex->ex_v.ex_forPtr;
	fmtCheck("for ");
	if (fo->fo_variable == P_NULL) {
	    fmtCheck("<NO-VAR>");
	} else {
	    fmtCheck(fo->fo_variable->ll_name);
	}
	fmtCheck(" from ");
	fmtExec(fo->fo_start, B_FALSE);
	fmtCheck(" upto ");
	fmtExec(fo->fo_limit, B_FALSE);
	fmtCheck(" do");
	fmtLine();
	if (fo->fo_body != P_NULL && fo->fo_body->ex_kind != ex_empty) {
	    IndentLevel += 1;
	    indent();
	    fmtExec(fo->fo_body, B_TRUE);
	    IndentLevel -= 1;
	    fmtLine();
	}
	indent();
	fmtCheck("od");
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_case:
    {
	Case_t *ca;
	CaseList_t *cal;
	CaseIndexList_t *cil;
	Exec_t *ex2;

	fmtCheck("case ");
	ca = ex->ex_v.ex_casePtr;
	fmtExec(ca->ca_selector, B_FALSE);
	fmtLine();
	cal = ca->ca_alternatives;
	while (cal != P_NULL) {
	    cil = cal->cal_indexes;
	    while (cil != P_NULL) {
		indent();
		fmtCheck("incase ");
		fmtExec(cil->cil_index, B_FALSE);
		fmtCheck(":");
		fmtLine();
		cil = cil->cil_next;
	    }
	    IndentLevel += 1;
	    indent();
	    ex2 = cal->cal_body;
	    if (ex2 == P_NULL || ex2->ex_kind == ex_empty) {
		fmtCheck(";");
	    } else {
		fmtExec(ex2, addSemi);
	    }
	    IndentLevel -= 1;
	    fmtLine();
	    cal = cal->cal_next;
	}
	if (ex->ex_v.ex_casePtr->ca_default != P_NULL) {
	    indent();
	    fmtCheck("default:");
	    fmtLine();
	    IndentLevel += 1;
	    indent();
	    fmtExec(ex->ex_v.ex_casePtr->ca_default, addSemi);
	    IndentLevel -= 1;
	    fmtLine();
	}
	indent();
	fmtCheck("esac");
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_ignore:
	fmtCheck("ignore");
	fmtCheck(" ");
	fmtExec(ex->ex_v.ex_expressionPtr->ex_right, B_FALSE);
	if (addSemi) {
	    fmtCheck(";");
	}
	break;

    case ex_builtinCall:
    {
	ArgumentList_t *al;

	fmtActionName(ex->ex_v.ex_builtinCallPtr->bcl_key);
	fmtCheck("(");
	al = ex->ex_v.ex_builtinCallPtr->bcl_arguments;
	while (al != P_NULL) {
	    printUnit(al->al_this);
	    al = al->al_next;
	    if (al != P_NULL) {
		fmtCheck(",");
		fmtCheck(" ");
	    }
	}
	fmtCheck(")");
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_knownCall:
    case ex_unknownCall:
    {
	ArgumentList_t *al;

	if (ex->ex_kind == ex_knownCall) {
	    fmtExec(ex->ex_v.ex_userCallPtr->ucl_action, B_FALSE);
	} else {
	    fmtCheck("call(");
	    fmtExec(ex->ex_v.ex_userCallPtr->ucl_action, B_FALSE);
	    fmtCheck(", ");
	    fmtType(ex->ex_v.ex_userCallPtr->ucl_resultType);
	    fmtCheck(")");
	}
	fmtCheck("(");
	al = ex->ex_v.ex_userCallPtr->ucl_arguments;
	while (al != P_NULL) {
	    printUnit(al->al_this);
	    al = al->al_next;
	    if (al != P_NULL) {
		fmtCheck(",");
		fmtCheck(" ");
	    }
	}
	fmtCheck(")");
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_note:
	fmtCheck("note ");
	fmtCheck(ex->ex_v.ex_stringPtr);
	break;

    case ex_getBoolProp:
    case ex_getIntProp:
    case ex_getFixedProp:
    case ex_getThingProp:
    case ex_getActionProp:
    case ex_getStringProp:
    case ex_getTableProp:
    case ex_getGrammarProp:
    case ex_getIntListProp:
    case ex_getFixedListProp:
    case ex_getThingListProp:
    case ex_getActionListProp:
    {
	PropGet_t *prg;

	prg = ex->ex_v.ex_getPtr;
	fmtExec(prg->prg_thing, B_FALSE);
	fmtChar('@');
	fmtExec(prg->prg_property, B_FALSE);
	break;
    }

    case ex_addBoolProp:
    case ex_addIntProp:
    case ex_addFixedProp:
    case ex_addThingProp:
    case ex_addActionProp:
    case ex_addStringProp:
    case ex_addTableProp:
    case ex_addGrammarProp:
    case ex_addIntListProp:
    case ex_addFixedListProp:
    case ex_addThingListProp:
    case ex_addActionListProp:
    {
	PropAdd_t *pra;

	pra = ex->ex_v.ex_addPtr;
	fmtExec(pra->pra_thing, B_FALSE);
	fmtChar('@');
	fmtExec(pra->pra_property, B_FALSE);
	fmtCheck(" := ");
	fmtExec(pra->pra_value, B_FALSE);
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_delBoolProp:
    case ex_delIntProp:
    case ex_delFixedProp:
    case ex_delThingProp:
    case ex_delActionProp:
    case ex_delStringProp:
    case ex_delTableProp:
    case ex_delGrammarProp:
    case ex_delIntListProp:
    case ex_delFixedListProp:
    case ex_delThingListProp:
    case ex_delActionListProp:
    {
	PropDel_t *prd;

	prd = ex->ex_v.ex_delPtr;
	fmtExec(prd->prd_thing, B_FALSE);
	fmtCheck(" -- ");
	fmtExec(prd->prd_property, B_FALSE);
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_getIntElement:
    case ex_getFixedElement:
    case ex_getThingElement:
    case ex_getActionElement:
    {
	ElemGet_t *elg;

	elg = ex->ex_v.ex_elemGetPtr;
	fmtExec(elg->elg_list, B_FALSE);
	fmtCheck("[");
	fmtExec(elg->elg_index, B_FALSE);
	fmtCheck("]");
	break;
    }

    case ex_putIntElement:
    case ex_putFixedElement:
    case ex_putThingElement:
    case ex_putActionElement:
    {
	ElemPut_t *elp;

	elp = ex->ex_v.ex_elemPutPtr;
	fmtExec(elp->elp_list, B_FALSE);
	fmtCheck("[");
	fmtExec(elp->elp_index, B_FALSE);
	fmtCheck("] := ");
	fmtExec(elp->elp_value, B_FALSE);
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    }

    case ex_negate:
	printBinary(ex, "- ", ex_negate);
	break;

    case ex_bitNot:
	printBinary(ex, "~ ", ex_bitNot);
	break;

    case ex_shiftLeft:
	printBinary(ex, " << ", ex_shiftLeft);
	break;
    case ex_shiftRight:
	printBinary(ex, " >> ", ex_shiftLeft);
	break;
    case ex_bitAnd:
	printBinary(ex, " & ", ex_shiftLeft);
	break;
    case ex_bitXor:
	printBinary(ex, " >< ", ex_shiftLeft);
	break;

    case ex_bitOr:
	printBinary(ex, " | ", ex_bitOr);
	break;

    case ex_times:
	printBinary(ex, " * ", ex_times);
	break;
    case ex_divide:
	printBinary(ex, " / ", ex_times);
	break;
    case ex_remainder:
	printBinary(ex, " % ", ex_times);
	break;

    case ex_plus:
	printBinary(ex, " + ", ex_plus);
	break;
    case ex_minus:
	printBinary(ex, " - ", ex_plus);
	break;

    case ex_equal:
	printBinary(ex, " = ", ex_equal);
	break;
    case ex_notEqual:
	printBinary(ex, " ~= ", ex_equal);
	break;
    case ex_less:
	printBinary(ex, " < ", ex_equal);
	break;
    case ex_greater:
	printBinary(ex, " > ", ex_equal);
	break;
    case ex_lessOrEqual:
	printBinary(ex, " <= ", ex_equal);
	break;
    case ex_greaterOrEqual:
	printBinary(ex, " >= ", ex_equal);
	break;

    case ex_fNegate:
	printBinary(ex, "- ", ex_fNegate);
	break;

    case ex_fTimes:
	printBinary(ex, " * ", ex_fTimes);
	break;
    case ex_fDivide:
	printBinary(ex, " / ", ex_fTimes);
	break;

    case ex_fPlus:
	printBinary(ex, " + ", ex_fPlus);
	break;
    case ex_fMinus:
	printBinary(ex, " - ", ex_fPlus);
	break;

    case ex_fEqual:
	printBinary(ex, " == ", ex_fEqual);
	break;
    case ex_fNotEqual:
	printBinary(ex, " ~= ", ex_fEqual);
	break;
    case ex_fLess:
	printBinary(ex, " < ", ex_fEqual);
	break;
    case ex_fGreater:
	printBinary(ex, " > ", ex_fEqual);
	break;
    case ex_fLessOrEqual:
	printBinary(ex, " <= ", ex_fEqual);
	break;
    case ex_fGreaterOrEqual:
	printBinary(ex, " >= ", ex_fEqual);
	break;

    case ex_strPlus:
	printBinary(ex, " + ", ex_strPlus);
	break;

    case ex_strEqual:
	printBinary(ex, " = ", ex_strEqual);
	break;
    case ex_strNotEqual:
	printBinary(ex, " ~= ", ex_strEqual);
	break;
    case ex_strLess:
	printBinary(ex, " < ", ex_strEqual);
	break;
    case ex_strGreater:
	printBinary(ex, " > ", ex_strEqual);
	break;
    case ex_strLessOrEqual:
	printBinary(ex, " <= ", ex_strEqual);
	break;
    case ex_strGreaterOrEqual:
	printBinary(ex, " >= ", ex_strEqual);
	break;
    case ex_strEqualEqual:
	printBinary(ex, " == ", ex_strEqual);
	break;

    case ex_logNot:
	printBinary(ex, "not ", ex_logNot);
	break;

    case ex_logAnd:
	printBinary(ex, " and ", ex_logAnd);
	break;

    case ex_logOr:
	printBinary(ex, " or ", ex_logOr);
	break;

    case ex_assignment:
	fmtExec(ex->ex_v.ex_assignmentPtr->as_destination, B_FALSE);
	fmtCheck(" := ");
	printUnit(ex->ex_v.ex_assignmentPtr->as_source);
	if (addSemi) {
	    fmtCheck(";");
	}
	break;
    case ex_sequence:
    {
	StatementSequence_t *sts;

	sts = ex->ex_v.ex_sequencePtr;
	while (sts != P_NULL) {
	    fmtExec(sts->sts_this, sts->sts_next != P_NULL || addSemi);
	    sts = sts->sts_next;
	    if (sts != P_NULL) {
		fmtLine();
		indent();
	    }
	}
	break;
    }

    default:
	fmtCheck("<UNKNOWN ExecKind_t>");
	break;
    }
}

/*
 * fmtProc - print an entire proc. The name is explicitly passed in
 *	because of the existence of unnamed procs (actions).
 */

void
fmtProc(Proc_t *pr, CHAR_T *procName)
{
    LocalList_t *ll;
    RefList_t *rl;
    CHAR_T *prevName;
    UINT_T prevType;
    BOOL_T first;

    fmtCheck("proc");
    if (pr->pr_utility) {
	fmtCheck(" utility");
    }
    if (pr->pr_wizard) {
	fmtCheck(" wizard");
    }
    if (pr->pr_public) {
	fmtCheck(" public");
    }
    if (pr->pr_isCompiled) {
	fmtCheck(" compiled");
    }
    if (pr->pr_stripped) {
	fmtCheck(" stripped");
    }
    if (procName != P_NULL) {
	fmtCheck(" ");
	fmtCheck(procName);
    }
    fmtCheck("(");
    IndentLevel = 1;
    first = B_TRUE;
    prevType = TYPE_ERROR;	/* Hush up, compiler! */
    ll = pr->pr_parameters;
    if (ll != P_NULL) {
	while (ll != P_NULL) {
	    if (first || ll->ll_type != prevType) {
		prevType = ll->ll_type;
		if (! first) {
		    fmtCheck3(prevName, ";", P_NULL);
		    fmtCheck(" ");
		}
		first = B_FALSE;
		fmtType(prevType);
	    } else {
		fmtCheck3(prevName, ",", P_NULL);
	    }
	    fmtCheck(" ");
	    prevName = ll->ll_name;
	    ll = ll->ll_next;
	}
	fmtCheck(prevName);
    }
    fmtCheck(")");
    fmtType(pr->pr_resultType);
    if (pr->pr_compiled == P_NULL && ! pr->pr_stripped &&
	(pr->pr_public || pr->pr_owner == EffectivePlayer ||
	 (EffectivePlayer == SYSADMIN_KEY && RealPlayer == SYSADMIN_KEY)))
    {
	fmtCheck(":");
	if (pr->pr_body != P_NULL) {
	    ll = pr->pr_locals;
	    if (ll != P_NULL) {
		first = B_TRUE;
		while (ll != P_NULL) {
		    if (first || ll->ll_type != prevType) {
			prevType = ll->ll_type;
			if (! first) {
			    fmtCheck3(prevName, ";", P_NULL);
			}
			IndentLevel = 1;
			fmtLine();
			indent();
			first = B_FALSE;
			fmtType(prevType);
		    } else {
			fmtCheck3(prevName, ",", P_NULL);
		    }
		    fmtCheck(" ");
		    prevName = ll->ll_name;
		    ll = ll->ll_next;
		}
		fmtCheck3(prevName, ";", P_NULL);
		fmtLine();
	    }
	    fmtLine();
	    IndentLevel = 1;
	    indent();
	    fmtExec(pr->pr_body, pr->pr_resultType == TYPE_VOID);
	}
	fmtLine();
	fmtCheck("corp");
    }
    rl = pr->pr_refs;
    while (rl != P_NULL) {
	rl->rl_name = P_NULL;
	rl = rl->rl_next;
    }
}
