#include <string.h>

#include "MUD.h"
#include "types.h"
#include "Request.h"
#include "MUDLib.h"
#include "MUDLibPrivate.h"
#include "MUDServ.h"
#include "ClientStuff.h"
#include "telnetFlags.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * telnet.c - support for telnet-style text-only clients.
 */

#define DEBUG			B_TRUE
#define DEBUG_TO_USER		B_TRUE
#define INITIAL_HELP		B_FALSE

#define INPUT_BUFFER_SIZE	1024
#define LINE_BUFFER_SIZE	128
#define OUTPUT_BUFFER_SIZE	1024
#define SB_BUFFER_SIZE		512
#define DEBUG_BUFFER_SIZE	1024

/* Telnet protocol commands. Commands must start with IAC, followed by
   the command byte. Other occurrences of command bytes (not after IAC)
   are passed through. Options consist of IAC, one of the five
   negotiation bytes (DONT, DO, WONT, WILL, SB), then the option code. */

#define TELNET_IAC	255	/* Interprete As Command */
#define TELNET_DONT	254	/* Don't do option */
#define TELNET_DO	253	/* Do option */
#define TELNET_WONT	252	/* Won't do option */
#define TELNET_WILL	251	/* Will do option */
#define TELNET_SB	250	/* Subnegotiation begin */
#define TELNET_GA	249	/* Go Ahead */
#define TELNET_EL	248	/* Erase Line */
#define TELNET_EC	247	/* Erase Character */
#define TELNET_AYT	246	/* Are You There? */
#define TELNET_AO	245	/* Abort Output */
#define TELNET_IP	244	/* Interrupt Process */
#define TELNET_BRK	243	/* BREAK */
#define TELNET_DM	242	/* Data Mark */
#define TELNET_NOP	241	/* No Operation */
#define TELNET_SE	240	/* End of Subnegotiation parameters */
#define TELNET_EOR	239	/* End Of Record */
#define TELNET_ABORT	238
#define TELNET_SUSP	237	/* Suspend */
#define TELNET_EOF	236	/* End Of File */

    /* Telnet option codes */

#define BINARY		0	/* transmit binary (full 8 bits) */
#define ECHO		1	/* enable echoing from remote end */
#define SUPGA		3	/* SUPress Go Ahead */
#define STATUS		5	/* query status of options on other end */
#define TIMING_MARK	6	/* for flushing of pre-typed input, etc. */
#define NAOCRD		10
#define NAOHTS		11
#define NAOHTD		12
#define NAOFFD		13
#define NAOVTS		14
#define NAOVTD		15
#define NAOLFD		16
#define XTNDASC		17
#define LOGOUT		18
#define BYTEMACRO	19	/* can support byte macros */
#define DET		20	/* Data Entry Terminal */
#define SUPDUP		21
#define SUPDUPOUT	22
#define SENDLOC		23
#define TERMINALTYPE	24	/* Can send terminal type. */
#define ENDOFRECORD	25
#define OUTMRK		27
#define TTYLOC		28
#define MODE3270	29
#define X3PAD		30
#define NAWS		31	/* Negotiate About Window Size */
#define TTYSPEED	32
#define FLOWCNTL	33
#define LINEMODE	34
#define XDISPLOC	35
#define ENVIRON		36	/* username, etc. */
#define AUTH		37
#define NEWENVIRON	39
#define MAX_OPTION	39	/* for handling with arrays */
#define EXOPL		255	/* Extended option negotiation */

#define CNTL_R		0x12
#define CNTL_U		0x15
#define CNTL_X		0x18

typedef struct TelnetClient {
    /* Stuff for telnet clients */
    Client_t *tncl_client;
    ULONG_T tncl_flags;
    ULONG_T tncl_outputPos;
    ULONG_T tncl_outputCurrent;
    UINT_T tncl_linePos;
    UINT_T tncl_SBPos;
#if DEBUG
    UINT_T tncl_debugPos;
    char tncl_debugBuffer[DEBUG_BUFFER_SIZE + 2];
    char tncl_staticBuffer[20];
    BOOL_T tncl_debugCmds;
    BOOL_T tncl_debugPrompt;
    BOOL_T tncl_debugInput;
#endif
    BOOL_T tncl_echoOn;
    BOOL_T tncl_supGA;
    BOOL_T tncl_doEOR;
    BYTE_T tncl_inputBuffer[INPUT_BUFFER_SIZE];
    BYTE_T tncl_SBBuffer[SB_BUFFER_SIZE];
    CHAR_T tncl_lineBuffer[LINE_BUFFER_SIZE + 1];
    BYTE_T tncl_outputBuffer[OUTPUT_BUFFER_SIZE];
    CHAR_T *tncl_prompt;		/* temporary prompt */
    CHAR_T *tncl_erase;			/* for erasing prompt */
    CHAR_T *tncl_terminalType;
    UINT_T tncl_promptLen;
    BOOL_T tncl_sentOptionDO[MAX_OPTION + 1];
    BOOL_T tncl_sentOptionWILL[MAX_OPTION + 1];
    BOOL_T tncl_sentOptionDONT[MAX_OPTION + 1];
    BOOL_T tncl_sentOptionWONT[MAX_OPTION + 1];
    BOOL_T tncl_promptShown;
    BOOL_T tncl_lastWasCR;
    BOOL_T tncl_meDoEcho;
    BOOL_T tncl_sendBinary;
    BOOL_T tncl_receiveBinary;
    enum {
	tntst_normal,
	tntst_command,
	tntst_SB
    } tncl_telnetState;
    BOOL_T tncl_lastWasIAC;
    BYTE_T tncl_command;
    BOOL_T tncl_writeWaiting;
    BOOL_T tncl_doneOutput;
    BOOL_T tncl_wantIdleAction;
    CHAR_T tncl_playerName[PLAYER_NAME_LEN];
    CHAR_T tncl_password[PASSWORD_LEN];	/* scratch buffer only */
    enum {
	tnst_name,		/* asking for player name */
	tnst_plPassword,	/* asking for player password */
	tnst_newPlayerNCP,	/* asking if want new player, no creation pw */
	tnst_newPlayerWCP,	/* asking if want new player, with "" pw */
	tnst_crPassword,	/* asking for creation password */
	tnst_npPassword1,	/* asked for new player password */
	tnst_npPassword2,	/* re-asked for verification */
#if 0
	tnst_parsing,		/* directly executing in wizard mode */
#endif
	tnst_normal,		/* normal - input goes to server */
	tnst_exiting,		/* waiting for a return to exit */
	tnst_changePassword1,	/* server wants us to enter old password */
	tnst_changePassword2,	/* server wants us to enter new password */
	tnst_changePassword3	/* verify new password */
    } tncl_state, tncl_saveState;
    UINT_T tncl_nameCount;	/* how many attempts at character name */
    UINT_T tncl_ynCount;	/* how many attempts at a Y/N */
} TelnetClient_t;

#if DEBUG

#define CMD_DEBUG(tncl, m) \
    if (tncl->tncl_debugCmds) { \
	debug(tncl, m); \
    }

#define CMD_DEBUG_CMD(tncl, m, cmd, byte)   debugCmd(tncl, m, cmd, byte)

#define CMD_DEBUG_SHORT_CMD(tncl, m, cmd)   debugShortCmd(tncl, m, cmd)

#define DEBUG_PROMPT(tncl, m)		    debugPrompt(tncl, m)

#else

#define CMD_DEBUG(tncl, m)
#define CMD_DEBUG_CMD(tncl, m, cmd, byte)
#define CMD_DEBUG_SHORT_CMD(tncl, m, cmd)
#define DEBUG_PROMPT(tncl, m)

#endif

#if DEBUG

static char *(OptionNames[]) = {
    "BINARY", "ECHO", "2", "SUPGA", "4", "STATUS", "TIMING_MARK",
    "7", "8", "9", "NAOCRD", "NAOHTS", "NAOHTD", "NAOFFD", "NAOVTS",
    "NAOVTD", "NAOLFD", "XTNDASC", "LOGOUT", "BYTEMACRO", "DET",
    "SUPDUP", "SUPDUPOUT", "SENDLOC", "TERMINALTYPE", "ENDOFRECORD",
    "26", "OUTMRK", "TTYLOC", "MODE3270", "X3PAD", "NAWS", "TTYSPEED",
    "FLOWCNTL", "LINEMODE", "XDISPLOC", "ENVIRON", "AUTH", "38",
    "NEWENVIRON"
};

static char *(CmdNames[]) = {
    "EOF", "SUSP", "ABORT", "EOR", "SE", "NOP", "DM", "BRK", "IP", "AO",
    "AYT", "EC", "EL", "GA", "SB", "WILL", "WONT", "DO", "DONT", "IAC"
};

/*
 * debugEnd - end of debugging sequence.
 */

static void
debugEnd(TelnetClient_t *tncl)
{
    static void sendChars(TelnetClient_t *tncl, CHAR_T *p, UINT_T len);

    if (tncl->tncl_debugPos != 0) {
	tncl->tncl_debugBuffer[tncl->tncl_debugPos] = '\n';
#if DEBUG_TO_USER
	sendChars(tncl, &tncl->tncl_debugBuffer[0], tncl->tncl_debugPos + 1);
#else
	tncl->tncl_debugBuffer[tncl->tncl_debugPos + 1] = '\0';
	mess(&tncl->tncl_debugBuffer[0]);
#endif
    }
}

/*
 * debugChar - add a character to the debug buffer.
 */

static void
debugChar(TelnetClient_t *tncl, char ch)
{
    static void debugString(TelnetClient_t *tncl, char *m);

    if (tncl->tncl_debugPos == DEBUG_BUFFER_SIZE) {
	debugEnd(tncl);
	debugString(tncl, "**+ ");
    }
    tncl->tncl_debugBuffer[tncl->tncl_debugPos] = ch;
    tncl->tncl_debugPos += 1;
}

/*
 * debugString - add a string to the debug output.
 */

static void
debugString(TelnetClient_t *tncl, char *m)
{

    while (*m != '\0') {
	debugChar(tncl, *m);
	++m;
    }
}

/*
 * debugStart - first step in a debug sequence.
 */

static void
debugStart(TelnetClient_t *tncl)
{

    tncl->tncl_debugPos = 0;
    debugString(tncl, "** ");
}

/*
 * debugByte - add the decimal form of a byte to the debug output.
 */

static void
debugByte(TelnetClient_t *tncl, BYTE_T b)
{

    if (b >= 100) {
	debugChar(tncl, b / 100 + '0');
	b %= 100;
    }
    if (b >= 10) {
	debugChar(tncl, b / 10 + '0');
	b %= 10;
    }
    debugChar(tncl, b + '0');
}

/*
 * debug - used during debugging.
 */

static void
debug(TelnetClient_t *tncl, char *message)
{

    debugStart(tncl);
    debugString(tncl, message);
    debugEnd(tncl);
}

/*
 * debugWithBuffer - common for promptDebug and doInputLine.
 */

static void
debugWithBuffer(TelnetClient_t *tncl, char *m, char *buf)
{

    debugStart(tncl);
    debugString(tncl, m);
    debugString(tncl, ": ");
    debugString(tncl, buf);
    debugEnd(tncl);
}

/*
 * fmtByte - utility for optionName and cmdName.
 */

static char *
fmtByte(TelnetClient_t *tncl, BYTE_T b)
{
    char *p;

    p = &tncl->tncl_staticBuffer[0];
    if (b >= 100) {
	*p++ = b / 100 + '0';
	b %= 100;
    }
    if (b >= 10) {
	*p++ = b / 10 + '0';
	b %= 10;
    }
    *p++ = b % 10 + '0';
    *p = '\0';
    return &tncl->tncl_staticBuffer[0];
}

/*
 * optionName - return the name for an option.
 */

static char *
optionName(TelnetClient_t *tncl, BYTE_T option)
{

    if (option <= NEWENVIRON) {
	return OptionNames[option];
    } else if (option == EXOPL) {
	return "EXOPL";
    } else {
	return fmtByte(tncl, option);
    }
}

/*
 * cmdName - return the name for a short command.
 */

static char *
cmdName(TelnetClient_t *tncl, BYTE_T cmd)
{

    if (cmd >= TELNET_EOF) {
	return CmdNames[cmd - TELNET_EOF];
    } else {
	return fmtByte(tncl, cmd);
    }
}

/*
 * debugCmd - used during debugging.
 */

static void
debugCmd(TelnetClient_t *tncl, char *m, BYTE_T cmd, BYTE_T option)
{

    if (tncl->tncl_debugCmds) {
	debugStart(tncl);
	debugString(tncl, m);
	debugChar(tncl, ' ');
	switch (cmd) {
	case TELNET_DONT:
	    debugString(tncl, "DONT");
	    break;
	case TELNET_DO:
	    debugString(tncl, "DO");
	    break;
	case TELNET_WONT:
	    debugString(tncl, "WONT");
	    break;
	case TELNET_WILL:
	    debugString(tncl, "WILL");
	    break;
	default:
	    debugString(tncl, "???");
	    break;
	}
	debugChar(tncl, ' ');
	debugString(tncl, optionName(tncl, option));
	debugEnd(tncl);
    }
}

/*
 * debugShortCmd - ditto.
 */

static void
debugShortCmd(TelnetClient_t *tncl, char *m, BYTE_T cmd)
{

    if (tncl->tncl_debugCmds) {
	debugStart(tncl);
	debugString(tncl, m);
	debugChar(tncl, ' ');
	debugString(tncl, cmdName(tncl, cmd));
	debugEnd(tncl);
    }
}

/*
 * debugPrompt - used during debugging.
 */

static void
debugPrompt(TelnetClient_t *tncl, char *message)
{

    if (tncl->tncl_debugPrompt) {
	debugWithBuffer(tncl, message, tncl->tncl_prompt);
    }
}

#endif

/*
 * pushBuffer - try to make progress with sending from the client buffer
 *	out over the socket. Return B_TRUE if we made room in the buffer.
 */

static BOOL_T
pushBuffer(TelnetClient_t *tncl)
{
    UINT_T current;
    LONG_T wantLen, sentLen;

    current = tncl->tncl_outputCurrent;
    wantLen = tncl->tncl_outputPos - current;
    if (wantLen == 0) {
	return B_TRUE;
    }
    sentLen = doSend(tncl->tncl_client,
		     &tncl->tncl_outputBuffer[current], wantLen);
    if (sentLen > 0) {
	if (sentLen == wantLen) {
	    /* Buffer is now empty. */
	    tncl->tncl_outputCurrent = 0;
	    tncl->tncl_outputPos = 0;
	} else {
	    /* Made progress, but did not empty the buffer. */
	    tncl->tncl_outputCurrent = current + sentLen;
	}
	return B_TRUE;
    }
    return B_FALSE;
}

/*
 * putByte - bottom level interface for sending data to client. If we are
 *	stuck, because our buffer is full, then return B_FALSE, and the
 *	stall notification must propagate all the way back up.
 */

static BOOL_T
putByte(TelnetClient_t *tncl, BYTE_T b)
{
    ULONG_T pos, current;

    /* We allow a few bytes of slop, so that we can reasonably handle the
       cases where we have to escape a character or send commands */
    pos = tncl->tncl_outputPos;
    if (pos >= OUTPUT_BUFFER_SIZE - 5) {
	/* Reached end of output buffer - can we shift it down? */
	current = tncl->tncl_outputCurrent;
	if (current != 0) {
	    /* Yes, that makes space. */
	    memcpy(&tncl->tncl_outputBuffer[0],
		   &tncl->tncl_outputBuffer[current], pos - current);
	    pos -= current;
	    tncl->tncl_outputPos = pos;
	    tncl->tncl_outputCurrent = 0;
	} else {
	    /* No, have to try to flush to player. */
	    if (! tncl->tncl_writeWaiting) {
		if (! pushBuffer(tncl)) {
		    addToWriteWaiters(tncl->tncl_client);
		    tncl->tncl_writeWaiting = B_TRUE;
		}
	    }
	}
    }
    pos = tncl->tncl_outputPos;
    if (pos < OUTPUT_BUFFER_SIZE) {
	/* Lose any bytes we have no room for. This should only happen
	   in awkward places like prompt changes, when the socket is
	   filled up. I could fix this with a couple more variables
	   in the TelnetClient structure, marking where in the prompt
	   or erase I am, and handle those first in 'telnetWrite'. */
	tncl->tncl_outputBuffer[pos] = b;
	pos += 1;
	tncl->tncl_outputPos = pos;
	tncl->tncl_doneOutput = B_TRUE;
    }
    return (BOOL_T) ! tncl->tncl_writeWaiting;
}

/*
 * sendShortCmd - send one of the one-byte commands.
 */

static void
sendShortCmd(TelnetClient_t *tncl, BYTE_T cmd)
{

    (void) putByte(tncl, TELNET_IAC);
    (void) putByte(tncl, cmd);
    CMD_DEBUG_SHORT_CMD(tncl, "send", cmd);
}

/*
 * sendCmd - send a telnet command.
 */

static void
sendCmd(TelnetClient_t *tncl, BYTE_T cmd, BYTE_T option)
{

    (void) putByte(tncl, TELNET_IAC);
    (void) putByte(tncl, cmd);
    (void) putByte(tncl, option);
    CMD_DEBUG_CMD(tncl, "send", cmd, option);
    (void) pushBuffer(tncl);
}

/*
 * sendOptionDO - request that a given option be enabled.
 */

static void
sendOptionDO(TelnetClient_t *tncl, BYTE_T option)
{

    sendCmd(tncl, TELNET_DO, option);
    tncl->tncl_sentOptionDO[option] = B_TRUE;
}

/*
 * sendOptionWILL - offer to enable an option.
 */

static void
sendOptionWILL(TelnetClient_t *tncl, BYTE_T option)
{

    sendCmd(tncl, TELNET_WILL, option);
    tncl->tncl_sentOptionWILL[option] = B_TRUE;
}

/*
 * sendOptionDONT - tell the other guy to not do something.
 */

static void
sendOptionDONT(TelnetClient_t *tncl, BYTE_T option)
{

    sendCmd(tncl, TELNET_DONT, option);
    tncl->tncl_sentOptionDONT[option] = B_TRUE;
}

/*
 * sendOptionWONT - refuse to do something.
 */

static void
sendOptionWONT(TelnetClient_t *tncl, BYTE_T option)
{

    sendCmd(tncl, TELNET_WONT, option);
    tncl->tncl_sentOptionWONT[option] = B_TRUE;
}

#if DEBUG

/*
 * debugFlag - helper for doDebugCommand.
 */

static void
debugFlag(TelnetClient_t *tncl, char *name, BOOL_T value)
{

    debugStart(tncl);
    debugString(tncl, name);
    debugString(tncl, ": ");
    debugString(tncl, value ? "TRUE" : "FALSE");
    debugEnd(tncl);
}

/*
 * debugBit - another helper for doDebugCommand
 */

static void
debugBit(TelnetClient_t *tncl, char *name, ULONG_T bit)
{

    debugStart(tncl);
    debugString(tncl, name);
    debugString(tncl, ": ");
    debugString(tncl, (tncl->tncl_flags & bit) != 0 ? "TRUE" : "FALSE");
    debugEnd(tncl);
}

/*
 * debugToggleFlag - another helper for doDebugCommand
 */

static void
debugToggleFlag(TelnetClient_t *tncl, char *name, BOOL_T *pFlag)
{

    debugStart(tncl);
    debugString(tncl, name);
    debugString(tncl, " display turned ");
    if (*pFlag) {
	debugString(tncl, "off");
	*pFlag = B_FALSE;
    } else {
	debugString(tncl, "on");
	*pFlag = B_TRUE;
    }
    debugEnd(tncl);
}

/*
 * debugToggleBit - another helper for doDebugCommand
 */

static void
debugToggleBit(TelnetClient_t *tncl, char *name, ULONG_T bit)
{

    debugStart(tncl);
    debugString(tncl, name);
    debugString(tncl, " turned ");
    if ((tncl->tncl_flags & bit) != 0) {
	debugString(tncl, "off");
	tncl->tncl_flags &= ~bit;
    } else {
	debugString(tncl, "on");
	tncl->tncl_flags |= bit;
    }
    debugEnd(tncl);
}

/*
 * doDebugCommand - remote user has issued a debug command.
 */

static void
doDebugCommand(TelnetClient_t *tncl)
{
    char *p;

    p = &tncl->tncl_lineBuffer[2];
    tncl->tncl_lineBuffer[tncl->tncl_linePos] = '\0';
    if (strcmp(p, "?") == 0) {
	debug(tncl, "Telnet mode commands:\n\n"
	      "**? - show this info\n"
	      "**st - show current server telnet status\n"
	      "**cmd - toggle telnet option negotiation display\n"
	      "**prompt - toggle prompt activity display\n"
	      "**input - toggle input line display\n"
	      "**ga - toggle any sending of GA\n"
	      "**pwblank - toggle attempts to blank passwords\n"
	      "**pwsupga - toggle use of SUPGA in password blanking\n"
	      "**eraseprompt - toggle attempts to erase prompts\n"
	      "**binary - renegotiate binary\n"
	      "**echo - renegotiate echo\n"
	      "**eor - renegotiate ENDOFRECORD\n"
	      "**naws - force NAWS redo\n"
	      "**ttytype - force TERMINALTYPE redo\n"
	      "**fbinary - force binary on/off\n"
	      "**fecho - force server echo on/off");
    } else if (strcmp(p, "st") == 0) {
	debugFlag(tncl, "sendBinary", tncl->tncl_sendBinary);
	debugFlag(tncl, "receiveBinary", tncl->tncl_receiveBinary);
	debugFlag(tncl, "server echo", tncl->tncl_meDoEcho);
	debugFlag(tncl, "suppress goahead", tncl->tncl_supGA);
	debugFlag(tncl, "use end of record", tncl->tncl_doEOR);
	debugStart(tncl);
	debugString(tncl, "terminal type is '");
	debugString(tncl, tncl->tncl_terminalType);
	debugChar(tncl, '\'');
	debugEnd(tncl);
	debugStart(tncl);
	debugString(tncl, "display height ");
	debugByte(tncl, tncl->tncl_client->cl_textHeight);
	debugString(tncl, " width ");
	debugByte(tncl, tncl->tncl_client->cl_textWidth);
	debugEnd(tncl);
	debugBit(tncl, "ever send GA", B_SEND_GA);
	debugBit(tncl, "attempt password blanking", B_PW_BLANK);
	debugBit(tncl, "use SUPGA on password blanking", B_PW_BLANK_SUPGA);
	debugBit(tncl, "erase prompts", B_ERASE_PROMPT);
    } else if (strcmp(p, "cmd") == 0) {
	debugToggleFlag(tncl, "telnet option negotiation",
			&tncl->tncl_debugCmds);
    } else if (strcmp(p, "prompt") == 0) {
	debugToggleFlag(tncl, "prompt activity", &tncl->tncl_debugPrompt);
    } else if (strcmp(p, "input") == 0) {
	debugToggleFlag(tncl, "input line", &tncl->tncl_debugInput);
    } else if (strcmp(p, "ga") == 0) {
	debugToggleBit(tncl, "send of GA's", B_SEND_GA);
    } else if (strcmp(p, "pwblank") == 0) {
	debugToggleBit(tncl, "blanking of passwords", B_PW_BLANK);
    } else if (strcmp(p, "pwsupga") == 0) {
	debugToggleBit(tncl, "use of SUPGA in password blanking",
		       B_PW_BLANK_SUPGA);
    } else if (strcmp(p, "eraseprompt") == 0) {
	debugToggleBit(tncl, "attempts to erase prompt", B_ERASE_PROMPT);
    } else if (strcmp(p, "binary") == 0) {
	debug(tncl, "attempting to enable binary data");
	sendOptionWILL(tncl, BINARY);
	sendOptionDO(tncl, BINARY);
    } else if (strcmp(p, "echo") == 0) {
	if (tncl->tncl_meDoEcho) {
	    debug(tncl, "asking client to echo");
	    sendOptionWONT(tncl, ECHO);
	    if ((tncl->tncl_flags & B_PW_BLANK_SUPGA) != 0) {
		debug(tncl, "asking for no SUPGA");
		sendOptionWONT(tncl, SUPGA);
		sendOptionDONT(tncl, SUPGA);
	    }
	} else {
	    debug(tncl, "offering to echo");
	    sendOptionWILL(tncl, ECHO);
	    if ((tncl->tncl_flags & B_PW_BLANK_SUPGA) != 0) {
		debug(tncl, "asking for SUPGA");
		sendOptionWILL(tncl, SUPGA);
		sendOptionDO(tncl, SUPGA);
	    }
	}
    } else if (strcmp(p, "eor") == 0) {
	if (tncl->tncl_doEOR) {
	    debug(tncl, "trying to turn EOR off");
	    sendOptionWONT(tncl, ENDOFRECORD);
	    sendOptionDONT(tncl, ENDOFRECORD);
	} else {
	    debug(tncl, "trying to turn EOR on");
	    sendOptionWILL(tncl, ENDOFRECORD);
	    sendOptionDO(tncl, ENDOFRECORD);
	}
    } else if (strcmp(p, "naws") == 0) {
	debug(tncl, "re-requesting NAWS");
	sendOptionDO(tncl, NAWS);
    } else if (strcmp(p, "ttytype") == 0) {
	debug(tncl, "re-requesting TTYTYPE");
	sendOptionDO(tncl, TERMINALTYPE);
    } else if (strcmp(p, "fbinary") == 0) {
	if (tncl->tncl_sendBinary) {
	    debug(tncl, "forcing BINARY off");
	    tncl->tncl_sendBinary = B_FALSE;
	} else {
	    debug(tncl, "forcing BINARY on");
	    tncl->tncl_sendBinary = B_TRUE;
	}
	tncl->tncl_receiveBinary = tncl->tncl_sendBinary;
    } else if (strcmp(p, "fecho") == 0) {
	if (tncl->tncl_meDoEcho) {
	    debug(tncl, "server will no longer echo");
	    tncl->tncl_meDoEcho = B_FALSE;
	} else {
	    debug(tncl, "server will start echoing");
	    tncl->tncl_meDoEcho = B_TRUE;
	}
    } else {
	debug(tncl, "Invalid telnet mode command. **? for help");
    }
}

#endif

/*
 * sendChar - bottom level of putting characters to the client. This
 *	handles any translation needed on the character for the telnet
 *	protocol rules.
 */

static BOOL_T
sendChar(TelnetClient_t *tncl, CHAR_T ch)
{

    if (ch == TELNET_IAC) {
	/* Yes - single '&' - want both executed. */
	return (putByte(tncl, TELNET_IAC) & putByte(tncl, TELNET_IAC));
    }
    if ((ch & 0x80) != 0 && ! tncl->tncl_sendBinary) {
	ch &= 0x7f;
    }
    if (ch == '\n') {
	return (putByte(tncl, '\r') & putByte(tncl, '\n'));
    }
    return putByte(tncl, ch);
}

/*
 * sendChars - send a known-length array of characters. Ignore stalling.
 */

static void
sendChars(TelnetClient_t *tncl, CHAR_T *p, UINT_T len)
{

    while (len != 0) {
	(void) sendChar(tncl, *p);
	++p;
	len -= 1;
    }
}

/*
 * sendString - send a string.
 */

static void
sendString(TelnetClient_t *tncl, char *st)
{

    while (*st != '\0') {
	(void) sendChar(tncl, *st);
	++st;
    }
}

/*
 * complain - a complain from the telnet code to the user.
 */

static void
complain(TelnetClient_t *tncl, char *message)
{

    sendString(tncl, "** ");
    sendString(tncl, message);
    sendChar(tncl, '\n');
}

/*
 * showPrompt - show the current prompt.
 */

static void
showPrompt(TelnetClient_t *tncl, BOOL_T doGA)
{

    sendChars(tncl, tncl->tncl_prompt, tncl->tncl_promptLen);
    tncl->tncl_promptShown = B_TRUE;
    if (doGA) {
	if (tncl->tncl_doEOR) {
	    sendShortCmd(tncl, TELNET_EOR);
	} else if (! tncl->tncl_supGA && (tncl->tncl_flags & B_SEND_GA) != 0) {
	    sendShortCmd(tncl, TELNET_GA);
	}
    }
}

/*
 * unshowPrompt - remove the current prompt.
 */

static void
unshowPrompt(TelnetClient_t *tncl)
{

    sendChars(tncl, tncl->tncl_erase, tncl->tncl_promptLen + 4);
    tncl->tncl_promptShown = B_FALSE;
}

/*
 * processOutgoing - a server rt_message request. Process the text
 *	of the message through the telnet protocol, buffering it in
 *	the client buffer. Then, try sending the data. If we managed
 *	to send all of it, return B_TRUE, else return B_FALSE.
 */

static BOOL_T
processOutgoing(Client_t *cl, Request_t *rq, unsigned int startPos)
{
    TelnetClient_t *tncl;
    CHAR_T *p;
    unsigned int i;

    tncl = cl->cl_telnet;
    if (startPos == 0 && tncl->tncl_promptShown &&
	(tncl->tncl_flags & B_ERASE_PROMPT) != 0)
    {
	DEBUG_PROMPT(tncl, "processOutgoing unshowing");
	unshowPrompt(tncl);
    }
    p = &rq->rq_u.ru_text[startPos];
    for (i = startPos; i != rq->rq_usedLen; i += 1) {
	if (! sendChar(tncl, *p)) {
	    /* We have plugged up. Must pause here. */
	    cl->cl_outGoingPos = i;
	    return B_FALSE;
	}
	++p;
    }
    /* All buffered up. If the buffer is nearly full, try sending it
       right away. This is a heuristic to try to keep the user seeing
       full output messages, rather than partial ones. */
    if (tncl->tncl_outputPos >= OUTPUT_BUFFER_SIZE - 40) {
	return pushBuffer(tncl);
    }
    return B_TRUE;
}

/*
 * telnetWrite - called when we can write to this client socket.
 */

void
telnetWrite(Client_t *cl)
{
    TelnetClient_t *tncl;
    Request_t *rq, *rqTemp;
    BOOL_T newLineAtEnd;

    TRACE("telnetWrite", cl);
    tncl = cl->cl_telnet;

    /* Is there something in the output buffer? */
    if (tncl->tncl_outputCurrent != tncl->tncl_outputPos) {
	/* Yes, try to push it out. */
	if (! pushBuffer(tncl)) {
	    return;
	}
    }

    newLineAtEnd = B_FALSE;
    /* Is there a partially sent message sitting around? */
    rq = cl->cl_outGoing;
    if (rq != P_NULL && cl->cl_outGoingPos != 0) {
	/* Yes, try to process the remainder. */
	if (! processOutgoing(cl, rq, cl->cl_outGoingPos)) {
	    return;
	}
	rqTemp = rq;
	rq = rq->rq_next;
	if (rqTemp->rq_u.ru_text[rqTemp->rq_usedLen - 1] == '\n') {
	    newLineAtEnd = B_TRUE;
	}
	freeRequestE(&MS, rqTemp);
    }

    /* OK, try any other queued up messages. */
    while (rq != P_NULL && ! cl->cl_deleted) {
	if (! processOutgoing(cl, rq, 0)) {
	    break;
	}
	rqTemp = rq;
	rq = rq->rq_next;
	if (rqTemp->rq_u.ru_text[rqTemp->rq_usedLen - 1] == '\n') {
	    newLineAtEnd = B_TRUE;
	}
	freeRequestE(&MS, rqTemp);
    }
    cl->cl_outGoing = rq;
    if (rq == P_NULL) {
	if (newLineAtEnd) {
	    /* The end of the last outgoing text request was a newline. This
	       is a good place to show the prompt. */
	    DEBUG_PROMPT(tncl, "telnetWrite showing");
	    showPrompt(tncl, B_TRUE);
	}
	removeFromWriteWaiters(cl);
    }
}

/*
 * setPrompt - set in a new prompt.
 */

static void
setPrompt(TelnetClient_t *tncl, CHAR_T *prompt)
{
    char *p;
    UINT_T len;
    BOOL_T promptWasShown;

    promptWasShown = tncl->tncl_promptShown;
    if (promptWasShown) {
	if ((tncl->tncl_flags & B_ERASE_PROMPT) != 0) {
	    DEBUG_PROMPT(tncl, "setPrompt unshowing");
	    unshowPrompt(tncl);
	} else {
	    sendChar(tncl, '\n');
	}
    }
    memFreeE(&MS, tncl->tncl_erase, tncl->tncl_promptLen + 4);
    memFreeE(&MS, tncl->tncl_prompt, tncl->tncl_promptLen + 1);
    len = strlen(prompt);
    p = MEM_ALLOC_E(&MS, len + 4, "ERSE");
    tncl->tncl_erase = p;
    p[0] = '\r';
    p[1] = '\0';
    p = strTail(&p[2], prompt);
    p[0] = '\r';
    p[1] = '\0';
    p = MEM_ALLOC_E(&MS, len + 1, "PRPT");
    tncl->tncl_prompt = p;
    *strTail(p, prompt) = '\0';
    tncl->tncl_promptLen = len;
    if (promptWasShown) {
	DEBUG_PROMPT(tncl, "setPrompt showing");
	showPrompt(tncl, B_TRUE);
    }
    if (! tncl->tncl_writeWaiting) {
	(void) pushBuffer(tncl);
    }
}

/*
 * setEcho - control echoing, for password entering.
 */

static void
setEcho(TelnetClient_t *tncl, BOOL_T echoOn)
{

    /* Seems backwards but it is the other end's echoing we are
       controlling. It is "conventional" to link ECHO with SUPGA, so doing
       so here will hopefully allow us to work properly with more TELNET
       clients. */
    if (echoOn) {
	if (tncl->tncl_meDoEcho && (tncl->tncl_flags & B_PW_BLANK) != 0) {
	    sendOptionWONT(tncl, ECHO);
	    if ((tncl->tncl_flags & B_PW_BLANK_SUPGA) != 0) {
		sendOptionWONT(tncl, SUPGA);
		sendOptionDONT(tncl, SUPGA);
	    }
	}
	CMD_DEBUG(tncl, "setting EchoOn");
    } else {
	if (! tncl->tncl_meDoEcho && (tncl->tncl_flags & B_PW_BLANK) != 0) {
	    sendOptionWILL(tncl, ECHO);
	    if ((tncl->tncl_flags & B_PW_BLANK_SUPGA) != 0) {
		sendOptionWILL(tncl, SUPGA);
		sendOptionDO(tncl, SUPGA);
	    }
	}
	CMD_DEBUG(tncl, "clearing EchoOn");
    }
    tncl->tncl_echoOn = echoOn;
}

/*
 * logPlayer - handy little routine.
 */

static void
logPlayer(TelnetClient_t *tncl, CHAR_T *m1, CHAR_T *m2)
{
    CHAR_T buffer[PLAYER_NAME_LEN + 80];

    *strTail(strTail(strTail(strTail(&buffer[0], m1),
			    &tncl->tncl_playerName[0]), m2), "\n") = '\0';
    log(&buffer[0]);
}

/*
 * outputPlayer - another handy one.
 */

static void
outputPlayer(TelnetClient_t *tncl, CHAR_T *m1, CHAR_T *m2)
{

    if (m1 != P_NULL) {
	sendString(tncl, m1);
    }
    (void) sendChar(tncl, '\"');
    sendString(tncl, &tncl->tncl_playerName[0]);
    (void) sendChar(tncl, '\"');
    sendString(tncl, m2);
    sendString(tncl, ".\n");
}

/*
 * resetAfterPW - little utility.
 */

static void
resetAfterPW(TelnetClient_t *tncl)
{

    setEcho(tncl, B_TRUE);
    if (tncl->tncl_state == tnst_normal) {
	setPrompt(tncl, tncl->tncl_client->cl_prompt);
    } else {
	setPrompt(tncl, "> ");
    }
}

/*
 * startPlayer - all OK, start the player running.
 */

static void
startPlayer(TelnetClient_t *tncl)
{

    log2NL("startPlayer for character ", &tncl->tncl_playerName[0]);
    tncl->tncl_state = tnst_normal;
    resetAfterPW(tncl);
    fullyInsertClient(tncl->tncl_client);	/* BEFORE the doStartPlayer */
    doStartPlayerStuff(tncl->tncl_client);
}

/*
 * newPlayer - we are creating a new player.
 */

static void
newPlayer(TelnetClient_t *tncl)
{

    setEcho(tncl, B_FALSE);
    tncl->tncl_state = tnst_npPassword1;
    setPrompt(tncl, "Enter password for new character: ");
}

/*
 * setExiting - common code used hither and yon.
 */

static void
setExiting(TelnetClient_t *tncl)
{

    tncl->tncl_state = tnst_exiting;
    setPrompt(tncl, "Type a return to exit: ");
}

/*
 * nameRetry - ask for player name again?
 */

static void
nameRetry(TelnetClient_t *tncl)
{

    if (tncl->tncl_nameCount == NAME_TRIES) {
	setExiting(tncl);
    } else {
	tncl->tncl_nameCount += 1;
    }
}

/*
 * nameAgain - check if want to go round to asking for name again.
 */

static void
nameAgain(TelnetClient_t *tncl, BOOL_T needPrompt)
{

    if (tncl->tncl_nameCount == NAME_TRIES) {
	setExiting(tncl);
    } else {
	tncl->tncl_state = tnst_name;
	tncl->tncl_nameCount += 1;
	if (needPrompt) {
	    setPrompt(tncl, "Enter character name: ");
	}
    }
}

/*
 * requestPlayer - send a request for the connection of the given player.
 */

static void
requestPlayer(TelnetClient_t *tncl, CHAR_T *name, UINT_T len)
{
    Request_t *rq;
    Client_t *cl;

    rq = allocRequestE(&MS, len + 1);
    memcpy(&rq->rq_u.ru_text[0], name, len);
    rq->rq_u.ru_text[len] = '\0';
    rq->rq_usedLen = len + 1;
    cl = tncl->tncl_client;
    rq->rq_key = cl->cl_isLocal ? 0 : 1;
    askPlayer(cl, rq);
    switch (rq->rq_key) {
    case 0:
	/* no such player and automatic creation disallowed */
	outputPlayer(tncl, "Character ", " does not exist");
	logPlayer(tncl, "Character ", " creation disallowed");
	nameRetry(tncl);
	break;
    case 1:
	/* no such player, but it is ok to create a new one, and there
	   is no creation password */
	tncl->tncl_state = tnst_newPlayerNCP;
	tncl->tncl_ynCount = 1;
	outputPlayer(tncl, "Character ", " does not exist");
	setPrompt(tncl, "Do you wish to create it (N/y)? ");
	break;
    case 2:
	/* no such player, but it is ok to create a new one, and there
	   is no creation password */
	tncl->tncl_state = tnst_newPlayerWCP;
	tncl->tncl_ynCount = 1;
	outputPlayer(tncl, "Character ", " does not exist");
	setPrompt(tncl, "Do you wish to create it (N/y)? ");
	break;
    case 3:
	/* the given name is not a player in the character table */
	outputPlayer(tncl, P_NULL, " is not a character");
	logPlayer(tncl, "Non-character ", " used as name");
	nameRetry(tncl);
	break;
    case 4:
	/* remote login of SysAdmin denied */
	sendString(tncl, "Remote login of SysAdmin denied.\n");
	setExiting(tncl);
	log("Remote login of SysAdmin denied.\n");
	break;
    case 5:
	/* it is a player, now get the password. */
	tncl->tncl_state = tnst_plPassword;
	setPrompt(tncl, "Enter password: ");
	setEcho(tncl, B_FALSE);
	break;
    case 6:
	/* player is already connected */
	outputPlayer(tncl, P_NULL, " is already connected");
	logPlayer(tncl, "Attempt to reconnect connected character ", "");
	nameRetry(tncl);
	break;
    }
    freeRequestE(&MS, rq);
}

/*
 * handleInputLine - state-sensitive handling of player input.
 */

static void
handleInputLine(TelnetClient_t *tncl, CHAR_T *p, ULONG_T len)
{
    Client_t *cl;
    Request_t *rq;
    CHAR_T *q;
    CHAR_T ch;

    cl = tncl->tncl_client;
    switch (tncl->tncl_state) {
    case tnst_name:
	if (len >= PLAYER_NAME_LEN) {
	    len = PLAYER_NAME_LEN - 1;
	}
	memcpy(&tncl->tncl_playerName[0], p, len);
	tncl->tncl_playerName[len] = '\0';
	q = &tncl->tncl_playerName[0];
	do {
	    ch = *q++;
	} while (ch != '\0' && ch != ' ' && ch != '\t' &&
		 ch != '.' && ch != ';' && ch != ',' && ch != '\"');
	if (len == 0) {
	    (void) pushBuffer(tncl);
	    endClient(cl);
	    tncl->tncl_state = tnst_exiting;
	} else if (ch != '\0') {
	    sendString(tncl,
		"Character names cannot contain spaces or punctuation.\n");
	    logPlayer(tncl, "Character name '",
		      "' contained space or punctuation");
	    nameRetry(tncl);
	} else {
	    requestPlayer(tncl, p, len);
	}
	break;
    case tnst_plPassword:
	rq = allocRequestE(&MS, PASSWORD_LEN);
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	copy(&rq->rq_u.ru_text[0], p, len);
	rq->rq_u.ru_text[len] = '\0';
	MUDCrypt(&rq->rq_u.ru_text[0], B_FALSE, cl->cl_sessionKey);
	loginPlayer(tncl->tncl_client, rq);
	switch (rq->rq_key) {
	case 0:
	    /* Password OK. Start playing. */
	    startPlayer(tncl);
	    break;
	case 1:
	    /* pw wrong, try again. */
	    sendString(tncl, "Password incorrect. Try again.\n");
	    break;
	case 2:
	    /* pw wrong, run denied */
	    sendString(tncl, "Password incorrect.\n");
	    setEcho(tncl, B_TRUE);
	    nameAgain(tncl, B_TRUE);
	    break;
	}
	freeRequestE(&MS, rq);
	break;
    case tnst_newPlayerNCP:
    case tnst_newPlayerWCP:
	while (len != 0 && *p == ' ') {
	    len -= 1;
	    ++p;
	}
	if (len != 0 && (*p == 'y' || *p == 'Y')) {
	    if (tncl->tncl_state == tnst_newPlayerNCP) {
		/* no creation password needed */
		newPlayer(tncl);
	    } else {
		tncl->tncl_state = tnst_crPassword;
		setPrompt(tncl, "Enter character creation password: ");
		setEcho(tncl, B_FALSE);
	    }
	} else if (len == 0 || *p == 'n' || *p == 'N') {
	    logPlayer(tncl, "New character ", " creation declined");
	    nameAgain(tncl, B_TRUE);
	} else {
	    if (tncl->tncl_ynCount == YN_TRIES) {
		setExiting(tncl);
	    } else {
		tncl->tncl_ynCount += 1;
	    }
	}
	break;
    case tnst_crPassword:
	rq = allocRequestE(&MS, PASSWORD_LEN);
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	copy(&rq->rq_u.ru_text[0], p, len);
	rq->rq_u.ru_text[len] = '\0';
	MUDCrypt(&rq->rq_u.ru_text[0], B_FALSE, cl->cl_sessionKey);
	creationCheck(cl, rq);
	switch (rq->rq_key) {
	case 0:
	    /* Password OK. Create the new player. */
	    newPlayer(tncl);
	    break;
	case 1:
	    /* pw wrong, try again. */
	    sendString(tncl, "Creation password incorrect. Try again.\n");
	    break;
	case 2:
	    /* pw wrong, run denied */
	    sendString(tncl, "Creation password incorrect.\n");
	    setEcho(tncl, B_TRUE);
	    nameAgain(tncl, B_TRUE);
	    break;
	}
	freeRequestE(&MS, rq);
	break;
    case tnst_npPassword1:
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	} else {
	    tncl->tncl_password[len] = '\0';
	}
	copy(&tncl->tncl_password[0], p, len);
	MUDCrypt(&tncl->tncl_password[0], B_FALSE, cl->cl_sessionKey);
	setPrompt(tncl, "Re-enter password for verification: ");
	tncl->tncl_state = tnst_npPassword2;
	break;
    case tnst_npPassword2:
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	*(p + len) = '\0';
	q = &tncl->tncl_password[0];
	MUDCrypt(q, B_TRUE, cl->cl_sessionKey);
	if (strCmp(p, q) != 0) {
	    memset(&tncl->tncl_password[0], 0, PASSWORD_LEN);
	    sendString(tncl, "Password not verified.\n");
	    logPlayer(tncl, "New character ", " password not verified");
	    setEcho(tncl, B_TRUE);
	    nameAgain(tncl, B_TRUE);
	} else {
	    rq = allocRequestE(&MS, PASSWORD_LEN + PLAYER_NAME_LEN);
	    MUDCrypt(&tncl->tncl_password[0], B_FALSE,
		     tncl->tncl_client->cl_sessionKey);
	    memcpy(&rq->rq_u.ru_text[0], &tncl->tncl_password[0],PASSWORD_LEN);
	    memset(&tncl->tncl_password[0], 0, PASSWORD_LEN);
	    strCpy(&rq->rq_u.ru_text[PASSWORD_LEN], &tncl->tncl_playerName[0]);
	    createPlayer(cl, rq);
	    if (rq->rq_key == 0) {
		/* hmmm. Someone got there before us! */
		sendString(tncl, "Sorry! Someone just created it!\n");
		setExiting(tncl);
		logPlayer(tncl,
			  "New character ", " creation failed - duplicate");
	    } else {
		outputPlayer(tncl, "New character ", " created");
		logPlayer(tncl, "New character ", " created");
		startPlayer(tncl);
	    }
	    freeRequestE(&MS, rq);
	}
	break;
#if 0
    case tnst_parsing:
	if (ms->FileList == P_NULL) {
	    ms->EditLineNumber += 1;
	    /* when reading from files, this is done in 'getChar' */
	    if (len != 0) {
		if (len >= INPUT_BUFFER_SIZE) {
		    len = INPUT_BUFFER_SIZE - 1;
		}
		blockCopy(&ms->InputLine[0], p, len);
		ms->InputLine[len] = '\n';
		len += 1;
		ms->InputTotal = len;
		ms->InputPos = 0;
	    }
	}
	break;
#endif
    case tnst_normal:
	/* no other check on 'len' here - should be OK, even in server */
	if (len == 0xffff) {
	    len = 0xfffe;
	}
	if (cl->cl_inputAction != NIL_ACTION) {
	    CHAR_T *q;

	    /* Note: empty lines are passed - they could matter. */
	    q = MEM_ALLOC_E(&MS, len + 1, "ISTR");
	    copy(q, p, len);
	    q[len] = '\0';
	    setupForCall(cl);
	    (void) doProc1Top(cl->cl_inputAction, q);
	    cleanupAfterCall();
	}
	break;
    case tnst_exiting:
	(void) pushBuffer(tncl);
	endClient(tncl->tncl_client);
	break;
    case tnst_changePassword1:
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	copy(&tncl->tncl_password[0], p, len);
	tncl->tncl_password[len] = '\0';
	MUDCrypt(&tncl->tncl_password[0], B_FALSE, cl->cl_sessionKey);
	if (doChangePassword(cl, &tncl->tncl_password[0])) {
	    tncl->tncl_state = tnst_changePassword2;
	    if (cl->cl_playerPassword) {
		setPrompt(tncl, "Enter new character password: ");
	    } else {
		setPrompt(tncl, "Enter new creation password: ");
	    }
	} else {
	    sendString(tncl, "Password incorrect. No change done.\n");
	    tncl->tncl_state = tncl->tncl_saveState;
	    resetAfterPW(tncl);
	}
	break;
    case tnst_changePassword2:
	/* The first copy of the new password. */
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	copy(&tncl->tncl_password[0], p, len);
	tncl->tncl_password[len] = '\0';
	MUDCrypt(&tncl->tncl_password[0], B_FALSE, cl->cl_sessionKey);
	setPrompt(tncl, "Re-enter password for verification: ");
	tncl->tncl_state = tnst_changePassword3;
	break;
    case tnst_changePassword3:
	/* The second copy of the new password. */
	if (len >= PASSWORD_LEN) {
	    len = PASSWORD_LEN - 1;
	}
	*(p + len) = '\0';
	tncl->tncl_state = tncl->tncl_saveState;
	MUDCrypt(&tncl->tncl_password[0], B_TRUE, cl->cl_sessionKey);
	if (strCmp(&tncl->tncl_password[0], p) == 0) {
	    MUDCrypt(&tncl->tncl_password[0], B_FALSE, cl->cl_sessionKey);
	    (void) doChangePassword(cl, &tncl->tncl_password[0]);
	    sendString(tncl, "Password changed.\n");
	} else {
	    sendString(tncl, "Verification failed. Password not changed.\n");
	}
	resetAfterPW(tncl);
	break;
    }
}
    
/*
 * doInputLine - stuff for sending a line for further processing.
 */

static void
doInputLine(TelnetClient_t *tncl)
{

    if (tncl->tncl_meDoEcho) {
	(void) sendChar(tncl, '\n');
    } else {
	(void) sendChar(tncl, '\r');
    }
#if DEBUG
    if (tncl->tncl_linePos >= 3 && tncl->tncl_lineBuffer[0] == '*' &&
	tncl->tncl_lineBuffer[1] == '*')
    {
	doDebugCommand(tncl);
	messageFlush();
	tncl->tncl_linePos = 0;
	return;
    }
    if (tncl->tncl_debugInput) {
	/* lineBuffer is one longer than we ever put into it */
	tncl->tncl_lineBuffer[tncl->tncl_linePos] = '\0';
	debugWithBuffer(tncl, "input", &tncl->tncl_lineBuffer[0]);
    }
#endif
    tncl->tncl_promptShown = B_FALSE;
    handleInputLine(tncl, &tncl->tncl_lineBuffer[0], tncl->tncl_linePos);
    messageFlush();
    tncl->tncl_linePos = 0;
    if (tncl->tncl_wantIdleAction) {
	tncl->tncl_wantIdleAction = B_FALSE;
	if (tncl->tncl_meDoEcho) {
	    (void) sendChar(tncl, '\n');
	} else {
	    (void) sendChar(tncl, '\r');
	}
	(void) pushBuffer(tncl);
	endClient(tncl->tncl_client);
    }
}

/*
 * doBackup - erase a character.
 */

static void
doBackup(TelnetClient_t *tncl)
{

    (void) sendChar(tncl, '\b');
    (void) sendChar(tncl, ' ');
    (void) sendChar(tncl, '\b');
}

/*
 * cookChar - handle a single input character. Return B_TRUE if the
 *	character was a carriage return, and we have just called
 *	doInputLine to process the user input, else return B_FALSE.
 */

static BOOL_T
cookChar(TelnetClient_t *tncl, char ch)
{
    UINT_T i;

    switch (ch) {
    case '\r':
	doInputLine(tncl);
	tncl->tncl_lastWasCR = B_TRUE;
	return B_TRUE;
    case '\n':
	if (! tncl->tncl_lastWasCR) {
	    doInputLine(tncl);
	    return B_TRUE;
	}
	break;
    case '\b':
    case 0x7f:
	if (tncl->tncl_linePos != 0) {
	    tncl->tncl_linePos -= 1;
	    if (tncl->tncl_meDoEcho) {
		doBackup(tncl);
	    }
	}
	break;
    case CNTL_X:
    case CNTL_U:
	if (tncl->tncl_linePos != 0) {
	    if (tncl->tncl_meDoEcho) {
		for (i = 0; i != tncl->tncl_linePos; i += 1) {
		    doBackup(tncl);
		}
	    }
	    tncl->tncl_linePos = 0;
	}
	break;
    case CNTL_R:
	if (tncl->tncl_meDoEcho && tncl->tncl_linePos != 0) {
	    sendChar(tncl, '^');
	    sendChar(tncl, 'R');
	    sendChar(tncl, '\n');
	    DEBUG_PROMPT(tncl, "CNTL_R showing");
	    showPrompt(tncl, B_FALSE);
	    for (i = 0; i != tncl->tncl_linePos; i += 1) {
		sendChar(tncl, tncl->tncl_lineBuffer[i]);
	    }
	}
	break;
    default:
	if (tncl->tncl_linePos != LINE_BUFFER_SIZE) {
	    tncl->tncl_lineBuffer[tncl->tncl_linePos] = ch;
	    tncl->tncl_linePos += 1;
	    if (tncl->tncl_meDoEcho) {
		if (tncl->tncl_echoOn) {
		    sendChar(tncl, ch);
		} else {
		    sendChar(tncl, ' ');
		}
	    }
	}
	break;
    }
    tncl->tncl_lastWasCR = B_FALSE;
    return B_FALSE;
}

/*
 * processSB - process a subnegotiation sequence in tncl_SBBuffer.
 */

static void
processSB(TelnetClient_t *tncl)
{
    UINT_T i, j, len;
    BYTE_T *p;

#if DEBUG
    if (tncl->tncl_debugCmds) {
	debugStart(tncl);
	debugString(tncl, "SB ");
	if (tncl->tncl_SBPos == 0) {
	    debugString(tncl, "len = 0");
	} else {
	    debugString(tncl, optionName(tncl, tncl->tncl_SBBuffer[0]));
	    if (tncl->tncl_SBPos != 1) {
		for (i = 1; i < tncl->tncl_SBPos; i += 1) {
		    debugChar(tncl, ' ');
		    debugByte(tncl, tncl->tncl_SBBuffer[i]);
		}
	    }
	}
	debugEnd(tncl);
    }
#endif

    if (tncl->tncl_SBPos != 0) {
	switch (tncl->tncl_SBBuffer[0]) {
	case NAWS:
	    /* RFC 1073 */
	    if (tncl->tncl_SBPos == 5) {
		/* telnet protocol does everything big-endian */
		i = (UINT_T) tncl->tncl_SBBuffer[1] << 8 |
		    (UINT_T) tncl->tncl_SBBuffer[2];
		j = (UINT_T) tncl->tncl_SBBuffer[3] << 8 |
		    (UINT_T) tncl->tncl_SBBuffer[4];
		if (i == 0) {
		    i = 80;
		}
		if (j == 0) {
		    j = 24;
		}
#if DEBUG
		if (tncl->tncl_debugCmds) {
		    debugStart(tncl);
		    debugString(tncl, "NAWS got height ");
		    debugByte(tncl, j);
		    debugString(tncl, ", width ");
		    debugByte(tncl, i);
		    debugEnd(tncl);
		}
#endif
		tncl->tncl_client->cl_textWidth = i;
		tncl->tncl_client->cl_textHeight = j;
		tncl->tncl_client->cl_gotTextSize = B_TRUE;
	    } else {
		complain(tncl, "invalid NAWS SB received");
	    }
	    break;
	case TERMINALTYPE:
	    /* RFC 1091 */
	    if (tncl->tncl_SBPos >= 4 && tncl->tncl_SBBuffer[1] == 0) {
		p = &tncl->tncl_SBBuffer[2];
		len = tncl->tncl_SBPos - 2;
		/* TERMINAL-TYPE IS ... IAC */
		while (*p != TELNET_IAC && len != 0) {
		    ++p;
		    len -= 1;
		}
		*p = '\0';
		freeString(tncl->tncl_terminalType);
		tncl->tncl_terminalType = stringCopy(&tncl->tncl_SBBuffer[2]);
#if DEBUG
		if (tncl->tncl_debugCmds) {
		    debugStart(tncl);
		    debugString(tncl, "Got terminal type '");
		    debugString(tncl, tncl->tncl_terminalType);
		    debugChar(tncl, '\'');
		    debugEnd(tncl);
		}
#endif
	    } else {
		complain(tncl, "invalid TERMINALTYPE SB received");
	    }
	    break;
	default:
	    complain(tncl, "unknown SB command received");
	}
    }
}

/*
 * cookByte - first level of interpreting data from user. This
 *	level handles all of the TELNET protocol stuff.
 */

static BOOL_T
cookByte(TelnetClient_t *tncl, BYTE_T b)
{
    BOOL_T sentLine;

    sentLine = B_FALSE;
    switch (tncl->tncl_telnetState) {
    case tntst_normal:
	if (tncl->tncl_lastWasIAC) {
	    switch (b) {
	    case TELNET_IAC:
		/* Should only get these if other end is in binary, i.e. if
		   "ReceiveBinary" is set. */
		CMD_DEBUG_SHORT_CMD(tncl, "recv", TELNET_IAC);
		(void) cookChar(tncl, (char) b);
		tncl->tncl_telnetState = tntst_normal;
		break;
	    case TELNET_AYT:
		/* Yep, I'm here. */
		CMD_DEBUG_SHORT_CMD(tncl, "recv", TELNET_AYT);
		sendShortCmd(tncl, TELNET_NOP);
		tncl->tncl_telnetState = tntst_normal;
		break;
	    case TELNET_GA:
	    case TELNET_AO:
	    case TELNET_DM:
	    case TELNET_NOP:
	    case TELNET_EOR:
	    case TELNET_SUSP:
		/* Just ignore these. */
		CMD_DEBUG_SHORT_CMD(tncl, "recv", b);
		tncl->tncl_telnetState = tntst_normal;
		break;
	    case TELNET_DONT:
	    case TELNET_DO:
	    case TELNET_WONT:
	    case TELNET_WILL:
		tncl->tncl_command = b;
		tncl->tncl_telnetState = tntst_command;
		break;
	    case TELNET_SB:
		tncl->tncl_telnetState = tntst_SB;
		tncl->tncl_SBPos = 0;
		break;
	    case TELNET_SE:
		complain(tncl, "invalid SE without matching SB");
		tncl->tncl_telnetState = tntst_normal;
		break;
	    case TELNET_EC:
		CMD_DEBUG_SHORT_CMD(tncl, "recv", TELNET_EC);
		(void) cookChar(tncl, '\b');
		tncl->tncl_telnetState = tntst_normal;
		break;
	    case TELNET_EL:
		CMD_DEBUG_SHORT_CMD(tncl, "recv", TELNET_EL);
		(void) cookChar(tncl, CNTL_X);
		tncl->tncl_telnetState = tntst_normal;
		break;
	    case TELNET_IP:
	    case TELNET_BRK:
	    case TELNET_ABORT:
	    case TELNET_EOF:
		/* Do it gently */
		CMD_DEBUG_SHORT_CMD(tncl, "recv", b);
		if (tncl->tncl_state == tnst_normal) {
		    (void) pushBuffer(tncl);
		    endClient(tncl->tncl_client);
		}
		tncl->tncl_state = tnst_exiting;
		break;
	    }
	    tncl->tncl_lastWasIAC = B_FALSE;
	} else {
	    if (b == TELNET_IAC) {
		tncl->tncl_lastWasIAC = B_TRUE;
	    } else if (b != 0) {
		sentLine = cookChar(tncl, (char) b);
	    }
	}
	break;
    case tntst_command:
	tncl->tncl_telnetState = tntst_normal;
	CMD_DEBUG_CMD(tncl, "recv", tncl->tncl_command, b);
	switch (tncl->tncl_command) {
	case TELNET_DO:
	    if (b > MAX_OPTION) {
		/* Just ignore it. */
		break;
	    }
	    if (tncl->tncl_sentOptionWILL[b] || tncl->tncl_sentOptionWONT[b]) {
		tncl->tncl_sentOptionWILL[b] = B_FALSE;
		tncl->tncl_sentOptionWONT[b] = B_FALSE;
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "setting SendBinary");
		    tncl->tncl_sendBinary = B_TRUE;
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "setting MeDoEcho");
		    tncl->tncl_meDoEcho = B_TRUE;
		    break;
		case SUPGA:
		    CMD_DEBUG(tncl, "setting SupGA");
		    tncl->tncl_supGA = B_TRUE;
		    break;
		case ENDOFRECORD:
		    CMD_DEBUG(tncl, "setting DoEOR");
		    tncl->tncl_doEOR = B_TRUE;
		}
	    } else {
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "setting SendBinary");
		    tncl->tncl_sendBinary = B_TRUE;
		    sendCmd(tncl, TELNET_WILL, b);
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "setting MeDoEcho");
		    tncl->tncl_meDoEcho = B_TRUE;
		    sendCmd(tncl, TELNET_WILL, b);
		    break;
		case SUPGA:
		    CMD_DEBUG(tncl, "setting SupGA");
		    tncl->tncl_supGA = B_TRUE;
		    sendCmd(tncl, TELNET_WILL, b);
		    break;
		case ENDOFRECORD:
		    CMD_DEBUG(tncl, "setting DoEOR");
		    tncl->tncl_doEOR = B_TRUE;
		    break;
		case NAWS:
		    sendCmd(tncl, TELNET_WILL, b);
		    break;
		default:
		    /* No, I won't do it! */
		    sendCmd(tncl, TELNET_WONT, b);
		    break;
		}
	    }
	    break;
	case TELNET_WILL:
	    if (b > MAX_OPTION) {
		break;
	    }
	    if (tncl->tncl_sentOptionDO[b] || tncl->tncl_sentOptionDONT[b]) {
		tncl->tncl_sentOptionDO[b] = B_FALSE;
		tncl->tncl_sentOptionDONT[b] = B_FALSE;
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "setting ReceiveBinary");
		    tncl->tncl_receiveBinary = B_TRUE;
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "clearing MeDoEcho");
		    tncl->tncl_meDoEcho = B_FALSE;
		    break;
		case SUPGA:
		case NAWS:
		    break;
		case TERMINALTYPE:
		    CMD_DEBUG(tncl, "asking for terminal type");
		    putByte(tncl, TELNET_IAC);
		    putByte(tncl, TELNET_SB);
		    putByte(tncl, TERMINALTYPE);
		    putByte(tncl, 1);	/* SEND = 1 */
		    putByte(tncl, TELNET_IAC);
		    putByte(tncl, TELNET_SE);
		    break;
		default:
		    sendCmd(tncl, TELNET_DONT, b);
		    break;
		}
	    } else {
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "setting ReceiveBinary");
		    tncl->tncl_receiveBinary = B_TRUE;
		    sendCmd(tncl, TELNET_DO, b);
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "clearing MeDoEcho");
		    tncl->tncl_meDoEcho = B_FALSE;
		    sendCmd(tncl, TELNET_DO, b);
		    break;
		case SUPGA:
		case NAWS:
		case TERMINALTYPE:
		    /* Sure, do these - I understand them. */
		    sendCmd(tncl, TELNET_DO, b);
		    break;
		default:
		    /* Don't need to send a DONT, since that is default,
		       but do it in case someone needs it. */
		    sendCmd(tncl, TELNET_DONT, b);
		    break;
		}
	    }
	    break;
	case TELNET_DONT:
	    if (b > MAX_OPTION) {
		break;
	    }
	    if (tncl->tncl_sentOptionWONT[b] || tncl->tncl_sentOptionWILL[b]) {
		tncl->tncl_sentOptionWONT[b] = B_FALSE;
		tncl->tncl_sentOptionWILL[b] = B_FALSE;
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "clearing SendBinary");
		    tncl->tncl_sendBinary = B_FALSE;
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "clearing MeDoEcho");
		    tncl->tncl_meDoEcho = B_FALSE;
		    break;
		case SUPGA:
		    CMD_DEBUG(tncl, "clearing SupGA");
		    tncl->tncl_supGA = B_FALSE;
		    break;
		case ENDOFRECORD:
		    CMD_DEBUG(tncl, "clearing DoEOR");
		    tncl->tncl_doEOR = B_FALSE;
		    break;
		}
	    } else {
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "clearing SendBinary");
		    tncl->tncl_sendBinary = B_FALSE;
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "clearing MeDoEcho");
		    tncl->tncl_meDoEcho = B_FALSE;
		    break;
		case SUPGA:
		    CMD_DEBUG(tncl, "clearing SupGA");
		    tncl->tncl_supGA = B_FALSE;
		    break;
		case ENDOFRECORD:
		    CMD_DEBUG(tncl, "clearing DoEOR");
		    tncl->tncl_doEOR = B_FALSE;
		    break;
		}
		sendCmd(tncl, TELNET_WONT, b);
	    }
	    break;
	case TELNET_WONT:
	    if (b > MAX_OPTION) {
		break;
	    }
	    if (tncl->tncl_sentOptionDO[b] || tncl->tncl_sentOptionDONT[b]) {
		tncl->tncl_sentOptionDO[b] = B_FALSE;
		tncl->tncl_sentOptionDONT[b] = B_FALSE;
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "clearing ReceiveBinary");
		    tncl->tncl_receiveBinary = B_FALSE;
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "setting MeDoEcho");
		    tncl->tncl_meDoEcho = B_TRUE;
		    break;
		}
	    } else {
		switch (b) {
		case BINARY:
		    CMD_DEBUG(tncl, "clearing ReceiveBinary");
		    tncl->tncl_receiveBinary = B_FALSE;
		    break;
		case ECHO:
		    CMD_DEBUG(tncl, "setting MeDoEcho");
		    tncl->tncl_meDoEcho = B_TRUE;
		    break;
		}
		sendCmd(tncl, TELNET_DONT, b);
	    }
	}
	break;
    case tntst_SB:
	if (tncl->tncl_lastWasIAC) {
	    if (b == TELNET_IAC) {
		if (tncl->tncl_SBPos < SB_BUFFER_SIZE) {
		    tncl->tncl_SBBuffer[tncl->tncl_SBPos] = b;
		    tncl->tncl_SBPos += 1;
		}
	    } else if (b == TELNET_SE) {
		tncl->tncl_telnetState = tntst_normal;
		processSB(tncl);
	    } else {
		complain(tncl, "invalid IAC in SB");
		tncl->tncl_telnetState = tntst_normal;
	    }
	    tncl->tncl_lastWasIAC = B_FALSE;
	} else if (b == TELNET_IAC) {
	    tncl->tncl_lastWasIAC = B_TRUE;
	} else {
	    if (tncl->tncl_SBPos < SB_BUFFER_SIZE) {
		tncl->tncl_SBBuffer[tncl->tncl_SBPos] = b;
		tncl->tncl_SBPos += 1;
	    }
	}
	break;
    }
    return sentLine;
}

/*
 * telnetRead - called when there is something to read on this client.
 */

void
telnetRead(Client_t *cl)
{
    TelnetClient_t *tncl;
    CHAR_T *p;
    LONG_T len;
    BOOL_T sentLine;

    TRACE("telnetRead", cl);
    sentLine = B_FALSE;
    tncl = cl->cl_telnet;
    tncl->tncl_doneOutput = B_FALSE;
    p = &tncl->tncl_inputBuffer[0];
    len = doRecv(cl, p, INPUT_BUFFER_SIZE);
    sentLine = B_FALSE;
    while (len > 0) {
	if (cookByte(tncl, *p)) {
	    sentLine = B_TRUE;
	}
	len -= 1;
	++p;
    }
    if (sentLine && ! tncl->tncl_promptShown) {
	DEBUG_PROMPT(tncl, "telnetRead showing");
	showPrompt(tncl, B_TRUE);
    }
    if (tncl->tncl_doneOutput && ! tncl->tncl_writeWaiting) {
	(void) pushBuffer(tncl);
    }
}

/*
 * telnetAllocClient - a new connection has arrived - setup our
 *	client structure for it, returning a pointer to it.
 */

TelnetClient_t *
telnetAllocClient(Client_t *cl, int flags)
{
    TelnetClient_t *tncl;
    unsigned int i;

    tncl = MEM_ALLOC_E(&MS, sizeof(TelnetClient_t), "tncl");
    tncl->tncl_client = cl;
    tncl->tncl_flags = flags;
    tncl->tncl_outputPos = 0;
    tncl->tncl_outputCurrent = 0;
    tncl->tncl_linePos = 0;
#if DEBUG
    tncl->tncl_debugPos = 0;
    tncl->tncl_debugCmds = B_FALSE;
    tncl->tncl_debugPrompt = B_FALSE;
    tncl->tncl_debugInput = B_FALSE;
#endif

    tncl->tncl_promptLen = 0;
    tncl->tncl_erase = stringCopy("\rX\r");
    tncl->tncl_erase[1] = '\0';
    tncl->tncl_prompt = stringCopy("");
    tncl->tncl_terminalType = stringCopy("UNKNOWN");
    for (i = 0; i != MAX_OPTION; i += 1) {
	tncl->tncl_sentOptionDO[i] = B_FALSE;
	tncl->tncl_sentOptionWILL[i] = B_FALSE;
	tncl->tncl_sentOptionDONT[i] = B_FALSE;
	tncl->tncl_sentOptionWONT[i] = B_FALSE;
    }
    tncl->tncl_echoOn = B_TRUE;
    tncl->tncl_promptShown = B_TRUE;	/* forces the first one out */
    tncl->tncl_lastWasCR = B_FALSE;
    tncl->tncl_telnetState = tntst_normal;
    tncl->tncl_lastWasIAC = B_FALSE;
    tncl->tncl_writeWaiting = B_FALSE;
    tncl->tncl_wantIdleAction = B_FALSE;
    tncl->tncl_state = tnst_name;
    tncl->tncl_nameCount = 1;

    /* Set these to all B_FALSE, as per telnet specs */
    tncl->tncl_sendBinary = B_FALSE;
    tncl->tncl_receiveBinary = B_FALSE;
    tncl->tncl_meDoEcho = B_FALSE;
    tncl->tncl_supGA = B_FALSE;
    tncl->tncl_doEOR = B_FALSE;

    if ((tncl->tncl_flags & B_NEG_BINARY) != 0) {
	sendOptionWILL(tncl, BINARY);
	sendOptionDO(tncl, BINARY);
    }
    if ((tncl->tncl_flags & B_NEG_PSEUDO) != 0) {
	sendOptionWONT(tncl, ECHO);
	sendOptionWONT(tncl, SUPGA);
    } else {
	if ((tncl->tncl_flags & B_NEG_ECHO) != 0) {
	    sendOptionWILL(tncl, ECHO);
	}
	if ((tncl->tncl_flags & B_NEG_SUPGA) != 0) {
	    sendOptionWILL(tncl, SUPGA);
	    sendOptionDO(tncl, SUPGA);
	}
    }
    if ((tncl->tncl_flags & B_NEG_TTYTYPE) != 0) {
	sendOptionDO(tncl, TERMINALTYPE);
    }
    if ((tncl->tncl_flags & B_NEG_EORMODE) != 0) {
	/* We will do EOR stuff, be we don't want it. */
	sendOptionWILL(tncl, ENDOFRECORD);
	sendOptionDONT(tncl, ENDOFRECORD);
    }
    if ((tncl->tncl_flags & B_NEG_NAWS) != 0) {
	sendOptionDO(tncl, NAWS);
    }

    /* Startup overrides of proper telnet protocol defaults */
    if ((tncl->tncl_flags & B_INIT_ECHO) != 0) {
	tncl->tncl_meDoEcho = B_TRUE;
    }
    if ((tncl->tncl_flags & B_INIT_EORMODE) != 0) {
	tncl->tncl_doEOR = B_TRUE;
    }

#if INITIAL_HELP
    sendString(tncl, "** Thanks for trying out the telnet code.\n"
	       "You can type **? for help, **st for status, etc.\n"
	       "After this message you should get a prompt for "
	       "character name.\n");
#else
    sendString(tncl, "** cgMUD telnet mode. **? for help, **st for status\n");
#endif

    setPrompt(tncl, "Enter character name: ");

    TRACE("telnetAllocClient", tncl);
    return tncl;
}

/*
 * telnetStopClient - normal shutdown of a telnet client.
 */

void
telnetStopClient(TelnetClient_t *tncl, BOOL_T doShutDown)
{

    if (doShutDown) {
	tncl->tncl_wantIdleAction = B_TRUE;
    }
}

/*
 * telnetFreeClient - the outside world is done with this client.
 */

void
telnetFreeClient(TelnetClient_t *tncl)
{

    TRACE("telnetFreeClient", tncl);
    memFreeE(&MS, tncl->tncl_erase, tncl->tncl_promptLen + 4);
    memFreeE(&MS, tncl->tncl_prompt, tncl->tncl_promptLen + 1);
    freeString(tncl->tncl_terminalType);
    memFreeE(&MS, tncl, sizeof(TelnetClient_t));
}

/*
 * telnetChangePassword - external interface by which we can go through
 *	a sequence of asking for a given type of password.
 */

void
telnetChangePassword(Client_t *cl)
{
    TelnetClient_t *tncl;

    TRACE("telnetAskPassword", cl);
    tncl = cl->cl_telnet;
    tncl->tncl_saveState = tncl->tncl_state;
    tncl->tncl_state = tnst_changePassword1;
    if (cl->cl_playerPassword) {
	setPrompt(tncl, "Enter old character password: ");
    } else {
	setPrompt(tncl, "Enter old creation password: ");
    }
    setEcho(tncl, B_FALSE);
}

/*
 * telnetSetPrompt - external interface by which we can change the
 *	prompt that is being shown.
 */

void
telnetSetPrompt(Client_t *cl, CHAR_T *newPrompt)
{

    TRACE("telnetSetPrompt", cl);
    setPrompt(cl->cl_telnet, newPrompt);
}

/*
 * telnetPutText - interface to send and/or queue an rt_message request.
 */

void
telnetPutText(Client_t *cl, Request_t *rq, BOOL_T flush)
{
    TelnetClient_t *tncl;
    Request_t **pRq;

    TRACE("telnetPutText", cl);
    tncl = cl->cl_telnet;
    if (tncl->tncl_writeWaiting) {
	/* Already plugged up - just queue the new message. */
	pRq = &cl->cl_outGoing;
	while (*pRq != P_NULL) {
	    pRq = &(*pRq)->rq_next;
	}
	*pRq = rq;
    } else {
	/* Try it right away. */
	if (processOutgoing(cl, rq, 0)) {
	    /* Sent/buffered the whole thing right away. */
	    if (rq->rq_u.ru_text[rq->rq_usedLen - 1] == '\n') {
		DEBUG_PROMPT(tncl, "telnetPutText showing");
		showPrompt(tncl, B_TRUE);
	    }
	    freeRequestE(&MS, rq);
	    if (flush) {
		/* We later will do a pushBuffer on the client on behalf
		   of which we are running now, but not for others. So,
		   since we want the data to get to the various clients
		   without too much delay, we write this stuff to this
		   other client right away. */
		(void) pushBuffer(tncl);
	    }
	} else {
	    /* Still some left - have to queue it. tncl_writeWaiting will
	       have been set. */
	    rq->rq_next = P_NULL;
	    cl->cl_outGoing = rq;
	}
    }
}
