/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

typedef enum {
    b_negBINARY,	/* negotiate BINARY on startup */
    b_negECHO,		/* negotiate ECHO on startup */
    b_negPseudo,	/* "negotiate" pseudo-linemode on startup */
    b_negSUPGA, 	/* negotiate SUPGA on startup */
    b_negEORMODE,	/* negotiate ENDOFRECORD on startup */
    b_negTTYTYPE,	/* negotiate TERMINALTYPE on startup */
    b_negNAWS,		/* negotiate NAWS on startup */
    b_initEcho, 	/* start echoing right away */
    b_initEORMODE,	/* use EORMODE right away */
    b_sendGA,		/* send GA after prompts */
    b_pwBlank,		/* try to do password echo suppression */
    b_pwBlankSUPGA,	/* do SUPGA as part of pw echo stuff */
    b_erasePrompt,	/* try to erase prompt as needed */
} CmdBitNum_t;

#define B_NEG_BINARY	(1 << (b_negBINARY	- b_negBINARY))	/* 0x0001 */
#define B_NEG_ECHO	(1 << (b_negECHO	- b_negBINARY))	/* 0x0002 */
#define B_NEG_PSEUDO	(1 << (b_negPseudo	- b_negBINARY)) /* 0x0004 */
#define B_NEG_SUPGA 	(1 << (b_negSUPGA	- b_negBINARY))	/* 0x0008 */
#define B_NEG_EORMODE	(1 << (b_negEORMODE	- b_negBINARY))	/* 0x0010 */
#define B_NEG_TTYTYPE	(1 << (b_negTTYTYPE	- b_negBINARY)) /* 0x0020 */
#define B_NEG_NAWS	(1 << (b_negNAWS	- b_negBINARY))	/* 0x0040 */
#define B_INIT_ECHO 	(1 << (b_initEcho	- b_negBINARY))	/* 0x0080 */
#define B_INIT_EORMODE	(1 << (b_initEORMODE	- b_negBINARY))	/* 0x0100 */
#define B_SEND_GA	(1 << (b_sendGA		- b_negBINARY))	/* 0x0200 */
#define B_PW_BLANK	(1 << (b_pwBlank	- b_negBINARY))	/* 0x0400 */
#define B_PW_BLANK_SUPGA (1 << (b_pwBlankSUPGA	- b_negBINARY))	/* 0x0800 */
#define B_ERASE_PROMPT	(1 << (b_erasePrompt	- b_negBINARY))	/* 0x1000 */

/* Default is for a decent normal telnet client. 't' */

#define DEFAULT_TELNET_FLAGS \
    (B_NEG_BINARY | B_NEG_PSEUDO | B_NEG_SUPGA | \
     B_NEG_EORMODE | B_NEG_TTYTYPE | B_NEG_NAWS | \
     B_ERASE_PROMPT | B_PW_BLANK | B_PW_BLANK_SUPGA)

/* Flags for a client that doesn't understand telnet at all. 's' */

#define SIMPLE_TELNET_FLAGS	0x0

/* Flags for a client that normally works with LPMUD. 'l' */

#define LPMUD_TELNET_FLAGS \
    (B_NEG_BINARY | B_NEG_TTYTYPE | B_NEG_NAWS | B_SEND_GA | B_PW_BLANK)
