#include <stdio.h>

#include "MUD.h"
#include "types.h"

/*
 * Amiga MUD
 *
 * Copyright (c) 1997 by Chris Gray
 *
 * CGMud Copyright (c) 2001 by Chris Gray
 */

/*
 * MUDCre - create an empty, but complete MUD database.
 */

#define DEBUG_IN	B_FALSE

typedef struct {
    UINT_T ix_slotLength;	/* full length of slot */
    UINT_T ix_usedLength;	/* space used by this entry */
    ULONG_T ix_offset;		/* byte offset in file */
} Index_t;

#define PUBLIC_SYMBOL_OFF	0
#define PUBLIC_SYMBOL_LEN	69
#define PNAME_OFF	(PUBLIC_SYMBOL_OFF + PUBLIC_SYMBOL_LEN)
#define PNAME_LEN	sizeof(Property_t)
#define CRE_PASS_OFF	(PNAME_OFF + PNAME_LEN)
#define CRE_PASS_LEN	17
#define MACHINE_LIST_OFF	(CRE_PASS_OFF + CRE_PASS_LEN)
#define MACHINE_LIST_LEN	sizeof(ItemList_t)
#define SA_PLAYER_OFF	(MACHINE_LIST_OFF + MACHINE_LIST_LEN)
#define SA_PLAYER_LEN	sizeof(Player_t)
#define SA_THING_OFF	(SA_PLAYER_OFF + SA_PLAYER_LEN)
#define SA_THING_LEN	(sizeof(Thing_t) + sizeof(Attr_t) + sizeof(Attr_t))
#define SA_SYM_OFF	(SA_THING_OFF + SA_THING_LEN)
#define SA_SYM_LEN	10
#define SA_NAME_OFF	(SA_SYM_OFF + SA_SYM_LEN)
#define SA_NAME_LEN	9
#define SA_PROMPT_OFF	(SA_NAME_OFF + SA_NAME_LEN)
#define SA_PROMPT_LEN	8
#define PLAYER_TAB_OFF	(SA_PROMPT_OFF + SA_PROMPT_LEN)
#define PLAYER_TAB_LEN	25
#define BUILTIN_TAB_OFF (PLAYER_TAB_OFF + PLAYER_TAB_LEN)
#define BUILTIN_TAB_LEN 10
#define PICON_OFF	(BUILTIN_TAB_OFF + BUILTIN_TAB_LEN)
#define PICON_LEN	sizeof(Property_t)
#define SA_ICON_OFF	(PICON_OFF + PICON_LEN)
#define SA_ICON_LEN	(sizeof(ItemList_t) + 32)

#define TOTAL_LEN	(SA_ICON_OFF + SA_ICON_LEN)

#define CHAR_LEN	(TOTAL_LEN)	/* keep it even */

#define T_SHIFT 	(8 - (32 - TYPE_SHIFT))

#define T_PLAYER =		    (TYPE_PLAYER << T_SHIFT)
#define T_THING =		    (TYPE_THING << T_SHIFT)
#define T_ACTION =		    (TYPE_ACTION << T_SHIFT)
#define T_STRING =		    (TYPE_STRING << T_SHIFT)
#define T_TABLE =		    (TYPE_TABLE << T_SHIFT)
#define T_LIST_INTEGER =	    (TYPE_LIST_INTEGER << T_SHIFT)
#define T_PROP_STRING = 	    (TYPE_PROP_STRING << T_SHIFT)
#define T_PROP_LIST_INTEGER =	    (TYPE_PROP_LIST_INTEGER << T_SHIFT)

#define INDEX_COUNT	14

Index_t IndexData[INDEX_COUNT] = {
    { 0,  0,  0},		/* dummy entry, always present */
    {PUBLIC_SYMBOL_LEN, PUBLIC_SYMBOL_LEN, PUBLIC_SYMBOL_OFF},
    {PNAME_LEN, PNAME_LEN, PNAME_OFF},
    {CRE_PASS_LEN, CRE_PASS_LEN, CRE_PASS_OFF},
    {MACHINE_LIST_LEN, MACHINE_LIST_LEN, MACHINE_LIST_OFF},
    {SA_PLAYER_LEN, SA_PLAYER_LEN, SA_PLAYER_OFF},
    {SA_THING_LEN, SA_THING_LEN, SA_THING_OFF},
    {SA_SYM_LEN, SA_SYM_LEN, SA_SYM_OFF},
    {SA_NAME_LEN, SA_NAME_LEN, SA_NAME_OFF},
    {SA_PROMPT_LEN, SA_PROMPT_LEN, SA_PROMPT_OFF},
    {PLAYER_TAB_LEN, PLAYER_TAB_LEN, PLAYER_TAB_OFF},
    {BUILTIN_TAB_LEN, BUILTIN_TAB_LEN, BUILTIN_TAB_OFF},
    {PICON_LEN, PICON_LEN, PICON_OFF},
    {SA_ICON_LEN, SA_ICON_LEN, SA_ICON_OFF}
};

#if DEBUG_IN

static void
showUnit(ULONG_T offset, ULONG_T len, char *name)
{
    static ULONG_T current = 0;

    printf("[%08x]%08x/%04d: %s\n", current, offset, len, name);
    current += len;
}

#define SHOW(o, l, n)	showUnit(o, l, n)

#else

#define SHOW(o, l, n)

#endif

FILE *Fd;

static void
writeByte(BYTE_T b)
{

    (void) fwrite(&b, 1, 1, Fd);
}

static void
writeUInt(UINT_T u)
{
    unsigned char ch;

    ch = (u >> 8) & 0xff;
    writeByte(ch);
    writeByte(u & 0xff);
}

static void
writeULong(ULONG_T n)
{
    UINT_T ui;

    ui = (n >> 16) & 0xffff;
    writeUInt(ui);
    writeUInt(n & 0xffff);
}

static void
writeString(char *p)
{

    (void) fputs(p, Fd);
    (void) fwrite("", 1, 1, Fd);
}

static void
writeDataFile(void)
{

    /* 1 */	/* the initial PublicSymbolTable */
    SHOW(PUBLIC_SYMBOL_OFF, PUBLIC_SYMBOL_LEN, "public symbol table");
    writeUInt(8);					/* shift = 8 */
    writeUInt(4);					/* usedCount = 4 */
    writeULong(SYSADMIN_KEY);				/* tb_owner SysAdmin */
    writeUInt(1);					/* tb_useCount = 1 */
    writeUInt(8);					/* nameLen = 8 */
    writeString("p_pName");
    writeULong(NAME_KEY);
    /* no 2-byte pad in data file */
    writeUInt(8);					/* nameLen = 8 */
    writeString("p_pIcon");
    writeULong(ICON_KEY);
    /* no 2-byte pad in data file */
    writeUInt(11);					/* nameLen = 11 */
    writeString("Characters");
    writeULong(NIL_TABLE | 10);
    /* no 2-byte pad in data file */
    writeUInt(8);					/* nameLen = 8 */
    writeString("Builtin");
    writeULong(NIL_TABLE | 11);
    /* no 2-byte pad in data file */

    /* 2 */	/* the Property_t for p_pName */
    SHOW(PNAME_OFF, PNAME_LEN, "p_pName");
    writeULong(SYSADMIN_KEY);				/* prp_owner */
    writeUInt(2);					/* prp_useCount */
    writeUInt(0);					/* padding */

    /* 3 */	/* the string which is the player creation password */
    SHOW(CRE_PASS_OFF, CRE_PASS_LEN, "creation password");
    writeString("CreationPassword");

    /* 4 */	/* the initial list of machines */
    SHOW(MACHINE_LIST_OFF, MACHINE_LIST_LEN, "machine list");
    writeULong(SYSADMIN_KEY);				/* owner = SysAdmin */
    writeUInt(1);					/* useCount = 1 */
    writeUInt(0);					/* elementCount = 0 */

    /* 5 */	/* the initial Player_t for SysAdmin */
    SHOW(SA_PLAYER_OFF, SA_PLAYER_LEN, "SysAdmin Player_t");
    writeString("SysAdminPassword");			/* password */
    writeByte(0); writeByte(0); writeByte(0);
    writeULong(NIL_THING | 6);				/* thing key */
    writeULong(NIL_TABLE | 7);				/* symtab key */
    writeULong(SYSADMIN_KEY);				/* sponsor SysAdmin */
    writeULong(NIL_THING);				/* location = nil */
    writeULong(NIL_ACTION);				/* inputAction = nil */
    writeULong(NIL_ACTION);				/* rawKeyAction */
    writeULong(NIL_ACTION);				/* mouseUpAction */
    writeULong(NIL_ACTION);				/* buttonAction */
    writeULong((TYPE_STRING << TYPE_SHIFT) | 9);	/* prompt key */
    writeULong(0);					/* lastOn time */
    writeULong(0);					/* logon count */
    writeULong(0);					/* total seconds */
    writeULong(NIL_ACTION);				/* idleAction */
    writeULong(NIL_ACTION);				/* activeAction */
    writeULong(NIL_ACTION);				/* effectDoneAction */
    writeUInt(TEXT_COLUMNS);				/* textWidth */
    writeUInt(23);					/* textHeight */
    writeUInt(0);					/* flags */
    writeByte(ps_wizard);				/* status */
    writeByte(B_TRUE);					/* wizardMode */
    writeByte(B_TRUE);					/* newPlayer */
    writeByte(0); writeUInt(0);				/* pad */

    /* 6 */	/* the initial Thing_t for SysAdmin */
    SHOW(SA_THING_OFF, SA_THING_LEN, "SysAdmin Thing_t");
    writeULong(NIL_THING);				/* th_parent = nil */
    writeULong(SYSADMIN_KEY);				/* owner = SysAdmin */
    writeUInt(1);					/* th_useCount = 1 */
    writeByte(2);					/* propListCount = 2 */
    writeByte(ts_public);				/* th_status public */
    writeULong(NAME_KEY);
    writeULong((TYPE_STRING << TYPE_SHIFT) | 8);
    writeULong(ICON_KEY);
    writeULong(NIL_LIST_INTEGER | 13);

    /* 7 */	/* the initial PrivateSymbolTable for SysAdmin */
    SHOW(SA_SYM_OFF, SA_SYM_LEN, "SysAdmin table");
    writeUInt(1);					/* tb_shift = 1 */
    writeUInt(0);					/* tb_usedCount = 0 */
    writeULong(SYSADMIN_KEY);				/* tb_owner SysAdmin */
    writeUInt(1);					/* tb_useCount = 1 */

    /* 8 */	/* the string which is SysAdmin's p_pName */
    SHOW(SA_NAME_OFF, SA_NAME_LEN, "SysAdmin name");
    writeString("SysAdmin");

    /* 9 */	/* the prompt string for SysAdmin */
    SHOW(SA_PROMPT_OFF, SA_PROMPT_LEN, "SysAdmin prompt");
    writeString("input> ");

    /* 10 */	/* the initial Characters table */
    SHOW(PLAYER_TAB_OFF, PLAYER_TAB_LEN, "Characters");
    writeUInt(4);					/* shift = 4 */
    writeUInt(1);					/* usedCount = 1 */
    writeULong(SYSADMIN_KEY);				/* tb_owner SysAdmin */
    writeUInt(1);					/* tb_useCount = 1 */
    writeUInt(9);					/* nameLen = 9 */
    writeString("SysAdmin");
    writeULong(SYSADMIN_KEY);
    /* no 2-byte pad in data file */

    /* 11 */	/* entry for the builtin table */
    SHOW(BUILTIN_TAB_OFF, BUILTIN_TAB_LEN, "Builtin");
    writeUInt(9);					/* shift = 9 */
    writeUInt(0);					/* usedCount = 0 */
    writeULong(SYSADMIN_KEY);				/* tb_owner SysAdmin */
    writeUInt(1);					/* tb_useCount = 1 */

    /* 12 */	/* the Property_t for p_pIcon */
    SHOW(PICON_OFF, PICON_LEN, "p_pIcon");
    writeULong(SYSADMIN_KEY);				/* prp_owner SysAdmin*/
    writeUInt(2);					/* prp_useCount */
    writeUInt(0);					/* padding */

    /* 13 */	/* SysAdmin's p_pIcon */
    SHOW(SA_ICON_OFF, SA_ICON_LEN, "SysAdmin icon");
    writeULong(SYSADMIN_KEY);				/* owner = SysAdmin */
    writeUInt(1);					/* useCount = 1 */
    writeUInt(8);					/* elementCount = 8 */
    /* bah! need binary constants here! */
/*
    0b0000000000000000
    0b0011111111111100
    0b0100000000000010
    0b0100001111000010
    0b0100011111100010
    0b0100011000000010
    0b0010001111000100
    0b0010000001100100
    0b0010011111100100
    0b0001001111001000
    0b0001000000001000
    0b0000100000010000
    0b0000010000100000
    0b0000001001000000
    0b0000000110000000
    0b0000000000000000
*/
    writeByte(0x00); writeByte(0x00); writeByte(0x3f); writeByte(0xfc);
    writeByte(0x40); writeByte(0x02); writeByte(0x43); writeByte(0xc2);
    writeByte(0x47); writeByte(0xe2); writeByte(0x46); writeByte(0x02);
    writeByte(0x23); writeByte(0xc4); writeByte(0x20); writeByte(0x64);
    writeByte(0x27); writeByte(0xe4); writeByte(0x13); writeByte(0xc8);
    writeByte(0x10); writeByte(0x08); writeByte(0x08); writeByte(0x10);
    writeByte(0x04); writeByte(0x20); writeByte(0x02); writeByte(0x40);
    writeByte(0x01); writeByte(0x80); writeByte(0x00); writeByte(0x00);
}

static void
writeIndexFile(void)
{

    writeULong((VERSION1 - '0') * 10 + (VERSION2 - '0'));
    writeULong(INDEX_COUNT);
    writeULong(CHAR_LEN);
    writeULong(NIL_TABLE | 1);
    writeULong(NIL_ACTION);
    writeULong((TYPE_STRING << TYPE_SHIFT) | 3);
    writeULong(0x0000001c);		 /* flags */
    writeULong(NIL_LIST_INTEGER | 4);
    writeULong(10);			 /* default abort timeout */
    (void) fwrite(&IndexData, INDEX_COUNT, sizeof(Index_t), Fd);
}

int
main(int argc, char *argv[])
{
    char reply[100];
    BOOL_T autoYes;

    autoYes = B_FALSE;
    argc -= 1;
    ++argv;
    while (argc != 0) {
	argc -= 1;
	if (strcmp(*argv, "-y") == 0) {
	    autoYes = B_TRUE;
	} else {
	    fprintf(stderr, "Use is: MUDCre [-y]\n");
	    return 1;
	}
	++argv;
    }

    Fd = fopen(INDEX_NAME, "r");
    if (Fd != NULL) {
	(void) fclose(Fd);
	if (! autoYes) {
	    printf("Database files already exist. Overwrite? [N/y] ");
	    fflush(stdout);
	    if (fgets(&reply[0], 100, stdin) == NULL || reply[0] == '\n' ||
		reply[0] == '\0' || (reply[0] != 'y' && reply[0] != 'Y'))
	    {
		fprintf(stderr, "Old files not removed.\n");
		return 1;
	    }
	}
	(void) remove(INDEX_NAME);
	(void) remove(DATA_NAME);
	(void) remove(LOG_NAME);
	printf("Removed old files.\n");
    }
    Fd = fopen(INDEX_NAME, "wb");
    if (Fd == NULL) {
	fprintf(stderr, "Could not create index file '%s'!\n", INDEX_NAME);
	return 1;
    }
    writeIndexFile();
    (void) fclose(Fd);
    printf("Wrote index file '%s'\n", INDEX_NAME);
    Fd = fopen(DATA_NAME, "wb");
    if (Fd == NULL) {
	fprintf(stderr, "Could not create data file '%s'!\n", DATA_NAME);
	return 1;
    }
    writeDataFile();
    fclose(Fd);
    printf("Wrote data file '%s'\n", DATA_NAME);
    return 0;
}
