#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
/*#include <sys/select.h> for AIX */
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <fcntl.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "request.h"

#define COUNT		10000
#define INTERVAL	100
#define MESSAGE		"thisIsAnIdentifierToLookup"

int Socket;

static void
doAbort(void)
{
    shutdown(Socket, 2);
    close(Socket);
    exit(1);
}

static void
doPerror(char *where)
{
    perror(where);
    doAbort();
}

static void
sendOne(unsigned char kind, void *p, unsigned int len)
{
    Request_t rq;

    rq.rq_type = kind;
    rq.rq_usedLen = htons(len);
    if (len != 0) {
	memcpy(&rq.rq_text[0], p, len);
    }
    len += REQ_LEN;
    if (send(Socket, (char *) &rq, len, 0) != len) {
	doPerror("send");
    }
}

static void
waitForReply(void)
{
    struct timeval tv;
    fd_set readFds, exceptFds;
    Request_t rq;
    int res;

    FD_ZERO(&readFds);
    FD_SET(Socket, &readFds);
    exceptFds = readFds;
    tv.tv_sec = 10;
    tv.tv_usec = 0;
    res = select(Socket + 1, &readFds, NULL, &exceptFds, &tv);
    if (res < 0) {
	doPerror("select");
    }
    if (recv(Socket, (char *) &rq, REQ_LEN, 0) != REQ_LEN) {
	doPerror("recv 1");
    }
    res = ntohs(rq.rq_usedLen);
    if (res != 0) {
	if (res > MAX_LEN) {
	    printf("len is %d?\n", res);
	    doAbort();
	}
	if (recv(Socket, &rq.rq_text[0], res, 0) != res) {
	    doPerror("recv 2");
	}
    }
}

int
main(int argc, char **argv)
{
    struct sockaddr_in addr;
    struct hostent *hp;
    int i, on;
    enum {
	noReply,
	withReply,
	mixed
    } messageType;
    static char *(typeNames[]) = {"noReply", "withReply", "mixed"};

    messageType = withReply;
    if (argc > 1) {
	if (*argv[1] == 'n') {
	    messageType = noReply;
	} else if (*argv[1] == 'm') {
	    messageType = mixed;
	} else if (*argv[1] == 'w') {
	    messageType = withReply;
	} else {
	    fprintf(stderr, "Usage: client [n|m|w]\n");
	    exit(1);
	}
    }
    hp = gethostbyname("localhost");
    if (hp == NULL) {
	perror("gethostbyname");
	exit(1);
    }
    Socket = socket(AF_INET, SOCK_STREAM, 0);
    if (Socket < 0) {
	perror("socket");
	exit(1);
    }
#if 1
    on = 1;
    if (setsockopt(Socket, IPPROTO_TCP, TCP_NODELAY, (char *) &on,
		   sizeof(int)) != 0)
    {
	perror("setsockopt(TCP_NODELAY)");
	close(Socket);
	exit(1);
    }
#endif
    memset(&addr, 0, sizeof(struct sockaddr_in));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(6667);
    addr.sin_addr.s_addr = *(unsigned long int *) hp->h_addr;
    printf("connecting to 0x%08x/%d\n", ntohl(addr.sin_addr.s_addr), 6667);
    if (connect(Socket, (struct sockaddr *) &addr,
		sizeof(struct sockaddr_in)) < 0)
    {
	doPerror("connect");
    }
    printf("connected, doing %d %s operations\n",COUNT,typeNames[messageType]);
    for (i = 0; i != COUNT; i += 1) {
	if (messageType != withReply) {
	    sendOne(rt_noReply, NULL, 0);
	    /* No reply needed */
	}
	if (messageType != noReply) {
	    sendOne(rt_withReply, MESSAGE, strlen(MESSAGE));
	    /* wait for reply */
	    waitForReply();
	}
	if (i % INTERVAL == 0) {
	    printf("%d ", i);
	    fflush(stdout);
	}
    }
    /* signal server to go away */
    sendOne(rt_shutdown, NULL, 0);
    shutdown(Socket, 2);
    close(Socket);
    printf("client all done\n");
    return 0;
}
