#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
/*#include <sys/select.h> for AIX */
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>

#include "request.h"

#define INTERVAL	100

static void
handleSession(int sock, int s)
{
    Request_t rq;
    fd_set readFds;
    struct timeval tv;
    int res, len, count, sent, received;

    count = 0;
    sent = 0;
    received = 0;
    while (1) {
	FD_ZERO(&readFds);
	FD_SET(sock, &readFds);
	FD_SET(s, &readFds);
	tv.tv_sec = 10;
	tv.tv_usec = 0;
	res = select(s + 1, &readFds, NULL, NULL, &tv);
	if (res < 0) {
	    perror("select");
	    return;
	}
	if (FD_ISSET(s, &readFds)) {
	    if (recv(s, (char *) &rq, REQ_LEN, 0) != REQ_LEN) {
		perror("recv 1");
		return;
	    }
	    count += 1;
	    if (count % INTERVAL == 0) {
		printf("%d ", count);
		fflush(stdout);
	    }
	    if (rq.rq_type == rt_shutdown) {
		/* client will now terminate */
		break;
	    }
	    len = ntohs(rq.rq_usedLen);
	    if (len != 0) {
		received += len;
		if (recv(s, &rq.rq_text[0], len, 0) != len) {
		    perror("recv 2");
		    return;
		}
	    }
	    if (rq.rq_type == rt_withReply) {
		len = (rand() / 1000) % MAX_LEN;
		rq.rq_usedLen = htons(len);
		sent += len;
		len += REQ_LEN;
		if (send(s, (char *) &rq, len, 0) != len) {
		    perror("send");
		    return;
		}
	    } /* otherwise, no reply */
	} else if (FD_ISSET(sock, &readFds)) {
	    printf("a second connection!\n");
	}
    }
    printf("sent %d, received %d\n", sent, received);
}

static void
acceptConnections(int sock)
{
    struct sockaddr_in myAddr, from;
    fd_set readFds;
    int fromLen, s, res, on;

    memset(&myAddr, 0, sizeof(struct sockaddr_in));
    myAddr.sin_port = htons(6667);
    myAddr.sin_family = AF_INET;
    if (bind(sock, (struct sockaddr *) &myAddr,
	sizeof(struct sockaddr_in)) < 0)
    {
	perror("bind");
	return;
    }
    if (listen(sock, 5) < 0) {
	perror("listen");
	return;
    }
    printf("server all set, awaiting connection\n");
    FD_ZERO(&readFds);
    FD_SET(sock, &readFds);
    res = select(sock + 1, &readFds, NULL, NULL, NULL);
    if (res < 0) {
	perror("select");
	return;
    }
    fromLen = sizeof(struct sockaddr_in);
    s = accept(sock, (struct sockaddr *) &from, &fromLen);
    if (s < 0) {
	perror("accept");
	return;
    }
#if 0
    on = 1;
    if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *)&on,sizeof(int)) != 0){
	perror("setsockopt(TCP_NODELAY)");
	close(s);
	return;
    }
#endif
    printf("server got a connection, fd %d\n", s);
    handleSession(sock, s);
    shutdown(s, 2);
    close(s);
}

static void
handler(int sig)
{
    printf("interrupt!\n");
}

int
main(int argc, char **argv)
{
    int sock;

    (void) signal(SIGINT, handler);
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	perror("socket");
	exit(1);
    }
    /* ioctl(sock, FIONBIO ...) has not noticeable effect */
    acceptConnections(sock);
    shutdown(sock, 2);
    close(sock);
    printf("server done\n");
    return 0;
}
